
using System;
using System.IO;

namespace Espresso.OMR
{
  abstract class BaseMObjects
	{
		protected PageManager page_manager_;

		bool empirical_filename_valid_;
		string empirical_filename_;

		static bool definitive_convert_ = false;
		static bool remove_text_pending_ = false;
		static bool text_removed_ = false;
		static bool single_staff_lines_ = false;
		static int  insist_sl_num_ = 0;

		protected GCharList gcharacters_;
		protected StaffSystemList staff_systems_;
		protected GObjectList non_staff_systems_;
		protected StaffInfo staff_info_;

		protected GCharList def_gcharacters_;
		protected StaffSystemList def_staff_systems_;
		protected GObjectList def_non_staff_systems_;
		protected StaffInfo def_staff_info_;


		public BaseMObjects(PageManager page_manager, int x_dim, int y_dim)
		{

			page_manager_ = page_manager; 
			empirical_filename_valid_ = false;
			empirical_filename_ = "";

			gcharacters_ = new GCharList();
			staff_systems_ = new StaffSystemList();
			non_staff_systems_ = new GObjectList();
			staff_info_ = new StaffInfo();

			def_gcharacters_ = new GCharList();
			def_staff_systems_ = new StaffSystemList();
			def_non_staff_systems_ = new GObjectList();
			def_staff_info_ = new StaffInfo();

		}


		public virtual void resize(int x_dim, int y_dim)
		{
			// nothing in base object currently needs to be resized
		}

		public virtual void reset_and_resize(int x_dim, int y_dim)
		{
			gcharacters_.reset();
			staff_systems_.reset();
			non_staff_systems_.reset();

			def_gcharacters_.reset();
			def_staff_systems_.reset();
			def_non_staff_systems_.reset();
		}

		/*-------------------------------------------------------------------------*/


		public static void set_definitive_convert(bool def_conv)
		{
			definitive_convert_ = def_conv;
		}

		public static bool definitive_convert()
		{
			return definitive_convert_;
		}

		public static void set_remove_text_pending(bool pending)
		{
			remove_text_pending_ = pending;
		}

		public static bool text_removed()
		{
			return text_removed_;
		}

		public static void set_single_staff_lines(bool ssl)
		{
			single_staff_lines_ = ssl;
		}

		public static bool single_staff_lines()
		{
			return single_staff_lines_;
		}

		public static void set_insist_sl_num(int insist_sl_num)
		{
			insist_sl_num_ = insist_sl_num;
		}

		public static int insist_sl_num()
		{
			return insist_sl_num_;
		}


		/*-------------------------------------------------------------------------*/


		public int no_staff_system_objects(StaffSystemList ssl)
		{
			int no_objects = 0;

			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				BaseStaffSystem staff_system = current_ss.val();
				no_objects+=staff_system.staff_system_objects().length();
				current_ss = ssl.next();
			}

			return no_objects;
		}

		public int no_staffs(StaffSystemList ssl)
		{
			int no_staffs = 0;

			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				BaseStaffSystem staff_system = current_ss.val();
				no_staffs+=staff_system.staffs().length();
				current_ss = ssl.next();
			}

			return no_staffs;
		}



		public int no_def_staff_system_objects()
		{
			int no_objects = 0;

			StaffSystemList ssl = def_staff_systems_;
			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				BaseStaffSystem staff_system = current_ss.val();
				no_objects+=staff_system.staff_system_objects().length();
				current_ss = ssl.next();
			}

			return no_objects;
		}

		public int no_def_staffs()
			   {
				   int no_staffs = 0;

				   StaffSystemList ssl = def_staff_systems_;
				   StaffSystemNode current_ss = ssl.head();
				   while (current_ss!=null)
				   {
					   BaseStaffSystem staff_system = current_ss.val();
					   no_staffs+=staff_system.staffs().length();
					   current_ss = ssl.next();
				   }

				   return no_staffs;
			   }

		/*-------------------------------------------------------------------------*/

		protected void build_original_dsep_bitmap(int page_x_dim,
			int page_y_dim)
		{
			// build up page in current_fbitmap

			FileBitmap current_fbitmap = page_manager_.current_fbitmap();
			current_fbitmap.resize(page_x_dim,page_y_dim);
			current_fbitmap.clear_all();

			if (def_gcharacters_.length()>0)
			{
				if (!remove_text_pending_)
				{
					GCharNode current_gc = def_gcharacters_.head();
					while (current_gc!=null)
					{
						GChar gchar = current_gc.val();
						current_fbitmap.inlay_bitmap(gchar);
						current_gc = def_gcharacters_.next();
					}
				}
				else
				{
					text_removed_ = true;
					remove_text_pending_ = false;
				}
			}

			StaffSystemList ssl = def_staff_systems_; // shorthand
			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();

				GObject ss_gobject = staff_system.gobject();
				current_fbitmap.inlay_bitmap(ss_gobject);

				GObjectList sso = staff_system.staff_system_objects(); // staff sys obs 

				current_ss = ssl.next();
			}

			GObjectList nssl = def_non_staff_systems_; // shorthand
			GObjectNode current_nss = nssl.head();
			while (current_nss!=null)
			{
				GObject non_ss_object = current_nss.val();
				current_fbitmap.inlay_bitmap(non_ss_object);
				current_nss = nssl.next();
			}
		}




		protected void build_original_convert_dsep_bitmap(int page_x_dim,
			int page_y_dim)
		{
			// build up page in current_fbitmap

			FileBitmap current_fbitmap = page_manager_.current_fbitmap();
			current_fbitmap.resize(page_x_dim,page_y_dim);
			current_fbitmap.clear_all();

			if (gcharacters_.length()>0)
			{
				if (!remove_text_pending_)
				{
					GCharNode current_gc = gcharacters_.head();
					while (current_gc!=null)
					{
						GChar gchar = current_gc.val();
						current_fbitmap.inlay_bitmap(gchar);
						current_gc = gcharacters_.next();
					}
				}
				else
				{
					text_removed_ = true;
					remove_text_pending_ = false;
				}
			}

			StaffSystemList ssl = staff_systems_; // shorthand
			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();

				GObject ss_gobject = staff_system.gobject();
				current_fbitmap.inlay_bitmap(ss_gobject);
#if _DSEP
	      GObject without_staffs = staff_system.without_staffs();
	      without_staffs.copy_sub_bitmap(ss_gobject);
#endif

				current_ss = ssl.next();
			}

			GObjectList nssl = non_staff_systems_; // shorthand
			GObjectNode current_nss = nssl.head();
			while (current_nss!=null)
			{
				GObject non_ss_object = current_nss.val();
				current_fbitmap.inlay_bitmap(non_ss_object);
				current_nss = nssl.next();
			}
		}



		protected void build_original_sep_bitmap(int page_x_dim,
			int page_y_dim)
		{
			// build up page in current_fbitmap

			FileBitmap current_fbitmap = page_manager_.current_fbitmap();
			current_fbitmap.resize(page_x_dim,page_y_dim);
			current_fbitmap.clear_all();

			GCharNode current_gc = gcharacters_.head();
			while (current_gc!=null)
			{
				GChar gchar = current_gc.val();
				current_fbitmap.inlay_bitmap(gchar);
				current_gc = gcharacters_.next();
			}

			StaffSystemList ssl = staff_systems_; // shorthand
			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();
				GObjectList sso = staff_system.staff_system_objects(); // staff sys obs 

				GObjectNode current_sso = sso.head();
				while (current_sso!=null)
				{
					current_fbitmap.inlay_bitmap(current_sso.val());
					current_sso = sso.next();
				}

				StaffList stl = staff_system.staffs();
				StaffNode current_st = stl.head();
				while (current_st!=null)
				{
					BaseStaff staff = (BaseStaff)current_st.val();
					StaffLines staff_lines = staff.staff_lines();
					SubBitmapArray1d bitmap_array = staff_lines.bitmaps();
					for (int i=0; i<staff.no_staff_lines(); i++)
						current_fbitmap.inlay_bitmap(bitmap_array[i]);
	  
					current_st = stl.next();
				}

				current_ss = ssl.next();
			}

			GObjectList nssl = non_staff_systems_; // shorthand
			GObjectNode current_nss = nssl.head();
			while (current_nss!=null)
			{
				GObject non_ss_object = current_nss.val();
				current_fbitmap.inlay_bitmap(non_ss_object);
				current_nss = nssl.next();
			}
		}


		public Bitmap bitmap() 
		{ return page_manager_.current_fbitmap(); }

		public virtual int page_x_dim()
		{ return page_manager_.current_fbitmap().x_dim(); }

		public virtual int page_y_dim()
		{ return page_manager_.current_fbitmap().y_dim(); }


		public GCharList gcharacters() 
		{ return gcharacters_; }

		public StaffSystemList staff_systems() 
		{ return staff_systems_; }

		public GObjectList non_staff_systems() 
		{ return non_staff_systems_; }

		public StaffInfo staff_info() 
		{ return staff_info_; }

		public int no_gcharacters()
		{ return gcharacters_.length(); }

		public int no_non_staff_system_objects()
		{ return non_staff_systems_.length(); }


		public GCharList def_gcharacters() 
		{ return def_gcharacters_; }

		public StaffSystemList def_staff_systems() 
		{ return def_staff_systems_; }

		public GObjectList def_non_staff_systems() 
		{ return def_non_staff_systems_; }

		public StaffInfo def_staff_info() 
		{ return def_staff_info_; }

		public int no_def_gcharacters()
		{ return def_gcharacters_.length(); }

		public int no_def_non_staff_system_objects()
		{ return def_non_staff_systems_.length(); }

		public bool empirical_filename_valid()
		{ return empirical_filename_valid_; }

		public string empirical_filename()
		{ return empirical_filename_; }


		/*---------------------------------------------------------------------------*/

		protected void output_header(BinaryWriter bwout, String header,
				   int restrict_y_top,
				   int restrict_y_bot)
		  {
		    IOSupport.write_line(bwout,header);

		    IOSupport.write_line(bwout,"dpi = " + page_manager_.dpi().ToString());

		    IOSupport.write(bwout,"page_x_dim = " + page_x_dim().ToString() + " , "); // virtual function call
		    if (restrict_y_bot==0)
		      {
			IOSupport.write_line(bwout,"page_y_dim = " + page_y_dim().ToString());  // virtual function call
		      }
		    else
		      {
			int ry_dim = restrict_y_bot - restrict_y_top +1;
			IOSupport.write_line(bwout,"page_y_dim = " + ry_dim.ToString());
		      }
		  }

		protected void output_header(BinaryWriter bwout, String header)
		{
			output_header(bwout,header,0,0);
		}

		protected void output_sep_gchars(BinaryWriter bwout,
				       String pre_label,
				       GCharList gcharacters,
				       int restrict_y_top,
				       int restrict_y_bot)
		  {

		    int no_gcharacters = 0;
		    if (restrict_y_bot>0)
		      {
			for (GCharNode cg=gcharacters.head(); cg!=null; cg=gcharacters.next())
			  {
			    GChar gchar = cg.val();
			    int gchar_yt = gchar.y_top();
			    int gchar_yb = gchar.y_bot();
			    if ((gchar_yt>=restrict_y_top) && (gchar_yt<=restrict_y_bot))
			      {
				int yb = gchar.y_bot();
				if (yb<=restrict_y_bot)
				  {
				    no_gcharacters++;
				  }
			      }
			  }
		      }
		    else
		      {
			no_gcharacters = gcharacters.length();
		      }

		    IOSupport.write_line(bwout,pre_label + "gcharacters :");
		    IOSupport.write_line(bwout,"no_gcharacters = " + no_gcharacters.ToString());

		    for (GCharNode cg=gcharacters.head(); cg!=null; cg=gcharacters.next())
		      {
			if (restrict_y_bot==0)
			  {
				GChar gchar = cg.val();
			    gchar.write(bwout);
			  }
			else
			  {
			    GChar gchar = cg.val();
			    int gchar_yt = gchar.y_top();
			    if ((gchar_yt>=restrict_y_top) && (gchar_yt<=restrict_y_bot))
			      {
				int yb = gchar.y_bot();
				if (yb<=restrict_y_bot)
				  {
				    gchar.add_to_y_org(-restrict_y_top);
				    gchar.write(bwout);
				    gchar.add_to_y_org(+restrict_y_top);
				  }
			      }
			  }
		      }
		  }

		protected void output_sep_gchars(BinaryWriter bwout,
			String pre_label,
			GCharList gcharacters)
		{
			output_sep_gchars(bwout,pre_label,gcharacters,0,0);
		}


		protected void output_sep_ss(BinaryWriter bwout, String pre_label,
				   StaffSystemList ssl, 
				   bool is_complete,
				   int restrict_y_top,
				   int restrict_y_bot)
		  {
		    // * if updating this function, keep 'output_given_pr_objects' consistent. *

		    int no_staff_systems = 0;
		    if (restrict_y_bot>0)
		      {
			for (StaffSystemNode ss=ssl.head(); ss!=null; ss=ssl.next())
			  {
			    StaffSystem staff_system = ss.val();
			    GObject ss_gobject = staff_system.gobject();
			    int ss_yt = ss_gobject.y_top();

			    if ((ss_yt>=restrict_y_top) && (ss_yt<=restrict_y_bot))
			      {
				no_staff_systems++;
			      }
			  }
		      }
		    else
		      {
			no_staff_systems = ssl.length();
		      }

		    IOSupport.write_line(bwout,pre_label + "staff_systems :");
		    IOSupport.write_line(bwout,"no_staff_systems = " + no_staff_systems.ToString());

		    for (StaffSystemNode ss=ssl.head(); ss!=null; ss=ssl.next())
		      {
			StaffSystem staff_system = ss.val();
			GObject ss_gobject = staff_system.gobject();
			int ss_yt = ss_gobject.y_top();

			if ((restrict_y_bot==0) ||
			    ((ss_yt>=restrict_y_top) && (ss_yt<=restrict_y_bot)))
			  {
			    if (is_complete)
			      {
				staff_system.write_complete_staff_system(bwout,restrict_y_top);
			      }
			    else
			      {
				staff_system.write_staff_system(bwout,restrict_y_top);
			      }
			  }
		      }
		  }

		protected void output_sep_ss(BinaryWriter bwout, String pre_label,
			StaffSystemList ssl, 
			bool is_complete)
		{
			output_sep_ss(bwout,pre_label,ssl,is_complete,0,0);
		}

		protected void output_sep_nss_and_sinfo(BinaryWriter bwout, 
					      String pre_label,
					      GObjectList nssl,
					      StaffInfo staff_info,
					      int restrict_y_top,
					      int restrict_y_bot)
		  {
		    // * if updating this function, keep 'output_given_pr_objects' consistent. *

		    int no_non_staff_systems = 0;
		    if (restrict_y_bot>0)
		      {
			for (GObjectNode nss=nssl.head(); nss!=null; nss=nssl.next())
			  {
			    GObject nss_gobject = nss.val();
			    int yt = nss_gobject.y_top();
			    if ((yt>=restrict_y_top) && (yt<=restrict_y_bot))
			      {
				int yb = nss_gobject.y_bot();
				if (yb<=restrict_y_bot)
				  {
				    no_non_staff_systems++;
				  }
			      }
			  }
		      }
		    else
		      {
			no_non_staff_systems = nssl.length();
		      }

		    IOSupport.write_line(bwout,pre_label + "non_staff_systems :");
		    IOSupport.write_line(bwout,"no_non_staff_systems = " + no_non_staff_systems.ToString());

		    for (GObjectNode nss=nssl.head(); nss!=null; nss=nssl.next())
		      {
			if (restrict_y_bot==0)
			  {
				GObject nss_gobject = nss.val();
			    nss_gobject.write(bwout);
			  }
			else
			  {
			    GObject nss_gobject = nss.val();
			    int yt = nss_gobject.y_top();
			    if ((yt>=restrict_y_top) && (yt<=restrict_y_bot))
			      {
				int yb = nss_gobject.y_bot();
				if (yb<=restrict_y_bot)
				  {
				    nss_gobject.add_to_y_org(-restrict_y_top);
				    nss_gobject.write(bwout);
				    nss_gobject.add_to_y_org(+restrict_y_top);
				  }
			      }
			  }
		      }

		    staff_info.write(bwout);
		  }

		protected void output_sep_nss_and_sinfo(BinaryWriter bwout, 
			String pre_label,
			GObjectList nssl,
			StaffInfo staff_info)
		{
			output_sep_nss_and_sinfo(bwout,pre_label,nssl,staff_info,0,0);
		}



		public void output_dseparated_objects(String dseperate_filename)
		  {
		    // * if updating this function, keep 'output_given_pr_objects' consistent. *

		    BinaryWriter bwout =FilenameManager.open_output_pipe(dseperate_filename);

		    output_header(bwout,"Definitive Seperate File Format 1.2");

		    output_sep_gchars(bwout,"def_",def_gcharacters_);
		    output_sep_ss(bwout,"def_",def_staff_systems_,true);
		    output_sep_nss_and_sinfo(bwout,"def_",def_non_staff_systems_,def_staff_info_);

		    FilenameManager.close_output_pipe();
		  }


		public void output_separated_objects(String seperate_filename,
					      StaffSystemList ssl,
					      GObjectList nssl,
					      int restrict_y_top,
					      int restrict_y_bot)
		  {
		    // * if updating this function, keep 'output_given_pr_objects' consistent. *

		    BinaryWriter bwout = FilenameManager.open_output_pipe(seperate_filename);

		    int ryt = restrict_y_top;
		    int ryb = restrict_y_bot;
		    output_header(bwout,"Seperate File Format 1.2",ryt,ryb);

		    output_sep_gchars(bwout,"",gcharacters_,ryt,ryb);
		    output_sep_ss(bwout,"",ssl,false,ryt,ryb);
		    output_sep_nss_and_sinfo(bwout,"",nssl,staff_info_,ryt,ryb);

		    output_sep_gchars(bwout,"def_",gcharacters_,ryt,ryb);
		    output_sep_ss(bwout,"def_",def_staff_systems_,false,ryt,ryb);
		    output_sep_nss_and_sinfo(bwout,"def_",def_non_staff_systems_,def_staff_info_,
					     ryt,ryb);

		    FilenameManager.close_output_pipe();
		  }


		public void output_separated_objects(String separate_filename,
			StaffSystemList ssl,
			GObjectList nssl)
		{
			output_separated_objects(separate_filename,ssl,nssl,0,0);
		}

		/*------------------------*/

		protected bool input_sep_gchars(BinaryReader brin, String pre_label,
				      GCharList gcharacters,
				      int y_offset)
		  {
		    string gchar_expected_token = pre_label + "gcharacters";

		    if (!IOSupport.read_token(brin,gchar_expected_token)) return false;
		    if (!IOSupport.read_token(brin,":")) return false;
      
		    int no_gcharacters;
		    if (!IOSupport.read_token(brin,"no_gcharacters")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out no_gcharacters)) return false;
    
		    for (int i=0; i<no_gcharacters; i++)
		      {
			GChar gchar = new GChar();
			if (!gchar.read(brin)) return false; // does resize
			gchar.add_to_y_org(y_offset);
			gcharacters.append(gchar); // responisbilty passed over to list
		      }

		    return true;
		  }



		protected bool input_sep_ss_and_nss(BinaryReader brin, String pre_label,
					  StaffSystemList ssl, 
					  GObjectList nssl,
					  StaffInfo staff_info,
					  int y_offset)
		  {
		    String ss_expected_token = pre_label + "staff_systems";

		    if (!IOSupport.read_token(brin,ss_expected_token)) return false;
		    if (!IOSupport.read_token(brin,":")) return false;
  
		    int no_staff_systems;
		    if (!IOSupport.read_token(brin,"no_staff_systems")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out no_staff_systems)) return false;
  
		    for (int i=0; i<no_staff_systems; i++)
		      {
			StaffSystem staff_system  // call to a virtual fn
			  = generate_empty_staff_system();
			if (!staff_system.read_staff_system(brin,y_offset)) return false;
			ssl.append(staff_system); // responisbilty passed over to list
		      }
  
		    String nss_expected_token = pre_label + "non_staff_systems";

		    if (!IOSupport.read_token(brin,nss_expected_token)) return false;
		    if (!IOSupport.read_token(brin,":")) return false;
  
		    int no_non_staff_systems;
		    if (!IOSupport.read_token(brin,"no_non_staff_systems")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out no_non_staff_systems)) return false;
  
		    for (int i=0; i<no_non_staff_systems; i++)
		      {
			GObject nss_ob = new GObject();
			if (!nss_ob.read(brin)) return false;
			nss_ob.add_to_y_org(y_offset);
			nss_ob.resize_projects(nss_ob.x_dim(),nss_ob.y_dim());
			nssl.append(nss_ob); // responisbilty passed over to list
		      }
  
		    if (!staff_info.read(brin)) return false;

		    return true;
		  }



		public bool input_dseparated_objects(BinaryReader brin, int y_offset)
		  {

		    double format_no;
		    if (!IOSupport.read_token(brin,"Definitive" )) return false;
		    if (!IOSupport.read_token(brin,"Seperate" )) return false;
		    if (!IOSupport.read_token(brin,"File" )) return false;
		    if (!IOSupport.read_token(brin,"Format" )) return false;
		    if (!IOSupport.read_double(brin,out format_no)) return false;
  
		    int dpi;
		    if (!IOSupport.read_token(brin,"dpi")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out dpi)) return false;
		    page_manager_.set_dpi(dpi);

		    String next_keyword;
		    int page_x_dim;
		    if (!IOSupport.read_string(brin,out next_keyword)) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (next_keyword=="empiricals")
		      {
			if (!IOSupport.read_string(brin,out empirical_filename_)) return false;
			empirical_filename_valid_ = true;

			if (!IOSupport.read_token(brin,"page_x_dim")) return false;
			if (!IOSupport.read_token(brin,"=")) return false;
		      }
		    else
		      {
			empirical_filename_valid_ = false;
			empirical_filename_ = "";
		      }
		    if (!IOSupport.read_cardinal(brin,out page_x_dim)) return false; 

		    if (!IOSupport.read_token(brin,",")) return false;
  
		    int page_y_dim;
		    if (!IOSupport.read_token(brin,"page_y_dim")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out page_y_dim)) return false;

		    if (y_offset==0)
		      {
			reset_and_resize(page_x_dim,page_y_dim);
		      }
		    else
		      {
			int curr_page_x_dim = page_manager_.current_fbitmap().x_dim();
			page_x_dim = Math.Max(curr_page_x_dim,page_x_dim);
			page_y_dim += y_offset;
			resize(page_x_dim,page_y_dim);
		      }
  
		    if (format_no >= 1.2)
		      {
			if (!definitive_convert_)
			  {
			    if (format_no >= 1.1) // unnecessary ? //****
			      {
				if (!input_sep_gchars(brin,"def_",def_gcharacters_,y_offset)) 
				  return false;
			      }
			  }
			else
			  {
			    if (format_no >= 1.1) // unnecessary ? //****
			      {
				if (!input_sep_gchars(brin,"def_",gcharacters_,y_offset)) 
				  return false;
			      }
			  }

			if (!definitive_convert_)
			  {
			    if (!input_sep_ss_and_nss(brin,"def_",def_staff_systems_,
						      def_non_staff_systems_,
						      def_staff_info_,y_offset)) return false;
			  }
			else
			  {
			    if (!input_sep_ss_and_nss(brin,"def_",staff_systems_,
						      non_staff_systems_,
						      staff_info_,y_offset)) return false;
			  }
		      }

		    // build up page in current_fbitmap
		    if (!definitive_convert_)
		      {
			build_original_dsep_bitmap(page_x_dim,page_y_dim);
		      }
		    else
		      {
			build_original_convert_dsep_bitmap(page_x_dim,page_y_dim);
		      }

		    return true;
		  }


	  public bool input_dseparated_objects(BinaryReader brin)
	  {
		  return input_dseparated_objects(brin,0);
	  }

		public bool input_separated_objects(BinaryReader brin, int y_offset)
		  {
		    double format_no;
		    if (!IOSupport.read_token(brin,"Seperate" )) return false;
		    if (!IOSupport.read_token(brin,"File" )) return false;
		    if (!IOSupport.read_token(brin,"Format" )) return false;
		    if (!IOSupport.read_double(brin,out format_no)) return false;
  
		    int dpi;
		    if (!IOSupport.read_token(brin,"dpi")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out dpi)) return false;
		    page_manager_.set_dpi(dpi);

		    String next_keyword;
		    int page_x_dim;
		    if (!IOSupport.read_string(brin,out next_keyword)) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (next_keyword=="empiricals")
		      {
			if (!IOSupport.read_string(brin,out empirical_filename_)) return false;
			empirical_filename_valid_ = true;

			if (!IOSupport.read_token(brin,"page_x_dim")) return false;
			if (!IOSupport.read_token(brin,"=")) return false;
		      }
		    else
		      {
			empirical_filename_valid_ = false;
			empirical_filename_ = "";
		      }
		    if (!IOSupport.read_cardinal(brin,out page_x_dim)) return false; 

		    if (!IOSupport.read_token(brin,",")) return false;
  
		    int page_y_dim;
		    if (!IOSupport.read_token(brin,"page_y_dim")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out page_y_dim)) return false;

		    if (y_offset==0)
		      {
			reset_and_resize(page_x_dim,page_y_dim);
		      }
		    else
		      {
			int curr_page_x_dim = page_manager_.current_fbitmap().x_dim();
			page_x_dim = Math.Max(curr_page_x_dim,page_x_dim);
			page_y_dim += y_offset;
			resize(page_x_dim,page_y_dim);
		      }

		    if (format_no >= 1.1) 
		      {
			if (!input_sep_gchars(brin,"",gcharacters_,y_offset)) return false;
		      }
		    if (!input_sep_ss_and_nss(brin,"",staff_systems_,non_staff_systems_,
					      staff_info_,y_offset)) return false;
  
		    if (format_no >= 1.2)
		      {
			if (format_no >= 1.1) // unnecessary ? //****
			  {
			    if (!input_sep_gchars(brin,"def_",gcharacters_,y_offset)) return false;
			  }
			if (!input_sep_ss_and_nss(brin,"def_",def_staff_systems_,
						  def_non_staff_systems_,
						  def_staff_info_,y_offset)) return false;
		      }
  
		    // build up page in current_fbitmap
		    build_original_sep_bitmap(page_x_dim,page_y_dim);

		    return true;
		  }


		protected abstract StaffSystem generate_empty_staff_system();


	}
}
