
using System;

namespace Espresso.OMR
{
	abstract class BasePageManager
	{
		protected StringList filenames_;
		protected int dpi_;
		protected TimedFileBitmap current_fbitmap_;

		public BasePageManager()
		{
			filenames_ = new StringList();
			dpi_ = 0;
			current_fbitmap_ = new TimedFileBitmap();
		}

		public BasePageManager(int argc, string[] argv)
		{
			filenames_ = new StringList();
			dpi_ = 300;
			current_fbitmap_ = new TimedFileBitmap();

			for (int i=1; i<argc; i++)
			{
				string fname = argv[i];
				filenames_.append(fname);
			}
		}

		public abstract void replace_page(string rp_filename);

		public BasePageManager(StringArray1d filenames)
		{
			filenames_ = new StringList();
			dpi_ = 300;
			current_fbitmap_ = new TimedFileBitmap();

			int no_filenames = filenames.length();
			for (int i=1; i<no_filenames; i++)
			{
				string filename = filenames[i];
				filenames_.append(filename);
			}
		}

		public TimedFileBitmap next_page()
		{
			StringNode new_page_node = filenames_.next();

			if (new_page_node!=null) 
			{
				string new_page = new_page_node.val();
				replace_page(new_page);
			}

			return current_fbitmap_;
		}


		public void replace_filenames(StringArray1d filenames)
		{
			filenames_.reset();

			for (int i=0; i<filenames.length(); i++)
			{
				string fname = filenames[i];
				filenames_.append(fname);
			}
		}


		public TimedFileBitmap prev_page()
		{
			StringNode new_page_node = filenames_.prev();

			if (new_page_node!=null) 
			{
				string new_page = new_page_node.val();
				replace_page(new_page);
			}

			return current_fbitmap_;
		}

		public TimedFileBitmap first_page()
		{
			StringNode new_page_node = filenames_.head();

			if (new_page_node!=null) 
			{
				string new_page = new_page_node.val();
				replace_page(new_page);
			}

			return current_fbitmap_;
		}

		public TimedFileBitmap last_page()
		{
			StringNode new_page_node = filenames_.tail();

			if (new_page_node!=null) 
			{
				string new_page = new_page_node.val();
				replace_page(new_page);
			}

			return current_fbitmap_;
		}

		public void set_dpi(int dpi) { dpi_ = dpi; }
		public int dpi() { return dpi_; }
		public int no_pages() { return filenames_.length(); }

		public TimedFileBitmap current_fbitmap() { return current_fbitmap_;}
		
		public string current_filename() 
		{ 
			// shouldn't be called if no_pages() == 0
			return filenames_.curr().val();
		}
	}
}
