
using System;
using System.IO;

namespace Espresso.OMR
{

  abstract class BaseStaff
    {
      protected GObject staff_area_;

      protected int no_staff_lines_;
      protected StaffLines staff_lines_;

      protected StaffInfo info_;

      public BaseStaff()
	{
	  staff_area_ = new GObject();
	  staff_lines_ = new StaffLines();
	  info_ = new StaffInfo();
	}


      public BaseStaff(SubBitmap bitmap, int x_org, int y_org,
		       int x_dim, int y_dim)
	{

	  staff_area_ = new GObject(bitmap,x_org,y_org,x_dim,y_dim);
	  no_staff_lines_ = 0;
	  staff_lines_ = new StaffLines();
	  info_ = new StaffInfo();
	}

      public BaseStaff(int x_org, int y_org, int x_dim, int y_dim)
	{
	  staff_area_ = new GObject(x_org,y_org,x_dim,y_dim);
	  no_staff_lines_ = 0;
	  staff_lines_ = new StaffLines();
	  info_ = new StaffInfo();
	}


      public void reset_org(int x_org, int y_org)
	{
	  staff_area_.add_to_x_org(x_org);
	  staff_area_.add_to_y_org(y_org);
	}

      public GObject staff_area() { return staff_area_; }

      public int no_staff_lines() { return no_staff_lines_; }
      public StaffLines staff_lines() { return staff_lines_; }
      public StaffInfo info() { return info_; }


      /*---------------------------------------------------------------------------*/

      protected virtual void write_staff_area(BinaryWriter bwout, int y_offset)
	{
	  IOSupport.write_line(bwout,"staff_area_data :");
	  IOSupport.write_line(bwout,"x_off = " + staff_area_.x_off().ToString());
	  IOSupport.write(bwout,"x_org = " + staff_area_.x_org().ToString() + " , ");

	  int y_org_rel = staff_area_.y_org() - y_offset;
	  IOSupport.write_line(bwout,"y_org = " + y_org_rel.ToString());
	  IOSupport.write(bwout,"x_dim = " + staff_area_.x_dim().ToString() + " , ");
	  IOSupport.write_line(bwout,"y_dim = " + staff_area_.y_dim().ToString());
	}

      protected void write_staff_area(BinaryWriter bwout)
	{
	  write_staff_area(bwout,0);
	}

      protected virtual void write_complete_staff_area(BinaryWriter bwout, 
				     int y_offset)
	{
	  IOSupport.write_line(bwout,"complete_staff_area_data :");

	  SubBitmap sub_bitmap = (SubBitmap)staff_area_;
	  sub_bitmap.add_to_y_org(-y_offset);
	  sub_bitmap.write(bwout);
	  sub_bitmap.add_to_y_org(+y_offset);
	}

      protected void write_complete_staff_area(BinaryWriter bwout)
	{
	  write_complete_staff_area(bwout,0);
	}

      protected virtual void write_staff_segs(BinaryWriter bwout)
	{
	  IOSupport.write_line(bwout,"left_staff_seg :");
	  // write out dummy entry
	  IOSupport.write_line(bwout,"global_x_start = 0 , global_x_end = 0");

	  IOSupport.write_line(bwout,"right_staff_seg :");
	  // write out dummy entry
	  IOSupport.write_line(bwout,"global_x_start = 0 , global_x_end = 0");
	}

      protected virtual void write_staff_lines(BinaryWriter bwout, int y_offset)
	{
	  int no_staff_lines = no_staff_lines_;
	  IOSupport.write_line(bwout,"no_staff_lines = " + no_staff_lines.ToString());
	  staff_lines_.write_staff_lines(bwout,y_offset);
	}

      protected void write_staff_lines(BinaryWriter bwout)
	{
	  write_staff_lines(bwout,0);
	}

      public void write_staff(BinaryWriter bwout, int y_offset)
	{
	  write_staff_area(bwout,y_offset);
	  write_staff_segs(bwout);
	  write_staff_lines(bwout,y_offset);
 
	  info_.write(bwout);
	}

      public void write_staff(BinaryWriter bwout)
	{
	  write_staff(bwout,0);
	}

      public void write_complete_staff(BinaryWriter bwout, int y_offset)
	{
	  write_complete_staff_area(bwout,y_offset);
	  write_staff_segs(bwout);
	  write_staff_lines(bwout);
 
	  info_.write(bwout);
	}

      public void write_complete_staff(BinaryWriter bwout)
	{
	  write_complete_staff(bwout,0);
	}



      protected virtual bool read_staff_area(BinaryReader brin, int y_offset)
	{
	  String token;
	  if (!IOSupport.read_string(brin,out token)) return false;

	  bool is_complete; // safe
	  if (token=="complete_staff_area_data")
	    {
	      is_complete = true;
	    }
	  else if (token=="staff_area_data")
	    {
	      is_complete = false;
	    }
	  else
	    {
	      Console.Error.WriteLine("Error. Unexpected token: " + token);
	      return false;
	    }

	  if (!IOSupport.read_token(brin,":")) return false;

	  if (!is_complete)
	    {
	      int x_off;
	      if (!IOSupport.read_token(brin,"x_off")) return false;
	      if (!IOSupport.read_token(brin,"=")) return false;
	      if (!IOSupport.read_cardinal(brin,out x_off)) return false;
	      staff_area_.set_x_off(x_off);
      
	      int x_org;
	      if (!IOSupport.read_token(brin,"x_org")) return false;
	      if (!IOSupport.read_token(brin,"=")) return false;
	      if (!IOSupport.read_cardinal(brin,out x_org)) return false;
	      staff_area_.set_x_org(x_org);
      
	      if (!IOSupport.read_token(brin,",")) return false;
      
	      int y_org;
	      if (!IOSupport.read_token(brin,"y_org")) return false;
	      if (!IOSupport.read_token(brin,"=")) return false;
	      if (!IOSupport.read_cardinal(brin,out y_org)) return false;
	      y_org += y_offset;
	      staff_area_.set_y_org(y_org);
      
	      int x_dim;
	      if (!IOSupport.read_token(brin,"x_dim")) return false;
	      if (!IOSupport.read_token(brin,"=")) return false;
	      if (!IOSupport.read_cardinal(brin,out x_dim)) return false;
      
	      if (!IOSupport.read_token(brin,",")) return false;
      
	      int y_dim;
	      if (!IOSupport.read_token(brin,"y_dim")) return false;
	      if (!IOSupport.read_token(brin,"=")) return false;
	      if (!IOSupport.read_cardinal(brin,out y_dim)) return false;
      
	      staff_area_.resize(x_dim,y_dim);
	    }
	  else
	    {
	      if (!staff_area_.read(brin)) return false;
	    }

	  return true;
	}


      protected virtual bool read_staff_segs(BinaryReader brin)
	{
	  int dummy_global_x_start, dummy_global_x_end;

	  if (!IOSupport.read_token(brin,"left_staff_seg")) return false;
	  if (!IOSupport.read_token(brin,":")) return false;

	  // skip left staff seg entry
	  if (!IOSupport.read_token(brin,"global_x_start")) return false;
	  if (!IOSupport.read_token(brin,"=")) return false;
	  if (!IOSupport.read_cardinal(brin,out dummy_global_x_start)) return false;

	  if (!IOSupport.read_token(brin,",")) return false;

	  if (!IOSupport.read_token(brin,"global_x_end")) return false;
	  if (!IOSupport.read_token(brin,"=")) return false;
	  if (!IOSupport.read_cardinal(brin,out dummy_global_x_end)) return false;

	  if (!IOSupport.read_token(brin,"right_staff_seg")) return false;
	  if (!IOSupport.read_token(brin,":")) return false;

	  // skip right staff seg entry
	  if (!IOSupport.read_token(brin,"global_x_start")) return false;
	  if (!IOSupport.read_token(brin,"=")) return false;
	  if (!IOSupport.read_cardinal(brin,out dummy_global_x_start)) return false;

	  if (!IOSupport.read_token(brin,",")) return false;

	  if (!IOSupport.read_token(brin,"global_x_end")) return false;
	  if (!IOSupport.read_token(brin,"=")) return false;
	  if (!IOSupport.read_cardinal(brin,out dummy_global_x_end)) return false;

	  return true;
	}


      protected virtual bool read_staff_lines(BinaryReader brin, int y_offset)
	{
	  int no_staff_lines;
	  if (!IOSupport.read_token(brin,"no_staff_lines")) return false;
	  if (!IOSupport.read_token(brin,"=")) return false;
	  if (!IOSupport.read_cardinal(brin,out no_staff_lines)) return false;
	  no_staff_lines_ = no_staff_lines;
  
	  if (!staff_lines_.read_staff_lines(brin,y_offset)) return false;

	  return true;
	}


      public abstract bool read_staff(BinaryReader brin, int y_offset);

      public bool read_staff(BinaryReader brin)
	{
	  return read_staff(brin,0);
	}


    }
}









