
using System;
using System.IO;

namespace Espresso.OMR
{

	abstract class BaseStaffSystem
	{	
		protected GObject gobject_;
		protected StaffList staffs_;
		protected GObjectList staff_system_objects_;
		protected StaffInfo staff_info_;


		public BaseStaffSystem()
		{
			staffs_ = new StaffList();
			GObjectList staff_system_objects_ = new GObjectList();
			staff_info_ = new StaffInfo();

			gobject_ = new GObject(); // gobject_ initialised to a void gobject
		}

		public BaseStaffSystem(GObject gobject)
		{
			gobject_ = gobject;

			staffs_ = new StaffList();
			staff_system_objects_ = new GObjectList();
			staff_info_ = new StaffInfo();
		}


		public void copy_staff_info(StaffInfo staff_info)
		{
			staff_info_ = staff_info.DeepCopy();
		}

		public StaffInfo staff_info() { return staff_info_; }
		public GObject gobject() { return gobject_; }

		public StaffList staffs() { return staffs_; }
		public GObjectList staff_system_objects() { return staff_system_objects_; }



		/*---------------------------------------------------------------------------*/

		protected virtual void write_gobject(BinaryWriter bwout, int y_offset)
		  {
		    IOSupport.write_line(bwout,"staff_system_data :");
		    IOSupport.write_line(bwout,"x_off = " + gobject_.x_off().ToString());
		    IOSupport.write(bwout,"x_org = " + gobject_.x_org().ToString() + " , ");

		    int y_org_rel = gobject_.y_org() - y_offset;
		    IOSupport.write_line(bwout,"y_org = " + y_org_rel.ToString());
		    IOSupport.write(bwout,"x_dim = " + gobject_.x_dim() + " , ");
		    IOSupport.write_line(bwout,"y_dim = " + gobject_.y_dim().ToString());
		  }

		protected virtual void write_complete_gobject(BinaryWriter bwout, 
							     int y_offset)
		  {
		    IOSupport.write_line(bwout,"complete_staff_system_data :");

		    SubBitmap sub_bitmap = (SubBitmap)gobject_;
		    sub_bitmap.add_to_y_org(-y_offset);
		    sub_bitmap.write(bwout);
		    sub_bitmap.add_to_y_org(+y_offset);
		  }

		protected virtual void write_staffs(BinaryWriter bwout, int y_offset)
		  {
		    int no_staffs = staffs_.length();
		    IOSupport.write_line(bwout,"no_staffs = " + no_staffs.ToString());

		    for (StaffNode staff=staffs_.head(); staff!=null; staff=staffs_.next())
		      staff.val().write_staff(bwout,y_offset);
		  }

		protected virtual void write_complete_staffs(BinaryWriter bwout,
							    int y_offset)
		  {
		    int no_staffs = staffs_.length();
		    IOSupport.write_line(bwout,"no_staffs = " + no_staffs.ToString());

		    for (StaffNode staff=staffs_.head(); staff!=null; staff=staffs_.next())
		      staff.val().write_complete_staff(bwout,y_offset);
		  }

		protected virtual void write_staff_system_objects(BinaryWriter bwout,
								 int y_offset)
		  {
		    GObjectList ss_obs = staff_system_objects_; // shorthand

		    int no_ss_obs = staff_system_objects_.length();
		    IOSupport.write_line(bwout,"no_staff_system_objects = " + no_ss_obs.ToString());

		    for (GObjectNode ss_ob=ss_obs.head(); ss_ob!=null; ss_ob=ss_obs.next())
		      {
			SubBitmap sub_bitmap = (SubBitmap)(ss_ob.val());
			sub_bitmap.add_to_y_org(-y_offset);
			sub_bitmap.write(bwout);
			sub_bitmap.add_to_y_org(+y_offset);
		      }
		  }

		protected virtual void write_staff_system_objects(BinaryWriter bwout,
								 GObjectList ss_obs,
								 int y_offset)
		  {
		    int no_ss_obs = ss_obs.length();
		    IOSupport.write_line(bwout,"no_staff_system_objects = " + no_ss_obs.ToString());

		    for (GObjectNode ss_ob=ss_obs.head(); ss_ob!=null; ss_ob=ss_obs.next())
		      {
			SubBitmap sub_bitmap = (SubBitmap)(ss_ob.val());
			sub_bitmap.add_to_y_org(-y_offset);
			sub_bitmap.write(bwout);
			sub_bitmap.add_to_y_org(+y_offset);
		      }
		  }


		public void write_staff_system(BinaryWriter bwout, 
							 GObjectList ss_obs,
							 int y_offset)
		  {
		    // link this up with << staff_system (extra_stream.cc) better // ****
		    write_gobject(bwout,y_offset);
		    write_staffs(bwout,y_offset);
		    write_staff_system_objects(bwout,ss_obs,y_offset);
		    staff_info_.write(bwout);
		  }


		public void write_staff_system(BinaryWriter bwout,
							 int y_offset)
		  {
		    // NB: due to virtual call on staff_system_objects, this is not the
		    //     same as doing an inline call to above function
		    write_gobject(bwout,y_offset);
		    write_staffs(bwout,y_offset);
		    write_staff_system_objects(bwout,y_offset);
		    staff_info_.write(bwout);
		  }

		public void write_complete_staff_system(BinaryWriter bwout, 
								  GObjectList ss_obs,
								  int y_offset)
		  {
		    // link this up with << staff_system (extra_stream.cc) better // ****
		    write_complete_gobject(bwout,y_offset);
		    write_complete_staffs(bwout,y_offset);
		    write_staff_system_objects(bwout,ss_obs,y_offset);
		    staff_info_.write(bwout);
		  }

		public void write_complete_staff_system(BinaryWriter bwout,
								  int y_offset)
		  {
		    // NB: due to virtual call on staff_system_objects, this is not the
		    //     same as doing an inline call to above function
		    write_complete_gobject(bwout,y_offset);
		    write_complete_staffs(bwout,y_offset);
		    write_staff_system_objects(bwout,y_offset);
		    staff_info_.write(bwout);
		  }

		protected virtual bool read_gobject(BinaryReader brin, int y_offset)
		  {
		    String token;
		    if (!IOSupport.read_string(brin,out token)) return false;

		    bool is_complete; // safe
		    if (token=="complete_staff_system_data")
		      {
			is_complete = true;
		      }
		    else if (token=="staff_system_data")
		      {
			is_complete = false;
		      }
		    else
		      {
			Console.Error.WriteLine("Error. Unexpected token: " + token);
			return false;
		      }

		    if (!IOSupport.read_token(brin,":")) return false;

		    if (!is_complete)
		      {
			int x_off;
			if (!IOSupport.read_token(brin,"x_off")) return false;
			if (!IOSupport.read_token(brin,"=")) return false;
			if (!IOSupport.read_cardinal(brin,out x_off)) return false;
			gobject_.set_x_off(x_off);
      
			int x_org;
			if (!IOSupport.read_token(brin,"x_org")) return false;
			if (!IOSupport.read_token(brin,"=")) return false;
			if (!IOSupport.read_cardinal(brin,out x_org)) return false;
			gobject_.set_x_org(x_org);
      
			if (!IOSupport.read_token(brin,",")) return false;
      
			int y_org;
			if (!IOSupport.read_token(brin,"y_org")) return false;
			if (!IOSupport.read_token(brin,"=")) return false;
			if (!IOSupport.read_cardinal(brin,out y_org)) return false;
			y_org += y_offset;
			gobject_.set_y_org(y_org);
      
			int x_dim;
			if (!IOSupport.read_token(brin,"x_dim")) return false;
			if (!IOSupport.read_token(brin,"=")) return false;
			if (!IOSupport.read_cardinal(brin,out x_dim)) return false;
      
			if (!IOSupport.read_token(brin,",")) return false;
      
			int y_dim;
			if (!IOSupport.read_token(brin,"y_dim")) return false;
			if (!IOSupport.read_token(brin,"=")) return false;
			if (!IOSupport.read_cardinal(brin,out y_dim)) return false;
      
			gobject_.resize(x_dim,y_dim);
		      }
		    else
		      {
			if (!gobject_.read(brin)) return false;
		      }

		    return true;
		  }

		protected virtual bool read_staffs(BinaryReader brin, int y_offset)
		  {
		    int no_staffs;
		    if (!IOSupport.read_token(brin,"no_staffs")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out no_staffs)) return false;

		    for (int i=0; i<no_staffs; i++)
		      {
			Staff staff = new Staff();
			if (!staff.read_staff(brin,y_offset)) return false;
			staffs_.append(staff); // responisbilty passed over to list
		      }

		    return true;
		  }

		protected virtual bool read_staff_system_objects(BinaryReader brin, 
								int y_offset)
		  {
		    int no_ss_obs;
		    if (!IOSupport.read_token(brin,"no_staff_system_objects")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out no_ss_obs)) return false;

		    GObjectList ss_obs = staff_system_objects_; // shorthand
		    for (int i=0; i<no_ss_obs; i++)
		      {
			GObject ss_ob = new GObject();
			if (!ss_ob.read(brin)) return false;
			ss_ob.add_to_y_org(y_offset);
			ss_ob.resize_projects(ss_ob.x_dim(),ss_ob.y_dim());
			ss_obs.append(ss_ob); // responisbilty passed over to list
		      }

		    return true;
		  }

		  abstract public bool read_staff_system(BinaryReader brin, int y_offset);


		public bool read_staff_system(BinaryReader brin)
		{
			return read_staff_system(brin,0);
		}





	}
}




