
using System;

namespace Espresso.OMR
{
	class BitmapOps
	{
		private Bitmap scratch_;
		private Bitmap isolated_;
		private PointStack stack_;

		public BopDeform Deform;
		public BopFlood  Flood;
		public BopFollow Follow;
		public BopRotate Rotate;
		public BopCount  Count;

		public BitmapOps(int x_dim, int y_dim)
		{
			scratch_  = new Bitmap(x_dim,y_dim);
			isolated_ = new Bitmap(x_dim,y_dim);

			stack_ = new PointStack(x_dim);

			scratch_.clear_all();
			isolated_.clear_all();

			Deform = new BopDeform();
			Flood  = new BopFlood(x_dim,y_dim,scratch_,isolated_,stack_);
			Follow = new BopFollow(x_dim,y_dim,scratch_,isolated_);
			Rotate = new BopRotate();
			Count  = new BopCount();
		}

		public void resize(int x_dim, int y_dim)
		{
			scratch_.resize(x_dim,y_dim);
			scratch_.clear_all();
	
			isolated_.resize(x_dim,y_dim);
			isolated_.clear_all();
	
			stack_.reset(x_dim);
		}


		public void exchange_buffer(Bitmap bitmap)
		{
			scratch_.exchange(bitmap);
		}


 
		/*---------------------*/


		private bool object_exists_above(SubBitmap src_bitmap,
			int src_x,
			int sl_top, int sl_bot,
			BopFollow.ModeEnum mode)
		{
			bool do_remove = false;
	
			int l_circle_thresh = 170*(sl_bot-sl_top+1)/100;

			switch (mode)
			{
				case BopFollow.ModeEnum.vertical_requirement:
					do_remove = Follow.check_clarke_requirement_above(src_bitmap,src_x,sl_top);
					break;

				case BopFollow.ModeEnum.circle_requirement:
					bool do_remove_top=false, do_remove_bot=false;
					Follow.check_circle_requirement(src_bitmap,src_x,sl_top,sl_bot,l_circle_thresh,
						ref do_remove_top,ref do_remove_bot);
					do_remove = do_remove_top;
					break;
			}

			return !do_remove;
		}


		private bool object_exists_below(SubBitmap src_bitmap,
			int src_x,
			int sl_top, int sl_bot,
			BopFollow.ModeEnum mode)
		{
			bool do_remove = false;

			int l_circle_thresh = 170*(sl_bot-sl_top+1)/100;

			switch (mode)
			{
				case BopFollow.ModeEnum.vertical_requirement:
					do_remove = Follow.check_clarke_requirement_below(src_bitmap,src_x,sl_bot);
					break;

				case BopFollow.ModeEnum.circle_requirement:
					bool do_remove_top = false, do_remove_bot = false; 
					Follow.check_circle_requirement(src_bitmap,src_x,sl_top,sl_bot,l_circle_thresh,
						ref do_remove_top,ref do_remove_bot);
					do_remove = do_remove_bot;
					break;
			}

			return !do_remove;
		}





		public GObject cross_staff_lines_flood_fill(SubBitmap src_bitmap, 
			int x_seed, int y_seed,
			BopFollow.ModeEnum mode)
		{
			int src_x_org = src_bitmap.x_org();
			int src_y_org = src_bitmap.y_org();

			Bitmap isl_bitmap = isolated_;
			Bitmap is_staff_line = scratch_;

			int src_x_dim = src_bitmap.x_dim();
			int src_y_dim = src_bitmap.y_dim();

			int x_left=x_seed;
			int x_right=x_seed;
			int y_top=y_seed;
			int y_bot=y_seed;

			stack_.push(new Point(x_seed,y_seed));
			isl_bitmap.set_pixel(x_seed,y_seed);
			src_bitmap.clear_pixel(x_seed,y_seed);

			while (!stack_.is_empty())
			{
				Point pt = stack_.pop();
				int x = pt.x();
				int y = pt.y();
				int glob_x = src_x_org + x;
				int glob_y = src_y_org + y;

				if (x>=1)
				{
					if ((src_bitmap[y,x-1]==Bitmap.PixelEnum.set) && 
						(is_staff_line[glob_y,glob_x-1]!=Bitmap.PixelEnum.set))
					{

						isl_bitmap.set_pixel(x-1,y);
						src_bitmap.clear_pixel(x-1,y);
						stack_.push(new Point(x-1,y));
						if (x-1<x_left) x_left=x-1;
					}
				}

				if (y>=1)
				{
					if (is_staff_line[glob_y-1,glob_x]==Bitmap.PixelEnum.set)
					{
						// cross over staff line;
						int sl_bot = y-1;
						do
						{
							glob_y--;
							y--;

							if (y<0) break;

						}
						while (is_staff_line[glob_y,glob_x]==Bitmap.PixelEnum.set);

						int sl_top = y+1;

						if ((y>=0) && 
							(object_exists_above(src_bitmap,x,sl_top,sl_bot,mode)))
						{
							for (int ry=sl_top; ry<=sl_bot; ry++)
							{
								isl_bitmap.set_pixel(x,ry);
								src_bitmap.clear_pixel(x,ry);
								is_staff_line.clear_pixel(glob_x,src_y_org+ry);
							}

							if (src_bitmap[y,x]==Bitmap.PixelEnum.set)
							{
								isl_bitmap.set_pixel(x,y);
								src_bitmap.clear_pixel(x,y);
								stack_.push(new Point(x,y));
								if (y<y_top) y_top=y;
							}
						}
						else
						{
							// copy staff line half way;
							int sl_mid = (sl_top + sl_bot)/2;
							for (int ry=sl_mid; ry<=sl_bot; ry++)
							{
								isl_bitmap.set_pixel(x,ry);
								src_bitmap.clear_pixel(x,ry);
								is_staff_line.clear_pixel(glob_x,src_y_org+ry);
							}
						}
					}
					else
					{
						if (src_bitmap[y-1,x]==Bitmap.PixelEnum.set)
						{
							isl_bitmap.set_pixel(x,y-1);
							src_bitmap.clear_pixel(x,y-1);
							stack_.push(new Point(x,y-1));
							if (y-1<y_top) y_top=y-1;
						}

						/* top left corner */
						if (x>=1)
						{
							if (src_bitmap[y-1,x-1]==Bitmap.PixelEnum.set)
							{
								isl_bitmap.set_pixel(x-1,y-1);
								src_bitmap.clear_pixel(x-1,y-1);
								stack_.push(new Point(x-1,y-1));
								if (x-1<x_left) x_left=x-1;
								if (y-1<y_top) y_top=y-1;
							}
						}

						/* top right corner */
						if (x<src_x_dim-1)
						{
							if (src_bitmap[y-1,x+1]==Bitmap.PixelEnum.set)
							{
								isl_bitmap.set_pixel(x+1,y-1);
								src_bitmap.clear_pixel(x+1,y-1);
								stack_.push(new Point(x+1,y-1));
								if (x+1>x_right) x_right=x+1;
								if (y-1<y_top) y_top=y-1;
							}
						}	      
					}
				}

				if (x<src_x_dim-1)
				{
					if ((src_bitmap[y,x+1]==Bitmap.PixelEnum.set) &&
						(is_staff_line[glob_y,glob_x+1]!=Bitmap.PixelEnum.set))
					{
						isl_bitmap.set_pixel(x+1,y);
						src_bitmap.clear_pixel(x+1,y);
						stack_.push(new Point(x+1,y));
						if (x+1>x_right) x_right=x+1;
					}
				}

				if (y<src_y_dim-1)
				{
					if (is_staff_line[glob_y+1,glob_x]==Bitmap.PixelEnum.set)
					{
						// cross over staff line;
						int sl_top = y+1;
						do
						{
							glob_y++;
							y++;

							if (y>=src_y_dim) break;
						}
						while (is_staff_line[glob_y,glob_x]==Bitmap.PixelEnum.set);

						int sl_bot = y-1;

						if ((y<src_y_dim) && 
							(object_exists_below(src_bitmap,x,sl_top,sl_bot,mode)))
						{
							for (int ry=sl_top; ry<=sl_bot; ry++)
							{
								isl_bitmap.set_pixel(x,ry);
								src_bitmap.clear_pixel(x,ry);
								is_staff_line.clear_pixel(glob_x,src_y_org+ry);
							}

							if (src_bitmap[y,x]==Bitmap.PixelEnum.set)
							{
								isl_bitmap.set_pixel(x,y);
								src_bitmap.clear_pixel(x,y);
								stack_.push(new Point(x,y));
								if (y>y_bot) y_bot=y;
							}
						}
						else
						{
							// copy staff line half way;
							int sl_mid = (sl_top + sl_bot)/2;
							for (int ry=sl_top; ry<=sl_mid; ry++)
							{
								isl_bitmap.set_pixel(x,ry);
								src_bitmap.clear_pixel(x,ry);
								is_staff_line.clear_pixel(glob_x,src_y_org+ry);
							}
						}
					}
					else
					{
						if (src_bitmap[y+1,x]==Bitmap.PixelEnum.set)
						{
							isl_bitmap.set_pixel(x,y+1);
							src_bitmap.clear_pixel(x,y+1);
							stack_.push(new Point(x,y+1));
							if (y+1>y_bot) y_bot=y+1;
						}

						/* bottom right */
						if (x<src_x_dim-1)
						{
							if (src_bitmap[y+1,x+1]==Bitmap.PixelEnum.set)
							{
								isl_bitmap.set_pixel(x+1,y+1);
								src_bitmap.clear_pixel(x+1,y+1);
								stack_.push(new Point(x+1,y+1));
								if (x+1>x_right) x_right=x+1;
								if (y+1>y_bot) y_bot=y+1;
							}
						}

						/* bottom left */
						if (x>=1)
						{
							if (src_bitmap[y+1,x-1]==Bitmap.PixelEnum.set)
							{
								isl_bitmap.set_pixel(x-1,y+1);
								src_bitmap.clear_pixel(x-1,y+1);
								stack_.push(new Point(x-1,y+1));
								if (x-1<x_left) x_left=x-1;
								if (y+1>y_bot) y_bot=y+1;
							}
						}
					}
				}
			}
  
			int x_dim = x_right-x_left+1;
			int y_dim = y_bot-y_top+1;

			GObject dst_bitmap 
				= new GObject(isl_bitmap,x_left,y_top,x_dim,y_dim,SubBitmap.Operation.remove);

			dst_bitmap.add_to_x_org(src_x_org);
			dst_bitmap.add_to_y_org(src_y_org);

			return dst_bitmap;
		}

		public Bitmap scratch() { return scratch_; }

		public Bitmap isolated() { return isolated_; }
    
	}
}
      
