
using System;

namespace Espresso.OMR
{
	class BopFlood
	{

		protected PointStack stack_;
		protected Bitmap scratch_;
		protected Bitmap isolated_;

		public BopFlood(int x_dim,int y_dim, Bitmap scratch, Bitmap isolated, PointStack stack)
		{
			scratch_ = scratch;
			isolated_ = isolated;

			stack_ = stack;
		}


		public void flood_fill_basic_into(Bitmap src_bitmap, 
			int x_seed, int y_seed,
			SubBitmap dst_sub_bitmap)
		{
			// Resizes a SubBitmap

			// whole function could be replaced by src_x_org=0, src_y_org=0 call

			Bitmap dst_bitmap = scratch_;
			Bitmap isl_bitmap = isolated_;

			int dst_x_dim = dst_bitmap.x_dim();
			int dst_y_dim = dst_bitmap.y_dim();

			int src_x_dim = src_bitmap.x_dim();
			int src_y_dim = src_bitmap.y_dim();

			if ((src_x_dim>dst_x_dim) && (src_y_dim>dst_y_dim))
			{
				scratch_.resize(src_x_dim,src_y_dim);  scratch_.clear_all();
				isolated_.resize(src_x_dim,src_y_dim); isolated_.clear_all();
				dst_x_dim = dst_bitmap.x_dim();
				dst_y_dim = dst_bitmap.y_dim();
			}
			else if (src_x_dim>dst_x_dim)
			{
				scratch_.resize(src_x_dim,dst_y_dim);  scratch_.clear_all();
				isolated_.resize(src_x_dim,dst_y_dim); isolated_.clear_all();
				dst_x_dim = dst_bitmap.x_dim();
			}
			else if (src_y_dim>dst_y_dim)
			{
				scratch_.resize(dst_x_dim,src_y_dim);  scratch_.clear_all();
				isolated_.resize(dst_x_dim,src_y_dim); isolated_.clear_all();
				dst_x_dim = dst_bitmap.x_dim();
				dst_y_dim = dst_bitmap.y_dim();
			}

			int x_left=x_seed;
			int x_right=x_seed;
			int y_top=y_seed;
			int y_bot=y_seed;

			stack_.push(new Point(x_seed,y_seed));
			dst_bitmap.set_pixel(x_seed,y_seed);
			isl_bitmap.set_pixel(x_seed,y_seed);
			src_bitmap.clear_pixel(x_seed,y_seed);

			while (!stack_.is_empty())
			{
				Point pt = stack_.pop();
				int x = pt.x();
				int y = pt.y();

				if (x>=1)
				{
					if (src_bitmap[y,x-1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y);
						isl_bitmap.set_pixel(x-1,y);
						src_bitmap.clear_pixel(x-1,y);
						stack_.push(new Point(x-1,y));
						if (x-1<x_left) x_left=x-1;
					}
				}

				if (y>=1)
				{
					if (src_bitmap[y-1,x]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x,y-1);
						isl_bitmap.set_pixel(x,y-1);
						src_bitmap.clear_pixel(x,y-1);
						stack_.push(new Point(x,y-1));
						if (y-1<y_top) y_top=y-1;
					}
				}

				if (x<src_x_dim-1)
				{
					if (src_bitmap[y,x+1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y);
						isl_bitmap.set_pixel(x+1,y);
						src_bitmap.clear_pixel(x+1,y);
						stack_.push(new Point(x+1,y));
						if (x+1>x_right) x_right=x+1;
					}
				}

				if (y<src_y_dim-1)
				{
					if (src_bitmap[y+1,x]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x,y+1);
						isl_bitmap.set_pixel(x,y+1);
						src_bitmap.clear_pixel(x,y+1);
						stack_.push(new Point(x,y+1));
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				/* and the corners */
				if ((x>=1) && (y>=1))
				{
					if (src_bitmap[y-1,x-1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y-1);
						isl_bitmap.set_pixel(x-1,y-1);
						src_bitmap.clear_pixel(x-1,y-1);
						stack_.push(new Point(x-1,y-1));
						if (x-1<x_left) x_left=x-1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<src_x_dim-1) && (y>=1))
				{
					if (src_bitmap[y-1,x+1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y-1);
						isl_bitmap.set_pixel(x+1,y-1);
						src_bitmap.clear_pixel(x+1,y-1);
						stack_.push(new Point(x+1,y-1));
						if (x+1>x_right) x_right=x+1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<src_x_dim-1) && (y<src_y_dim-1))
				{
					if (src_bitmap[y+1,x+1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y+1);
						isl_bitmap.set_pixel(x+1,y+1);
						src_bitmap.clear_pixel(x+1,y+1);
						stack_.push(new Point(x+1,y+1));
						if (x+1>x_right) x_right=x+1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				if ((x>=1) && (y<src_y_dim-1))
				{
					if (src_bitmap[y+1,x-1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y+1);
						isl_bitmap.set_pixel(x-1,y+1);
						src_bitmap.clear_pixel(x-1,y+1);
						stack_.push(new Point(x-1,y+1));
						if (x-1<x_left) x_left=x-1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}
			}
  
			int x_dim = x_right-x_left+1;
			int y_dim = y_bot-y_top+1;

			dst_sub_bitmap.set_x_off(x_left%8);
			dst_sub_bitmap.resize(x_dim,y_dim);

			dst_sub_bitmap.extract(isl_bitmap,x_left,y_top,x_dim,y_dim);
			isl_bitmap.clear_area(x_left,y_top,x_dim,y_dim);

			dst_sub_bitmap.set_x_org(x_left);
			dst_sub_bitmap.set_y_org(y_top);
		}

		public GObject flood_fill_basic(Bitmap src_bitmap, 
			int x_seed, int y_seed)
		{
			// Creates a SubBitmap

			// whole function could be replaced by src_x_org=0, src_y_org=0 call

			GObject flood_gobject = new GObject();
			flood_fill_basic_into(src_bitmap,x_seed,y_seed,flood_gobject);

			int fgo_x_dim = flood_gobject.x_dim();
			int fgo_y_dim = flood_gobject.y_dim();
			flood_gobject.resize_projects(fgo_x_dim,fgo_y_dim);

			return flood_gobject;
		}

		public GObject flood_fill_basic(Bitmap src_bitmap, 
			int x_seed, int y_seed,
			int src_x_org, int src_y_org)
		{
			// Creates a SubBitmap

			Bitmap dst_bitmap = scratch_;
			Bitmap isl_bitmap = isolated_;

			int dst_x_dim = dst_bitmap.x_dim();
			int dst_y_dim = dst_bitmap.y_dim();
			int src_x_dim = src_bitmap.x_dim();
			int src_y_dim = src_bitmap.y_dim();

			if ((src_x_org+src_x_dim>dst_x_dim) && (src_y_org+src_y_dim>dst_y_dim))
			{
				scratch_.resize(src_x_org+src_x_dim,src_y_org+src_y_dim);  
				scratch_.clear_all();
				isolated_.resize(src_x_org+src_x_dim,src_y_org+src_y_dim); 
				isolated_.clear_all();
				dst_x_dim = dst_bitmap.x_dim();
				dst_y_dim = dst_bitmap.y_dim();
			}
			else if (src_x_org+src_x_dim>dst_x_dim)
			{
				scratch_.resize(src_x_org+src_x_dim,dst_y_dim); scratch_.clear_all();
				isolated_.resize(src_x_org+src_x_dim,dst_y_dim); isolated_.clear_all();
				dst_x_dim = dst_bitmap.x_dim();
			}
			else if (src_y_org+src_y_dim>dst_y_dim)
			{
				scratch_.resize(dst_x_dim,src_y_org+src_y_dim); scratch_.clear_all();
				isolated_.resize(dst_x_dim,src_y_org+src_y_dim); isolated_.clear_all();
				dst_y_dim = dst_bitmap.y_dim();
			}

			int x_left=x_seed;
			int x_right=x_seed;
			int y_top=y_seed;
			int y_bot=y_seed;

			stack_.push(new Point(x_seed,y_seed));
			dst_bitmap.set_pixel(src_x_org+x_seed,src_y_org+y_seed);
			isl_bitmap.set_pixel(src_x_org+x_seed,src_y_org+y_seed);
			src_bitmap.clear_pixel(x_seed,y_seed);

			while (!stack_.is_empty())
			{
				Point pt = stack_.pop();
				int x = pt.x();
				int y = pt.y();

				if (x>=1)
				{
					if (src_bitmap[y,x-1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(src_x_org+x-1,src_y_org+y);
						isl_bitmap.set_pixel(src_x_org+x-1,src_y_org+y);
						src_bitmap.clear_pixel(x-1,y);
						stack_.push(new Point(x-1,y));
						if (x-1<x_left) x_left=x-1;
					}
				}

				if (y>=1)
				{
					if (src_bitmap[y-1,x]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(src_x_org+x,src_y_org+y-1);
						isl_bitmap.set_pixel(src_x_org+x,src_y_org+y-1);
						src_bitmap.clear_pixel(x,y-1);
						stack_.push(new Point(x,y-1));
						if (y-1<y_top) y_top=y-1;
					}
				}

				if (x<src_x_dim-1)
				{
					if (src_bitmap[y,x+1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(src_x_org+x+1,src_y_org+y);
						isl_bitmap.set_pixel(src_x_org+x+1,src_y_org+y);
						src_bitmap.clear_pixel(x+1,y);
						stack_.push(new Point(x+1,y));
						if (x+1>x_right) x_right=x+1;
					}
				}

				if (y<src_y_dim-1)
				{
					if (src_bitmap[y+1,x]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(src_x_org+x,src_y_org+y+1);
						isl_bitmap.set_pixel(src_x_org+x,src_y_org+y+1);
						src_bitmap.clear_pixel(x,y+1);
						stack_.push(new Point(x,y+1));
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				/* and the corners */
				if ((x>=1) && (y>=1))
				{
					if (src_bitmap[y-1,x-1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(src_x_org+x-1,src_y_org+y-1);
						isl_bitmap.set_pixel(src_x_org+x-1,src_y_org+y-1);
						src_bitmap.clear_pixel(x-1,y-1);
						stack_.push(new Point(x-1,y-1));
						if (x-1<x_left) x_left=x-1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<src_x_dim-1) && (y>=1))
				{
					if (src_bitmap[y-1,x+1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(src_x_org+x+1,src_y_org+y-1);
						isl_bitmap.set_pixel(src_x_org+x+1,src_y_org+y-1);
						src_bitmap.clear_pixel(x+1,y-1);
						stack_.push(new Point(x+1,y-1));
						if (x+1>x_right) x_right=x+1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<src_x_dim-1) && (y<src_y_dim-1))
				{
					if (src_bitmap[y+1,x+1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(src_x_org+x+1,src_y_org+y+1);
						isl_bitmap.set_pixel(src_x_org+x+1,src_y_org+y+1);
						src_bitmap.clear_pixel(x+1,y+1);
						stack_.push(new Point(x+1,y+1));
						if (x+1>x_right) x_right=x+1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				if ((x>=1) && (y<src_y_dim-1))
				{
					if (src_bitmap[y+1,x-1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(src_x_org+x-1,src_y_org+y+1);
						isl_bitmap.set_pixel(src_x_org+x-1,src_y_org+y+1);
						src_bitmap.clear_pixel(x-1,y+1);
						stack_.push(new Point(x-1,y+1));
						if (x-1<x_left) x_left=x-1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}
			}
  
			int x_dim = x_right-x_left+1;
			int y_dim = y_bot-y_top+1;

			GObject flood_gobject
				= new GObject(isl_bitmap,src_x_org+x_left,src_y_org+y_top,
				x_dim,y_dim,GObject.Operation.remove);

			return flood_gobject;
		}

		public GObject flood_fill_basic(SubBitmap src_bitmap, 
			int x_seed, int y_seed)
		{
			int src_x_org = src_bitmap.x_org();
			int src_y_org = src_bitmap.y_org();

			return flood_fill_basic(src_bitmap,x_seed,y_seed,src_x_org,src_y_org);
		}


		public GObject flood_fill_2bit(Bitmap src_bitmap, 
			int x_seed, int y_seed)
		{
			// Creates a SubBitmap

			Bitmap dst_bitmap = scratch_;
			Bitmap isl_bitmap = isolated_;

			int dst_x_dim = dst_bitmap.x_dim();
			int dst_y_dim = dst_bitmap.y_dim();

			int src_x_dim = src_bitmap.x_dim();
			int src_y_dim = src_bitmap.y_dim();

			if ((src_x_dim>dst_x_dim) && (src_y_dim>dst_y_dim))
			{
				scratch_.resize(src_x_dim,src_y_dim);  scratch_.clear_all();
				isolated_.resize(src_x_dim,src_y_dim); isolated_.clear_all();
				dst_x_dim = dst_bitmap.x_dim();
				dst_y_dim = dst_bitmap.y_dim();
			}
			else if (src_x_dim>dst_x_dim)
			{
				scratch_.resize(src_x_dim,dst_y_dim);  scratch_.clear_all();
				isolated_.resize(src_x_dim,dst_y_dim); isolated_.clear_all();
				dst_x_dim = dst_bitmap.x_dim();
			}
			else if (src_y_dim>dst_y_dim)
			{
				scratch_.resize(dst_x_dim,src_y_dim);  scratch_.clear_all();
				isolated_.resize(dst_x_dim,src_y_dim); isolated_.clear_all();
				dst_y_dim = dst_bitmap.y_dim();
			}

			int x_left=x_seed;
			int x_right=x_seed;
			int y_top=y_seed;
			int y_bot=y_seed;

			stack_.push(new Point(x_seed,y_seed));
			dst_bitmap.set_pixel(x_seed,y_seed);
			isl_bitmap.set_pixel(x_seed,y_seed);
			src_bitmap.clear_pixel(x_seed,y_seed);

			while (!stack_.is_empty())
			{
				Point pt = stack_.pop();
				int x = pt.x();
				int y = pt.y();

				if (x>=1)
				{
					Pixel pixel_l = src_bitmap[y,x-1];
					if (pixel_l==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y);
						isl_bitmap.set_pixel(x-1,y);
						pixel_l.clear_pixel();
						stack_.push(new Point(x-1,y));
						if (x-1<x_left) x_left=x-1;

						if (x>=2)
						{
							if (src_bitmap[y,x-2]==Pixel.PixelEnum.clear)
							{
								dst_bitmap.set_pixel(x-2,y);
								isl_bitmap.set_pixel(x-2,y);
								stack_.push(new Point(x-2,y));
								if (x-2<x_left) x_left=x-2;
							}
						}
					}
				}

				if (y>=1)
				{
					Pixel pixel_t = src_bitmap[y-1,x];
					if (pixel_t==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x,y-1);
						isl_bitmap.set_pixel(x,y-1);
						pixel_t.clear_pixel();
						stack_.push(new Point(x,y-1));
						if (y-1<y_top) y_top=y-1;
					}
				}

				if (x<src_x_dim-1)
				{
					Pixel pixel_r = src_bitmap[y,x+1];
					if (pixel_r==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y);
						isl_bitmap.set_pixel(x+1,y);
						pixel_r.clear_pixel();
						stack_.push(new Point(x+1,y));
						if (x+1>x_right) x_right=x+1; // optimise


						if (x<src_x_dim-2)
						{
							if (src_bitmap[y,x+2]==Pixel.PixelEnum.clear)
							{
								dst_bitmap.set_pixel(x+2,y);
								isl_bitmap.set_pixel(x+2,y);
								stack_.push(new Point(x+2,y));
								if (x+2>x_right) x_right=x+2;
							}
						}
					}
				}

				if (y<src_y_dim-1)
				{
					Pixel pixel_b = src_bitmap[y+1,x];
					if (pixel_b==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x,y+1);
						isl_bitmap.set_pixel(x,y+1);
						pixel_b.clear_pixel();
						stack_.push(new Point(x,y+1));
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				/* and the corners */

				if ((x>=1) && (y>=1))
				{
					Pixel pixel_tl = src_bitmap[y-1,x-1];
					if (pixel_tl==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y-1);
						isl_bitmap.set_pixel(x-1,y-1);
						pixel_tl.clear_pixel();
						stack_.push(new Point(x-1,y-1));
						if (x-1<x_left) x_left=x-1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<src_x_dim-1) && (y>=1))
				{
					Pixel pixel_tr = src_bitmap[y-1,x+1];
					if (pixel_tr==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y-1);
						isl_bitmap.set_pixel(x+1,y-1);
						pixel_tr.clear_pixel();
						stack_.push(new Point(x+1,y-1));
						if (x+1>x_right) x_right=x+1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<src_x_dim-1) && (y<src_y_dim-1))
				{
					Pixel pixel_br = src_bitmap[y+1,x+1];
					if (pixel_br==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y+1);
						isl_bitmap.set_pixel(x+1,y+1);
						pixel_br.clear_pixel();
						stack_.push(new Point(x+1,y+1));
						if (x+1>x_right) x_right=x+1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				if ((x>=1) && (y<src_y_dim-1))
				{
					Pixel pixel_bl = src_bitmap[y+1,x-1];
					if (pixel_bl==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y+1);
						isl_bitmap.set_pixel(x-1,y+1);
						pixel_bl.clear_pixel();
						stack_.push(new Point(x-1,y+1));
						if (x-1<x_left) x_left=x-1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}
			}
  
			int x_dim = x_right-x_left+1;
			int y_dim = y_bot-y_top+1;

			GObject flood_gobject
				= new GObject(isl_bitmap,x_left,y_top,x_dim,y_dim,GObject.Operation.remove);

			return flood_gobject;
		}




		public GObject flood_fill_nbit(Bitmap src_bitmap, 
			int x_seed, int y_seed,
			int n)
		{
			// Creates a SubBitmap

			Bitmap dst_bitmap = scratch_;
			Bitmap isl_bitmap = isolated_;

			int dst_x_dim = dst_bitmap.x_dim();
			int dst_y_dim = dst_bitmap.y_dim();

			int src_x_dim = src_bitmap.x_dim();
			int src_y_dim = src_bitmap.y_dim();

			if ((src_x_dim>dst_x_dim) && (src_y_dim>dst_y_dim))
			{
				scratch_.resize(src_x_dim,src_y_dim);  scratch_.clear_all();
				isolated_.resize(src_x_dim,src_y_dim); isolated_.clear_all();
				dst_x_dim = dst_bitmap.x_dim();
				dst_y_dim = dst_bitmap.y_dim();
			}
			else if (src_x_dim>dst_x_dim)
			{
				scratch_.resize(src_x_dim,dst_y_dim);  scratch_.clear_all();
				isolated_.resize(src_x_dim,dst_y_dim); isolated_.clear_all();
				dst_x_dim = dst_bitmap.x_dim();
			}
			else if (src_y_dim>dst_y_dim)
			{
				scratch_.resize(dst_x_dim,src_y_dim);  scratch_.clear_all();
				isolated_.resize(dst_x_dim,src_y_dim); isolated_.clear_all();
				dst_y_dim = dst_bitmap.y_dim();
			}

			int x_left=x_seed;
			int x_right=x_seed;
			int y_top=y_seed;
			int y_bot=y_seed;

			stack_.push(new Point(x_seed,y_seed));
			dst_bitmap.set_pixel(x_seed,y_seed);
			isl_bitmap.set_pixel(x_seed,y_seed);
			src_bitmap.clear_pixel(x_seed,y_seed);

			while (!stack_.is_empty())
			{
				int i;

				Point pt = stack_.pop();
				int x = pt.x();
				int y = pt.y();

				for (i=1; i<=n; i++)
				{
					if (x>=i)
					{
						int xmi = x - i;
						Pixel pixel_l = src_bitmap[y,xmi];
						if (pixel_l==Pixel.PixelEnum.set)
						{
							dst_bitmap.set_pixel(xmi,y);
							isl_bitmap.set_pixel(xmi,y);
							pixel_l.clear_pixel();
							stack_.push(new Point(xmi,y));
							if (xmi<x_left) x_left=xmi;
						}
					}
					else break;
				}

				if (y>=1)
				{
					Pixel pixel_t = src_bitmap[y-1,x];
					if (pixel_t==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x,y-1);
						isl_bitmap.set_pixel(x,y-1);
						pixel_t.clear_pixel();
						stack_.push(new Point(x,y-1));
						if (y-1<y_top) y_top=y-1;
					}
				}

				for (i=1; i<=n; i++)
				{
					if (x<src_x_dim-i)
					{
						int xpi = x + i;
						Pixel pixel_r = src_bitmap[y,xpi];
						if (pixel_r==Pixel.PixelEnum.set)
						{
							dst_bitmap.set_pixel(xpi,y);
							isl_bitmap.set_pixel(xpi,y);
							pixel_r.clear_pixel();
							stack_.push(new Point(xpi,y));
							if (xpi>x_right) x_right=xpi; // optimise
						}
					}
					else break;
				}

				if (y<src_y_dim-1)
				{
					Pixel pixel_b = src_bitmap[y+1,x];
					if (pixel_b==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x,y+1);
						isl_bitmap.set_pixel(x,y+1);
						pixel_b.clear_pixel();
						stack_.push(new Point(x,y+1));
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				/* and the corners */
				/* any hoizontal gaps in y-1, y+1 picked up by adjacent line spread */

				if ((x>=1) && (y>=1))
				{
					Pixel pixel_tl = src_bitmap[y-1,x-1];
					if (pixel_tl==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y-1);
						isl_bitmap.set_pixel(x-1,y-1);
						pixel_tl.clear_pixel();
						stack_.push(new Point(x-1,y-1));
						if (x-1<x_left) x_left=x-1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<src_x_dim-1) && (y>=1))
				{
					Pixel pixel_tr = src_bitmap[y-1,x+1];
					if (pixel_tr==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y-1);
						isl_bitmap.set_pixel(x+1,y-1);
						pixel_tr.clear_pixel();
						stack_.push(new Point(x+1,y-1));
						if (x+1>x_right) x_right=x+1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<src_x_dim-1) && (y<src_y_dim-1))
				{
					Pixel pixel_br = src_bitmap[y+1,x+1];
					if (pixel_br==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y+1);
						isl_bitmap.set_pixel(x+1,y+1);
						pixel_br.clear_pixel();
						stack_.push(new Point(x+1,y+1));
						if (x+1>x_right) x_right=x+1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				if ((x>=1) && (y<src_y_dim-1))
				{
					Pixel pixel_bl = src_bitmap[y+1,x-1];
					if (pixel_bl==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y+1);
						isl_bitmap.set_pixel(x-1,y+1);
						pixel_bl.clear_pixel();
						stack_.push(new Point(x-1,y+1));
						if (x-1<x_left) x_left=x-1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}
			}
  
			int x_dim = x_right-x_left+1;
			int y_dim = y_bot-y_top+1;

			GObject flood_gobject
				= new GObject(isl_bitmap,x_left,y_top,x_dim,y_dim,GObject.Operation.remove);

			return flood_gobject;
		}




		public GObject flood_fill_byte(ByteArray2d src_array, 
			int x_seed, int y_seed)
		{
			// Creates a SubBitmap

			ByteArray2d dst_array = scratch_.array();
			ByteArray2d isl_array = isolated_.array();

			int dax_dim = dst_array.x_dim();
			int day_dim = dst_array.y_dim();

			int sax_dim = src_array.x_dim();
			int say_dim = src_array.y_dim();
			if ((sax_dim>dax_dim) || (say_dim>day_dim))
			{
				int max_ax_dim = Math.Max(sax_dim,dax_dim);
				int max_ay_dim = Math.Max(say_dim,day_dim);

				dst_array.resize(max_ax_dim,max_ay_dim); scratch_.clear_all();
				isl_array.resize(max_ax_dim,max_ay_dim); isolated_.clear_all();
			}

			int x_left=x_seed;
			int x_right=x_seed;
			int y_top=y_seed;
			int y_bot=y_seed;

			stack_.push(new Point(x_seed,y_seed));
			dst_array[y_seed,x_seed] = src_array[y_seed,x_seed];
			isl_array[y_seed,x_seed] = dst_array[y_seed,x_seed];
			src_array[y_seed,x_seed] = 0;


			while (!stack_.is_empty())
			{
				Point pt = stack_.pop();
				int x = pt.x();
				int y = pt.y();

				if (x>=1)
				{
					byte byte_l = src_array[y,x-1];
					if (byte_l!=0)
					{
						dst_array[y,x-1] = byte_l;
						isl_array[y,x-1] = byte_l;
						src_array[y,x-1] = 0;
						stack_.push(new Point(x-1,y));
						if (x-1<x_left) x_left=x-1;

					}
				}

				if (y>=1)
				{
					byte byte_t = src_array[y-1,x];
					if (byte_t!=0)
					{
						dst_array[y-1,x] = byte_t;
						isl_array[y-1,x] = byte_t;
						src_array[y-1,x] = 0;
						stack_.push(new Point(x,y-1));
						if (y-1<y_top) y_top=y-1;
					}
				}

				if (x<sax_dim-1)
				{
					byte byte_r = src_array[y,x+1];
					if (byte_r!=0)
					{
						dst_array[y,x+1] = byte_r;
						isl_array[y,x+1] = byte_r;
						src_array[y,x+1] = 0;
						stack_.push(new Point(x+1,y));
						if (x+1>x_right) x_right=x+1; // optimise
	      
					}
				}

				if (y<say_dim-1)
				{
					byte byte_b = src_array[y+1,x];
					if (byte_b!=0)
					{
						dst_array[y+1,x] = byte_b;
						isl_array[y+1,x] = byte_b;
						src_array[y+1,x] = 0;
						stack_.push(new Point(x,y+1));
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				/* and the corners */

				if ((x>=1) && (y>=1))
				{
					byte byte_tl = src_array[y-1,x-1];
					if (byte_tl!=0)
					{
						dst_array[y-1,x-1] = byte_tl;
						isl_array[y-1,x-1] = byte_tl;
						src_array[y-1,x-1] = 0;
						stack_.push(new Point(x-1,y-1));
						if (x-1<x_left) x_left=x-1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<sax_dim-1) && (y>=1))
				{
					byte byte_tr = src_array[y-1,x+1];
					if (byte_tr!=0)
					{
						dst_array[y-1,x+1] = byte_tr;
						isl_array[y-1,x+1] = byte_tr;
						src_array[y-1,x+1] = 0;
						stack_.push(new Point(x+1,y-1));
						if (x+1>x_right) x_right=x+1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<sax_dim-1) && (y<say_dim-1))
				{
					byte byte_br = src_array[y+1,x+1];
					if (byte_br!=0)
					{
						dst_array[y+1,x+1] = byte_br;
						isl_array[y+1,x+1] = byte_br;
						src_array[y+1,x+1] = 0;
						stack_.push(new Point(x+1,y+1));
						if (x+1>x_right) x_right=x+1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				if ((x>=1) && (y<say_dim-1))
				{
					byte byte_bl = src_array[y+1,x-1];
					if (byte_bl!=0)
					{
						dst_array[y+1,x-1] = byte_bl;
						isl_array[y+1,x-1] = byte_bl;
						src_array[y+1,x-1] = 0;
						stack_.push(new Point(x-1,y+1));
						if (x-1<x_left) x_left=x-1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}
			}
  
			int x_dim = x_right-x_left+1;
			int y_dim = y_bot-y_top+1;

			int dst_x_dim = scratch_.x_dim();
			int bit_x_left = x_left<<3;
			int bit_x_dim = x_dim<<3;

			if (bit_x_left+bit_x_dim>dst_x_dim)
				bit_x_dim = dst_x_dim - bit_x_left;

			GObject flood_gobject
				= new GObject(isolated_,bit_x_left,y_top,bit_x_dim,y_dim,GObject.Operation.remove);

			return flood_gobject;
		}




		public GObject dual_flood_fill(Bitmap guide_bitmap, Bitmap use_bitmap,
			int x_seed, int y_seed)
		{
			// Creates a SubBitmap

			Bitmap dst_bitmap = scratch_;
			Bitmap isl_bitmap = isolated_;

			int dst_x_dim = dst_bitmap.x_dim();
			int dst_y_dim = dst_bitmap.y_dim();

			int guide_x_dim = guide_bitmap.x_dim();
			int guide_y_dim = guide_bitmap.y_dim();

			if ((guide_x_dim>dst_x_dim) && (guide_y_dim>dst_y_dim))
			{
				scratch_.resize(guide_x_dim,guide_y_dim);  scratch_.clear_all();
				isolated_.resize(guide_x_dim,guide_y_dim); isolated_.clear_all();
				dst_x_dim = dst_bitmap.x_dim();
				dst_y_dim = dst_bitmap.y_dim();
			}
			else if (guide_x_dim>dst_x_dim) 
			{
				scratch_.resize(guide_x_dim,dst_y_dim);  scratch_.clear_all();
				isolated_.resize(guide_x_dim,dst_y_dim); isolated_.clear_all();
				dst_x_dim = dst_bitmap.x_dim();
			}
			else if (guide_y_dim>dst_y_dim)
			{
				scratch_.resize(dst_x_dim,guide_y_dim);  scratch_.clear_all();
				isolated_.resize(dst_x_dim,guide_y_dim); isolated_.clear_all();
				dst_y_dim = dst_bitmap.y_dim();
			}

			int x_left=x_seed;
			int x_right=x_seed;
			int y_top=y_seed;
			int y_bot=y_seed;

			stack_.push(new Point(x_seed,y_seed));
			dst_bitmap.set_pixel(x_seed,y_seed);
			isl_bitmap.set_pixel(x_seed,y_seed);
			guide_bitmap.clear_pixel(x_seed,y_seed);

			while (!stack_.is_empty())
			{
				Point pt = stack_.pop();
				int x = pt.x();
				int y = pt.y();

				if (x>=1)
				{
					if (guide_bitmap[y,x-1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y);
						if (use_bitmap[y,x-1]==Bitmap.PixelEnum.set)
							isl_bitmap.set_pixel(x-1,y);
						guide_bitmap.clear_pixel(x-1,y);
						stack_.push(new Point(x-1,y));
						if (x-1<x_left) x_left=x-1;
					}
				}

				if (y>=1)
				{
					if (guide_bitmap[y-1,x]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x,y-1);
						if (use_bitmap[y-1,x]==Bitmap.PixelEnum.set)
							isl_bitmap.set_pixel(x,y-1);
						guide_bitmap.clear_pixel(x,y-1);
						stack_.push(new Point(x,y-1));
						if (y-1<y_top) y_top=y-1;
					}
				}

				if (x<guide_x_dim-1)
				{
					if (guide_bitmap[y,x+1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y);
						if (use_bitmap[y,x+1]==Bitmap.PixelEnum.set)
							isl_bitmap.set_pixel(x+1,y);
						guide_bitmap.clear_pixel(x+1,y);
						stack_.push(new Point(x+1,y));
						if (x+1>x_right) x_right=x+1;
					}
				}

				if (y<guide_y_dim-1)
				{
					if (guide_bitmap[y+1,x]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x,y+1);
						if (use_bitmap[y+1,x]==Bitmap.PixelEnum.set)
							isl_bitmap.set_pixel(x,y+1);
						guide_bitmap.clear_pixel(x,y+1);
						stack_.push(new Point(x,y+1));
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				/* and the corners */
				if ((x>=1) && (y>=1))
				{
					if (guide_bitmap[y-1,x-1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y-1);
						if (use_bitmap[y-1,x-1]==Bitmap.PixelEnum.set)
							isl_bitmap.set_pixel(x-1,y-1);
						guide_bitmap.clear_pixel(x-1,y-1);
						stack_.push(new Point(x-1,y-1));
						if (x-1<x_left) x_left=x-1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<guide_x_dim-1) && (y>=1))
				{
					if (guide_bitmap[y-1,x+1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y-1);
						if (use_bitmap[y-1,x+1]==Bitmap.PixelEnum.set)
							isl_bitmap.set_pixel(x+1,y-1);
						guide_bitmap.clear_pixel(x+1,y-1);
						stack_.push(new Point(x+1,y-1));
						if (x+1>x_right) x_right=x+1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<guide_x_dim-1) && (y<guide_y_dim-1))
				{
					if (guide_bitmap[y+1,x+1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y+1);
						if (use_bitmap[y+1,x+1]==Bitmap.PixelEnum.set)
							isl_bitmap.set_pixel(x+1,y+1);
						guide_bitmap.clear_pixel(x+1,y+1);
						stack_.push(new Point(x+1,y+1));
						if (x+1>x_right) x_right=x+1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				if ((x>=1) && (y<guide_y_dim-1))
				{
					if (guide_bitmap[y+1,x-1]==Bitmap.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y+1);
						if (use_bitmap[y+1,x-1]==Bitmap.PixelEnum.set)
							isl_bitmap.set_pixel(x-1,y+1);
						guide_bitmap.clear_pixel(x-1,y+1);
						stack_.push(new Point(x-1,y+1));
						if (x-1<x_left) x_left=x-1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}
			}
  
			int x_dim = x_right-x_left+1;
			int y_dim = y_bot-y_top+1;

			GObject flood_gobject
				= new GObject(isl_bitmap,x_left,y_top,x_dim,y_dim,GObject.Operation.remove);

			return flood_gobject;
		}



		public GObject flood_fill_for_hough(Bitmap hough_bitmap, 
			ByteArray2d hough_accumulator,
			int x_seed, int y_seed,
			ref int max_pat_match)
		{
			// Creates a GObject

			Bitmap dst_bitmap = scratch_;
			Bitmap isl_bitmap = isolated_;

			int src_x_dim = hough_bitmap.x_dim();
			int src_y_dim = hough_bitmap.y_dim();

			int x_left=x_seed;
			int x_right=x_seed;
			int y_top=y_seed;
			int y_bot=y_seed;

			stack_.push(new Point(x_seed,y_seed));
			dst_bitmap.set_pixel(x_seed,y_seed);
			isl_bitmap.set_pixel(x_seed,y_seed);
			hough_bitmap.clear_pixel(x_seed,y_seed);

			while (!stack_.is_empty())
			{
				Point pt = stack_.pop();
				int x = pt.x();
				int y = pt.y();

				byte pat_match = hough_accumulator[y,x];
				if (pat_match>max_pat_match) max_pat_match=pat_match;

				if (x>=1)
				{
					Pixel pixel_l = hough_bitmap[y,x-1];
					if (pixel_l==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y);
						isl_bitmap.set_pixel(x-1,y);
						pixel_l.clear_pixel();
						stack_.push(new Point(x-1,y));
						if (x-1<x_left) x_left=x-1;

						if (x>=2)
						{
							if (hough_bitmap[y,x-2]==Pixel.PixelEnum.clear)
							{
								dst_bitmap.set_pixel(x-2,y);
								isl_bitmap.set_pixel(x-2,y);
								stack_.push(new Point(x-2,y));
								if (x-2<x_left) x_left=x-2;
							}
						}
					}
				}

				if (y>=1)
				{
					Pixel pixel_t = hough_bitmap[y-1,x];
					if (pixel_t==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x,y-1);
						isl_bitmap.set_pixel(x,y-1);
						pixel_t.clear_pixel();
						stack_.push(new Point(x,y-1));
						if (y-1<y_top) y_top=y-1;
					}
				}

				if (x<src_x_dim-1)
				{
					Pixel pixel_r = hough_bitmap[y,x+1];
					if (pixel_r==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y);
						isl_bitmap.set_pixel(x+1,y);
						pixel_r.clear_pixel();
						stack_.push(new Point(x+1,y));
						if (x+1>x_right) x_right=x+1; // optimise


						if (x<src_x_dim-2)
						{
							if (hough_bitmap[y,x+2]==Pixel.PixelEnum.clear)
							{
								dst_bitmap.set_pixel(x+2,y);
								isl_bitmap.set_pixel(x+2,y);
								stack_.push(new Point(x+2,y));
								if (x+2>x_right) x_right=x+2;
							}
						}
					}
				}

				if (y<src_y_dim-1)
				{
					Pixel pixel_b = hough_bitmap[y+1,x];
					if (pixel_b==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x,y+1);
						isl_bitmap.set_pixel(x,y+1);
						pixel_b.clear_pixel();
						stack_.push(new Point(x,y+1));
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				/* and the corners */

				if ((x>=1) && (y>=1))
				{
					Pixel pixel_tl = hough_bitmap[y-1,x-1];
					if (pixel_tl==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y-1);
						isl_bitmap.set_pixel(x-1,y-1);
						pixel_tl.clear_pixel();
						stack_.push(new Point(x-1,y-1));
						if (x-1<x_left) x_left=x-1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<src_x_dim-1) && (y>=1))
				{
					Pixel pixel_tr = hough_bitmap[y-1,x+1];
					if (pixel_tr==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y-1);
						isl_bitmap.set_pixel(x+1,y-1);
						pixel_tr.clear_pixel();
						stack_.push(new Point(x+1,y-1));
						if (x+1>x_right) x_right=x+1;
						if (y-1<y_top) y_top=y-1;
					}
				}

				if ((x<src_x_dim-1) && (y<src_y_dim-1))
				{
					Pixel pixel_br = hough_bitmap[y+1,x+1];
					if (pixel_br==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x+1,y+1);
						isl_bitmap.set_pixel(x+1,y+1);
						pixel_br.clear_pixel();
						stack_.push(new Point(x+1,y+1));
						if (x+1>x_right) x_right=x+1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}

				if ((x>=1) && (y<src_y_dim-1))
				{
					Pixel pixel_bl = hough_bitmap[y+1,x-1];
					if (pixel_bl==Pixel.PixelEnum.set)
					{
						dst_bitmap.set_pixel(x-1,y+1);
						isl_bitmap.set_pixel(x-1,y+1);
						pixel_bl.clear_pixel();
						stack_.push(new Point(x-1,y+1));
						if (x-1<x_left) x_left=x-1;
						if (y+1>y_bot) y_bot=y+1;
					}
				}
			}
  
			int x_dim = x_right-x_left+1;
			int y_dim = y_bot-y_top+1;

			GObject flood_gobject
				= new GObject(isl_bitmap,x_left,y_top,x_dim,y_dim,GObject.Operation.remove);

			return flood_gobject;

		}


		/*---------------------------------------------------------------------------*/



		protected void move_clockwise(ref int x, ref int y)
		{
			/* Given displacement from a pivot -1 <= x <= +1, -1 <= y <= +1.
			   Rotate clockwise and return new displacement in x, y.          */

			switch(x)
			{
				case -1 :
				switch(y)
				{
					case -1 : x =  0; return;
					case  0 : y = -1; return;
					case  1 : y =  0; return;
				}
				break;

				case  0 :
				switch(y)
				{
					case -1 : x =  1; return;
					case  1 : x = -1; return;
				}
				break;

				case  1 :
				switch(y)
				{
					case -1 : y = 0; return;
					case  0 : y = 1; return;
					case  1 : x = 0; return;
				}
				break;
			}
		}

		protected void next_clockwise(int xpivot, int ypivot, ref int x, ref int y)
		{
			/* Given a pivot (xpivot, ypivot) and a location (x, y) such that
			   |xpivot - x| <= 1 and |ypivot - y| <= 1 rotate location one step
			   aroung pivot and return new location in x, y.  i.e
     
			   ...	      ...
		   .@.	==>   .@.
		   ..+	      .+.
     
		   @ = (xpivot, ypivot)
		   + = (x, y)                                                         */

			int dx = x - xpivot;
			int dy = y - ypivot;
  
			move_clockwise(ref dx, ref dy);
  
			x = xpivot + dx;
			y = ypivot + dy;
		}


		public GObject contour_trace(Bitmap src_bitmap, int x, int y)
		{
			/* Given co-ords (x, y) of seed pixel in b, isolate the mark that the 
			   seed pixel belongs to.  Record the Start and End
			   positions of the mark on each line, and return them.                */

			int x_left = x;
			int x_right = x;
			int y_top = y;
			int y_bottom = y;

			int px, py;		/* Previous pixel. */
			int cx, cy;		/* Current pixel. */

			/* The Current state.
     
			   ................,......
			   ................@#
			   ###
     
			   . = white
			   @ = seed
			   # = one of these will be the next pixel
			   , = what we will call the previous pixel even though it is white.
     
			   Everywhere else there be dragons.                                       */
  
			cx = x;
			cy = y;
  
			px = x;
			py = y - 1;
  
			/* The plan, in general
    
			  6*+	   
			  5@1
			  432
    
			  * = previous pixel.
			  @ = current pixel.
			  1..6 the candidates for the next pixel in the order they will be checked.
			  If they are all white then *is the next pixel
			  + = pixel that does not need to be checked, since it was examined last 
			  iteration and is known to be white.
			  If it wasn't white then it would be the current pixel.
    
			  Rotate diagram for all possibilities.                                    */
  
			/* sl records the extent of the mark on each row, by recording the leftmost 
			   and rightmost pixel of the mark
			   on each scan line.  The list is indexed by the y-offset from the seed 
			   pixel of mark.                                                          */
    
			do 
			{
				int nx = px, ny = py;	/* The candidate pixels. */
				int i, off;
      
				/* Get starting candidate pixel */
				next_clockwise(cx, cy, ref nx, ref ny); 		
				next_clockwise(cx, cy, ref nx, ref ny);
      
				for(i = 0; i < 6; i++)
				{
					if (src_bitmap.coord_in_bitmap(nx,ny))
					{
						if (src_bitmap[ny,nx]==Bitmap.PixelEnum.set) 
							break; /* Candidate is black, stop. */
					}
					next_clockwise(cx, cy, ref nx, ref ny);  /* Get next candidate.       */
				}
      
				if(i == 6) /* None of the candidates was black. */
				{
					if (src_bitmap[py,px]==Bitmap.PixelEnum.set) 
					{
						/* Check (px, py) is black, in case it is actually the
					   first previous pixel which is white.                 */
						int s;
						s = cx; cx = px; px = s;
						s = cy; cy = py; py = s;
	      
						/* continue; ? */
					}
					else break;	/* Get out. There is only one pixel in the mark. */
				}
				else
				{
					px = cx; cx = nx;
					py = cy; cy = ny;
				}
            
				off = cy - y;
      
				/* fprintf(stderr, "off = %d, cx = %d, cy = %d\n", off, cx, cy); */

				if (cx<x_left) x_left=cx;
				if (cx>x_right) x_right=cx;
				if (cy<y_top) y_top=cy;
				if (cy>y_bottom) y_bottom = cy;

			} while(!(cx == x && cy == y));


			int x_dim = x_right-x_left+1;
			int y_dim = y_bottom-y_top+1;

			GObject traced_gobject
				=new GObject(src_bitmap,x_left,y_top,x_dim,y_dim,GObject.Operation.remove);

			scratch_.inlay_bitmap(traced_gobject);
  
			return traced_gobject;
		}
  

		/*-------------*/



		public GObject trim(SubBitmap src_bitmap)
		{
			// Creates a GObject

			int src_x_dim = src_bitmap.x_dim();
			int src_y_dim = src_bitmap.y_dim();

			int x,y;
			bool match;

			match=false;
			int x_left = 0;
			for (x=0; x<src_x_dim; x++)
			{
				for (y=0; y<src_y_dim; y++)
				{
					if (src_bitmap[y,x]==SubBitmap.PixelEnum.set)
					{
						x_left=x;
						match=true;
						break;
					}
				}
				if (match) break;
			}

			match=false;
			int x_right = src_x_dim-1;
			for (x=src_x_dim-1; x>=0; x--)
			{
				for (y=0; y<src_y_dim; y++)
				{
					if (src_bitmap[y,x]==SubBitmap.PixelEnum.set)
					{
						x_right=x;
						match=true;
						break;
					}
				}
				if (match) break;
			}

			match=false;
			int y_top = 0;
			for (y=0; y<src_y_dim; y++)
			{
				for (x=0; x<src_x_dim; x++)
				{
					if (src_bitmap[y,x]==SubBitmap.PixelEnum.set)
					{
						y_top=y;
						match=true;
						break;
					}
				}
				if (match) break;
			}

			match=false;
			int y_bot = src_y_dim-1;
			for (y=src_y_dim-1; y>=0; y--)
			{
				for (x=0; x<src_x_dim; x++)
				{
					if (src_bitmap[y,x]==SubBitmap.PixelEnum.set)
					{
						y_bot = y;
						match=true;
						break;
					}
				}
				if (match) break;
			}


			int x_dim = x_right-x_left+1;
			int y_dim = y_bot-y_top+1;

			GObject trimmed_gobject
				= new GObject(src_bitmap,x_left,y_top,x_dim,y_dim,GObject.Operation.nop);

			trimmed_gobject.add_to_x_org(src_bitmap.x_org());
			trimmed_gobject.add_to_y_org(src_bitmap.y_org());
  

			return trimmed_gobject;
		}
	}
}
