
using System;

namespace Espresso.OMR
{

  delegate bool BopFollowCheck(SubBitmap bitmap, int src_x, int src_y);

  class BopFollow
    {
      public enum DirectionEnum { forwards, backwards };
      public enum ModeEnum { vertical_requirement, circle_requirement };
      public enum SLCopyEnum { delete, keep };

      protected Bitmap scratch_;
      protected Bitmap isolated_;

      protected BresenCircle circle_;
      protected BresenLineArray1d lines_;

      public BopFollowCheck check_prerau_requirement_above;
      public BopFollowCheck check_prerau_requirement_below;
      public BopFollowCheck check_simple_requirement_above;
      public BopFollowCheck check_simple_requirement_below;
      public BopFollowCheck check_clarke_requirement_above;
      public BopFollowCheck check_clarke_requirement_below;

      public BopFollow(int x_dim,int y_dim, Bitmap scratch, Bitmap isolated)
	{
	  scratch_ = scratch;
	  isolated_ = isolated;

	  circle_ = null;
	  lines_ = new BresenLineArray1d();

	  // set up callback methods as delegates
	  check_prerau_requirement_above 
	    = new BopFollowCheck(check_prerau_requirement_above_callback);
	  check_prerau_requirement_below
	    = new BopFollowCheck(check_prerau_requirement_below_callback);
	  check_simple_requirement_above
	    = new BopFollowCheck(check_simple_requirement_above_callback);
	  check_simple_requirement_below
	    = new BopFollowCheck(check_simple_requirement_below_callback);
	  check_clarke_requirement_above
	    = new BopFollowCheck(check_clarke_requirement_above_callback);
	  check_clarke_requirement_below
	    = new BopFollowCheck(check_clarke_requirement_below_callback);

	}


      public void resize_circle(int radius)
	{
	  int p;

	  circle_ = new BresenCircle(radius);

	  int circ_array_len = circle_.length();

	  lines_.resize(circ_array_len);

	  int i=0;
	  int p_low, p_high;
	  circle_.arc_range(90.0,270.0,out p_low,out p_high);

	  for (p=p_low; p<p_high; p++)
	    {
	      Point pt = circle_.data(p);
	      lines_[i++].generate_line(0,0,pt.x(),pt.y());
	    }
  
	  circle_.arc_range(270.0,450.0,out p_low, out p_high);
	  for (p=p_low; p<p_high; p++)
	    {
	      Point pt = circle_.data(p);
	      lines_[i++].generate_line(0,0,pt.x(),pt.y());
	    }
	}



      public void follow_line_straight(SubBitmap staff_system_copy,
				       LineInfo line_info,
				       SubBitmap line_bitmap)
	{
	  int ss_x_org = staff_system_copy.x_org();
	  int ss_y_org = staff_system_copy.y_org();

	  int delta_x = line_info.delta_x();
	  int delta_y = line_info.delta_y();
	  int incr_e  = 2*delta_y;
	  int incr_ns = 2*(delta_y - delta_x);
	  int diff = 2*delta_y - delta_x;

	  double grad = line_info.grad();
	  int y_move = (MathExtra.sign_bit(grad)) ? 1 : -1;

	  int y_start = line_info.y_top(); 
	  int thick = line_info.avg_thickness();
	  int half_thick = MathExtra.irint((double)thick/2.0);

	  int x_start = line_info.x_left();
	  int x_end = line_info.x_right();

	  int ssc_y_dim = staff_system_copy.y_dim();

	  for (int x=x_start; x<=x_end; x++)
	    {
	      int y;

	      if (y_start<0) break;

	      int y_mid =y_start+half_thick;
	      int y_end =y_start+thick-1;

	      // copy staff line slither

	      // add guard !? // ****
	      for (y=y_start-1; y<=y_end+1; y++)
		{
		  int ssc_y = y-ss_y_org;
		  if (staff_system_copy[ssc_y,x]==Bitmap.PixelEnum.set)
		    scratch_.set_pixel(ss_x_org+x,y);
		}

	      if (diff <=0) diff += incr_e;
	      else { diff += incr_ns; y_start+=y_move; }
     
	    }

	  int sl_x_start = ss_x_org+x_start;
	  int sl_y_start = line_info.y_top()-delta_y;

	  int sl_x_dim = delta_x;
	  int sl_y_dim = delta_y+thick;

	  line_bitmap.set_x_off(sl_x_start%8);
	  line_bitmap.resize(sl_x_dim,sl_y_dim);

	  line_bitmap.extract(scratch_,sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);

	  line_bitmap.add_to_x_org(sl_x_start);
	  line_bitmap.add_to_y_org(sl_y_start);
	}






      public void follow_line_straight_prerau(SubBitmap staff_system_copy,
					      LineInfo line_info,
					      SubBitmap line_bitmap)
	{
	  int ss_x_org = staff_system_copy.x_org();
	  int ss_y_org = staff_system_copy.y_org();

	  int delta_x = line_info.delta_x();
	  int delta_y = line_info.delta_y();
	  int incr_e  = 2*delta_y;
	  int incr_ns = 2*(delta_y - delta_x);
	  int diff = 2*delta_y - delta_x;

	  double grad = line_info.grad();
	  int y_move = (MathExtra.sign_bit(grad)) ? 1 : -1;

	  int y_start = line_info.y_top(); 
	  int thick = line_info.avg_thickness();
	  int half_thick = MathExtra.irint((double)thick/2.0);

	  int x_start = line_info.x_left();
	  int x_end = line_info.x_right();

	  int ssc_y_dim = staff_system_copy.y_dim();

	  for (int x=x_start; x<=x_end; x++)
	    {
	      int y;

	      if (y_start<0) break;

	      int y_mid =y_start+half_thick;
	      int y_end =y_start+thick-1;

	      // copy staff line slither

	      for (y=y_start; y<=y_end; y++)
		{
		  int ssc_y = y-ss_y_org;
		  if (staff_system_copy[ssc_y,x]==Bitmap.PixelEnum.set)
		    scratch_.set_pixel(ss_x_org+x,y);
		}

	      // remove staff line slither (tolerance -1,+1)

	      // add guard !? // ****

	      for (y=y_start-1; y<=y_end+1; y++)
		{
		  int ssc_y = y-ss_y_org;
		  if ((ssc_y>=0)  && (ssc_y<ssc_y_dim))
		    staff_system_copy.clear_pixel(x,ssc_y);
		}

	      if (diff <=0) diff += incr_e;
	      else { diff += incr_ns; y_start+=y_move; }
     
	    }

	  int sl_x_start = ss_x_org+x_start;
	  int sl_y_start = line_info.y_top()-delta_y;

	  int sl_x_dim = delta_x;
	  int sl_y_dim = delta_y+thick;

	  line_bitmap.set_x_off(sl_x_start%8);
	  line_bitmap.resize(sl_x_dim,sl_y_dim);

	  line_bitmap.extract(scratch_,sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);

	  line_bitmap.add_to_x_org(sl_x_start);
	  line_bitmap.add_to_y_org(sl_y_start);

	  scratch_.clear_area(sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);
	}


      protected bool check_prerau_requirement_above_callback(SubBitmap src_bitmap,
					  int src_x, 
					  int src_y)
	{
	  return true;
	}

      protected bool check_prerau_requirement_below_callback(SubBitmap src_bitmap,
					  int src_x, 
					  int src_y)
	{
	  return true;
	}


      protected bool check_simple_requirement_above_callback(SubBitmap src_bitmap,
					  int src_x, 
					  int src_y)
	{
	  // check top half;
  
	  int src_x_dim = src_bitmap.x_dim();

	  int ym1=src_y-1;
	  int ym2=src_y-2;

	  bool only_staff=true;

	  if (ym2>=0)
	    if (src_bitmap[ym1,src_x]==Bitmap.PixelEnum.set)
	      {
		only_staff = (src_bitmap[ym2,src_x]==Bitmap.PixelEnum.clear);
	      }
  
	  return only_staff;
	}

      protected bool check_simple_requirement_below_callback(SubBitmap src_bitmap,
					  int src_x, 
					  int src_y)
	{
	  // check bottom half;
	  int src_x_dim = src_bitmap.x_dim();
	  int src_y_dim = src_bitmap.y_dim();

	  int yp1=src_y+1;
	  int yp2=src_y+2;

	  bool only_staff=true;

	  if (yp2<src_y_dim)
	    if (src_bitmap[yp1,src_x]==Bitmap.PixelEnum.set)
	      {
		only_staff = (src_bitmap[yp2,src_x]==Bitmap.PixelEnum.clear);	
	      }

	  return only_staff;
	}


      protected bool check_clarke_requirement_above_callback(SubBitmap src_bitmap,
					  int src_x, 
					  int src_y)
	{
	  // check top half;
  
	  int src_x_dim = src_bitmap.x_dim();

	  int ym1=src_y-1;
	  int ym2=src_y-2;

	  bool only_staff=true;

	  if (ym2>=0)
	    if (src_bitmap[ym1,src_x]==Bitmap.PixelEnum.set)
	      {
		bool has_object_c = (src_bitmap[ym2,src_x]==Bitmap.PixelEnum.set);
		bool has_object_m1
		  = (src_x-1>=0) ? (src_bitmap[ym2,src_x-1]==Bitmap.PixelEnum.set) : false;
		bool has_object_p1
		  = (src_x+1<src_x_dim) ? (src_bitmap[ym2,src_x+1]==Bitmap.PixelEnum.set) : false;
	
		only_staff = !(has_object_m1 || has_object_c || has_object_p1);
	      }
  
	  return only_staff;
	}

      protected bool check_clarke_requirement_below_callback(SubBitmap src_bitmap,
					  int src_x, 
					  int src_y)
	{
	  // check bottom half;
	  int src_x_dim = src_bitmap.x_dim();
	  int src_y_dim = src_bitmap.y_dim();

	  int yp1=src_y+1;
	  int yp2=src_y+2;

	  bool only_staff=true;

	  if (yp2<src_y_dim)
	    if (src_bitmap[yp1,src_x]==Bitmap.PixelEnum.set)
	      {
		bool has_object_c = (src_bitmap[yp2,src_x]==Bitmap.PixelEnum.set);
		bool has_object_m1
		  = (src_x-1>=0) ? (src_bitmap[yp2,src_x-1]==Bitmap.PixelEnum.set) : false;
		bool has_object_p1
		  = (src_x+1<src_x_dim) ? (src_bitmap[yp2,src_x+1]==Bitmap.PixelEnum.set) : false;
	
		only_staff = !(has_object_m1 || has_object_c || has_object_p1);
	      }

	  return only_staff;
	}



      public void follow_line_straight_clarke(SubBitmap staff_system_copy,
					      LineInfo line_info,
					      SubBitmap line_bitmap)
	{
	  int ss_x_org = staff_system_copy.x_org();
	  int ss_y_org = staff_system_copy.y_org();

	  int delta_x = line_info.delta_x();
	  int delta_y = line_info.delta_y();
	  int incr_e  = 2*delta_y;
	  int incr_ns = 2*(delta_y - delta_x);
	  int diff = 2*delta_y - delta_x;

	  double grad = line_info.grad();
	  int y_move = (MathExtra.sign_bit(grad)) ? 1 : -1;

	  int y_start = line_info.y_top(); 
	  int thick = line_info.avg_thickness();
	  int half_thick = MathExtra.irint((double)thick/2.0);

	  int x_start = line_info.x_left();
	  int x_end = line_info.x_right();

	  for (int x=x_start; x<=x_end; x++)
	    {
	      int y;

	      if (y_start<0) break;

	      int y_mid =y_start+half_thick;
	      int y_end =y_start+thick-1;

	      // check top half
	      int ym1=y_start-ss_y_org-1;
	      int ym2=y_start-ss_y_org-2;
	      bool only_staff=true;
	      if (ym2>=0)
		if (staff_system_copy[ym1,x]==Bitmap.PixelEnum.set)
		  {
		    bool has_object_c
		      = (staff_system_copy[ym2,x]==Bitmap.PixelEnum.set);
		    bool has_object_m1
		      = (x-1>=x_start) ? (staff_system_copy[ym2,x-1]==Bitmap.PixelEnum.set) 
		      : false;
		    bool has_object_p1
		      = (x+1<=x_end)   ? (staff_system_copy[ym2,x+1]==Bitmap.PixelEnum.set) 
		      : false;

		    only_staff = !(has_object_m1 || has_object_c || has_object_p1);
		  }

	      for (y=y_start; y<y_mid; y++)
		{
		  int ssc_y = y-ss_y_org;
		  if (staff_system_copy[ssc_y,x]==Bitmap.PixelEnum.set)
		    scratch_.set_pixel(ss_x_org+x,y);
		}

	      if (only_staff)
		{
		  for (y=y_start-1; y<y_mid; y++)
		    {
		      int ssc_y = y-ss_y_org;
		      if (ssc_y>=0) staff_system_copy.clear_pixel(x,ssc_y);
		    }
		}

	      // check bottom half
	      int ssc_y_dim = staff_system_copy.y_dim();
	      int yp1=y_end-ss_y_org+1;
	      int yp2=y_end-ss_y_org+2;
	      only_staff=true;
	      if (yp2<ssc_y_dim)
		if (staff_system_copy[yp1,x]==Bitmap.PixelEnum.set)
		  {
		    bool has_object_c 
		      = (staff_system_copy[yp2,x]==Bitmap.PixelEnum.set);
		    bool has_object_m1
		      = (x-1>=x_start) ? (staff_system_copy[yp2,x-1]==Bitmap.PixelEnum.set) 
		      : false;
		    bool has_object_p1
		      = (x+1<=x_end)   ? (staff_system_copy[yp2,x+1]==Bitmap.PixelEnum.set) 
		      : false;

		    only_staff = !(has_object_m1 || has_object_c || has_object_p1);
		  }

	      for (y=y_mid; y<=y_end; y++)
		{
		  int ssc_y = y-ss_y_org;
		  if (staff_system_copy[ssc_y,x]==Bitmap.PixelEnum.set)
		    scratch_.set_pixel(ss_x_org+x,y);
		}

	      if (only_staff)
		{
		  for (y=y_mid; y<=y_end+1; y++)
		    {
		      int ssc_y = y-ss_y_org;
		      if (ssc_y<ssc_y_dim) staff_system_copy.clear_pixel(x,ssc_y);
		    }
		}

	      if (diff <=0) diff += incr_e;
	      else { diff += incr_ns; y_start+=y_move; }
     
	    }

	  int sl_x_start = ss_x_org+x_start;
	  int sl_y_start = line_info.y_top()-delta_y;

	  int sl_x_dim = delta_x;
	  int sl_y_dim = delta_y+thick;

	  line_bitmap.set_x_off(sl_x_start%8);
	  line_bitmap.resize(sl_x_dim,sl_y_dim);

	  line_bitmap.extract(scratch_,sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);

	  line_bitmap.add_to_x_org(sl_x_start);
	  line_bitmap.add_to_y_org(sl_y_start);

	  scratch_.clear_area(sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);
	}





      /*---------------------------------------------------------------------------*/

      public void check_circle_requirement(SubBitmap src_bitmap,
				    int ssc_x_org, 
				    int ssc_y_top, 
				    int ssc_y_bot,
				    int y_thresh,
				    ref bool do_remove_top, 
				    ref bool do_remove_bot)
	{
	  int l;

	  int ss_x_dim = src_bitmap.x_dim();
	  int ss_y_dim = src_bitmap.y_dim();

	  int l_thick = ssc_y_bot - ssc_y_top +1;
	  int ssc_y_midb = ssc_y_top + l_thick/2;
	  int ssc_y_midt = (l_thick%2==1) ? ssc_y_midb : ssc_y_midb-1;

	  int lines_len = lines_.length();

	  // top half
	  do_remove_top = true;

	  for (l=0; l<lines_len/2; l++)
	    {
	      BresenLine line = lines_[l];
	      int line_len=line.length();
	      int p=0;
	      while (p<line_len)
		{	  
		  Point pt = line.data(p);
		  int ssc_x = ssc_x_org+pt.x();
		  int ssc_y = ssc_y_midt+pt.y();
		  if ((ssc_x<0) || (ssc_x>=ss_x_dim)) break;
		  if ((ssc_y<0) || (ssc_y>=ss_y_dim)) break;
		  if (src_bitmap[ssc_y,ssc_x]==Bitmap.PixelEnum.clear) break;
		  p++;
		}
	      p--;

	      Point end_pt = line.data(p);
	      if (ssc_y_midt+end_pt.y()<ssc_y_top-y_thresh)
		{
		  do_remove_top=false;
		  break;
		}
	    }

	  // bottom half
	  do_remove_bot = true;
	  for (l=lines_len/2; l<lines_len; l++)
	    {
	      BresenLine line = lines_[l];
	      int line_len=line.length();
	      int p=0;
	      while (p<line_len)
		{
		  Point pt = line.data(p);
		  int ssc_x = ssc_x_org+pt.x();
		  int ssc_y = ssc_y_midb+pt.y();
		  if ((ssc_x<0) || (ssc_x>=ss_x_dim)) break;
		  if ((ssc_y<0) || (ssc_y>=ss_y_dim)) break;

		  if (src_bitmap[ssc_y,ssc_x]==Bitmap.PixelEnum.clear) break;
		  p++;
		}
	      p--;

	      Point end_pt = line.data(p);
	      if (ssc_y_midb+end_pt.y()>ssc_y_bot+y_thresh)
		{
		  do_remove_bot=false;
		  break;
		}
	    }
	}

  
      /*-----------------------*/

      private void follow_line_section_wobble(SubBitmap staff_system,
				      SubBitmap staff_system_copy,
				      int x_start,int x_end,
				      int start_y_top, 
				      int start_y_bot,
				      int avg_line_thick,
				      double lvt_mul,
				      int delta_x,int delta_y,
				      double grad,
				      DirectionEnum direction,
				      ModeEnum mode,
				      SLCopyEnum sl_copy,
				      bool do_remove,
				      BopFollowCheck check_fn_above,
				      BopFollowCheck check_fn_below,
				      ref int min_y_top, ref int max_y_bot)
	{  
	  int ss_x_org = staff_system_copy.x_org();
	  int ss_y_org = staff_system_copy.y_org();
	  int ss_y_dim = staff_system_copy.y_dim();

	  int l_vert_thresh = MathExtra.irint(lvt_mul*(double)avg_line_thick);
	  int l_circle_thresh = 170*avg_line_thick/100;

	  int ssc_y_top = start_y_top-ss_y_org;
	  int ssc_y_bot = start_y_bot-ss_y_org;

	  int ssc_y_centre_line = (ssc_y_top+ssc_y_bot)/2;

	  // values for bresenham line plotting
	  int incr_e  = 2*delta_y;
	  int incr_ns = 2*(delta_y - delta_x);
	  int diff = 2*delta_y - delta_x;
	  int y_move = (MathExtra.sign_bit(grad)) ? 1 : -1;

	  int p,p_move;

	  if (direction==DirectionEnum.forwards)
	    {
	      p=x_start;
	      p_move=1;
	    }
	  else
	    {
	      p=x_end;
	      p_move=-1;
	      y_move*=-1; // change sign
	    }


	  int c_range = x_end - x_start +1;
	  int c=0;

	  int old_ssc_y_top = ssc_y_top;
	  int old_ssc_y_bot = ssc_y_bot;

	  while (c<c_range)
	    {
	      // fan out ssc_y_top and ssc_y_bot, based on previous values

	      if (ssc_y_top>0) ssc_y_top--;
	      int snap_shot_ssc_y_top = ssc_y_top;

	      if (ssc_y_bot<ss_y_dim-1) ssc_y_bot++;

	      // if one pixel above black, extend up until no longer black
	      if (staff_system_copy[ssc_y_top,p]==Bitmap.PixelEnum.set)
		{
		  if (ssc_y_top>0)
		    {
		      while (staff_system_copy[ssc_y_top,p]==Bitmap.PixelEnum.set)
			{
			  ssc_y_top--;
			  if (ssc_y_top<0) break;
			}
		    }
		  ssc_y_top++;
		}
	      else
		{
		  // pixel was not black, so search down until a black pixel is found
		  if (ssc_y_top<ssc_y_bot)
		    {
		      while (staff_system_copy[++ssc_y_top,p]==Bitmap.PixelEnum.clear)
			if (ssc_y_top==ssc_y_bot) break;
		    }
		}


	      // repeat for ssc_y_bot
	      if (staff_system_copy[ssc_y_bot,p]==Bitmap.PixelEnum.set)
		{
		  if (ssc_y_bot<ss_y_dim-1)
		    {
		      while (staff_system_copy[ssc_y_bot,p]==Bitmap.PixelEnum.set)
			{
			  ssc_y_bot++;
			  if (ssc_y_bot>=ss_y_dim) break;
			}
		    }
		  ssc_y_bot--;
		}
	      else
		{
		  if (ssc_y_bot>snap_shot_ssc_y_top)
		    {
		      while (staff_system_copy[--ssc_y_bot,p]==Bitmap.PixelEnum.clear)
			if (ssc_y_bot==snap_shot_ssc_y_top) break;
		    }
		}
      
	      // Check if it is definitely, Only a staff line
	      int c_thick = ssc_y_bot-ssc_y_top +1;
	      if ((c_thick<=avg_line_thick) && (c_thick>0))
		{
		  old_ssc_y_top = ssc_y_top;
		  old_ssc_y_bot = ssc_y_bot;
		}


	      // look for continuous segments

	      int closest_ysm_dist = 0;
	      int closest_yst = ssc_y_top; // incase no segment found
	      int closest_ysb = ssc_y_bot;

	      bool done_first_seg=false;
	      int y=ssc_y_top;
	      while (y<=ssc_y_bot)
		{
		  int seg_y_top=y;
		  do 
		    {
		      y++;
		      if (y>ssc_y_bot) break;
		    }
		  while (staff_system_copy[y,p]==Bitmap.PixelEnum.set);

		  int seg_y_bot=y-1;
		  int seg_y_mid=(seg_y_top+seg_y_bot)/2;

		  if (!done_first_seg)
		    {
		      closest_yst = seg_y_top;
		      closest_ysm_dist = Math.Abs(seg_y_mid-ssc_y_centre_line);
		      closest_ysb = seg_y_bot;
		      done_first_seg=true;
		    }
		  else
		    {
		      int seg_y_top_dist = Math.Abs(seg_y_top-ssc_y_centre_line);
		      int seg_y_mid_dist = Math.Abs(seg_y_mid-ssc_y_centre_line);
		      int seg_y_bot_dist = Math.Abs(seg_y_bot-ssc_y_centre_line);

		      if (seg_y_top_dist<closest_ysm_dist)
			{
			  closest_yst = seg_y_top;
			  closest_ysm_dist = seg_y_top_dist;
			  closest_ysb = seg_y_bot;
			}
		      if (seg_y_mid_dist<closest_ysm_dist)
			{
			  closest_yst = seg_y_top;
			  closest_ysm_dist = seg_y_top_dist;
			  closest_ysb = seg_y_bot;
			}
		      if (seg_y_bot_dist<closest_ysm_dist)
			{
			  closest_yst = seg_y_top;
			  closest_ysm_dist = seg_y_bot_dist;
			  closest_ysb = seg_y_bot;
			}

		    }

		  // skip white
		  do 
		    {
		      y++;
		      if (y>ssc_y_bot) break;
		    }
		  while (staff_system_copy[y,p]==Bitmap.PixelEnum.clear);

		}


	      if (do_remove)
		{
		  // look at new width;
		  bool do_remove_top = false;
		  bool do_remove_bot = false;

		  ssc_y_top = closest_yst;
		  ssc_y_bot = closest_ysb;
		  int l_thick = ssc_y_bot-ssc_y_top +1;
		  int ssc_y_midb = ssc_y_top + l_thick/2;
		  int ssc_y_midt = (l_thick%2==1) ? ssc_y_midb : ssc_y_midb-1;
	  
		  if (((l_thick<=l_vert_thresh) && (l_thick>0)) &&
		      ((ssc_y_top<old_ssc_y_bot+1) || (ssc_y_bot>old_ssc_y_top-1)))
		    {
		      if (mode==ModeEnum.circle_requirement) 
			check_circle_requirement(staff_system,p,ssc_y_top,ssc_y_bot,
						 l_circle_thresh,
						 ref do_remove_top,ref do_remove_bot);
		      else 
			{
			  do_remove_top 
			    = check_fn_above(staff_system_copy,p,ssc_y_top);

			  do_remove_bot 
			    = check_fn_below(staff_system_copy,p,ssc_y_bot);

			  // see <= backup 4 for different control of remove_top/bot;
			}
		    }

		  if (do_remove_top || do_remove_bot)
		    {
		      int ssc_y;
	      
		      // guard added // ****
		      if (ssc_y_top<0) ssc_y_top=0;
		      if (ssc_y_bot>=ss_y_dim) ssc_y_bot=ss_y_dim-1;

		      for (ssc_y=ssc_y_top; ssc_y<=ssc_y_midt; ssc_y++)
			{
			  if (staff_system_copy[ssc_y,p]==Bitmap.PixelEnum.set)
			    scratch_.set_pixel(ss_x_org+p,ss_y_org+ssc_y);
			}

		      for (ssc_y=ssc_y_midb; ssc_y<=ssc_y_bot; ssc_y++)
			{
			  if (staff_system_copy[ssc_y,p]==Bitmap.PixelEnum.set)
			    scratch_.set_pixel(ss_x_org+p,ss_y_org+ssc_y);
			}

		      if (do_remove_top)
			for (ssc_y=ssc_y_top; ssc_y<=ssc_y_midt; ssc_y++)
			  staff_system_copy.clear_pixel(p,ssc_y);
	      
		      if (do_remove_bot)
			for (ssc_y=ssc_y_midb; ssc_y<=ssc_y_bot; ssc_y++)
			  staff_system_copy.clear_pixel(p,ssc_y);
	  
		      if (diff <=0) diff += incr_e;
		      else 
			{
			  diff += incr_ns; 
			  ssc_y_centre_line+=y_move;
			}
		    }
		  else
		    {
		      for (int ssc_y=old_ssc_y_top; ssc_y<=old_ssc_y_bot; ssc_y++)
			{
			  int lb_y = ssc_y+ss_y_org;
			  if (staff_system_copy[ssc_y,p]==Bitmap.PixelEnum.set)
			    {
			      scratch_.set_pixel(ss_x_org+p,lb_y);
			      if (sl_copy==SLCopyEnum.delete)
				{
				  staff_system_copy.clear_pixel(p,ssc_y);
				}
			    }
			}

		      if (diff <=0) diff += incr_e;
		      else 
			{ 
			  diff += incr_ns; 
			  ssc_y_centre_line+=y_move;
			  old_ssc_y_top+=y_move; 
			  old_ssc_y_bot+=y_move;
			}
		      ssc_y_top = old_ssc_y_top;
		      ssc_y_bot = old_ssc_y_bot;	 
		    }
		}
	      else
		{
		  // look at new width;

		  ssc_y_top = closest_yst;
		  ssc_y_bot = closest_ysb;
		  int l_thick = ssc_y_bot-ssc_y_top +1;
		  int ssc_y_midb = ssc_y_top + l_thick/2;
		  int ssc_y_midt = (l_thick%2==1) ? ssc_y_midb : ssc_y_midb-1;
	  
		  if (((l_thick<=l_vert_thresh) && (l_thick>0)) &&
		      ((ssc_y_top<old_ssc_y_bot+1) || (ssc_y_bot>old_ssc_y_top-1)))
		    {
		      // guard added // ****
		      if (ssc_y_top<0) ssc_y_top=0;
		      if (ssc_y_bot>=ss_y_dim) ssc_y_bot=ss_y_dim-1;

		      int ssc_y;
		      for (ssc_y=ssc_y_top; ssc_y<=ssc_y_midt; ssc_y++)
			{
			  if (staff_system_copy[ssc_y,p]==Bitmap.PixelEnum.set)
			    {
			      scratch_.set_pixel(ss_x_org+p,ss_y_org+ssc_y);
			      if (sl_copy==SLCopyEnum.delete)
				{
				  staff_system_copy.clear_pixel(p,ssc_y);
				}
			    }
			}

		      for (ssc_y=ssc_y_midb; ssc_y<=ssc_y_bot; ssc_y++)
			{
			  if (staff_system_copy[ssc_y,p]==Bitmap.PixelEnum.set)
			    {
			      scratch_.set_pixel(ss_x_org+p,ss_y_org+ssc_y);
			      if (sl_copy==SLCopyEnum.delete)
				{
				  staff_system_copy.clear_pixel(p,ssc_y);
				}
			    }
			}
	  
		      if (diff <=0) diff += incr_e;
		      else 
			{
			  diff += incr_ns; 
			  ssc_y_centre_line+=y_move;
			}
		    }
		  else
		    {
		      // make default staff line thicker for sl cross algorithm

		      // guard;
		      int tssc_y_top = old_ssc_y_top-1;
		      int tssc_y_bot = old_ssc_y_bot+1;

		      if (tssc_y_top<0) tssc_y_top=0;
		      if (tssc_y_bot>=ss_y_dim) tssc_y_bot=ss_y_dim-1;

		      for (int ssc_y=tssc_y_top; ssc_y<=tssc_y_bot; ssc_y++)
			{
			  int lb_y = ssc_y+ss_y_org;
			  if (staff_system_copy[ssc_y,p]==Bitmap.PixelEnum.set)
			    {
			      scratch_.set_pixel(ss_x_org+p,lb_y);
			      if (sl_copy==SLCopyEnum.delete)
				{
				  staff_system_copy.clear_pixel(p,ssc_y);
				}
			    }
			}

		      if (diff <=0) diff += incr_e;
		      else 
			{ 
			  diff += incr_ns; 
			  ssc_y_centre_line+=y_move;
			  old_ssc_y_top+=y_move; 
			  old_ssc_y_bot+=y_move;
			}
		      ssc_y_top = old_ssc_y_top;
		      ssc_y_bot = old_ssc_y_bot;	 
		    }
		}

	      // check min and max if part of staffline
	      int lb_y_top = ssc_y_top+ss_y_org;
	      int lb_y_bot = ssc_y_bot+ss_y_org;

	      if (lb_y_top<min_y_top) min_y_top=lb_y_top;
	      if (lb_y_bot>max_y_bot) max_y_bot=lb_y_bot;

	      c++; p+=p_move;
	    }
	}



      public void follow_line_wobble_prerau(SubBitmap staff_system,
				     SubBitmap staff_system_copy,
				     LineInfo line_info,
				     SubBitmap line_bitmap, 
				     ModeEnum mode,
				     bool do_remove)
	{
	  int x_left = line_info.x_left();
	  int x_midl = line_info.x_midl();
	  int x_midr = line_info.x_midr();
	  int x_right = line_info.x_right();

	  int y_topl = line_info.y_topl();
	  int y_botl = line_info.y_botl();
	  int y_topr = line_info.y_topr();
	  int y_botr = line_info.y_botr();

	  int delta_x = line_info.delta_x();
	  int delta_y = line_info.delta_y();
	  double grad = line_info.grad();

	  int min_y_top = y_topl;
	  int max_y_bot = y_botl;

	  int avg_line_thick = line_info.avg_thickness();

	  // section left
	  follow_line_section_wobble(staff_system,staff_system_copy,x_left,x_midl,
				     y_topl,y_botl,avg_line_thick,2.0,
				     delta_x,delta_y,grad,DirectionEnum.backwards,mode,
				     SLCopyEnum.delete,do_remove,
				     check_prerau_requirement_above,
				     check_prerau_requirement_below,
				     ref min_y_top, ref max_y_bot);
	  // section middle
	  follow_line_section_wobble(staff_system,staff_system_copy,x_midl,x_midr,
				     y_topl,y_botl,avg_line_thick,2.0,
				     delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				     SLCopyEnum.delete,do_remove,
				     check_prerau_requirement_above,
				     check_prerau_requirement_below,
				     ref min_y_top, ref max_y_bot);
	  // section right
	  follow_line_section_wobble(staff_system,staff_system_copy,x_midr,x_right,
				     y_topr,y_botr,avg_line_thick,2.0,
				     delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				     SLCopyEnum.delete,do_remove,
				     check_prerau_requirement_above,
				     check_prerau_requirement_below,
				     ref min_y_top, ref max_y_bot);


	  int ss_x_org = staff_system_copy.x_org();

	  int sl_x_start = ss_x_org+x_left;
	  int sl_y_start = min_y_top;

	  int sl_x_dim = x_right - x_left +1;
	  int sl_y_dim = max_y_bot - min_y_top +1;  

	  line_bitmap.set_x_off(sl_x_start%8);
	  line_bitmap.resize(sl_x_dim,sl_y_dim);

	  line_bitmap.extract(scratch_,sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);

	  line_bitmap.add_to_x_org(sl_x_start);
	  line_bitmap.add_to_y_org(sl_y_start);

	  if (do_remove)
	    scratch_.clear_area(sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);
	}


      public void follow_line_wobble_simple(SubBitmap staff_system,
				     SubBitmap staff_system_copy,
				     LineInfo line_info,
				     SubBitmap line_bitmap, 
				     ModeEnum mode,
				     bool do_remove)
	{
	  int x_left = line_info.x_left();
	  int x_midl = line_info.x_midl();
	  int x_midr = line_info.x_midr();
	  int x_right = line_info.x_right();

	  int y_topl = line_info.y_topl();
	  int y_botl = line_info.y_botl();
	  int y_topr = line_info.y_topr();
	  int y_botr = line_info.y_botr();

	  int delta_x = line_info.delta_x();
	  int delta_y = line_info.delta_y();
	  double grad = line_info.grad();

	  int min_y_top = y_topl;
	  int max_y_bot = y_botl;

	  int avg_line_thick = line_info.avg_thickness();

	  // section left
	  follow_line_section_wobble(staff_system,staff_system_copy,x_left,x_midl,
				     y_topl,y_botl,avg_line_thick,2.0,
				     delta_x,delta_y,grad,DirectionEnum.backwards,mode,
				     SLCopyEnum.keep,do_remove,
				     check_simple_requirement_above,
				     check_simple_requirement_below,
				     ref min_y_top, ref max_y_bot);
	  // section middle
	  follow_line_section_wobble(staff_system,staff_system_copy,x_midl,x_midr,
				     y_topl,y_botl,avg_line_thick,2.0,
				     delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				     SLCopyEnum.keep,do_remove,
				     check_simple_requirement_above,
				     check_simple_requirement_below,
				     ref min_y_top, ref max_y_bot);
	  // section right
	  follow_line_section_wobble(staff_system,staff_system_copy,x_midr,x_right,
				     y_topr,y_botr,avg_line_thick,2.0,
				     delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				     SLCopyEnum.keep,do_remove,
				     check_simple_requirement_above,
				     check_simple_requirement_below,
				     ref min_y_top, ref max_y_bot);

	  int ss_x_org = staff_system_copy.x_org();

	  int sl_x_start = ss_x_org+x_left;
	  int sl_y_start = min_y_top;

	  int sl_x_dim = x_right - x_left +1;
	  int sl_y_dim = max_y_bot - min_y_top +1;  

	  line_bitmap.set_x_off(sl_x_start%8);
	  line_bitmap.resize(sl_x_dim,sl_y_dim);

	  line_bitmap.extract(scratch_,sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);

	  line_bitmap.add_to_x_org(sl_x_start);
	  line_bitmap.add_to_y_org(sl_y_start);

	  if (do_remove)
	    scratch_.clear_area(sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);
	}



      public void follow_line_wobble(SubBitmap staff_system,
			      SubBitmap staff_system_copy,
			      LineInfo line_info,
			      SubBitmap line_bitmap, 
			      ModeEnum mode,
			      bool do_remove)
	{
	  int x_left = line_info.x_left();
	  int x_midl = line_info.x_midl();
	  int x_midr = line_info.x_midr();
	  int x_right = line_info.x_right();

	  int y_topl = line_info.y_topl();
	  int y_botl = line_info.y_botl();
	  int y_topr = line_info.y_topr();
	  int y_botr = line_info.y_botr();

	  int delta_x = line_info.delta_x();
	  int delta_y = line_info.delta_y();
	  double grad = line_info.grad();

	  int min_y_top = y_topl;
	  int max_y_bot = y_botl;

	  int avg_line_thick = line_info.avg_thickness();

	  // section left
	  follow_line_section_wobble(staff_system,staff_system_copy,x_left,x_midl,
				     y_topl,y_botl,avg_line_thick,2.0,
				     delta_x,delta_y,grad,DirectionEnum.backwards,mode,
				     SLCopyEnum.keep,do_remove,
				     check_clarke_requirement_above,
				     check_clarke_requirement_below,
				     ref min_y_top, ref max_y_bot);
	  // section middle
	  follow_line_section_wobble(staff_system,staff_system_copy,x_midl,x_midr,
				     y_topl,y_botl,avg_line_thick,2.0,
				     delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				     SLCopyEnum.keep,do_remove,
				     check_clarke_requirement_above,
				     check_clarke_requirement_below,
				     ref min_y_top, ref max_y_bot);
	  // section right
	  follow_line_section_wobble(staff_system,staff_system_copy,x_midr,x_right,
				     y_topr,y_botr,avg_line_thick,2.0,
				     delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				     SLCopyEnum.keep,do_remove,
				     check_clarke_requirement_above,
				     check_clarke_requirement_below,
				     ref min_y_top, ref max_y_bot);


	  int ss_x_org = staff_system_copy.x_org();

	  int sl_x_start = ss_x_org+x_left;
	  int sl_y_start = min_y_top;

	  int sl_x_dim = x_right - x_left +1;
	  int sl_y_dim = max_y_bot - min_y_top +1;  

	  line_bitmap.set_x_off(sl_x_start%8);
	  line_bitmap.resize(sl_x_dim,sl_y_dim);

	  line_bitmap.extract(scratch_,sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);

	  line_bitmap.add_to_x_org(sl_x_start);
	  line_bitmap.add_to_y_org(sl_y_start);

	  if (do_remove)
	    scratch_.clear_area(sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);
	}







      /*---------------------------------------------------------------------------*/


      private void follow_line_section_track(SubBitmap staff_system,
				     SubBitmap staff_system_copy,
				     int x_start,int x_end,
				     int start_y_top, 
				     int start_y_bot,
				     int avg_line_thick,
				     double lvt_mul,
				     int delta_x,int delta_y,
				     double grad,
				     DirectionEnum direction,
				     ModeEnum mode,
				     SLCopyEnum sl_copy,
				     bool do_remove,
				     BopFollowCheck check_fn_above,
				     BopFollowCheck check_fn_below,
				     ref int min_y_top, ref int max_y_bot)
	{  
	  int ss_x_org = staff_system_copy.x_org();
	  int ss_y_org = staff_system_copy.y_org();
	  int ss_y_dim = staff_system_copy.y_dim();

	  int l_vert_thresh = MathExtra.irint(lvt_mul*(double)avg_line_thick);
	  int l_circle_thresh = 170*avg_line_thick/100;

	  int half_alt=avg_line_thick/2;

	  int start_ssc_y_top = start_y_top-ss_y_org;
	  int start_ssc_y_bot = start_y_bot-ss_y_org;
	  int ssc_y_centre_line = (start_ssc_y_top+start_ssc_y_bot)/2;

	  int top_l_dist = ssc_y_centre_line - start_ssc_y_top; // no +1
	  int bot_l_dist = start_ssc_y_bot - ssc_y_centre_line; // no +1

	  // values for bresenham line plotting
	  int incr_e  = 2*delta_y;
	  int incr_ns = 2*(delta_y - delta_x);
	  int diff = 2*delta_y - delta_x;
	  int y_move = (MathExtra.sign_bit(grad)) ? 1 : -1;

	  int p,p_move;

	  if (direction==DirectionEnum.forwards)
	    {
	      p=x_start;
	      p_move=1;
	    }
	  else
	    {
	      p=x_end;
	      p_move=-1;
	      y_move*=-1; // change sign
	    }

	  int c_range = x_end - x_start +1;
	  int c=0;

	  int ssc_y_top = start_ssc_y_top;
	  int ssc_y_bot = start_ssc_y_bot;

	  while (c<c_range)
	    {
	      int true_ssc_y_top;
	      int true_ssc_y_bot;

	      bool ssc_seg_valid = false;
      
	      if ((ssc_y_centre_line>=0) && (ssc_y_centre_line<ss_y_dim))
		{
		  if (staff_system_copy[ssc_y_centre_line,p]==Bitmap.PixelEnum.set)
		    {
		      // find extent
		      int ssc_y_top_single=ssc_y_centre_line;
		      while (staff_system_copy[ssc_y_top_single,p]==Bitmap.PixelEnum.set)
			{
			  ssc_y_top_single--;
			  if (ssc_y_top_single<0) break;
			}
		      ssc_y_top_single++;
	      
		      int ssc_y_bot_single=ssc_y_centre_line;
		      while (staff_system_copy[ssc_y_bot_single,p]==Bitmap.PixelEnum.set)
			{
			  ssc_y_bot_single++;
			  if (ssc_y_bot_single>=ss_y_dim) break;
			}
		      ssc_y_bot_single--;
	      
		      // check that it is thin enough
		      int l_thick_single = ssc_y_bot_single-ssc_y_top_single+1;
		      true_ssc_y_top = ssc_y_top_single;
		      true_ssc_y_bot = ssc_y_bot_single;
		      if ((l_thick_single>0) && (l_thick_single<=l_vert_thresh))
			{
			  ssc_y_top = ssc_y_top_single;
			  ssc_y_bot = ssc_y_bot_single;
			  ssc_seg_valid = true;
			}		
		    }
		  else
		    {
		      bool l_thick_above_valid, l_thick_below_valid;
	      
		      // find seg above
		      int ssc_y_top_above=0; // l_thick_above_valid only true when this variable has been updated with a legitimate value
		      int ssc_y_bot_above=ssc_y_centre_line;

		      if (ssc_y_bot_above>0)
			{
			  while (staff_system_copy[--ssc_y_bot_above,p]
				 ==Bitmap.PixelEnum.clear)
			    if (ssc_y_bot_above==0) break;
		  
			  ssc_y_top_above=ssc_y_bot_above;
			  if (staff_system_copy[ssc_y_bot_above,p]==Bitmap.PixelEnum.set)
			    {
			      while (staff_system_copy[ssc_y_top_above,p]
				     ==Bitmap.PixelEnum.set)
				{
				  ssc_y_top_above--;
				  if (ssc_y_top_above<0) break;
				}
			      ssc_y_top_above++;
			    }
			  else ssc_y_top_above+=1;
		  
			  // check that seg is thin and close enough
			  l_thick_above_valid=false;
			  int l_thick_above=ssc_y_bot_above-ssc_y_top_above +1;
			  if ((l_thick_above>0) && (l_thick_above<=l_vert_thresh))
			    {
			      int l_bot_dist=ssc_y_centre_line-ssc_y_bot_above+1;
			      if (l_bot_dist<=l_vert_thresh) l_thick_above_valid=true;
			    }
		  
			}
		      else l_thick_above_valid=false;
	      
		      // find seg below
		      int ssc_y_bot_below=0; // l_thick_below_valid only true when this variable has been updated with a legitimate value
	      
		      int ssc_y_top_below=ssc_y_centre_line;
		      if (ssc_y_top_below<ss_y_dim-1)
			{
			  while (staff_system_copy[++ssc_y_top_below,p]
				 ==Bitmap.PixelEnum.clear)
			    if (ssc_y_top_below==ss_y_dim-1) break;
		  
			  ssc_y_bot_below=ssc_y_top_below;
			  if (staff_system_copy[ssc_y_top_below,p]==Bitmap.PixelEnum.set)
			    {
			      while (staff_system_copy[ssc_y_bot_below,p]==Bitmap.PixelEnum.set)
				{
				  ssc_y_bot_below++;
				  if (ssc_y_bot_below>=ss_y_dim) break;
				}
			      ssc_y_top_below--;
			    }
			  else ssc_y_bot_below-=1;
		  

			  // check that seg is thin and close enough
			  l_thick_below_valid=false;
			  int l_thick_below=ssc_y_bot_below-ssc_y_top_below +1;
			  if ((l_thick_below>0) && (l_thick_below<=l_vert_thresh))
			    {
			      int l_top_dist=ssc_y_top_below-ssc_y_centre_line+1;
			      if (l_top_dist<=l_vert_thresh) l_thick_below_valid=true;
			    }
		  
			}
		      else l_thick_below_valid=false;
	      

		      if (l_thick_above_valid)
			{
			  if (l_thick_below_valid)
			    {
			      // pick closest one
			      int l_bot_dist=ssc_y_centre_line-ssc_y_bot_above+1;
			      int l_top_dist=ssc_y_top_below-ssc_y_centre_line+1;

			      if (l_bot_dist<l_top_dist)
				{
				  ssc_y_top = ssc_y_top_above;
				  ssc_y_bot = ssc_y_bot_above;
				  true_ssc_y_top = ssc_y_top_above;
				  true_ssc_y_bot = ssc_y_bot_above;
				}
			      else
				{
				  ssc_y_top = ssc_y_top_below;
				  ssc_y_bot = ssc_y_bot_below;
				  true_ssc_y_top = ssc_y_top_below;
				  true_ssc_y_bot = ssc_y_bot_below;
				}
		      
			      ssc_seg_valid = true;		      
			    }
			  else
			    {
			      // only one
			      ssc_y_top = ssc_y_top_above;
			      ssc_y_bot = ssc_y_bot_above;
			      true_ssc_y_top = ssc_y_top_above;
			      true_ssc_y_bot = ssc_y_bot_above;
			      ssc_seg_valid = true;
			    }
			}
		      else
			{
			  if (l_thick_below_valid)
			    {
			      // only one
			      ssc_y_top = ssc_y_top_below;
			      ssc_y_bot = ssc_y_bot_below;
			      true_ssc_y_top = ssc_y_top_below;
			      true_ssc_y_bot = ssc_y_bot_below;
			      ssc_seg_valid = true;
			    }
			  else
			    {
			      // none : just pick one
			      true_ssc_y_top = ssc_y_top_below;
			      true_ssc_y_bot = ssc_y_bot_below;
			      ssc_seg_valid = false;
			    }
			}

		    }
		}
	      else ssc_seg_valid=false;

	      int sscl_y_top;
	      int sscl_y_bot;

	      if (do_remove)
		{
		  bool do_remove_top=false;
		  bool do_remove_bot=false;

		  int l_thick = ssc_y_bot-ssc_y_top +1;
		  int ssc_y_midb = ssc_y_top + l_thick/2;
		  int ssc_y_midt = (l_thick%2==1) ? ssc_y_midb : ssc_y_midb-1;
	  
		  if (ssc_seg_valid)
		    {
		      // Check if it is definitely, Only a staff line;
		      if (ssc_y_top<0) ssc_y_top = 0;
		      if (ssc_y_bot>=ss_y_dim) ssc_y_bot = ss_y_dim-1;
	      
		      if (mode==ModeEnum.circle_requirement) 
			check_circle_requirement(staff_system,p,ssc_y_top,ssc_y_bot,
						 l_circle_thresh,
						 ref do_remove_top, ref do_remove_bot);
		      else 
			{
			  do_remove_top 
			    = check_fn_above(staff_system_copy,p,ssc_y_top);

			  do_remove_bot 
			    = check_fn_below(staff_system_copy,p,ssc_y_bot);

			  // see <= backup 4 for different control of remove_top/bot;
			}

		      // decide on range of staff line seg to produce;

		      int ssc_cly_top = ssc_y_centre_line-top_l_dist;
		      int ssc_cly_bot = ssc_y_centre_line+bot_l_dist;
	  
		      if (ssc_cly_top<0) ssc_cly_top=0;
		      if (ssc_cly_bot>=ss_y_dim) ssc_cly_bot=ss_y_dim-1;
	  
		      int min_dist;
		      int ssc_y_mid = (ssc_y_top + ssc_y_bot)/2;
	  
		      int dist_top = Math.Abs(ssc_y_top-ssc_y_centre_line);
		      min_dist = dist_top;
		      sscl_y_top = ssc_y_top;
		      sscl_y_bot = sscl_y_top + avg_line_thick -1;
	  
	  
		      int dist_mid = Math.Abs(ssc_y_mid-ssc_y_centre_line);
		      if (dist_mid<min_dist)
			{
			  min_dist = dist_mid;
			  sscl_y_top = ssc_y_mid - top_l_dist;
			  sscl_y_bot = ssc_y_mid + bot_l_dist;	      
			}
	  
		      int dist_bot = Math.Abs(ssc_y_bot-ssc_y_centre_line);
		      if (dist_bot<min_dist)
			{
			  min_dist = dist_bot;
			  sscl_y_bot = ssc_y_bot;
			  sscl_y_top = sscl_y_bot - avg_line_thick +1;
			}
		    }
		  else
		    {
		      // only remove a small section if starts ends on a staff line;
		      // did not turn out to be such a good idea;
		      // see <=backup4 for code;

		      // an alternative;
		      sscl_y_top = ssc_y_centre_line-top_l_dist;
		      sscl_y_bot = ssc_y_centre_line+bot_l_dist;
		    }

      
		  if (do_remove_top || do_remove_bot)
		    {
		      int ssc_y;

		      // guard added // ****
		      if (sscl_y_top<0) sscl_y_top=0;
		      if (sscl_y_bot>=ss_y_dim) sscl_y_bot=ss_y_dim-1;

		      for (ssc_y=sscl_y_top; ssc_y<=sscl_y_bot; ssc_y++)
			if (staff_system_copy[ssc_y,p]==Bitmap.PixelEnum.set)		  
			  scratch_.set_pixel(ss_x_org+p,ss_y_org+ssc_y);

		      if (do_remove_top)
			for (ssc_y=ssc_y_top; ssc_y<=ssc_y_midt; ssc_y++)
			  staff_system_copy.clear_pixel(p,ssc_y);
	  
		      if (do_remove_bot)
			for (ssc_y=ssc_y_midb; ssc_y<=ssc_y_bot; ssc_y++)
			  staff_system_copy.clear_pixel(p,ssc_y);
	      
		      // check min and max if part of staffline;
		      int lb_y_top = sscl_y_top+ss_y_org;
		      int lb_y_bot = sscl_y_bot+ss_y_org;
	      
		      if (lb_y_top<min_y_top) min_y_top=lb_y_top;
		      if (lb_y_bot>max_y_bot) max_y_bot=lb_y_bot;
		    }
		  else
		    {
		      int ssc_cly_top = ssc_y_centre_line-top_l_dist;
		      int ssc_cly_bot = ssc_y_centre_line+bot_l_dist;
	      
		      if (ssc_cly_top<0) ssc_cly_top=0;
		      if (ssc_cly_bot>=ss_y_dim) ssc_cly_bot=ss_y_dim-1;
	      
		      // used to be sscl_y_top and sscl_y_bot (a mistake?) // ****
		      for (int ssc_y=ssc_cly_top; ssc_y<=ssc_cly_bot; ssc_y++)
			{
			  int lb_y = ssc_y+ss_y_org;
			  if (staff_system_copy[ssc_y,p]==Bitmap.PixelEnum.set)
			    {
			      scratch_.set_pixel(ss_x_org+p,lb_y);
			      if (sl_copy==SLCopyEnum.delete)
				{
				  staff_system_copy.clear_pixel(p,ssc_y);
				}
			    }
			}

		      // check min and max if part of staffline;
		      int lb_y_top = ssc_cly_top+ss_y_org;
		      int lb_y_bot = ssc_cly_bot+ss_y_org;

		      if (lb_y_top<min_y_top) min_y_top=lb_y_top;
		      if (lb_y_bot>max_y_bot) max_y_bot=lb_y_bot;
		    }
	  
		}
	      else
		{
		  // do not remove;

		  int l_thick = ssc_y_bot-ssc_y_top +1;
		  int ssc_y_midb = ssc_y_top + l_thick/2;
		  int ssc_y_midt = (l_thick%2==1) ? ssc_y_midb : ssc_y_midb-1;
	  
		  if (ssc_seg_valid)
		    {
		      if (ssc_y_top<0) ssc_y_top = 0;
		      if (ssc_y_bot>=ss_y_dim) ssc_y_bot = ss_y_dim-1;

		      // decide on range of staff line seg to produce;

		      int ssc_cly_top = ssc_y_centre_line-top_l_dist;
		      int ssc_cly_bot = ssc_y_centre_line+bot_l_dist;
	  
		      if (ssc_cly_top<0) ssc_cly_top=0;
		      if (ssc_cly_bot>=ss_y_dim) ssc_cly_bot=ss_y_dim-1;
	  
		      int min_dist;
		      int ssc_y_mid = (ssc_y_top + ssc_y_bot)/2;
	  
		      int dist_top = Math.Abs(ssc_y_top-ssc_y_centre_line);
		      min_dist = dist_top;
		      sscl_y_top = ssc_y_top;
		      sscl_y_bot = sscl_y_top + avg_line_thick -1;
	  
	  
		      int dist_mid = Math.Abs(ssc_y_mid-ssc_y_centre_line);
		      if (dist_mid<min_dist)
			{
			  min_dist = dist_mid;
			  sscl_y_top = ssc_y_mid - top_l_dist;
			  sscl_y_bot = ssc_y_mid + bot_l_dist;	      
			}
	  
		      int dist_bot = Math.Abs(ssc_y_bot-ssc_y_centre_line);
		      if (dist_bot<min_dist)
			{
			  min_dist = dist_bot;
			  sscl_y_bot = ssc_y_bot;
			  sscl_y_top = sscl_y_bot - avg_line_thick +1;
			}
		    }
		  else
		    {
		      // make default staff line thicker for sl cross algorithm
		      sscl_y_top = ssc_y_centre_line-top_l_dist-1;
		      sscl_y_bot = ssc_y_centre_line+bot_l_dist+1;
		    }

		  int ssc_y;

		  // guard added // ****
		  if (sscl_y_top<0) sscl_y_top=0;
		  if (sscl_y_bot>=ss_y_dim) sscl_y_bot=ss_y_dim-1;

		  for (ssc_y=sscl_y_top; ssc_y<=sscl_y_bot; ssc_y++)
		    if (staff_system_copy[ssc_y,p]==Bitmap.PixelEnum.set)
		      {
			scratch_.set_pixel(ss_x_org+p,ss_y_org+ssc_y);
			if (sl_copy==SLCopyEnum.delete)
			  {
			    staff_system_copy.clear_pixel(p,ssc_y);
			  }
		      }

	      
		  // check min and max if part of staffline;
		  int lb_y_top = sscl_y_top+ss_y_org;
		  int lb_y_bot = sscl_y_bot+ss_y_org;
	      
		  if (lb_y_top<min_y_top) min_y_top=lb_y_top;
		  if (lb_y_bot>max_y_bot) max_y_bot=lb_y_bot;
		}

	      if (diff <=0) diff += incr_e;
	      else 
		{ 
		  diff += incr_ns; 
		  ssc_y_centre_line+=y_move;
		}

	      c++; p+=p_move;
	    }
	}




      public void follow_line_track_prerau(SubBitmap staff_system,
				    SubBitmap staff_system_copy,
				    LineInfo line_info,
				    SubBitmap line_bitmap,
				    ModeEnum mode,
				    bool do_remove)
	{
	  int x_left = line_info.x_left();
	  int x_midl = line_info.x_midl();
	  int x_midr = line_info.x_midr();
	  int x_right = line_info.x_right();

	  int y_topl = line_info.y_topl();
	  int y_botl = line_info.y_botl();
	  int y_topr = line_info.y_topr();
	  int y_botr = line_info.y_botr();

	  int delta_x = line_info.delta_x();
	  int delta_y = line_info.delta_y();
	  double grad = line_info.grad();

	  int min_y_top = y_topl;
	  int max_y_bot = y_botl;

	  int avg_line_thick = line_info.avg_thickness();

	  // section left
	  follow_line_section_track(staff_system,staff_system_copy,x_left,x_midl,
				    y_topl,y_botl,avg_line_thick,2.0,
				    delta_x,delta_y,grad,DirectionEnum.backwards,mode,
				    SLCopyEnum.delete,do_remove,
				    check_prerau_requirement_above,
				    check_prerau_requirement_below,
				    ref min_y_top, ref max_y_bot);
	  // section middle
	  follow_line_section_track(staff_system,staff_system_copy,x_midl,x_midr,
				    y_topl,y_botl,avg_line_thick,2.0,
				    delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				    SLCopyEnum.delete,do_remove,
				    check_prerau_requirement_above,
				    check_prerau_requirement_below,
				    ref min_y_top, ref max_y_bot);
	  // section right
	  follow_line_section_track(staff_system,staff_system_copy,x_midr,x_right,
				    y_topr,y_botr,avg_line_thick,2.0,
				    delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				    SLCopyEnum.delete,do_remove,
				    check_prerau_requirement_above,
				    check_prerau_requirement_below,
				    ref min_y_top, ref max_y_bot);


	  int ss_x_org = staff_system_copy.x_org();

	  int sl_x_start = ss_x_org+x_left;
	  int sl_y_start = min_y_top;

	  int sl_x_dim = x_right - x_left +1;
	  int sl_y_dim = max_y_bot - min_y_top +1;  

	  line_bitmap.set_x_off(sl_x_start%8);
	  line_bitmap.resize(sl_x_dim,sl_y_dim);

	  line_bitmap.extract(scratch_,sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);

	  line_bitmap.add_to_x_org(sl_x_start);
	  line_bitmap.add_to_y_org(sl_y_start);

	  if (do_remove)
	    scratch_.clear_area(sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);
	}




      public void follow_line_track_simple(SubBitmap staff_system,
				    SubBitmap staff_system_copy,
				    LineInfo line_info,
				    SubBitmap line_bitmap,
				    ModeEnum mode,
				    bool do_remove)
	{
	  int x_left = line_info.x_left();
	  int x_midl = line_info.x_midl();
	  int x_midr = line_info.x_midr();
	  int x_right = line_info.x_right();

	  int y_topl = line_info.y_topl();
	  int y_botl = line_info.y_botl();
	  int y_topr = line_info.y_topr();
	  int y_botr = line_info.y_botr();

	  int delta_x = line_info.delta_x();
	  int delta_y = line_info.delta_y();
	  double grad = line_info.grad();

	  int min_y_top = y_topl;
	  int max_y_bot = y_botl;

	  int avg_line_thick = line_info.avg_thickness();

	  // section left
	  follow_line_section_track(staff_system,staff_system_copy,x_left,x_midl,
				    y_topl,y_botl,avg_line_thick,2.0,
				    delta_x,delta_y,grad,DirectionEnum.backwards,mode,
				    SLCopyEnum.keep,do_remove,
				    check_simple_requirement_above,
				    check_simple_requirement_below,
				    ref min_y_top, ref max_y_bot);
	  // section middle
	  follow_line_section_track(staff_system,staff_system_copy,x_midl,x_midr,
				    y_topl,y_botl,avg_line_thick,2.0,
				    delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				    SLCopyEnum.keep,do_remove,
				    check_simple_requirement_above,
				    check_simple_requirement_below,
				    ref min_y_top, ref max_y_bot);
	  // section right
	  follow_line_section_track(staff_system,staff_system_copy,x_midr,x_right,
				    y_topr,y_botr,avg_line_thick,2.0,
				    delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				    SLCopyEnum.keep,do_remove,
				    check_simple_requirement_above,
				    check_simple_requirement_below,
				    ref min_y_top, ref max_y_bot);


	  int ss_x_org = staff_system_copy.x_org();

	  int sl_x_start = ss_x_org+x_left;
	  int sl_y_start = min_y_top;

	  int sl_x_dim = x_right - x_left +1;
	  int sl_y_dim = max_y_bot - min_y_top +1;  

	  line_bitmap.set_x_off(sl_x_start%8);
	  line_bitmap.resize(sl_x_dim,sl_y_dim);

	  line_bitmap.extract(scratch_,sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);

	  line_bitmap.add_to_x_org(sl_x_start);
	  line_bitmap.add_to_y_org(sl_y_start);

	  if (do_remove)
	    scratch_.clear_area(sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);
	}


      public void follow_line_track(SubBitmap staff_system,
			     SubBitmap staff_system_copy,
			     LineInfo line_info,
			     SubBitmap line_bitmap,
			     ModeEnum mode,
			     bool do_remove)
	{
	  int x_left = line_info.x_left();
	  int x_midl = line_info.x_midl();
	  int x_midr = line_info.x_midr();
	  int x_right = line_info.x_right();

	  int y_topl = line_info.y_topl();
	  int y_botl = line_info.y_botl();
	  int y_topr = line_info.y_topr();
	  int y_botr = line_info.y_botr();

	  int delta_x = line_info.delta_x();
	  int delta_y = line_info.delta_y();
	  double grad = line_info.grad();

	  int min_y_top = y_topl;
	  int max_y_bot = y_botl;

	  int avg_line_thick = line_info.avg_thickness();

	  // section left
	  follow_line_section_track(staff_system,staff_system_copy,x_left,x_midl,
				    y_topl,y_botl,avg_line_thick,2.0,
				    delta_x,delta_y,grad,DirectionEnum.backwards,mode,
				    SLCopyEnum.keep,do_remove,
				    check_clarke_requirement_above,
				    check_clarke_requirement_below,
				    ref min_y_top, ref max_y_bot);
	  // section middle
	  follow_line_section_track(staff_system,staff_system_copy,x_midl,x_midr,
				    y_topl,y_botl,avg_line_thick,2.0,
				    delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				    SLCopyEnum.keep,do_remove,
				    check_clarke_requirement_above,
				    check_clarke_requirement_below,
				    ref min_y_top, ref max_y_bot);
	  // section right
	  follow_line_section_track(staff_system,staff_system_copy,x_midr,x_right,
				    y_topr,y_botr,avg_line_thick,2.0,
				    delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				    SLCopyEnum.keep,do_remove,
				    check_clarke_requirement_above,
				    check_clarke_requirement_below,
				    ref min_y_top, ref max_y_bot);


	  int ss_x_org = staff_system_copy.x_org();

	  int sl_x_start = ss_x_org+x_left;
	  int sl_y_start = min_y_top;

	  int sl_x_dim = x_right - x_left +1;
	  int sl_y_dim = max_y_bot - min_y_top +1;  

	  line_bitmap.set_x_off(sl_x_start%8);
	  line_bitmap.resize(sl_x_dim,sl_y_dim);

	  line_bitmap.extract(scratch_,sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);

	  line_bitmap.add_to_x_org(sl_x_start);
	  line_bitmap.add_to_y_org(sl_y_start);

	  if (do_remove)
	    scratch_.clear_area(sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);
	}




      public void follow_line_track_wide(SubBitmap staff_system,
				  SubBitmap staff_system_copy,
				  LineInfo line_info,
				  SubBitmap line_bitmap,
				  ModeEnum mode,
				  bool do_remove)
	{
	  int x_left = line_info.x_left();
	  int x_midl = line_info.x_midl();
	  int x_midr = line_info.x_midr();
	  int x_right = line_info.x_right();

	  int y_topl = line_info.y_topl();
	  int y_botl = line_info.y_botl();
	  int y_topr = line_info.y_topr();
	  int y_botr = line_info.y_botr();

	  int delta_x = line_info.delta_x();
	  int delta_y = line_info.delta_y();
	  double grad = line_info.grad();

	  int min_y_top = y_topl;
	  int max_y_bot = y_botl;

	  int avg_line_thick = line_info.avg_thickness();

	  // section left
	  follow_line_section_track(staff_system,staff_system_copy,x_left,x_midl,
				    y_topl,y_botl,avg_line_thick,2.5,
				    delta_x,delta_y,grad,DirectionEnum.backwards,mode,
				    SLCopyEnum.keep,do_remove,
				    check_clarke_requirement_above,
				    check_clarke_requirement_below,
				    ref min_y_top, ref max_y_bot);
	  // section middle
	  follow_line_section_track(staff_system,staff_system_copy,x_midl,x_midr,
				    y_topl,y_botl,avg_line_thick,2.5,
				    delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				    SLCopyEnum.keep,do_remove,
				    check_clarke_requirement_above,
				    check_clarke_requirement_below,
				    ref min_y_top, ref max_y_bot);
	  // section right
	  follow_line_section_track(staff_system,staff_system_copy,x_midr,x_right,
				    y_topr,y_botr,avg_line_thick,2.5,
				    delta_x,delta_y,grad,DirectionEnum.forwards,mode,
				    SLCopyEnum.keep,do_remove,
				    check_clarke_requirement_above,
				    check_clarke_requirement_below,
				    ref min_y_top, ref max_y_bot);


	  int ss_x_org = staff_system_copy.x_org();

	  int sl_x_start = ss_x_org+x_left;
	  int sl_y_start = min_y_top;

	  int sl_x_dim = x_right - x_left +1;
	  int sl_y_dim = max_y_bot - min_y_top +1;  

	  line_bitmap.set_x_off(sl_x_start%8);
	  line_bitmap.resize(sl_x_dim,sl_y_dim);

	  line_bitmap.extract(scratch_,sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);

	  line_bitmap.add_to_x_org(sl_x_start);
	  line_bitmap.add_to_y_org(sl_y_start);

	  if (do_remove)
	    scratch_.clear_area(sl_x_start,sl_y_start,sl_x_dim,sl_y_dim);
	}



      /*---------------------------------------------------------------------------*/

      private void cleanup_circ_req_track(GObject gobject,
				  int x_start,int x_end,
				  int start_y_top, 
				  int avg_line_thick,
				  double lvt_mul,
				  int delta_x,int delta_y,
				  double grad,
				  DirectionEnum direction)
	{
	  int go_x_org = gobject.x_org();
	  int go_y_org = gobject.y_org();
	  int go_y_dim = gobject.y_dim();

	  int start_y_bot = start_y_top+avg_line_thick-1;
	  int start_go_y_top = start_y_top-go_y_org;
	  int start_go_y_bot = start_y_bot-go_y_org;
	  int go_y_centre_line = (start_go_y_top+start_go_y_bot)/2;

	  int top_l_dist = go_y_centre_line - start_go_y_top; // no +1
	  int bot_l_dist = start_go_y_bot - go_y_centre_line; // no +1

	  int l_vert_thresh = MathExtra.irint(lvt_mul*(double)avg_line_thick);

	  // values for bresenham
	  int incr_e  = 2*delta_y;
	  int incr_ns = 2*(delta_y - delta_x);
	  int diff = 2*delta_y - delta_x;
	  int y_move = (MathExtra.sign_bit(grad)) ? 1 : -1;

	  int p,p_move; // make p offset absolute ***

	  if (direction==DirectionEnum.forwards)
	    {
	      p=x_start;
	      p_move=1;
	    }
	  else
	    {
	      p=x_end;
	      p_move=-1;
	      y_move*=-1; // change sign
	    }

	  int c_range = x_end - x_start +1;
	  int c=0;

	  int go_y_top = start_go_y_top;
	  int go_y_bot = start_go_y_bot;

	  while (c<c_range)
	    {
	      bool go_seg_valid = false;
	      bool go_seg_present = true;

	      if ((go_y_centre_line>=0) && (go_y_centre_line<go_y_dim))
		{
		  if (gobject[go_y_centre_line,p-go_x_org]==Bitmap.PixelEnum.set)
		    {
		      // find extent
		      int go_y_top_single=go_y_centre_line;
		      while (gobject[go_y_top_single,p-go_x_org]==Bitmap.PixelEnum.set)
			{
			  go_y_top_single--;
			  if (go_y_top_single<0) break;
			}
		      go_y_top_single++;
	      
		      int go_y_bot_single=go_y_centre_line;
		      while (gobject[go_y_bot_single,p-go_x_org]==Bitmap.PixelEnum.set)
			{
			  go_y_bot_single++;
			  if (go_y_bot_single>=go_y_dim) break;
			}
		      go_y_bot_single--;
	      
		      // check that it is thin enough
		      int l_thick_single = go_y_bot_single-go_y_top_single+1;
		      if ((l_thick_single>0) && (l_thick_single<=l_vert_thresh))
			{
			  go_y_top = go_y_top_single;
			  go_y_bot = go_y_bot_single;
			  go_seg_valid = true;
			}		
		    }
		  else
		    {
		      bool l_thick_above_valid, l_thick_below_valid;
	      
		      // find seg above
		      int go_y_top_above = 0; // l_thick_above_valid only true when this variable has been updated with a legitimate value 	      
		      int go_y_bot_above=go_y_centre_line;

		      if (go_y_bot_above>0)
			{
			  while (gobject[--go_y_bot_above,p-go_x_org]
				 ==Bitmap.PixelEnum.clear)
			    if (go_y_bot_above==0) break;
		  
			  go_y_top_above=go_y_bot_above;
			  if (gobject[go_y_bot_above,p-go_x_org]==Bitmap.PixelEnum.set)
			    {
			      while (gobject[go_y_top_above,p-go_x_org]
				     ==Bitmap.PixelEnum.set)
				{
				  go_y_top_above--;
				  if (go_y_top_above<0) break;
				}
			      go_y_top_above++;
			    }
			  else go_y_top_above+=1;
		  
			  // check that seg is thin and close enough
			  l_thick_above_valid=false;
			  int l_thick_above=go_y_bot_above-go_y_top_above +1;
			  if ((l_thick_above>0) && (l_thick_above<=l_vert_thresh))
			    {
			      int l_bot_dist=go_y_centre_line-go_y_bot_above+1;
			      if (l_bot_dist<=l_vert_thresh) l_thick_above_valid=true;
			    }
		  
			}
		      else l_thick_above_valid=false;
	      
		      // find seg below
		      int go_y_bot_below = 0; // l_thick_below_valid only true when this variable has been updated with a legitimate value
		      
		      int go_y_top_below=go_y_centre_line;

		      if (go_y_top_below<go_y_dim-1)
			{
			  while (gobject[++go_y_top_below,p-go_x_org]
				 ==Bitmap.PixelEnum.clear)
			    if (go_y_top_below==go_y_dim-1) break;
		  
			  go_y_bot_below=go_y_top_below;
			  if (gobject[go_y_top_below,p-go_x_org]==Bitmap.PixelEnum.set)
			    {
			      while (gobject[go_y_bot_below,p-go_x_org]
				     ==Bitmap.PixelEnum.set)
				{
				  go_y_bot_below++;
				  if (go_y_bot_below>=go_y_dim) break;
				}
			      go_y_top_below--;
			    }
			  else go_y_bot_below-=1;
		  

			  // check that seg is thin and close enough
			  l_thick_below_valid=false;
			  int l_thick_below=go_y_bot_below-go_y_top_below +1;
			  if ((l_thick_below>0) && (l_thick_below<=l_vert_thresh))
			    {
			      int l_top_dist=go_y_top_below-go_y_centre_line+1;
			      if (l_top_dist<=l_vert_thresh) l_thick_below_valid=true;
			    }
		  
			}
		      else l_thick_below_valid=false;
	      

		      if (l_thick_above_valid)
			{
			  if (l_thick_below_valid)
			    {
			      // pick closest one
			      int l_bot_dist=go_y_centre_line-go_y_bot_above+1;
			      int l_top_dist=go_y_top_below-go_y_centre_line+1;
			      if (l_bot_dist<l_top_dist)
				{
				  go_y_top = go_y_top_above;
				  go_y_bot = go_y_bot_above;
				}
			      else
				{
				  go_y_top = go_y_top_below;
				  go_y_bot = go_y_bot_below;
				}
		      
			      go_seg_valid = true;		      
			    }
			  else
			    {
			      // only one
			      go_y_top = go_y_top_above;
			      go_y_bot = go_y_bot_above;
			      go_seg_valid = true;
			    }
			}
		      else
			{
			  if (l_thick_below_valid)
			    {
			      // only one
			      go_y_top = go_y_top_below;
			      go_y_bot = go_y_bot_below;
			      go_seg_valid = true;
			    }
			  else
			    {
			      // none
			      go_seg_present = false;
			      go_seg_valid = false;
			    }
			}

		    }
		}
	      else go_seg_valid=false;

	      int l_thick = go_y_bot-go_y_top +1;
	      int go_y_midb = go_y_top + l_thick/2;
	      int go_y_midt = (l_thick%2==1) ? go_y_midb : go_y_midb-1;

	      if (go_seg_present)
		{
		  if (go_seg_valid)
		    {
		      if (go_y_top<0) go_y_top = 0;
		      if (go_y_bot>=go_y_dim) go_y_bot = go_y_dim-1;
	      
		      for (int go_y=go_y_top; go_y<=go_y_bot; go_y++)
			gobject.clear_pixel(p-go_x_org,go_y);
		    }
		  else break;
		}

	      if (diff <=0) diff += incr_e;
	      else 
		{ 
		  diff += incr_ns; 
		  go_y_centre_line+=y_move;
		}

	      c++; p+=p_move;
	    }
	}






      public void cleanup_circ_req_object(GObject gobject,
					  LineInfo line_info)
	{
	  int x_left = gobject.x_org();
	  int x_right = x_left + gobject.x_dim() -1;

	  // values for bresenham line plotting
	  int delta_x = line_info.delta_x();
	  int delta_y = line_info.delta_y();

	  double grad = line_info.grad();

	  int avg_line_thick = line_info.avg_thickness();

	  int yl_start = line_info.y_top(); 
	  int xl_start = line_info.x_left();

	  int y_left
	    = yl_start+MathExtra.irint((double)(x_left-xl_start)*grad);

	  int y_right
	    = yl_start+MathExtra.irint((double)(x_right-xl_start)*grad);

	  cleanup_circ_req_track(gobject,x_left,x_right,y_left,avg_line_thick,2.0,
				 delta_x,delta_y,grad,DirectionEnum.forwards);
	  cleanup_circ_req_track(gobject,x_left,x_right,y_right,avg_line_thick,2.0,
				 delta_x,delta_y,grad,DirectionEnum.backwards);
	}
      /*---------------------------------------------------------------------------*/

    }
}
      
