
using System;

namespace Espresso.OMR
{
	class BopRotate
	{

		public BopRotate()
		{}

		protected void rotate_bitmap_bot_left(Bitmap src_bitmap,
			double cos_theta,
			double sin_theta,
			Bitmap dst_bitmap,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			int dy;

			// grad is +ve
			int sx_dim = src_bitmap.x_dim();
			int sy_dim = src_bitmap.y_dim();
			int dx_dim = dst_bitmap.x_dim();
			int dy_dim = dst_bitmap.y_dim();

			double area = (double)(dx_dim*dy_dim);
			int mod_step = MathExtra.irint(area/crange);
			if (mod_step==0) mod_step = 1;
			double comp_add = crange*(double)mod_step/area;
			double computed = (double)cbase;
			int dc = 0;

			double sx_start=0.0;
			double sy_start=(double)(sy_dim-1);

			for (dy=sy_dim-1; dy>=0; dy--)
			{
				double sx = sx_start;
				double sy = sy_start;
				for (int dx=0; dx<dx_dim; dx++)
				{
					if (((++dc)%mod_step)==0)
					{
						if (circle!=null) 
						{
							computed += comp_add;
							circle.computed_so_far(MathExtra.irint(computed));
						}
					}

					int isx = MathExtra.irint(sx);
					int isy = MathExtra.irint(sy);
					if ((isx<0) || (isy<0) || (isx>=sx_dim) || (isy>=sy_dim))
						dst_bitmap.clear_pixel(dx,dy); // clear rest of bitmap line
					else dst_bitmap[dy,dx]=src_bitmap[isy,isx];
					sx += cos_theta;
					sy -= sin_theta;
				}
				sx_start -= sin_theta;
				sy_start -= cos_theta;

			}

			sx_start= -sin_theta;
			sy_start=(double)(sy_dim-1)+cos_theta;

			for (dy=sy_dim; dy<dy_dim; dy++)
			{
				double sx = sx_start;
				double sy = sy_start;
				for (int dx=0; dx<dx_dim; dx++)
				{
					if (((++dc)%mod_step)==0)
					{
						if (circle!=null) 
						{
							computed += comp_add;
							circle.computed_so_far(MathExtra.irint(computed));
						}
					}

					int isx = MathExtra.irint(sx);
					int isy = MathExtra.irint(sy);
					if ((isx<0) || (isy<0) || (isx>=sx_dim) || (isy>=sy_dim))
						dst_bitmap.clear_pixel(dx,dy); // clear rest of bitmap line
					else dst_bitmap[dy,dx]=src_bitmap[isy,isx];
					sx += cos_theta;
					sy -= sin_theta;
				}
				sx_start -= sin_theta;
				sy_start += cos_theta;

			}
		}


		protected void rotate_bitmap_bot_left(Bitmap src_bitmap,
			double cos_theta,
			double sin_theta,
			Bitmap dst_bitmap,
			ComputingCircle circle)
		{
		  rotate_bitmap_bot_left(src_bitmap,cos_theta,sin_theta,
					 dst_bitmap,circle,0,100.0);
		}


		protected void rotate_bitmap_top_right(Bitmap src_bitmap,
			double cos_theta,
			double sin_theta,
			Bitmap dst_bitmap,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			int dy;

			// grad is -ve
			int sx_dim = src_bitmap.x_dim();
			int sy_dim = src_bitmap.y_dim();
			int dx_dim = dst_bitmap.x_dim();
			int dy_dim = dst_bitmap.y_dim();

			double area = (double)(dx_dim*dy_dim);
			int mod_step = MathExtra.irint(area/crange);
			if (mod_step==0) mod_step = 1;
			double comp_add = crange*(double)mod_step/area;
			double computed = (double)cbase;
			int dc = 0;

			double sx_start=0;
			double sy_start=0;;

			for (dy=dy_dim-sy_dim; dy<dy_dim; dy++)
			{
				double sx = sx_start;
				double sy = sy_start;
				for (int dx=0; dx<dx_dim; dx++)
				{
					if (((++dc)%mod_step)==0)
					{
						if (circle!=null) 
						{
							computed += comp_add;
							circle.computed_so_far(MathExtra.irint(computed));
						}
					}

					int isx = MathExtra.irint(sx);
					int isy = MathExtra.irint(sy);
					if ((isx<0) || (isy<0) || (isx>=sx_dim) || (isy>=sy_dim))
						dst_bitmap.clear_pixel(dx,dy); // clear rest of bitmap line
					else dst_bitmap[dy,dx]=src_bitmap[isy,isx];
					sx += cos_theta;
					sy += sin_theta;
				}
				sx_start -= sin_theta;
				sy_start += cos_theta;

			}

			sx_start= -sin_theta;
			sy_start= -cos_theta;;

			for (dy=dy_dim-sy_dim-1; dy>=0; dy--)
			{
				double sx = sx_start;
				double sy = sy_start;
				for (int dx=0; dx<dx_dim; dx++)
				{
					if (((++dc)%mod_step)==0)
					{
						if (circle!=null) 
						{
							computed += comp_add;
							circle.computed_so_far(MathExtra.irint(computed));
						}
					}

					int isx = MathExtra.irint(sx);
					int isy = MathExtra.irint(sy);
					if ((isx<0) || (isy<0) || (isx>=sx_dim) || (isy>=sy_dim))
						dst_bitmap.clear_pixel(dx,dy); // clear rest of bitmap line
					else dst_bitmap[dy,dx]=src_bitmap[isy,isx];
					sx += cos_theta;
					sy += sin_theta;
				}
				sx_start -= sin_theta;
				sy_start -= cos_theta;

			}
		}

		protected void rotate_bitmap_top_right(Bitmap src_bitmap,
			double cos_theta,
			double sin_theta,
			Bitmap dst_bitmap,
			ComputingCircle circle)
		{
		  rotate_bitmap_top_right(src_bitmap,cos_theta,sin_theta,
					  dst_bitmap,circle,0,100.0);
		}


		public void calculate_rotate_dims(int sx_dim, int sy_dim, 
			double grad, 
			out int dx_dim, out int dy_dim)
		{
			double hyp = Math.Sqrt(1.0+MathExtra.sqr(grad));
			double cos_theta = 1.0/hyp;
			double sin_theta = grad/hyp;

			int sxtr = sx_dim -1;
			int sxbr = sx_dim -1;
			int sybr = sy_dim -1;
			int sybl = sy_dim -1;

			double sxtrd = (double)sxtr;
			double sxbrd = (double)sxbr;
			double sybrd = (double)sybr;
			double sybld = (double)sybl;

			int dxtr = MathExtra.irint(sxtrd*cos_theta);
			int dytr = MathExtra.irint(sxtrd*sin_theta);
			int dxbr = MathExtra.irint(sxbrd*cos_theta - sybrd*sin_theta);
			int dybr = MathExtra.irint(sxbrd*sin_theta + sybrd*cos_theta);
			int dxbl = MathExtra.irint(-sybld*sin_theta);
			int dybl = MathExtra.irint(sybld*cos_theta);

			int dxl = dxbl;
			int dxr = Math.Max(dxtr,dxbr);
			int dyt = dytr;
			int dyb = Math.Max(dybl,dybr);

			int super_xl = Math.Min(0,dxl);
			int super_xr = Math.Max(sxtr,dxr);
			int super_yt = Math.Min(0,dyt);
			int super_yb = Math.Max(sybr,dyb);

			int dx_dim1 
				= super_xr - super_xl +1 +1; // extra +1 to include rouding errors
			int dy_dim1
				= super_yb - super_yt +1 +1; // extra +1 to include rouding errors

			// something not quite right here, so er on the side of caution
			// ****
			int dx_dim2 = sx_dim + MathExtra.irint((double)sx_dim*sin_theta);
			int dy_dim2 = sy_dim + MathExtra.irint((double)sy_dim*sin_theta);

			dx_dim = Math.Max(dx_dim1,dx_dim2);
			dy_dim = Math.Max(dy_dim1,dy_dim2);
		}

		public Bitmap rotate_bitmap(Bitmap src_bitmap, 
			double grad,
			ComputingCircle circle,
			int cbase,
			double crange)
		{
			double hyp = Math.Sqrt(1.0+MathExtra.sqr(grad));
			double cos_theta = 1.0/hyp;
			double sin_theta = Math.Abs(grad)/hyp;

			int sx_dim = src_bitmap.x_dim();
			int sy_dim = src_bitmap.y_dim();

			int dx_dim,dy_dim;
			calculate_rotate_dims(sx_dim,sy_dim,grad,out dx_dim,out dy_dim);

			Bitmap dst_bitmap = new Bitmap(dx_dim,dy_dim);//  need to clear

			if (MathExtra.sign_bit(grad))
				rotate_bitmap_top_right(src_bitmap,cos_theta,sin_theta,dst_bitmap,circle);
			else
				rotate_bitmap_bot_left(src_bitmap,cos_theta,sin_theta,dst_bitmap,circle);

			return dst_bitmap;
		}

		public Bitmap rotate_bitmap(Bitmap src_bitmap, 
			double grad,
			ComputingCircle circle)
		{
			return rotate_bitmap(src_bitmap,grad,circle,0,100.0);
		}

		public SubBitmap rotate_subbitmap(SubBitmap src_subbitmap, 
			double grad,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			double hyp = Math.Sqrt(1.0+MathExtra.sqr(grad));
			double cos_theta = 1.0/hyp;
			double sin_theta = grad/hyp;
			double fabs_sin_theta = Math.Abs(sin_theta);

			int sxtl = src_subbitmap.x_left();
			int sytl = src_subbitmap.y_top();
			int sxtr = src_subbitmap.x_right();
			int sytr = src_subbitmap.y_top();
			int sxbr = src_subbitmap.x_right();
			int sybr = src_subbitmap.y_bot();
			int sxbl = src_subbitmap.x_left();
			int sybl = src_subbitmap.y_bot();

			double sxtld = (double)sxtl;
			double sytld = (double)sytl;
			double sxtrd = (double)sxtr;
			double sytrd = (double)sytr;
			double sxbrd = (double)sxbr;
			double sybrd = (double)sybr;
			double sxbld = (double)sxbl;
			double sybld = (double)sybl;

			int dxtl = MathExtra.irint(sxtld*cos_theta - sytld*sin_theta);
			int dytl = MathExtra.irint(sxtld*sin_theta + sytld*cos_theta);
			int dxtr = MathExtra.irint(sxtrd*cos_theta - sytrd*sin_theta);
			int dytr = MathExtra.irint(sxtrd*sin_theta + sytrd*cos_theta);
			int dxbr = MathExtra.irint(sxbrd*cos_theta - sybrd*sin_theta);
			int dybr = MathExtra.irint(sxbrd*sin_theta + sybrd*cos_theta);
			int dxbl = MathExtra.irint(sxbld*cos_theta - sybld*sin_theta);
			int dybl = MathExtra.irint(sxbld*sin_theta + sybld*cos_theta);

			int dxl = Math.Min(dxtl,dxbl);
			int dxr = Math.Max(dxtr,dxbr);
			int dyt = Math.Min(dytl,dytr);
			int dyb = Math.Max(dybl,dybr);

			int super_xl = Math.Min(sxtl,dxl);
			int super_xr = Math.Max(sxbr,dxr);
			int super_yt = Math.Min(sytl,dyt);
			int super_yb = Math.Max(sybr,dyb);

			int dx_org = super_xl;
			int dy_org = super_yt;
			int dx_dim = super_xr - super_xl +1;
			int dy_dim = super_yb - super_yt +1; 

			if (dx_org<0) 
			{
				Console.Error.WriteLine("Warning: Current rotate code can not cope with x-origin of " + dx_org.ToString());
				Console.Error.WriteLine(" => shifting origin back onto page.");
				dx_org = 0;
			}

			if (dy_org<0) 
			{
				Console.Error.WriteLine("Warning: Current rotate code can not cope with y-origin of " + dy_org.ToString());
				Console.Error.WriteLine(" => shifting origin back onto page.");
				dy_org = 0;
			}

			/*  int sx_org = MathExtra.irint((double)src_subbitmap.x_org()*(1.0+sin_theta));
				int sy_org = MathExtra.irint((double)src_subbitmap.y_org()*(1.0+cos_theta));
				int sx_dim = src_subbitmap.x_dim();
				int sy_dim = src_subbitmap.y_dim();
				int dx_dim = sx_dim + MathExtra.irint((double)sx_dim*sin_theta);
				int dy_dim = sy_dim + MathExtra.irint((double)sy_dim*Math.Abs(cos_theta)); */

			SubBitmap dst_subbitmap 
				= new SubBitmap(dx_org,dy_org,dx_dim,dy_dim); // no need to clear

			if (MathExtra.sign_bit(grad))
				rotate_bitmap_top_right(src_subbitmap,cos_theta,fabs_sin_theta,
					dst_subbitmap,circle,cbase,crange);
			else
				rotate_bitmap_bot_left(src_subbitmap,cos_theta,fabs_sin_theta,
					dst_subbitmap,circle,cbase,crange);

			return dst_subbitmap;
		}

		public GObject rotate_gobject(GObject src_gobject, 
			double grad,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			double hyp = Math.Sqrt(1.0+MathExtra.sqr(grad));
			double cos_theta = 1.0/hyp;
			double sin_theta = grad/hyp;
			double fabs_sin_theta = Math.Abs(sin_theta);

			int sxtl = src_gobject.x_left();
			int sytl = src_gobject.y_top();
			int sxtr = src_gobject.x_right();
			int sytr = src_gobject.y_top();
			int sxbr = src_gobject.x_right();
			int sybr = src_gobject.y_bot();
			int sxbl = src_gobject.x_left();
			int sybl = src_gobject.y_bot();

			double sxtld = (double)sxtl;
			double sytld = (double)sytl;
			double sxtrd = (double)sxtr;
			double sytrd = (double)sytr;
			double sxbrd = (double)sxbr;
			double sybrd = (double)sybr;
			double sxbld = (double)sxbl;
			double sybld = (double)sybl;

			int dxtl = MathExtra.irint(sxtld*cos_theta - sytld*sin_theta);
			int dytl = MathExtra.irint(sxtld*sin_theta + sytld*cos_theta);
			int dxtr = MathExtra.irint(sxtrd*cos_theta - sytrd*sin_theta);
			int dytr = MathExtra.irint(sxtrd*sin_theta + sytrd*cos_theta);
			int dxbr = MathExtra.irint(sxbrd*cos_theta - sybrd*sin_theta);
			int dybr = MathExtra.irint(sxbrd*sin_theta + sybrd*cos_theta);
			int dxbl = MathExtra.irint(sxbld*cos_theta - sybld*sin_theta);
			int dybl = MathExtra.irint(sxbld*sin_theta + sybld*cos_theta);

			int dxl = Math.Min(dxtl,dxbl);
			int dxr = Math.Max(dxtr,dxbr);
			int dyt = Math.Min(dytl,dytr);
			int dyb = Math.Max(dybl,dybr);

			int super_xl = Math.Min(sxtl,dxl);
			int super_xr = Math.Max(sxbr,dxr);
			int super_yt = Math.Min(sytl,dyt);
			int super_yb = Math.Max(sybr,dyb);

			int dx_org = super_xl;
			int dy_org = super_yt;
			int dx_dim = super_xr - super_xl +1;
			int dy_dim = super_yb - super_yt +1; 

			if (dx_org<0) 
			{
				Console.Error.WriteLine("Warning: Current rotate code can not cope with x-origin of " + dx_org.ToString());
				Console.Error.WriteLine(" => shifting origin back onto page.");
				dx_org = 0;
			}

			if (dy_org<0) 
			{
				Console.Error.WriteLine("Warning: Current rotate code can not cope with y-origin of " + dy_org.ToString());
				Console.Error.WriteLine(" => shifting origin back onto page.");
				dy_org = 0;
			}

			/*  int sx_org = MathExtra.irint((double)src_gobject.x_org()*(1.0+sin_theta));
				int sy_org = MathExtra.irint((double)src_gobject.y_org()*(1.0+cos_theta));
				int sx_dim = src_gobject.x_dim();
				int sy_dim = src_gobject.y_dim();
				int dx_dim = sx_dim + MathExtra.irint((double)sx_dim*sin_theta);
				int dy_dim = sy_dim + MathExtra.irint((double)sy_dim*Math.Abs(cos_theta)); */

			GObject dst_gobject 
				= new GObject(dx_org,dy_org,dx_dim,dy_dim); // no need to clear

			if (MathExtra.sign_bit(grad))
				rotate_bitmap_top_right(src_gobject,cos_theta,fabs_sin_theta,
					dst_gobject,circle,cbase,crange);
			else
				rotate_bitmap_bot_left(src_gobject,cos_theta,fabs_sin_theta,
					dst_gobject,circle,cbase,crange);

			return dst_gobject;
		}



		/*---------------------------------------------------------------------------*/

  
		protected void shear_bitmap_bot_left(Bitmap src_bitmap,
			double grad,
			Bitmap dst_bitmap,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			int dy;

			// grad is +ve
			int sx_dim = src_bitmap.x_dim();
			int sy_dim = src_bitmap.y_dim();
			int dx_dim = dst_bitmap.x_dim();
			int dy_dim = dst_bitmap.y_dim();

			double area = (double)(dx_dim*dy_dim);
			int mod_step = MathExtra.irint(area/crange);
			if (mod_step==0) mod_step = 1;
			double comp_add = crange*(double)mod_step/area;
			double computed = (double)cbase;
			int dc = 0;

			int delta_x = dx_dim;
			int delta_y = MathExtra.irint((double)dx_dim*grad);
			int incr_e  = 2*delta_y;
			int incr_ne = 2*(delta_y - delta_x);
			int start_diff = 2*delta_y - delta_x;

			int isx_start=0;
			int isy_start=sy_dim-1;

			for (dy=sy_dim-1; dy>=0; dy--)
			{
				int isx = isx_start;
				int isy = isy_start;
				int diff=start_diff;
				for (int dx=0; dx<dx_dim; dx++)
				{
					if (((++dc)%mod_step)==0)
					{
						if (circle!=null) 
						{
							computed += comp_add;
							circle.computed_so_far(MathExtra.irint(computed));
						}
					}

					if ((isx<0) || (isy<0) || (isx>=sx_dim) || (isy>=sy_dim))
						dst_bitmap.clear_pixel(dx,dy); // clear rest of bitmap line
					else dst_bitmap[dy,dx]=src_bitmap[isy,isx];
	  
					if (diff <=0)
					{
						diff += incr_e;
						isx++;
					}
					else
					{
						diff += incr_ne;
						isx++;
						isy--;
					}
				}
				isy_start --;
			}

			isy_start=sy_dim;

			for (dy=sy_dim; dy<dy_dim; dy++)
			{
				int isx = isx_start;
				int isy = isy_start;
				int diff = start_diff;
				for (int dx=0; dx<dx_dim; dx++)
				{
					if (((++dc)%mod_step)==0)
					{
						if (circle!=null) 
						{
							computed += comp_add;
							circle.computed_so_far(MathExtra.irint(computed));
						}
					}

					if ((isx<0) || (isy<0) || (isx>=sx_dim) || (isy>=sy_dim))
						dst_bitmap.clear_pixel(dx,dy); // clear rest of bitmap line
					else dst_bitmap[dy,dx]=src_bitmap[isy,isx];

					if (diff <=0)
					{
						diff += incr_e;
						isx++;
					}
					else
					{
						diff += incr_ne;
						isx++;
						isy--;
					}
				}
				isy_start ++;
			}
		}


		protected void shear_bitmap_bot_left(Bitmap src_bitmap,
			double grad,
			Bitmap dst_bitmap,
			ComputingCircle circle)
		{
		  shear_bitmap_bot_left(src_bitmap,grad,
					  dst_bitmap,circle,0,100.0);
		}



		protected void shear_bitmap_top_right(Bitmap src_bitmap,
			double grad,
			Bitmap dst_bitmap,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			int dy;

			// grad is -ve
			int sx_dim = src_bitmap.x_dim();
			int sy_dim = src_bitmap.y_dim();
			int dx_dim = dst_bitmap.x_dim();
			int dy_dim = dst_bitmap.y_dim();

			double area = (double)(dx_dim*dy_dim);
			int mod_step = MathExtra.irint(area/crange);
			if (mod_step==0) mod_step = 1;
			double comp_add = crange*(double)mod_step/area;
			double computed = (double)cbase;
			int dc = 0;

			int delta_x = dx_dim;
			int delta_y = -MathExtra.irint((double)dx_dim*grad);
			int incr_e  = 2*delta_y;
			int incr_se = 2*(delta_y - delta_x);
			int start_diff = 2*delta_y - delta_x;

			int isx_start=0;
			int isy_start=0;

			for (dy=dy_dim-sy_dim; dy<dy_dim; dy++)
			{
				int isx = isx_start;
				int isy = isy_start;
				int diff = start_diff;
				for (int dx=0; dx<dx_dim; dx++)
				{
					if (((++dc)%mod_step)==0)
					{
						if (circle!=null) 
						{
							computed += comp_add;
							circle.computed_so_far(MathExtra.irint(computed));
						}
					}

					if ((isx<0) || (isy<0) || (isx>=sx_dim) || (isy>=sy_dim))
						dst_bitmap.clear_pixel(dx,dy); // clear rest of bitmap line
					else dst_bitmap[dy,dx]=src_bitmap[isy,isx];

					if (diff <=0)
					{
						diff += incr_e;
						isx++;
					}
					else
					{
						diff += incr_se;
						isx++;
						isy++;
					}
				}
				isy_start ++;
			}

			isy_start= -1;

			for (dy=dy_dim-sy_dim-1; dy>=0; dy--)
			{
				int isx = isx_start;
				int isy = isy_start;
				int diff = start_diff;
				for (int dx=0; dx<dx_dim; dx++)
				{
					if (((++dc)%mod_step)==0)
					{
						if (circle!=null) 
						{
							computed += comp_add;
							circle.computed_so_far(MathExtra.irint(computed));
						}
					}

					if ((isx<0) || (isy<0) || (isx>=sx_dim) || (isy>=sy_dim))
						dst_bitmap.clear_pixel(dx,dy); // clear rest of bitmap line
					else dst_bitmap[dy,dx]=src_bitmap[isy,isx];

					if (diff <=0)
					{
						diff += incr_e;
						isx++;
					}
					else
					{
						diff += incr_se;
						isx++;
						isy++;
					}
				}
				isy_start --;
			}
		}


		protected void shear_bitmap_top_right(Bitmap src_bitmap,
			double grad,
			Bitmap dst_bitmap,
			ComputingCircle circle)
		{
		  shear_bitmap_top_right(src_bitmap,grad,
					  dst_bitmap,circle,0,100.0);
		}



		/*---------------------*/



		public Bitmap shear_bitmap(Bitmap src_bitmap,
			double grad,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			int sx_dim = src_bitmap.x_dim();
			int sy_dim = src_bitmap.y_dim();
			int y_extra = MathExtra.irint((double)sx_dim*Math.Abs(grad));
			int dx_dim = sx_dim;
			int dy_dim = sy_dim + y_extra;

			Bitmap dst_bitmap = new Bitmap(dx_dim,dy_dim);//  no need to clear

			if (MathExtra.sign_bit(grad)) 
				shear_bitmap_top_right(src_bitmap,grad,dst_bitmap,circle,cbase,crange);
			else 
				shear_bitmap_bot_left(src_bitmap,grad,dst_bitmap,circle,cbase,crange);

			return dst_bitmap;
		}

		public Bitmap shear_bitmap(Bitmap src_bitmap,
			double grad,
			ComputingCircle circle)
		{
			return shear_bitmap(src_bitmap,grad,circle,0,100.0);
		}

		public SubBitmap shear_subbitmap(SubBitmap src_subbitmap, 
			double grad,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			int sx_org = src_subbitmap.x_org();
			int sy_org = src_subbitmap.y_org() + MathExtra.irint((double)sx_org*Math.Abs(grad));
			int sx_dim = src_subbitmap.x_dim();
			int sy_dim = src_subbitmap.y_dim();

			int y_extra_dim = MathExtra.irint((double)sx_dim*Math.Abs(grad));
			int dx_dim = sx_dim;
			int dy_dim = sy_dim + y_extra_dim;

			SubBitmap dst_subbitmap 
				= new SubBitmap(sx_org,sy_org,dx_dim,dy_dim); // no need to clear

			if (MathExtra.sign_bit(grad)) 
				shear_bitmap_top_right(src_subbitmap,grad,dst_subbitmap,
					circle,cbase,crange);
			else 
				shear_bitmap_bot_left(src_subbitmap,grad,dst_subbitmap,
					circle,cbase,crange);

			return dst_subbitmap;
		}

		public GObject shear_gobject(GObject src_gobject, 
			double grad,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			int sx_org = src_gobject.x_org();
			int sy_org = src_gobject.y_org() + MathExtra.irint((double)sx_org*Math.Abs(grad));
			int sx_dim = src_gobject.x_dim();
			int sy_dim = src_gobject.y_dim();

			int y_extra_dim = MathExtra.irint((double)sx_dim*Math.Abs(grad));
			int dx_dim = sx_dim;
			int dy_dim = sy_dim + y_extra_dim;

			GObject dst_gobject 
				= new GObject(sx_org,sy_org,dx_dim,dy_dim); // no need to clear

			if (MathExtra.sign_bit(grad)) 
				shear_bitmap_top_right(src_gobject,grad,dst_gobject,circle,cbase,crange);
			else 
				shear_bitmap_bot_left(src_gobject,grad,dst_gobject,circle,cbase,crange);

			return dst_gobject;
		}



		/*---------------------------------------------------------------------------*/


  
		protected void shear_bitmap_bot_left8(Bitmap src_bitmap,
			double grad,
			Bitmap dst_bitmap,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			int dy;

			// grad is +ve
			ByteArray2d src_array = src_bitmap.array();
			ByteArray2d dst_array = dst_bitmap.array();
			int sx_off = src_bitmap.x_off();

			int sx_dim = src_bitmap.x_dim();
			int sy_dim = src_bitmap.y_dim();
			int dx_dim = dst_bitmap.x_dim();
			int dy_dim = dst_bitmap.y_dim();

			double area = (double)(dx_dim*dy_dim);
			int mod_step = MathExtra.irint(area/crange);
			if (mod_step==0) mod_step = 1;
			double comp_add = crange*(double)mod_step/area;
			double computed = (double)cbase;
			int dc = 0;

			int delta_x = dx_dim;
			int delta_y = MathExtra.irint((double)dx_dim*grad);
			int incr_e  = 2*delta_y;
			int incr_ne = 2*(delta_y - delta_x);
			int start_diff = 2*delta_y - delta_x;

			int isx_start=0;
			int isy_start=sy_dim-1;

			for (dy=sy_dim-1; dy>=0; dy--)
			{
				int isx = isx_start; // same as dx
				int isy = isy_start;
				int sax=(isx_start+sx_off)>>3; if (sx_off!=0) sax++;
				int dax=(isx_start+sx_off)>>3; if (sx_off!=0) dax++; // same as sax
				int diff=start_diff;
				int dx=0;
				while (dx<dx_dim)
				{
					if ((sx_off+isx)%8==0)
					{
						if (diff+incr_e*7 <= 0) // do byte op
						{
							if ((isy<0) || (isy>=sy_dim))
								dst_array[dy,dax]=0;
							else
								dst_array[dy,dax]=src_array[isy,sax];

							diff += (8*incr_e);
							dx+=8;  isx+=8;
							dax++; sax++;

							if (diff>0)
							{
								diff += incr_ne;
								isy--;
							}

							if (dx<dx_dim) dc+=8;
							else dc+=(sx_off+dx_dim)%8;

							if (((dc)%mod_step)<8)
							{
								if (circle!=null) 
								{
									computed = cbase + (comp_add*(dc/mod_step));
									circle.computed_so_far(MathExtra.irint(computed));
								}
							}
							continue;
						}
						else
						{
							dax++; sax++;
						}
					}
					// do one bit
	  
					if (((++dc)%mod_step)==0)
					{
						if (circle!=null) 
						{
							computed += comp_add;
							circle.computed_so_far(MathExtra.irint(computed));
						}
					}
	  
					if ((isy<0) || (isy>=sy_dim))
						dst_bitmap.clear_pixel(dx,dy); // clear rest of bitmap line
					else dst_bitmap[dy,dx]=src_bitmap[isy,isx];
	  
					if (diff <=0)
					{
						diff += incr_e;
						isx++; dx++;
					}
					else
					{
						diff += incr_ne;
						isx++; dx++;
						isy--;
					}
				}
				isy_start --;
			}

			isy_start=sy_dim;

			for (dy=sy_dim; dy<dy_dim; dy++)
			{
				int isx = isx_start; // same as dx
				int isy = isy_start;
				int sax=(isx_start+sx_off)>>3;  if (sx_off!=0) sax++;
				int dax=(isx_start+sx_off)>>3;  if (sx_off!=0) dax++; // same as sax
				int diff=start_diff;
				int dx=0;
				while (dx<dx_dim)
				{
					if ((sx_off+isx)%8==0)
					{
						if (diff+incr_e*7 <= 0) // do byte op
						{
							if ((isy<0) || (isy>=sy_dim))
								dst_array[dy,dax]=0;
							else
								dst_array[dy,dax]=src_array[isy,sax];
		  
							diff += (8*incr_e);
							dx+=8; isx+=8;
							dax++; sax++;

							if (diff>0)
							{
								diff += incr_ne;
								isy--;
							} 

							if (dx<dx_dim) dc+=8;
							else dc+=(sx_off+dx_dim)%8;

							if (((dc)%mod_step)<8)
							{
								if (circle!=null) 
								{
									computed = cbase + (comp_add*(dc/mod_step));
									circle.computed_so_far(MathExtra.irint(computed));
								}
							}
							continue;
						}
						else
						{
							dax++; sax++;
						}
					}
					// do one bit

					if (((++dc)%mod_step)==0)
					{
						if (circle!=null) 
						{
							computed += comp_add;
							circle.computed_so_far(MathExtra.irint(computed));
						}
					}

					if ((isy<0) || (isy>=sy_dim))
						dst_bitmap.clear_pixel(dx,dy); // clear rest of bitmap line
					else dst_bitmap[dy,dx]=src_bitmap[isy,isx];

					if (diff <=0)
					{
						diff += incr_e;
						isx++; dx++;
					}
					else
					{
						diff += incr_ne;
						isx++; dx++;
						isy--;
					}
				}
				isy_start ++;
			} 
		}


		protected void shear_bitmap_bot_left8(Bitmap src_bitmap,
			double grad,
			Bitmap dst_bitmap,
			ComputingCircle circle)
		{
		  shear_bitmap_bot_left8(src_bitmap,grad,
					 dst_bitmap,circle,0,100.0);
		}



  
		protected void shear_bitmap_top_right8(Bitmap src_bitmap,
			double grad,
			Bitmap dst_bitmap,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			int dy;

			// grad is +ve
			ByteArray2d src_array = src_bitmap.array();
			ByteArray2d dst_array = dst_bitmap.array();
			int sx_off = src_bitmap.x_off();

			int sx_dim = src_bitmap.x_dim();
			int sy_dim = src_bitmap.y_dim();
			int dx_dim = dst_bitmap.x_dim();
			int dy_dim = dst_bitmap.y_dim();

			double area = (double)(dx_dim*dy_dim);
			int mod_step = MathExtra.irint(area/crange);
			if (mod_step==0) mod_step = 1;
			double comp_add = crange*(double)mod_step/area;
			double computed = (double)cbase;
			int dc = 0;

			int delta_x = dx_dim;
			int delta_y = -MathExtra.irint((double)dx_dim*grad);
			int incr_e  = 2*delta_y;
			int incr_ne = 2*(delta_y - delta_x);
			int start_diff = 2*delta_y - delta_x;

			int isx_start=0;
			int isy_start=0;

			for (dy=dy_dim-sy_dim; dy<dy_dim; dy++)
			{
				int isx = isx_start; // same as dx
				int isy = isy_start;
				int sax=(isx_start+sx_off)>>3; if (sx_off!=0) sax++;
				int dax=(isx_start+sx_off)>>3; if (sx_off!=0) dax++; // same as sax
				int diff=start_diff;
				int dx=0;
				while (dx<dx_dim)
				{
					if ((sx_off+isx)%8==0)
					{
						if (diff+incr_e*7 <= 0) // do byte op
						{
							if ((isy<0) || (isy>=sy_dim))
								dst_array[dy,dax]=0;
							else
								dst_array[dy,dax]=src_array[isy,sax];

							diff += (8*incr_e);
							dx+=8;  isx+=8;
							dax++; sax++;

							if (diff>0)
							{
								diff += incr_ne;
								isy++;
							}

							if (dx<dx_dim) dc+=8;
							else dc+=(sx_off+dx_dim)%8;

							if (((dc)%mod_step)<8)
							{
								if (circle!=null) 
								{
									computed = cbase + (comp_add*(dc/mod_step));
									circle.computed_so_far(MathExtra.irint(computed));
								}
							}
							continue;
						}
						else
						{
							dax++; sax++;
						}
					}
					// do one bit
	  
					if (((++dc)%mod_step)==0)
					{
						if (circle!=null) 
						{
							computed += comp_add;
							circle.computed_so_far(MathExtra.irint(computed));
						}
					}
	  
					if ((isy<0) || (isy>=sy_dim))
						dst_bitmap.clear_pixel(dx,dy); // clear rest of bitmap line
					else dst_bitmap[dy,dx]=src_bitmap[isy,isx];
	  
					if (diff <=0)
					{
						diff += incr_e;
						isx++; dx++;
					}
					else
					{
						diff += incr_ne;
						isx++; dx++;
						isy++;
					}
				}
				isy_start++;
			}

			isy_start=-1;

			for (dy=dy_dim-sy_dim-1; dy>=0; dy--)
			{
				int isx = isx_start; // same as dx
				int isy = isy_start;
				int sax=(isx_start+sx_off)>>3;  if (sx_off!=0) sax++;
				int dax=(isx_start+sx_off)>>3;  if (sx_off!=0) dax++; // same as sax
				int diff=start_diff;
				int dx=0;
				while (dx<dx_dim)
				{
					if ((sx_off+isx)%8==0)
					{
						if (diff+incr_e*7 <= 0) // do byte op
						{
							if ((isy<0) || (isy>=sy_dim))
								dst_array[dy,dax]=0;
							else
								dst_array[dy,dax]=src_array[isy,sax];
		  
							diff += (8*incr_e);
							dx+=8; isx+=8;
							dax++; sax++;

							if (diff>0)
							{
								diff += incr_ne;
								isy++;
							} 

							if (dx<dx_dim) dc+=8;
							else dc+=(sx_off+dx_dim)%8;

							if (((dc)%mod_step)<8)
							{
								if (circle!=null) 
								{
									computed = cbase + (comp_add*(dc/mod_step));
									circle.computed_so_far(MathExtra.irint(computed));
								}
							}
							continue;
						}
						else
						{
							dax++; sax++;
						}
					}
					// do one bit

					if (((++dc)%mod_step)==0)
					{
						if (circle!=null) 
						{
							computed += comp_add;
							circle.computed_so_far(MathExtra.irint(computed));
						}
					}

					if ((isy<0) || (isy>=sy_dim))
						dst_bitmap.clear_pixel(dx,dy); // clear rest of bitmap line
					else dst_bitmap[dy,dx]=src_bitmap[isy,isx];

					if (diff <=0)
					{
						diff += incr_e;
						isx++; dx++;
					}
					else
					{
						diff += incr_ne;
						isx++; dx++;
						isy++;
					}
				}
				isy_start--;
			} 
		}

		protected void shear_bitmap_top_right8(Bitmap src_bitmap,
			double grad,
			Bitmap dst_bitmap,
			ComputingCircle circle)
		{
		  shear_bitmap_top_right8(src_bitmap,grad,
					  dst_bitmap,circle,0,100.0);
		}



		/*---------------------*/



		public Bitmap shear_bitmap8(Bitmap src_bitmap,
			double grad,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			int sx_dim = src_bitmap.x_dim();
			int sy_dim = src_bitmap.y_dim();
			int y_extra = MathExtra.irint((double)sx_dim*Math.Abs(grad));
			int dx_dim = sx_dim;
			int dy_dim = sy_dim + y_extra;

			Bitmap dst_bitmap = new Bitmap(dx_dim,dy_dim);//  no need to clear

			if (MathExtra.sign_bit(grad)) 
				shear_bitmap_top_right8(src_bitmap,grad,dst_bitmap,circle,cbase,crange);
			else 
				shear_bitmap_bot_left8(src_bitmap,grad,dst_bitmap,circle,cbase,crange);

			return dst_bitmap;
		}

		public Bitmap shear_bitmap8(Bitmap src_bitmap,
			double grad,
			ComputingCircle circle)
		{
			return shear_bitmap8(src_bitmap,grad,circle,0,100.0);
		}

		public SubBitmap shear_subbitmap8(SubBitmap src_subbitmap, 
			double grad,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			int sx_org = src_subbitmap.x_org();
			int sy_org = src_subbitmap.y_org();
			int sx_dim = src_subbitmap.x_dim();
			int sy_dim = src_subbitmap.y_dim();
			int y_extra = MathExtra.irint((double)sx_dim*Math.Abs(grad));
			int dx_dim = sx_dim;
			int dy_dim = sy_dim + y_extra;

			SubBitmap dst_subbitmap 
				= new SubBitmap(sx_org,sy_org,dx_dim,dy_dim); // no need to clear

			if (MathExtra.sign_bit(grad)) 
				shear_bitmap_top_right8(src_subbitmap,grad,dst_subbitmap,
					circle,cbase,crange);
			else 
				shear_bitmap_bot_left8(src_subbitmap,grad,dst_subbitmap,
					circle,cbase,crange);

			return dst_subbitmap;
		}

		public GObject shear_gobject8(GObject src_gobject, 
			double grad,
			ComputingCircle circle,
			int cbase, 
			double crange)
		{
			int sx_org = src_gobject.x_org();
			int sy_org = src_gobject.y_org();
			int sx_dim = src_gobject.x_dim();
			int sy_dim = src_gobject.y_dim();
			int y_extra = MathExtra.irint((double)sx_dim*Math.Abs(grad));
			int dx_dim = sx_dim;
			int dy_dim = sy_dim + y_extra;

			GObject dst_gobject 
				= new GObject(sx_org,sy_org,dx_dim,dy_dim); // no need to clear

			if (MathExtra.sign_bit(grad)) 
				shear_bitmap_top_right8(src_gobject,grad,dst_gobject,circle,cbase,crange);
			else 
				shear_bitmap_bot_left8(src_gobject,grad,dst_gobject,circle,cbase,crange);

			return dst_gobject;
		}


	}
}
      
