using System;
using System.IO;
using System.Diagnostics;

namespace Espresso.OMR
{
	class LocalFileAccess : InputFileAccess
	{
		Process process_;
		BinaryReader brin_;

		bool using_stdin_;

		public LocalFileAccess()
		{
			process_ = null;
			brin_ = null;

			using_stdin_ = false;
		}

		~LocalFileAccess()
		{
			close_file();
		}

		public override bool file_exists(string filename)
		{
			return File.Exists(filename);
		}

		protected BinaryReader open_decompress_pipe(string filename)
		{
			process_ = new Process();
			process_.StartInfo.UseShellExecute = false;
			process_.StartInfo.RedirectStandardOutput = true;
			process_.StartInfo.FileName = @"c:\tmp\gzip\gzip.exe";
			process_.StartInfo.Arguments = "--decompress --stdout " + filename;
			process_.Start();

			Stream base_stream = process_.StandardOutput.BaseStream;

			brin_ = new BinaryReader(base_stream);
			return brin_;
		}
      
		protected void close_pipe()
		{
			process_.Close();
			process_ = null;
		}

		protected bool has_compressed_ending(string filename)
		{
			return (filename.EndsWith(".gz") || filename.EndsWith(".z") ||
					filename.EndsWith(".Z"));
		}

		public override BinaryReader open_file(string filename)
		{
			if (filename=="") 
			{
				Console.Error.WriteLine("Warning: LocalFileAccess.open_file() untested with reading from standard in!");
				Stream base_stream = Console.OpenStandardInput();
				brin_ = new BinaryReader(base_stream);
				using_stdin_ = true;
			}
			else 
			{
				if (has_compressed_ending(filename))
				{
					// pipe open
					brin_ = open_decompress_pipe(filename);
				}
				else
				{
					FileStream fsin = new FileStream(filename, FileMode.Open, System.IO.FileAccess.Read);
					brin_ = new BinaryReader(fsin);
				}
			}

			return brin_;
	  
		}

		public override void close_file()
		{
			if (using_stdin_)
			{
				// don't need to close anything
			}
			else 
			{
				if (brin_ != null)
				{
					brin_.Close();
					brin_ = null;
				}

				if (process_ != null)
				{
					close_pipe();
				}
			}
		}
	}

}
