using System;
using System.IO;
using System.Diagnostics;
using System.Net;

namespace Espresso.OMR
{
	class RemoteFileAccess : InputFileAccess
	{
		WebRequest web_request_; 
		WebResponse web_response_; 

		BinaryReader brin_;

		public RemoteFileAccess()
		{
			web_request_ = null;
			web_response_ = null;

			brin_ = null;
		}

		~RemoteFileAccess()
		{
			close_file();
		}

		public override bool file_exists(string url)
		{
			try 
			{
				WebRequest request = WebRequest.Create(url); 
				WebResponse response = request.GetResponse();
				return true;
			}
			catch (Exception)
			{
				return false;
			}
		}

		public override BinaryReader open_file(string url)
		{
			if (url=="") 
			{
				Console.Error.WriteLine("RemoteFileAccess.open_file() from standard in has no meaning!");
				throw new Exception("Opening remote standard in makes no sense!");
			}
			else 
			{
				bool has_compressed_ending = false;
				if (has_compressed_ending)
				{
					// consider downloading first in compressed form, and then running local gzip pipe over it
					brin_ = null;
				}
				else
				{
					web_request_ = WebRequest.Create(url); 
					web_response_ = web_request_.GetResponse();
			
					Stream responseStream = web_response_.GetResponseStream();

					brin_ = new BinaryReader(responseStream);
				}
			}

			return brin_;
	  
		}

		public override void close_file()
		{
			if (brin_ != null)
			{
				brin_.Close();
				brin_ = null;
			}
		}
	}
}
