
using System;

namespace Espresso.OMR
{


  
	class DeformData
	{
		protected IntArray2d count_;
		protected int max_value_;
      
		protected Bitmap thick_line_;
		protected Bitmap thin_line_;
		protected Bitmap fthin_line_;
		protected Bitmap efthin_line_ = null;
		protected IntArray1d transform_;
		protected IntArray1d smooth_transform_;
      
		protected int x_start_;
		protected int x_end_;
		protected int y_start_;
		protected int y_end_;
      
		protected int y_deform_;


		public DeformData()
		{
			count_ = new IntArray2d();
			max_value_ = 0;

			thick_line_ = new Bitmap();
			thin_line_ = new Bitmap();
			fthin_line_ = new Bitmap();

			x_start_ = 0;
			x_end_ = 0;
			y_start_ = 0;
			y_end_ = 0;
			y_deform_ = 0;

			transform_ = new IntArray1d();
			smooth_transform_ = new IntArray1d();
		}

		public void reset()
		{
			count_.resize(0,0);
			thick_line_.resize(0,0);
			thin_line_.resize(0,0);
			fthin_line_.resize(0,0);
			efthin_line_.resize(0,0);
			transform_.resize(0);
			smooth_transform_.resize(0);

			max_value_=0;
			x_start_=0;
			x_end_=0;
			y_start_=0;
			y_end_=0;
			y_deform_=0;
		}

		public void generate_count(StaffSystemList ssl,int max_x_dim)
		{
			int y;

			// find max dimensions for lines

			int max_y_dim = 0;

			StaffSystemNode current_ss = ssl.head();
			while(current_ss!=null)
			{
				StaffList stl = current_ss.val().staffs();
				StaffNode current_st = stl.head();
				while (current_st!=null)
				{
					Staff staff = current_st.val();
					int no_staff_lines = staff.no_staff_lines();
					StaffLines staff_lines = staff.staff_lines();
					SubBitmapArray1d bitmaps = staff_lines.bitmaps();
					for (int i=0; i<no_staff_lines; i++)
					{
						int y_dim = bitmaps[i].y_dim();
						if (y_dim>max_y_dim) max_y_dim = y_dim;
					}
					current_st = stl.next();
				}
				current_ss = ssl.next();
			}

			count_.resize(max_x_dim,max_y_dim);

			for (y=0; y<max_y_dim; y++)
				for (int x=0; x<max_x_dim; x++)
					count_[y,x]=0;

			// inlay all staff lines
			current_ss = ssl.head();
			while(current_ss!=null)
			{
				StaffList stl = current_ss.val().staffs();
				StaffNode current_st = stl.head();
				while (current_st!=null)
				{
					Staff staff = current_st.val();
					int no_staff_lines = staff.no_staff_lines();
					StaffLines staff_lines = staff.staff_lines();
					SubBitmapArray1d bitmaps = staff_lines.bitmaps();
					for (int i=0; i<no_staff_lines; i++)
					{
						SubBitmap bitmap = bitmaps[i];
						int sx_dim = bitmap.x_dim();
						int sy_dim = bitmap.y_dim();
						int sx_org = bitmap.x_org();

						for (int sy=0; sy<sy_dim; sy++)
							for (int sx=0,dx=sx_org; sx<sx_dim; sx++,dx++)		  
								if (bitmap[sy,sx]==Pixel.PixelEnum.set) 
								{
									count_[sy,dx]++;
									if (count_[sy,dx]>max_value_) max_value_++;
								} 
					}
					current_st = stl.next();
				}
				current_ss = ssl.next();
			}
		}


		public void generate_thick_line(double max_acc_thresh)
		{
			int y;

			int max_x_dim = count_.x_dim();
			int max_y_dim = count_.y_dim();
			Bitmap high_count = new Bitmap(max_x_dim,max_y_dim);

			for (y=0; y<max_y_dim; y++)
				for (int x=0; x<max_x_dim; x++)
					high_count.clear_pixel(x,y);
  
			int min_x=0,max_x=0;
			int min_y=0,max_y=0;
			bool min_max_invalid = true;

			// commmon staff line threshold
			int csl_thresh = MathExtra.irint((double)max_value_*max_acc_thresh); 
  
			for (y=0; y<max_y_dim; y++)
				for (int x=0; x<max_x_dim; x++)
				{
					if (count_[y,x]==csl_thresh)
					{
						high_count.set_pixel(x,y);
						if (min_max_invalid)
						{
							min_x = max_x = x;
							min_y = max_y = y;
							min_max_invalid = false;
						}
						else
						{
							if (x<min_x) min_x = x;
							if (x>max_x) max_x = x;
							if (y<min_y) min_y = y;
							if (y>max_y) max_y = y;
						}
					}
				}
	  
			x_start_ = min_x;
			x_end_ = max_x;
 
			int reduced_x_dim = max_x - min_x +1;
			int reduced_y_dim = max_y - min_y +1;

			thick_line_.resize(max_x_dim,reduced_y_dim);
			thick_line_.extract(high_count,0,min_y,max_x_dim,reduced_y_dim);
		}



		public void generate_thin_line_basic()
		{
			int y;

			int max_x_dim = thick_line_.x_dim();
			int max_y_dim = thick_line_.y_dim();
			Bitmap mid_line = new Bitmap(max_x_dim,max_y_dim);

			for (y=0; y<max_y_dim; y++)
				for (int x=0; x<max_x_dim; x++)
					mid_line.clear_pixel(x,y);

			int min_y=0,max_y=0;
			bool min_max_invalid = true;

			for (int x=x_start_; x<=x_end_; x++)
			{
				int y_start = max_y_dim; // initially invalid
				for (y=0; y<max_y_dim; y++)
					if (thick_line_[y,x]==Bitmap.PixelEnum.set)
					{
						y_start = y;
						break;
					}

				int y_end = -1; // initially invalid
				for (y=max_y_dim-1; y>=0; y--)
					if (thick_line_[y,x]==Pixel.PixelEnum.set)
					{
						y_end = y;
						break;
					}

				if (y_start<=y_end)
				{
					// got segment
					int y_mid = (y_start + y_end)/2; // rounds up
					mid_line.set_pixel(x,y_mid);

					if (min_max_invalid)
					{
						min_y = max_y = y_mid;
						min_max_invalid = false;
					}
					else
					{
						if (y<min_y) min_y = y;
						if (y>max_y) max_y = y;
					}
				}
			}
	  
			int reduced_y_dim = max_y - min_y +1;

			thin_line_.resize(max_x_dim,reduced_y_dim);
			thin_line_.extract(mid_line,0,min_y,max_x_dim,reduced_y_dim);
		}

		public void fill_in_gaps()
		{
			int y;

			int max_x_dim = thin_line_.x_dim();
			int max_y_dim = thin_line_.y_dim();

			fthin_line_.resize(max_x_dim,max_y_dim);
			fthin_line_.extract(thin_line_,0,0,max_x_dim,max_y_dim);

			int x = x_start_;

			while (x<=x_end_)
			{
				// look for start of gap
				int x_start=x_start_, y_start=y_start_;
				int x_end=x_end_, y_end=y_end_;

				bool match = false;

				for (y=0; y<max_y_dim; y++)
					if (fthin_line_[y,x]==Pixel.PixelEnum.set)
					{
						x_start = x;
						y_start = y;
						match=true;
						break;
					}

				x++;
				if (match) continue;

				while (x<=x_end_)
				{
					match = false;
					for (y=0; y<max_y_dim; y++)
						if (fthin_line_[y,x]==Pixel.PixelEnum.set)
						{
							x_end = x;
							y_end = y;
							match=true;
							break;
						}

					x++;
					if (!match) continue;
					x--;

					// fill in gap
					BresenLine bline = new BresenLine(x_start,y_start,x_end,y_end);
					int blen = bline.length();
					for (int i=0; i<blen; i++)
					{
						Point pt = bline.data(i);
						fthin_line_.set_pixel(pt.x(),pt.y());
					}
					break;
				}
			}
		}
	  

		public void extend_ends_angle()
		{
			int x,y;

			int max_x_dim = fthin_line_.x_dim();
			int max_y_dim = fthin_line_.y_dim();

			// left side
			// pick up start of line
			for (y=0; y<max_y_dim; y++)
				if (fthin_line_[y,x_start_]==Pixel.PixelEnum.set) break; //safe

			y_start_ = y;

			// track how far it extends horizontally
			for (x=x_start_+1; x<=x_end_; x++)
				if (fthin_line_[y_start_,x]==Pixel.PixelEnum.clear) break; // safe

			int lx_step = x - x_start_;
			int lgrad_sign
				= (y_start_>=1) 
				? (fthin_line_[y_start_-1,x]==Pixel.PixelEnum.set) ? 1 : -1  // safe
															: 0;
			int ly_step = ((x_start_-1)/lx_step)+1; // +ve

			// right side
			// pick up end of line
			for (y=0; y<max_y_dim; y++)
				if (fthin_line_[y,x_end_]==Pixel.PixelEnum.set) break; //safe

			y_end_ = y;

			// track how far it extends horizontally
			for (x=x_end_-1; x>=0; x--)
				if (fthin_line_[y_end_,x]==Pixel.PixelEnum.clear) break; // safe

			int rx_step = x_end_ - x;
			int rgrad_sign
				= (y_end_<max_y_dim-1)
				? (fthin_line_[y_end_+1,x]==Pixel.PixelEnum.set) ? 1 : -1  // safe
														  : 0; 
			int ry_step = ((max_x_dim-x_end_-2)/rx_step)+1; // +ve

			// extend bitmap to fit worst-case extended deform
			int ef_top=0;
			int ef_bot=max_y_dim;

			if (lgrad_sign<=0) ef_top = -ly_step;
			else ef_bot = max_y_dim + ly_step;

			if ((rgrad_sign>0) && (-ry_step<ef_top))
				ef_top = -ry_step;
			if ((rgrad_sign<=0) && (max_y_dim+ry_step>ef_bot)) 
				ef_bot = max_y_dim + ry_step;

			int ef_y_offset = -ef_top;

			int ef_x_dim = max_x_dim;
			int ef_y_dim = ef_bot - ef_top +1;
			efthin_line_.resize(ef_x_dim,ef_y_dim);

			for (y=0; y<ef_y_dim; y++)
				for (x=0; x<ef_x_dim; x++)
					efthin_line_.clear_pixel(x,y);

			efthin_line_.inlay_bitmap(fthin_line_,0,ef_y_offset);

			y_start_ += ef_y_offset;
			y_end_ += ef_y_offset;

			int lx1 = 0;
			int ly1 = (lgrad_sign>0) ? y_start_+ly_step-1: y_start_-ly_step+1;
			int lx2 = x_start_;
			int ly2 = y_start_;

			BresenLine lline = new BresenLine(lx1,ly1,lx2,ly2);
			int llen = lline.length();
			for (int li=0; li<llen; li++)
			{
				Point pt = lline.data(li);
				efthin_line_.set_pixel(pt.x(),pt.y());
			}

			int rx1 = ef_x_dim;
			int ry1 = (rgrad_sign<0) ? y_end_+ry_step-1 : y_end_-ry_step+1;
			int rx2 = x_end_;
			int ry2 = y_end_;

			BresenLine rline = new BresenLine(rx1,ry1,rx2,ry2);
			int rlen = rline.length();
			for (int ri=0; ri<rlen; ri++)
			{
				Point pt = rline.data(ri);
				efthin_line_.set_pixel(pt.x(),pt.y());
			}  
		}


		public void extend_ends_level()
		{
			int x,y;

			int max_x_dim = fthin_line_.x_dim();
			int max_y_dim = fthin_line_.y_dim();

			// left side
			// pick up start of line
			for (y=0; y<max_y_dim; y++)
				if (fthin_line_[y,x_start_]==Pixel.PixelEnum.set) break; //safe

			y_start_ = y;

			// right side
			// pick up end of line
			for (y=0; y<max_y_dim; y++)
				if (fthin_line_[y,x_end_]==Pixel.PixelEnum.set) break; //safe

			y_end_ = y;

			int ef_x_dim = max_x_dim; // for name sake
			int ef_y_dim = max_y_dim; // for name sake
			efthin_line_.resize(ef_x_dim,ef_y_dim);

			efthin_line_.clear_all();
			efthin_line_.inlay_bitmap(fthin_line_,0,0);

			// extend bitmap to fit worst-case extended deform
			for (x=0; x<x_start_; x++)
				efthin_line_.set_pixel(x,y_start_);

			for (x=x_end_+1; x<max_x_dim; x++)
				efthin_line_.set_pixel(x,y_end_);
		}

		public void generate_transform()
		{
			int ef_x_dim = efthin_line_.x_dim();
			int ef_y_dim = efthin_line_.y_dim();

			transform_.resize(ef_x_dim);

			for (int x=0; x<ef_x_dim; x++)
				for (int y=0; y<ef_y_dim; y++)
					if (efthin_line_[y,x]==Pixel.PixelEnum.set) // safe
													 {
														 transform_[x]=y;
														 break;
													 }

			y_deform_ = ef_y_dim;
		}

		public void generate_smooth_transform()
		{
			int tlen = transform_.length();
			smooth_transform_.resize(tlen);

			smooth_transform_[0]=transform_[0];
			for (int x=1; x<tlen-1; x++)
			{
				int total=0;
				for (int l=-1; l<=1; l++)
					total+=transform_[x+l]; //safe
				int avg = MathExtra.irint((double)total/3.0);
				smooth_transform_[x]=avg;
			}
			smooth_transform_[tlen-1]=smooth_transform_[tlen-1];
		}



		public Bitmap thick_line() { return thick_line_; }
		public Bitmap thin_line() { return thin_line_; }
		public Bitmap fthin_line() { return fthin_line_; }
		public Bitmap efthin_line() { return efthin_line_; }
		public IntArray1d transform() { return transform_; }
		public IntArray1d smooth_transform() { return smooth_transform_; }

		public int y_deform() { return y_deform_; }
	}
}
