
using System;

namespace Espresso.OMR
{

  class LineInfo
    {
      protected double grad_;
      protected int delta_x_;
      protected int delta_y_;

      protected int x_left_;
      protected int x_right_;

      protected int y_top_;
      protected int y_bot_;

      protected int x_midl_;
      protected int x_midr_;

      protected int y_topl_;
      protected int y_botl_;
      protected int y_topr_;
      protected int y_botr_;

      protected int avg_thick_;

      public LineInfo(int ss_x_dim,
	       StaffSeg left_staff_seg, StaffSeg right_staff_seg,
	       int avg_thick, int pos)
	{
	  int lx_mid = (left_staff_seg.x_start()+left_staff_seg.x_end())/2;
	  int rx_mid = (right_staff_seg.x_start()+right_staff_seg.x_end())/2;

	  StaffRegionArray1d left_lines = left_staff_seg.lines();
	  StaffRegionArray1d right_lines = right_staff_seg.lines();

	  int ly_top = left_lines[pos].top_edge();
	  int ry_top = right_lines[pos].top_edge();
	  int local_ave_thickness 
	    = MathExtra.irint((double)left_lines[pos].thickness()
			      +right_lines[pos].thickness()/2.0);

	  double delta_x = (double)(lx_mid - rx_mid);
	  double delta_y = (double)(ry_top - ly_top);
	  grad_ = delta_y/delta_x;

	  x_left_ = 0;
	  x_right_ = ss_x_dim-1;

	  // delta calcs used to subtract 1; // ****
	  // believe this to be an error => removed

	  delta_x_ = ss_x_dim;
	  delta_y_ = MathExtra.irint(ss_x_dim*Math.Abs(grad_));

	  y_top_ = ly_top+MathExtra.irint(lx_mid*grad_);
	  y_bot_ = y_top_ + local_ave_thickness -1;

	  // wobble info 

	  x_midl_ = (left_staff_seg.x_start()+left_staff_seg.x_end())/2;
	  x_midr_ = (right_staff_seg.x_start()+right_staff_seg.x_end())/2;

	  y_topl_ = left_lines[pos].top_edge();
	  y_botl_ = y_topl_ + left_lines[pos].thickness();

	  y_topr_ = right_lines[pos].top_edge();
	  y_botr_ = y_topr_ + right_lines[pos].thickness();

	  // general
	  avg_thick_ = avg_thick;
	}

      public double grad() { return grad_; }
      public int delta_x() { return delta_x_; }
      public int delta_y() { return delta_y_; }

      public int x_left() { return x_left_; }
      public int x_right() { return x_right_; }
      public int y_top() { return y_top_; }
      public int y_bot() { return y_bot_; }

      public int x_midl() { return x_midl_; }
      public int x_midr() { return x_midr_; }
      public int y_topl() { return y_topl_; }
      public int y_botl() { return y_botl_; }
      public int y_topr() { return y_topr_; }
      public int y_botr() { return y_botr_; }

      public int avg_thickness() { return avg_thick_; }

    }
}
