
using System;

namespace Espresso.OMR
{
	class MObjects : BaseMObjects
	{

		public enum GcharactersModeEnum 
		{ 
		        gc_ignore, gc_ocr, gc_definitive 
		}
		public enum FloodFillModeEnum 
		{
			flood_fill_basic, flood_fill_2bit,
			flood_fill_byte, contour_trace, ff_definitive };
		public enum StaffAreasModeEnum 
		{
			sa_locate_4_5, sa_locate_3_4, sa_locate_6_10,
			sg_min_gap_4, sg_min_gap_6, 
			sg_min_gap_12, sg_min_gap_30,
			sa_hproject, sa_hpeaks, 
			sa_staffs, sa_non_staffs, sa_definitive};
		public enum StaffSegModeEnum   
		{
			seg_locate_1_2_3, seg_locate_1_3_4,
			seg_locate_1_4_5, seg_locate_2_0_0,
			seg_hproject, seg_staffs, 
			seg_non_staffs, seg_definitive };
		public enum StaffLinesModeEnum   
		{
			sl_locate_strict, sl_locate_average,
			sl_locate_loose, sl_hproject, 
			sl_staffs, sl_non_staffs, sl_definitive };
		public enum CorrectSkewModeEnum 
		{
			rotate_page,    shear_page,    shear_page8,
			rotate_systems, shear_systems, shear_systems8,
			rotate_lines,   shear_lines,   shear_lines8 };
		public enum RemoveStaffsModeEnum 
		{
			remove_straight_prerau, 
			remove_straight_clarke, 
			remove_wobble, 
			remove_track, remove_track_wide,
			remove_wobble_simple, remove_track_simple,
			remove_wobble_prerau, remove_track_prerau,
			remove_wobble_circle, remove_track_circle,
			cross_straight, cross_wobble, cross_track,
			cross_wobble_circle, cross_track_circle};
		public enum CorrectDeformModeEnum 
		{
			deform_level_60, deform_level_70,
			deform_level_80, deform_angle_60,
			deform_angle_70, deform_angle_80 };

		public enum StaffSystemsFoundEnum { not_found, by_bit, by_byte };


		static bool definitive_compare_ = false;

		BitmapOps bitmap_ops_;
		OcrManager ocr_manager_;

		Bitmap corrected_page_;

		int staff_system_thresh_;

		GCharList corrected_gcharacters_ = null;
		StaffSystemList corrected_staff_systems_;
		GObjectList corrected_non_staff_systems_;

		IntArray1d title_hproject_;
		BoolArray1d title_is_vspan_;
		VSpanList title_vspans_;

		IntArray1d abcrop_hproject_;

		DeformData deform_;

		StaffSystemsFoundEnum staff_systems_found_;

		public MObjects(PageManager page_manager,int x_dim, int y_dim)
			: base(page_manager,x_dim,y_dim)
		{	  
			bitmap_ops_ = new BitmapOps(x_dim,y_dim);
			ocr_manager_ = new OcrManager();

			corrected_page_ = null;
			corrected_staff_systems_ = new StaffSystemList();
			corrected_non_staff_systems_ = new GObjectList();

			staff_system_thresh_ = 0;

			deform_ = new DeformData();

			staff_systems_found_ = StaffSystemsFoundEnum.not_found;
			title_hproject_ = new IntArray1d(y_dim);
			title_is_vspan_ = new BoolArray1d(y_dim);
			title_vspans_ = new VSpanList();

			abcrop_hproject_ = new IntArray1d(y_dim);
		}

		public MObjects(PageManager page_manager)
			: this(page_manager,0,0)
		{}


		public override void resize(int x_dim, int y_dim)
		{
			base.resize(x_dim,y_dim);

			title_hproject_.resize(y_dim);
			title_is_vspan_.resize(y_dim);

			abcrop_hproject_.resize(y_dim);

			if ((x_dim>0) && (y_dim>0))
				bitmap_ops_.resize(x_dim,y_dim);

			if (ocr_manager_.template_filename()!=ShareGlobals.ocr_template_filename_)
				ocr_manager_.reload_templates(ShareGlobals.ocr_template_filename_);

			// infact bitmap_ops.scratch() is the bitmap needed
			ocr_manager_.new_page(page_manager_.current_fbitmap()); // ***

			staff_system_thresh_ = page_manager_.dpi()/2; // hw // *****
		}

		public override void reset_and_resize(int x_dim, int y_dim)
		{
			base.reset_and_resize(x_dim,y_dim);
			corrected_staff_systems_.reset();
			corrected_non_staff_systems_.reset();

			corrected_page_=null;

			title_hproject_.resize(y_dim);
			title_is_vspan_.resize(y_dim);
			title_vspans_.reset();

			abcrop_hproject_.resize(y_dim);

			if ((x_dim>0) && (y_dim>0))
				bitmap_ops_.resize(x_dim,y_dim);

			if (ocr_manager_.template_filename()!=ShareGlobals.ocr_template_filename_)
				ocr_manager_.reload_templates(ShareGlobals.ocr_template_filename_);

			// infact bitmap_ops.scratch() is the bitmap needed
			ocr_manager_.new_page(page_manager_.current_fbitmap()); // ***

			staff_system_thresh_ = page_manager_.dpi()/2; // hw // *****
		}

		public void reset_and_resize_dsep(int x_dim, int y_dim)
		{
			corrected_staff_systems_.reset();
			corrected_non_staff_systems_.reset();

			corrected_page_=null;

			title_hproject_.resize(y_dim);
			title_is_vspan_.resize(y_dim);
			title_vspans_.reset();

			abcrop_hproject_.resize(y_dim);

			if ((x_dim>0) && (y_dim>0))
				bitmap_ops_.resize(x_dim,y_dim);

			if (ocr_manager_.template_filename()!=ShareGlobals.ocr_template_filename_)
				ocr_manager_.reload_templates(ShareGlobals.ocr_template_filename_);

			// infact bitmap_ops.scratch() is the bitmap needed
			ocr_manager_.new_page(page_manager_.current_fbitmap()); // ***

			staff_system_thresh_ = page_manager_.dpi()/2; // hw // *****
		}

		public void reset_and_resize_sep(int x_dim, int y_dim)
		{
			corrected_staff_systems_.reset();
			corrected_non_staff_systems_.reset();

			corrected_page_=null;

			title_hproject_.resize(y_dim);
			title_is_vspan_.resize(y_dim);
			title_vspans_.reset();

			abcrop_hproject_.resize(y_dim);

			if ((x_dim>0) && (y_dim>0))
				bitmap_ops_.resize(x_dim,y_dim);

			staff_system_thresh_ = page_manager_.dpi()/2; // hw // *****
		}

		public void reset_and_resize_prim(int x_dim, int y_dim)
		{
			// for future compatability!
		}

		/*-----------------*/


		public static void set_definitive_compare(bool def_comp)
		{
			definitive_compare_ = def_comp;
		}

		public static bool definitive_compare()
		{
			return definitive_compare_;
		}

		/*-----------------*/
      
		public void insert_order_xy(GCharList list, GChar gchar)
		{
			int gc_x_org = gchar.x_org();
			int gc_y_org = gchar.y_org();
 
			GCharNode current_gc = list.head();
			while (current_gc!=null)
			{
				GChar existing_gchar = current_gc.val();
				int egc_x_org = existing_gchar.x_org();
				int egc_y_org = existing_gchar.y_org();
      
				if (egc_x_org - ShareGlobals.order_xy_gap_tol_ > gc_x_org) 
				{
					// gone too far (beyond)
					list.insert_before(gchar);
					break;
				}
				else if (egc_x_org +  ShareGlobals.order_xy_gap_tol_ > gc_x_org) 
				{
					// within 
					if (egc_y_org>gc_y_org)
					{
						// gone too far
						list.insert_before(gchar);
						break;
					}
				}
				current_gc = list.next();
			}
 
			if (current_gc==null) list.append(gchar);
		}


		public void insert_order_xy(StaffSystemList list, 
			StaffSystem staff_system)
		{
			GObject ss_gobject = staff_system.gobject();
			int ss_x_org = ss_gobject.x_org();
			int ss_y_org = ss_gobject.y_org();
 
			StaffSystemNode current_ss = list.head();
			while (current_ss!=null)
			{
				StaffSystem existing_staff_system = current_ss.val();
				GObject ess_gobject = existing_staff_system.gobject();
				int ess_x_org = ess_gobject.x_org();
				int ess_y_org = ess_gobject.y_org();
      
				if (ess_x_org - ShareGlobals.order_xy_gap_tol_ > ss_x_org) 
				{
					// gone too far (beyond)
					list.insert_before(staff_system);
					break;
				}
				else if (ess_x_org +  ShareGlobals.order_xy_gap_tol_ > ss_x_org) 
				{
					// within 
					if (ess_y_org>ss_y_org)
					{
						// gone too far
						list.insert_before(staff_system);
						break;
					}
				}

				current_ss = list.next();
			}
 
			if (current_ss==null) list.append(staff_system);
		}


		public void insert_order_xy(GObjectList list, GObject gobject)
		{
			int go_x_org = gobject.x_org();
			int go_y_org = gobject.y_org();
 
			GObjectNode current_go = list.head();
			while (current_go!=null)
			{
				GObject existing_gobject = current_go.val();
				int ego_x_org = existing_gobject.x_org();
				int ego_y_org = existing_gobject.y_org();
      
				if (ego_x_org - ShareGlobals.order_xy_gap_tol_ > go_x_org) 
				{
					// gone too far (beyond)
					list.insert_before(gobject);
					break;
				}
				else if (ego_x_org +  ShareGlobals.order_xy_gap_tol_ > go_x_org) 
				{
					// within 
					if (ego_y_org>go_y_org)
					{
						// gone too far
						list.insert_before(gobject);
						break;
					}
				}
				current_go = list.next();
			}
 
			if (current_go==null) list.append(gobject);
		}
 




		/*-----------------*/


		public int no_potential_staffs(StaffSystemList ssl)
		{
			int no_staffs = 0;

			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();
				no_staffs+=staff_system.potential_staffs().length();
				current_ss = ssl.next();
			}

			return no_staffs;
		}

		public int no_second_pass_staffs(StaffSystemList ssl)
		{
			int no_staffs = 0;

			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();
				no_staffs+=staff_system.second_pass_staffs().length();
				current_ss = ssl.next();
			}

			return no_staffs;
		}

		/*-----------------*/

		public void generate_gcharacters(Bitmap bitmap,GcharactersModeEnum mode,
			int verbose_level,
			ComputingCircle computing,
			GCharList gcl)
		{
			if (mode==GcharactersModeEnum.gc_ignore) return;

			Bitmap orig_bitmap = (Bitmap) bitmap; // internally not a
			Bitmap scratch = bitmap_ops_.scratch();

			ocr_manager_.new_page(scratch); // ensures working with correct bitmap
			ocr_manager_.link_page();

			int x_dim = orig_bitmap.x_dim();
			int y_dim = orig_bitmap.y_dim();

			int area = x_dim*y_dim;
			int mod_step = area/100;
  
			int computed = 0;
			int c = 0; 

			// search from bottom right to top left to catch for dublin ocr package
			// search of extra area above potential shape included for letters like
			// 'i' and 'j'

			for (int y=y_dim-1; y>=0; y--)
				for (int x=x_dim-1; x>=0; x--)
				{
					if (((++c)%mod_step)==0) 
						BaseIvLink.computed_so_far(computing,++computed);

					if (orig_bitmap[y,x]==Pixel.PixelEnum.set)
					{
						// flood fill
						GObject isolated_gobject 
							= bitmap_ops_.Flood.flood_fill_basic(orig_bitmap,x,y);

						int igo_x_org = isolated_gobject.x_org();
						int igo_y_org = isolated_gobject.y_org();
						int igo_x_dim = isolated_gobject.x_dim();
						int igo_y_dim = isolated_gobject.y_dim();

						if (OcrFunction.ocr_check_bitmap_size(igo_x_dim,igo_y_dim))
						{

							// The Dublin OCR package actually works with rect regions,
							// so rather than the flood-fill that isolates shapes
							// simulate a rectangle cut

							SubBitmap also_in_rect 
								= new SubBitmap(orig_bitmap,igo_x_org,igo_y_org,igo_x_dim,igo_y_dim);
							isolated_gobject.inlay_bitmap(also_in_rect,0,0);
							scratch.inlay_bitmap(isolated_gobject);

							// basic match
							OcrMatch best_match
								=ocr_manager_.match(igo_x_org,igo_y_org,igo_x_dim,igo_y_dim);
							int best_x_org = igo_x_org;
							int best_y_org = igo_y_org;
							int best_x_dim = igo_x_dim;
							int best_y_dim = igo_y_dim;

							if (best_match.score()<OcrFunction.ocr_get_no_score())
							{
								GObjectList extra_gobjects = new GObjectList(); // initialised to empty

								int exl = igo_x_org;
								int exr = igo_x_org + igo_x_dim -1;
								int eyt = igo_y_org - igo_y_dim/3;
								int eyb = igo_y_org -1;
		    
								if (eyt<0) eyt=0;
								if (eyb<0) eyb=0;

								// check area above
								for (int ey=eyb; ey>=eyt; ey--)
									for (int ex=exr; ex>=exl; ex--)
									{
										// flood fill
										if (orig_bitmap[ey,ex]==Pixel.PixelEnum.set)
										{
											GObject extra_gobject 
												= bitmap_ops_.Flood.flood_fill_basic(orig_bitmap,
												ex,ey);

											extra_gobjects.append(extra_gobject);

											int ego_x_org = extra_gobject.x_org();
											int ego_y_org = extra_gobject.y_org();
											int ego_x_dim = extra_gobject.x_dim();
											int ego_y_dim = extra_gobject.y_dim();

											// Not strictly correct, may inadvertently 
											// include extra set pixels.  This section 
											// could be rewritten to avoid this if it 
											// is percieved to be a problem

											int pxl = Math.Min(igo_x_org,ego_x_org);
											int pyt = Math.Min(igo_y_org,ego_y_org);
											int pxr 
												=Math.Max(igo_x_org+igo_x_dim-1,
												ego_x_org+ego_x_dim-1);
											int pyb 
												=Math.Max(igo_y_org+igo_y_dim-1,
												ego_y_org+ego_y_dim-1);
											int pxd = pxr - pxl +1;
											int pyd = pyb - pyt +1;
			      
											OcrMatch poss_match 
												= ocr_manager_.match(pxl,pyt,pxd,pyd);

											if (poss_match.score()<best_match.score())
											{
												// bitwise assignment : may have associated
												// memory leakage on string
												best_match = poss_match;
												best_x_org = pxl;
												best_y_org = pyt;
												best_x_dim = pxd;
												best_y_dim = pyd;
											}
			      
										} /* if */
									} /* for */

								// put all the extra flood filled subbitmaps back
								// The balancing operation of removing from scratch is
								// not required

								GObjectNode current_go = extra_gobjects.head();
								while (current_go!=null)
								{
									orig_bitmap.inlay_bitmap(current_go.val());
									current_go = extra_gobjects.next();
								}

							} /* if */

							int accept_match_thresh
								= OcrFunction.ocr_get_accept_score()+ShareGlobals.ocr_accept_match_add_;
							if (best_match.score()<accept_match_thresh)
							{
								// got a match 
								// => remove bitmap from scatch and add it into gchar list
		    
								string name = best_match.name();

								int score = best_match.score();
								double certainty // increase certainty fn? // ******
									= 1.0 - ((double)score)/(double)accept_match_thresh;

								byte[,] library_graymap
									= best_match.library_graymap();
								int lg_x_dim = best_match.lg_x_dim();
								int lg_y_dim = best_match.lg_y_dim();
		    
								GChar best_gchar 
									= new GChar(name,scratch,
									best_x_org,best_y_org,best_x_dim,best_y_dim,
									certainty,library_graymap,lg_x_dim,lg_y_dim,
									GChar.Operation.remove);
		    
								for (int by=best_y_org; by<best_y_org+best_y_dim; by++)
									for (int bx=best_x_org; bx<best_x_org+best_x_dim; bx++)
										orig_bitmap.clear_pixel(bx,by);

								// working backwards is required for checking for 
								// additional dots, but does not necessarily give
								// xy ordering

								gcl.prepend(best_gchar);
							}
		
						} /* if pixel==set */
					}
				}


			BaseIvLink.computed_so_far(computing,100);

			bitmap_ops_.exchange_buffer(orig_bitmap);

			ocr_manager_.unlink_page();
		}



		public void reset_gcharacters()
		{
			FileBitmap current_fbitmap = page_manager_.current_fbitmap();

			GCharNode current_gc = gcharacters_.head();
			while (current_gc!=null)
			{
				GChar gchar = current_gc.val();
				current_fbitmap.inlay_bitmap(gchar);

				current_gc = gcharacters_.next();
			}

			gcharacters_.reset();
			staff_systems_.reset();
			non_staff_systems_.reset();

			corrected_gcharacters_.reset();
			corrected_staff_systems_.reset();
			corrected_non_staff_systems_.reset();
		}


		/*-----------------*/

		protected override StaffSystem generate_empty_staff_system()
		{
			return new StaffSystem();
		}

		/*-----------------*/


		private bool is_staff_system(GObject gobject)
		{
			return ((gobject.x_dim()>staff_system_thresh_) 
				&& (gobject.y_dim()>10)); // change this to 13 ? //*****
		}


		public void generate_staff_systems(Bitmap bitmap, FloodFillModeEnum mode,
			int verbose_level,
			ComputingCircle computing,
			StaffSystemList ssl, GObjectList nssl)
		{
			Bitmap orig_bitmap = (Bitmap) bitmap; // internally not a
			switch (mode)
			{
				case FloodFillModeEnum.flood_fill_basic:
				case FloodFillModeEnum.flood_fill_2bit:
				case FloodFillModeEnum.contour_trace:
				{
					int x_dim = orig_bitmap.x_dim();
					int y_dim = orig_bitmap.y_dim();

					int area = x_dim*y_dim;
					int mod_step = area/100;
	
					int computed = 0;
					int c = 0; 
	
					for (int y=0; y<y_dim; y++)
						for (int x=0; x<x_dim; x++)
						{
							if (((++c)%mod_step)==0) 
								BaseIvLink.computed_so_far(computing,++computed);

							if (orig_bitmap[y,x]==Bitmap.PixelEnum.set)
							{
								GObject isolated_gobject = null;
								switch (mode)
								{
									case FloodFillModeEnum.flood_fill_basic:
										isolated_gobject
											=bitmap_ops_.Flood.flood_fill_basic(orig_bitmap,x,y);
										break;
		      
									case FloodFillModeEnum.flood_fill_2bit:
										isolated_gobject
											=bitmap_ops_.Flood.flood_fill_2bit(orig_bitmap,x,y);
										break;

									case FloodFillModeEnum.contour_trace:
										isolated_gobject
											=bitmap_ops_.Flood.contour_trace(orig_bitmap,x,y);
										break;
								}
								GObject isolated_gobject_ref = isolated_gobject;

								// responisbilty for gobject passed over to list
								if (is_staff_system(isolated_gobject_ref))
								{
									StaffSystem poss_staff_system
										= new StaffSystem(isolated_gobject_ref);

									poss_staff_system.generate_peaks();
									poss_staff_system.generate_sorted_peak_diff(verbose_level);

									ssl.append(poss_staff_system);
								}
								else 
								{
									nssl.append(isolated_gobject_ref);
								}

								if (FilenameManager.file_type()!=FilenameManager.FileTypeEnum.dsep)
								{
									// place in definitive non_ss list
									GObject isolated_gobject_clone 
										= new GObject(isolated_gobject_ref);
									def_non_staff_systems_.append(isolated_gobject_clone);
								}

							}
						}
					staff_systems_found_ = StaffSystemsFoundEnum.by_bit;
					break;
				}
      
				case FloodFillModeEnum.flood_fill_byte:
				{
					ByteArray2d orig_array = orig_bitmap.array();
					int x_dim = orig_array.x_dim();
					int y_dim = orig_array.y_dim();

					int area = x_dim*y_dim;
					int mod_step = area/100;
	
					int computed = 0;
					int c = 0; 
	
					for (int y=0; y<y_dim; y++)
						for (int x=0; x<x_dim; x++)
						{
							if (((++c)%mod_step)==0)
								BaseIvLink.computed_so_far(computing,++computed);

							if (orig_array[y,x]!=0)
							{
								GObject isolated_gobject
									=bitmap_ops_.Flood.flood_fill_byte(orig_array,x,y);

								// responisbilty for gobject passed over to list
								if (is_staff_system(isolated_gobject))
								{
									StaffSystem poss_staff_system
										= new StaffSystem(isolated_gobject);

									poss_staff_system.generate_peaks();
									poss_staff_system.generate_sorted_peak_diff(verbose_level);
		      
									ssl.append(poss_staff_system);
								}
								else 
								{
									nssl.append(isolated_gobject);
								}

								if (FilenameManager.file_type()!=FilenameManager.FileTypeEnum.dsep)
								{
									// place in definitive non_ss list
									GObject isolated_gobject_clone 
										= new GObject(isolated_gobject);
									def_non_staff_systems_.append(isolated_gobject_clone);
								}

							}
						}
					staff_systems_found_ = StaffSystemsFoundEnum.by_byte;
					break;
				}
      
			}    
			BaseIvLink.computed_so_far(computing,100);

			bitmap_ops_.exchange_buffer(orig_bitmap); // restore page back to orignal
		}


		public void reset_staff_systems()
		{
			staff_systems_.reset();
			non_staff_systems_.reset();

			corrected_staff_systems_.reset();
			corrected_non_staff_systems_.reset();
		}



		/*---------------------*/

		protected int num_staffs(StaffSystemList ssl)
		{
			int n = 0;
  
			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				StaffList staffs = staff_system.staffs();
      
				n+=staffs.length();
      
				ss_current = ssl.next();
			} 
  
			return n;
		}

		protected void calc_staff_info(StaffSystemList ssl, int num_staffs,
			int verbose_level)
		{
			IntArray1d staff_heights = new IntArray1d(num_staffs);
			IntArray1d gap_heights   = new IntArray1d(num_staffs);
			IntArray1d line_heights  = new IntArray1d(num_staffs);
			DoubleArray1d skews      = new DoubleArray1d(num_staffs);

			int n=0;
			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				StaffList staffs = staff_system.staffs();
				StaffNode st_current = staffs.head();

				while (st_current!=null)
				{
					StaffInfo info = st_current.val().info();

					staff_heights[n]=info.staff_height();
					gap_heights[n]=info.gap_height();
					line_heights[n]=info.line_height();
					skews[n]=info.skew();

					if (verbose_level>=2)
					{
						Console.Out.Write("global      : ");
						Console.Out.Write("staff height = " + staff_heights[n] + " ");
						Console.Out.Write("gap height = " + gap_heights[n] + " ");
						Console.Out.Write("line height = " + line_heights[n] + " ");

						Console.Out.Write("skew = ");
						if (!MathExtra.sign_bit(skews[n])) Console.Out.Write("+");
						Console.Out.WriteLine("{0:f4}",skews[n]);
	      	  
					}

					n++;
					st_current = staffs.next();
				}
				ss_current = ssl.next();
			} 

			Statistics.IntValDelegate extract_int 
				= new Statistics.IntValDelegate(Statistics.get_int);

			Statistics.DoubleValDelegate extract_double 
				= new Statistics.DoubleValDelegate(Statistics.get_double);

			double glob_staff_height 
				= Statistics.int_median(staff_heights.data(),extract_int);

			double glob_gap_height 
				= Statistics.int_median(gap_heights.data(),extract_int);

			double glob_line_height 
				= Statistics.int_median(line_heights.data(),extract_int);

			double glob_skew 
				= Statistics.double_mean(skews.data(),extract_double);

			staff_info_.set_staff_height(MathExtra.irint(glob_staff_height));
			staff_info_.set_gap_height(MathExtra.irint(glob_gap_height));
			staff_info_.set_line_height(MathExtra.irint(glob_line_height));
			staff_info_.set_skew(glob_skew);
		}

		protected void add_gobject_into_ordered_list(GObject new_go,
			GObjectList gol)
		{
			int ng_x_org = new_go.x_org();
			int ng_y_org = new_go.y_org();

			GObjectNode current_go = gol.head();
			while (current_go!=null)
			{
				GObject local_go = current_go.val();
				int lg_x_org = local_go.x_org();
				int lg_y_org = local_go.y_org();

				if ((ng_y_org<lg_y_org) && (ng_x_org<lg_x_org)) // check this *****
					current_go = gol.next();
				else
				{	  
					gol.insert_before((GObject)new_go); // better if insert_before
					break;
				}

				if (current_go==null) 
				{
					// identical to append, but more consistent
					gol.tail();
					gol.insert_after((GObject)new_go); // better if insert_after
				}
			}
		}

		protected void down_size_gobjects(GObjectList gobjects)
		{
			int go_len = gobjects.length();

			int c=0;
			GObjectNode current_go = gobjects.head();
			while (current_go!=null)
			{
				c++;
				GObject gobject = current_go.val();
				int x_org = gobject.x_org();
				int y_org = gobject.y_org();
				int x_dim = gobject.x_dim();
				int y_dim = gobject.y_dim();

				current_go = gobjects.detach_forward(); 

				for (int y=0; y<y_dim; y++)
					for (int x=0; x<x_dim; x++)
					{
						if (gobject[y,x]==Pixel.PixelEnum.set)
						{
							GObject new_gobject 
								= bitmap_ops_.Flood.flood_fill_basic(gobject,x,y); // origins ok
							if (c<go_len)
								gobjects.insert_before(new_gobject);
							else
								gobjects.append(new_gobject);
						}
					}

				bitmap_ops_.scratch().clear_area(x_org,y_org,x_dim,y_dim);

				gobject = null; // not strictly neccessary, but marks where delete used to be in C++ code
			}
		}


		/*---------------------*/

		public void generate_staff_areas(Bitmap current_page,
			StaffSystemList ssl,GObjectList nssl,
			int sg_min_gap,
			int sa_prop_numer, 
			int sa_prop_denom,
			int verbose_level,
			VerboseExtra verb_extra,
			ComputingCircle computing)
		{
			int ss_len = ssl.length();
            int computing_step = 100 / ((ss_len == 0) ? 1 : ss_len);
			int computed = 0;

			if (verbose_level>=6) 
				IvLink.app_graphics_clear_all(verb_extra.graphics());

			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				GObject gobject = staff_system.gobject();
				int ss_x_org = gobject.x_org();
				int ss_y_org = gobject.y_org();

				if (staff_system.generate_staff_areas(current_page,
					sg_min_gap,
					sa_prop_numer,sa_prop_denom,
					verbose_level,
					verb_extra,
					computing,computed,
					computing_step)==false)
				{
					SubBitmap ss_bitmap = staff_system.gobject();
					GObject new_gobject 
						= new GObject(current_page, 
						ss_bitmap.x_org(), ss_bitmap.y_org(),
						ss_bitmap.x_dim(), ss_bitmap.y_dim());
					int new_gobject_x_org = new_gobject.x_org();
					int new_gobject_y_org = new_gobject.y_org();

					ss_current = ssl.detach_forward();
	  
					// add into non_staff_systems in order
					bool inserted = false;
					GObjectNode go_current = nssl.head(); // check this
					while (go_current!=null) // skip along y
					{
						GObject gobject_nss = go_current.val();

						if (gobject_nss.y_org() == new_gobject_y_org)
						{
							while (go_current!=null) // skip along x
							{
								if (gobject_nss.x_org() >= new_gobject_x_org)
								{
									nssl.insert_before(new_gobject);
									inserted = true;
									break;
								}
								go_current = nssl.next();
							}
							if (inserted) break;
						}
						else if (gobject_nss.y_org() > new_gobject_y_org)
						{
							nssl.insert_before(new_gobject);
							inserted = true;
							break;
						}
						go_current = nssl.next();
					}

					if (!inserted) 
					{
						nssl.tail();
						nssl.insert_after(new_gobject);
					}
				} 
				else ss_current = ssl.next();

				computed+=computing_step;
			} 

			if (verbose_level>=6) 
			{
				IvLink.app_graphics_flush(verb_extra.graphics());
				BaseIvLink.warning_dialog(verb_extra.main_window(),
					"Showing All Staff Segments");
			}

			if (staff_systems_found_==StaffSystemsFoundEnum.by_byte) down_size_gobjects(nssl);

			BaseIvLink.computed_so_far(computing,100);
		}


		protected void reset_given_staff_areas(StaffSystemList staff_systems)
		{
			StaffSystemNode current = staff_systems.head();
			while (current!=null)
			{
				StaffSystem staff_system = current.val();
				staff_system.potential_staffs().reset();
				current = staff_systems.next();
			}
		}


		public void reset_staff_areas()
		{ 
			reset_given_staff_areas(staff_systems_); 
			reset_given_staff_areas(corrected_staff_systems_); 
		}

		/*---------------------*/

		public void generate_staff_segs_first_pass(Bitmap current_page,
			StaffSystemList ssl,GObjectList nssl,
			double upper_thresh_mul,
			double lower_thresh_strict_mul,
			double lower_thresh_hyster_mul,
			int verbose_level,
			VerboseExtra verb_extra,
			ComputingCircle computing)
		{
			int dpi = page_manager_.dpi();
			int ss_len = ssl.length();

            int computing_step = 100 / ((ss_len == 0) ? 1 : ss_len);
			int computed = 0;

			if (verbose_level>=6) 
				IvLink.app_graphics_clear_all(verb_extra.graphics());

			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				GObject gobject = staff_system.gobject();
				int ss_x_org = gobject.x_org();
				int ss_y_org = gobject.y_org();

				staff_system.generate_staff_segs_first_pass(current_page,
					upper_thresh_mul,
					lower_thresh_strict_mul,
					lower_thresh_hyster_mul,
					verbose_level,
					verb_extra,
					computing,computed,computing_step,
					dpi);
				ss_current = ssl.next();

				computed+=computing_step;
			} 

			if (verbose_level>=6) 
			{
				IvLink.app_graphics_flush(verb_extra.graphics());
				BaseIvLink.warning_dialog(verb_extra.main_window(),
					"Showing All Staff Segments");
			}

			if (staff_systems_found_==StaffSystemsFoundEnum.by_byte) down_size_gobjects(nssl);

			BaseIvLink.computed_so_far(computing,100);
		}

		public void generate_staff_lines(Bitmap current_page,
			StaffSystemList ssl,GObjectList nssl,
			double sl_lines_vt,
			double sl_gaps_vt,
			int sl_lines_rt,
			int sl_gaps_rt,
			int verbose_level,
			VerboseExtra verb_extra,
			ComputingCircle computing)
		{
			int dpi = page_manager_.dpi();
			int ss_len = ssl.length();

            int computing_step = 100 / ((ss_len == 0) ? 1 : ss_len);
			int computed = 0;

			if (verbose_level>=6) 
				IvLink.app_graphics_clear_all(verb_extra.graphics());

			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				GObject gobject = staff_system.gobject();
				int ss_x_org = gobject.x_org();
				int ss_y_org = gobject.y_org();

				staff_system.generate_staff_lines(current_page,
					sl_lines_vt,sl_gaps_vt,
					sl_lines_rt,sl_gaps_rt,
					verbose_level,
					verb_extra,
					computing,computed,
					computing_step);

				ss_current = ssl.next();

				computed+=computing_step;
			} 

			calc_staff_info(ssl,num_staffs(ssl),verbose_level);

			if (verbose_level>=6) 
			{
				IvLink.app_graphics_flush(verb_extra.graphics());
				BaseIvLink.warning_dialog(verb_extra.main_window(),
					"Showing All Staff Segments");
			}

			BaseIvLink.computed_so_far(computing,100);
		}


		public void generate_staff_segs_second_pass(Bitmap current_page,
			StaffSystemList ssl,
			double upper_thresh_mul,
			double lower_thresh_strict_mul,
			double lower_thresh_hyster_mul,
			int verbose_level,
			VerboseExtra verb_extra,
			ComputingCircle computing,
			StaffInfo glob_staff_info,
			int med_no_staff_lines)
		{
			int dpi = page_manager_.dpi();
			int ss_len = ssl.length();

            int computing_step = 100 / ((ss_len == 0) ? 1 : ss_len);
			int computed = 0;

			if (verbose_level>=6) 
				IvLink.app_graphics_clear_all(verb_extra.graphics());

			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				GObject gobject = staff_system.gobject();
				int ss_x_org = gobject.x_org();
				int ss_y_org = gobject.y_org();

				staff_system.generate_staff_segs_second_pass(current_page,
					upper_thresh_mul,
					lower_thresh_strict_mul,
					lower_thresh_hyster_mul,
					verbose_level,
					verb_extra,
					computing,computed,computing_step,
					dpi,glob_staff_info,med_no_staff_lines);

				ss_current = ssl.next();

				computed+=computing_step;
			} 

			if (verbose_level>=6) 
			{
				IvLink.app_graphics_flush(verb_extra.graphics());
				BaseIvLink.warning_dialog(verb_extra.main_window(),
					"Showing All Staff Segments");
			}

			BaseIvLink.computed_so_far(computing,100);
		}



		protected void reset_given_staff_segs(StaffSystemList staff_systems)
		{
			// fn identical to 'reset_given_staff_lines' // ******
			StaffSystemNode current = staff_systems.head();
			while (current!=null)
			{
				StaffSystem staff_system = current.val();
				staff_system.staffs().reset();
				staff_system.second_pass_staffs().reset();
				current = staff_systems.next();
			}
		}



		protected void reset_given_staff_lines(StaffSystemList staff_systems)
		{
			StaffSystemNode current = staff_systems.head();
			while (current!=null)
			{
				StaffSystem staff_system = current.val();
				staff_system.staffs().reset();
				staff_system.second_pass_staffs().reset();
				current = staff_systems.next();
			}
		}


		protected void reset_staff_segs()
		{ 
			// fn still identical
			reset_given_staff_segs(staff_systems_); 
			reset_given_staff_segs(corrected_staff_systems_); 
		}


		protected void reset_staff_lines()
		{ 
			reset_given_staff_lines(staff_systems_); 
			reset_given_staff_lines(corrected_staff_systems_); 
		}


		public Bitmap inlay_def_staff_lines()
		{
			Bitmap scratch = bitmap_ops_.scratch();
  
			scratch.clear_all(); // probably not necessary, but play it safe

			StaffSystemList dssl = def_staff_systems_;
			StaffSystemNode current_dss = dssl.head();

			while (current_dss!=null)
			{
				StaffSystem dss = current_dss.val();

				StaffList dstl = dss.staffs();
				StaffNode current_dst = dstl.head();

				while (current_dst!=null)
				{
					Staff dst = current_dst.val();

					StaffLines staff_lines = dst.staff_lines();
					SubBitmapArray1d sl_bitmaps = staff_lines.bitmaps();
					int no_st_lines = sl_bitmaps.length();

					for (int i=0; i<no_st_lines; i++)
					{
						SubBitmap staff_line = sl_bitmaps[i];
						scratch.inlay_bitmap(staff_line);
					}

					current_dst = dstl.next();
				}

				current_dss = dssl.next();
			}

			return scratch;
		}


		/*---------------------*/

		public void generate_corrected_skew(CorrectSkewModeEnum mode,
			CpuUsage rusage,
			int verbose_level,
			ComputingCircle circle)
		{
			Bitmap current_page = page_manager_.current_fbitmap();
			int total_area = 0;
			int total_extra_time_user = 0;

			switch (mode)
			{
				case CorrectSkewModeEnum.rotate_page:
					corrected_page_ 
						= bitmap_ops_.Rotate.rotate_bitmap(current_page,staff_info_.skew(),circle);
					total_area = current_page.x_dim()*current_page.y_dim();
					break;
      
				case CorrectSkewModeEnum.shear_page:
					corrected_page_ 
						= bitmap_ops_.Rotate.shear_bitmap(current_page,staff_info_.skew(),circle);
					total_area = current_page.x_dim()*current_page.y_dim();
					break;

				case CorrectSkewModeEnum.shear_page8:
					corrected_page_ 
						= bitmap_ops_.Rotate.shear_bitmap8(current_page,staff_info_.skew(),circle);
					total_area = current_page.x_dim()*current_page.y_dim();
					break;

				case CorrectSkewModeEnum.rotate_systems:
				case CorrectSkewModeEnum.shear_systems:
				case CorrectSkewModeEnum.shear_systems8:
				{
					int ss_len = staff_systems_.length();
					int nss_len = non_staff_systems_.length();
					double comp_rn=100.0/(double)(ss_len+nss_len);
					double comp=0.0; int icomp=0;
					StaffSystemNode ss_current = staff_systems_.head();
					while (ss_current!=null)
					{
						StaffSystem staff_system = ss_current.val();
						double grad = staff_system.staff_info().skew();
						GObject gobject = staff_system.gobject();

						switch (mode)
						{
							case CorrectSkewModeEnum.rotate_systems:
							{
								GObject cr_gobject
									=bitmap_ops_.Rotate.rotate_gobject(gobject,grad,circle,icomp,comp_rn);
								int user_time_p1=rusage.snapshot_timer();
								StaffSystem cr_staff_system = new StaffSystem(cr_gobject);
								cr_staff_system.generate_peaks();
								cr_staff_system.generate_sorted_peak_diff(verbose_level);
								corrected_staff_systems_.append(cr_staff_system);

								int user_time_p2=rusage.snapshot_timer();
								total_extra_time_user += (user_time_p2 - user_time_p1);

								break;
							}

							case CorrectSkewModeEnum.shear_systems:
							{
								GObject cs_gobject
									=bitmap_ops_.Rotate.shear_gobject(gobject,grad,circle,icomp,comp_rn);

								int user_time_p1=rusage.snapshot_timer();

								StaffSystem cs_staff_system = new StaffSystem(cs_gobject);
								cs_staff_system.generate_peaks();
								cs_staff_system.generate_sorted_peak_diff(verbose_level);
								corrected_staff_systems_.append(cs_staff_system);

								int user_time_p2=rusage.snapshot_timer();
								total_extra_time_user += (user_time_p2 - user_time_p1);

								break;
							}

							case CorrectSkewModeEnum.shear_systems8:
							{
								GObject cs_gobject8
									=bitmap_ops_.Rotate.shear_gobject8(gobject,grad,circle,icomp,comp_rn);

								int user_time_p1=rusage.snapshot_timer();

								StaffSystem cs_staff_system8 = new StaffSystem(cs_gobject8);
								cs_staff_system8.generate_peaks();
								cs_staff_system8.generate_sorted_peak_diff(verbose_level);
								corrected_staff_systems_.append(cs_staff_system8);

								int user_time_p2=rusage.snapshot_timer();
								total_extra_time_user += (user_time_p2 - user_time_p1);

								break;
							}
						}

						total_area += gobject.x_dim()*gobject.y_dim();

						comp += comp_rn;
						icomp = MathExtra.irint(comp);
						ss_current = staff_systems_.next();
					}

					GObjectNode nss_current = non_staff_systems_.head();
					while (nss_current!=null)
					{
						GObject non_ssystem = nss_current.val();
						double grad = staff_info_.skew(); // global skew

						switch (mode)
						{
							case CorrectSkewModeEnum.rotate_systems:
							{
								GObject cr_gobject
									=bitmap_ops_.Rotate.rotate_gobject(non_ssystem,grad,circle,
									icomp,comp_rn);
								corrected_non_staff_systems_.append(cr_gobject);
								break;
							} 

							case CorrectSkewModeEnum.shear_systems:
							{
								GObject cs_gobject
									=bitmap_ops_.Rotate.shear_gobject(non_ssystem,grad,circle,
									icomp,comp_rn);
								corrected_non_staff_systems_.append(cs_gobject);
								break;
							}

							case CorrectSkewModeEnum.shear_systems8:
							{
								GObject cs_gobject8
									=bitmap_ops_.Rotate.shear_gobject8(non_ssystem,grad,circle,
									icomp,comp_rn);
								corrected_non_staff_systems_.append(cs_gobject8);
								break;
							}
						}

						total_area += non_ssystem.x_dim()*non_ssystem.y_dim();

						comp += comp_rn;
						icomp = MathExtra.irint(comp);
						nss_current = non_staff_systems_.next();
					}
					break;
				}


				case CorrectSkewModeEnum.rotate_lines:
				case CorrectSkewModeEnum.shear_lines:
				case CorrectSkewModeEnum.shear_lines8:
				{
					Console.Error.Write("Warning: relocation of origins not guaranetted to be correct");
					Console.Error.WriteLine(" => needs attention");

					int cp_x_dim = current_page.x_dim();
					int cp_y_dim = current_page.y_dim();
					double grad = staff_info_.skew();

					int user_time_p1=rusage.snapshot_timer();
      
					int cr_x_dim, cr_y_dim; // safe
					bitmap_ops_.Rotate.calculate_rotate_dims(cp_x_dim,cp_y_dim,grad,
						out cr_x_dim, out cr_y_dim);

					if ((object)corrected_page_==null)
					{
						corrected_page_ = new Bitmap(cr_x_dim,cr_y_dim);
					}
					else
					{
						corrected_page_.resize(cr_x_dim,cr_y_dim);
					}
					corrected_page_.clear_all();

					int user_time_p2=rusage.snapshot_timer();
					total_extra_time_user += (user_time_p2 - user_time_p1);

					int ss_len = staff_systems_.length();
					int nss_len = non_staff_systems_.length();
					double comp_rn=100.0/(double)(ss_len+nss_len);
					double comp=0.0; int icomp=0;

					StaffSystemNode ss_current = staff_systems_.head();
					while (ss_current!=null)
					{
						StaffSystem staff_system = ss_current.val();
						double ss_grad = staff_system.staff_info().skew();

						StaffList stl = staff_system.staffs();
						GObjectList ssol = staff_system.processed_staff_system_objects();

						int no_staffs = stl.length();
						int no_staff_system_objects = ssol.length();
						int no_obs_to_rot = no_staffs+no_staff_system_objects;

						double comp_strn 
							= (no_obs_to_rot>0) ? comp_rn/(double)no_obs_to_rot : comp_rn;
	  
						StaffNode current_st = stl.head();
						while (current_st!=null)
						{
							Staff staff = current_st.val();
							double st_grad = staff.info().skew();

							StaffLines sl = staff.staff_lines();	      
							SubBitmapArray1d sl_bitmaps = sl.bitmaps();
							int no_staff_lines = sl_bitmaps.length();
	      
							double comp_slrn 
								= (no_staff_lines>0) 
								? comp_strn/(double)(no_staff_lines) : comp_strn;

							for (int i=0; i<no_staff_lines; i++)
							{
								SubBitmap sl_bitmap = sl_bitmaps[i];
								switch (mode)
								{
									case CorrectSkewModeEnum.rotate_lines:
									{
										SubBitmap cr_sl_subbitmap
											=bitmap_ops_.Rotate.rotate_subbitmap(sl_bitmap,st_grad,
											circle,icomp,comp_strn);

										int rl_user_time_p1
											= rusage.snapshot_timer();

										corrected_page_.inlay_bitmap(cr_sl_subbitmap);
		      
										cr_sl_subbitmap = null; // marks C++ delete

										int rl_user_time_p2
											= rusage.snapshot_timer();
										total_extra_time_user 
											+= (rl_user_time_p2 - rl_user_time_p1);

										break;
									}

									case CorrectSkewModeEnum.shear_lines:
									{
										SubBitmap cs_sl_subbitmap
											=bitmap_ops_.Rotate.shear_subbitmap(sl_bitmap,st_grad,
											circle,icomp,comp_strn);

										int sl_user_time_p1 = rusage.snapshot_timer();

										corrected_page_.inlay_bitmap(cs_sl_subbitmap);

										cs_sl_subbitmap = null; // marks C++ delete

										int sl_user_time_p2 = rusage.snapshot_timer();
										total_extra_time_user += (sl_user_time_p2 - sl_user_time_p1);

										break;
									}

									case CorrectSkewModeEnum.shear_lines8:
									{
										SubBitmap cs_sl_subbitmap8
											=bitmap_ops_.Rotate.shear_subbitmap8(sl_bitmap,st_grad,
											circle,icomp,comp_strn);

										int sl_user_time_p1 = rusage.snapshot_timer();

										corrected_page_.inlay_bitmap(cs_sl_subbitmap8);

										cs_sl_subbitmap8 = null; // marks C++ delete

										int sl_user_time_p2 = rusage.snapshot_timer();
										total_extra_time_user += (sl_user_time_p2 - sl_user_time_p1);

										break;
									}

					    
								}

								total_area += sl_bitmap.x_dim()*sl_bitmap.y_dim();

							} // end of for loop;

							comp += comp_strn;
							icomp = MathExtra.irint(comp);

							current_st = stl.next();
						}

						GObjectNode sso_current = ssol.head();

						while (sso_current!=null)
						{
							GObject sso = sso_current.val();
							   
							switch (mode)
							{
								case CorrectSkewModeEnum.rotate_lines:
								{
									GObject cr_gobject
										=bitmap_ops_.Rotate.rotate_gobject(sso,ss_grad,circle,
										icomp,comp_rn);

									int rl_user_time_p1 
										= rusage.snapshot_timer();

									corrected_page_.inlay_bitmap(cr_gobject);

									cr_gobject = null; // marks C++ delete

									int rl_user_time_p2
										= rusage.snapshot_timer();
									total_extra_time_user 
										+= (rl_user_time_p2 - rl_user_time_p1);

									break;
								} 

								case CorrectSkewModeEnum.shear_lines:
								{
									GObject cs_gobject
										=bitmap_ops_.Rotate.shear_gobject(sso,ss_grad,circle,
										icomp,comp_rn);

									int sl_user_time_p1 
										= rusage.snapshot_timer();

									corrected_page_.inlay_bitmap(cs_gobject);

									cs_gobject = null; // marks C++ delete

									int sl_user_time_p2
										= rusage.snapshot_timer();
									total_extra_time_user 
										+= (sl_user_time_p2 - sl_user_time_p1);

									break;
								}

								case CorrectSkewModeEnum.shear_lines8:
								{
									GObject cs_gobject8
										=bitmap_ops_.Rotate.shear_gobject8(sso,ss_grad,circle,
										icomp,comp_rn);

									int sl_user_time_p1 
										= rusage.snapshot_timer();

									corrected_page_.inlay_bitmap(cs_gobject8);

									cs_gobject8 = null; // marks C++ delete

									int sl_user_time_p2
										= rusage.snapshot_timer();
									total_extra_time_user 
										+= (sl_user_time_p2 - sl_user_time_p1);

									break;
								}
							} // end of switch;

							total_area += sso.x_dim()*sso.y_dim();

							comp += comp_strn;
							icomp = MathExtra.irint(comp);
							sso_current = ssol.next();
						} // end of while;


						comp += comp_rn;
						icomp = MathExtra.irint(comp);
						ss_current = staff_systems_.next();
					}
      
					GObjectNode nss_current = non_staff_systems_.head();
					while (nss_current!=null)
					{
						GObject non_ssystem = nss_current.val();
						double  nss_grad = staff_info_.skew(); // global skew

						switch (mode)
						{
							case CorrectSkewModeEnum.rotate_lines:
							{
								GObject cr_gobject
									=bitmap_ops_.Rotate.rotate_gobject(non_ssystem,nss_grad,circle,
									icomp,comp_rn);

								int rl_user_time_p1=rusage.snapshot_timer();

								corrected_page_.inlay_bitmap(cr_gobject);

								cr_gobject = null; // marks C++ delete

								int rl_user_time_p2=rusage.snapshot_timer();
								total_extra_time_user += (rl_user_time_p2 - rl_user_time_p1);

								break;
							} 

							case CorrectSkewModeEnum.shear_lines:
							{
								GObject cs_gobject
									=bitmap_ops_.Rotate.shear_gobject(non_ssystem,nss_grad,circle,
									icomp,comp_rn);

								int sl_user_time_p1=rusage.snapshot_timer();

								corrected_page_.inlay_bitmap(cs_gobject);

								cs_gobject = null; // marks C++ delete

								int sl_user_time_p2=rusage.snapshot_timer();
								total_extra_time_user += (sl_user_time_p2 - sl_user_time_p1);

								break;
							}

							case CorrectSkewModeEnum.shear_lines8:
							{
								GObject cs_gobject8
									=bitmap_ops_.Rotate.shear_gobject8(non_ssystem,nss_grad,circle,
									icomp,comp_rn);

								int sl_user_time_p1=rusage.snapshot_timer();

								corrected_page_.inlay_bitmap(cs_gobject8);

								cs_gobject8 = null; // marks C++ delete

								int sl_user_time_p2=rusage.snapshot_timer();
								total_extra_time_user += (sl_user_time_p2 - sl_user_time_p1);

								break;
							}
						} // end of switch;

						total_area += non_ssystem.x_dim()*non_ssystem.y_dim();

						comp += comp_rn;
						icomp = MathExtra.irint(comp);
						nss_current = non_staff_systems_.next();
					} // end of while;
					break;
				}
			} // end of mode switch;

			BaseIvLink.computed_so_far(circle,100);

			// resize bitmap_op bitmaps to largest size

			int sx_dim = current_page.x_dim();
			int sy_dim = current_page.y_dim();
			double cp_grad = staff_info_.skew();
			int dx_dim,dy_dim;
			bitmap_ops_.Rotate.calculate_rotate_dims(sx_dim,sy_dim,cp_grad,out dx_dim,out dy_dim);
			bitmap_ops_.resize(dx_dim,dy_dim);

			if (verbose_level>=2)
			{
				Console.Out.Write("total area for skew correction: ");
				Console.Out.WriteLine(total_area + " pixels");
				Console.Out.WriteLine("total extra user time = " + total_extra_time_user);
			}

		}

		void reset_corrected_skew()
		{
			corrected_page_=null;
 
			reset_given_staff_areas(corrected_staff_systems_);
		}


		/*---------------------------------------------------------------------------*/


		public void filter_staff_systems_with_no_staffs(Bitmap current_page,
			StaffSystemList ssl,
			GObjectList nssl)
		{

			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				GObject gobject = staff_system.gobject();
				int ss_x_org = gobject.x_org();
				int ss_y_org = gobject.y_org();

				if (staff_system.staffs().length()==0)
				{
					SubBitmap ss_bitmap = staff_system.gobject();
					GObject new_gobject 
						= new GObject(current_page, 
						ss_bitmap.x_org(), ss_bitmap.y_org(),
						ss_bitmap.x_dim(), ss_bitmap.y_dim());
					int new_gobject_x_org = new_gobject.x_org();
					int new_gobject_y_org = new_gobject.y_org();

					ss_current = ssl.detach_forward();
	  
					// add into non_staff_systems in order
					bool inserted = false;
					GObjectNode go_current = nssl.head(); // check this // ****
					while (go_current!=null) // skip along y
					{
						GObject ss_gobject = go_current.val();

						if (ss_gobject.y_org() == new_gobject_y_org)
						{
							while (go_current!=null) // skip along x
							{
								if (ss_gobject.x_org() >= new_gobject_x_org)
								{
									nssl.insert_before(new_gobject);
									inserted = true;
									break;
								}
								go_current = nssl.next();
							}
							if (inserted) break;
						}
						else if (ss_gobject.y_org() > new_gobject_y_org)
						{
							nssl.insert_before(new_gobject);
							inserted = true;
							break;
						}
						go_current = nssl.next();
					}

					if (!inserted) 
					{
						nssl.tail();
						nssl.insert_after(new_gobject);
					}
				} 
				else ss_current = ssl.next();
			}

			if (staff_systems_found_==StaffSystemsFoundEnum.by_byte) down_size_gobjects(nssl);
		}





		protected void cleanup_staff_lines(GObject gobject,StaffSystem staff_system,
			RemoveStaffsModeEnum mode)
		{
			if (mode!=RemoveStaffsModeEnum.remove_wobble_circle && mode!=RemoveStaffsModeEnum.remove_track_circle) return; 

			GObject with_staffs = staff_system.gobject();
			int ss_x_dim = with_staffs.x_dim();

			StaffList staffs = staff_system.staffs();
			StaffNode st_current = staffs.head();

			while (st_current!=null)
			{
      
				Staff staff = st_current.val();
      
				StaffSeg left_staff_seg = staff.left_staff_seg();
				StaffSeg right_staff_seg = staff.right_staff_seg();
      
				int no_lines = staff.no_staff_lines();
      
				for (int i=0; i<no_lines; i++)
				{
					// follow staff line
					StaffInfo staff_info = staff.info(); // make reference? // ****
					int avg_line_thick = staff_info.line_height();
					LineInfo line_info = new LineInfo(ss_x_dim,left_staff_seg,right_staff_seg,
						avg_line_thick,i);
	  
					bitmap_ops_.Follow.cleanup_circ_req_object(gobject,line_info);
				}
				st_current = staffs.next();
			}
		}



		protected void generate_removed_staff_lines(GObject staff_system,
			GObject staff_system_copy, Staff staff,
			RemoveStaffsModeEnum mode)
		{
			int ss_x_org = staff_system_copy.x_org();
			int ss_y_org = staff_system_copy.y_org();
			int ss_x_dim = staff_system_copy.x_dim();
			int ss_y_dim = staff_system_copy.y_dim();

			StaffSeg left_staff_seg = staff.left_staff_seg();
			StaffSeg right_staff_seg = staff.right_staff_seg();

			int lx_start = (left_staff_seg.x_start()+left_staff_seg.x_end())/2;
			int rx_start = (right_staff_seg.x_start()+right_staff_seg.x_end())/2;

			StaffLines staff_lines = staff.staff_lines();
			SubBitmapArray1d staff_line_bitmaps=staff_lines.bitmaps();
			int no_lines = staff.no_staff_lines();
			staff_lines.resize(no_lines);

			for (int i=0; i<no_lines; i++)
			{
				// follow staff line
				StaffInfo staff_info = staff.info(); // make reference? // ****
				int avg_line_thick = staff_info.line_height();
				LineInfo line_info = new LineInfo(ss_x_dim,left_staff_seg,right_staff_seg,
					avg_line_thick,i);
				switch (mode)
				{
					case RemoveStaffsModeEnum.remove_straight_prerau:
						bitmap_ops_.Follow.follow_line_straight_prerau(staff_system_copy,line_info,
							staff_line_bitmaps[i]);
						break;

					case RemoveStaffsModeEnum.remove_straight_clarke:
						bitmap_ops_.Follow.follow_line_straight_clarke(staff_system_copy,line_info,
							staff_line_bitmaps[i]);
						break;

					case RemoveStaffsModeEnum.remove_wobble:
						bitmap_ops_.Follow.follow_line_wobble(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.vertical_requirement,true);
						break;

					case RemoveStaffsModeEnum.remove_track:
						bitmap_ops_.Follow.follow_line_track(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.vertical_requirement,true);
						break;

					case RemoveStaffsModeEnum.remove_track_wide:
						bitmap_ops_.Follow.follow_line_track_wide(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.vertical_requirement,true);
						break;

					case RemoveStaffsModeEnum.remove_wobble_simple:
						bitmap_ops_.Follow.follow_line_wobble_simple(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.vertical_requirement,true);
						break;

					case RemoveStaffsModeEnum.remove_track_simple:
						bitmap_ops_.Follow.follow_line_track_simple(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.vertical_requirement,true);
						break;

					case RemoveStaffsModeEnum.remove_wobble_prerau:
						bitmap_ops_.Follow.follow_line_wobble_prerau(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.vertical_requirement,true);
						break;

					case RemoveStaffsModeEnum.remove_track_prerau:
						bitmap_ops_.Follow.follow_line_track_prerau(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.vertical_requirement,true);
						break;

					case RemoveStaffsModeEnum.remove_wobble_circle:
						bitmap_ops_.Follow.follow_line_wobble(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.circle_requirement,true);
						break;

					case RemoveStaffsModeEnum.remove_track_circle:
						bitmap_ops_.Follow.follow_line_track(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.circle_requirement,true);
						break;

				}
			}
		}



		protected void generate_all_staff_lines_removed(StaffSystemList ssl,
			GObjectList nssl,
			RemoveStaffsModeEnum mode,
			int verbose_level,
			ComputingCircle circle)
		{
			int ss_len = staff_systems_.length();
			int ss_computing_step = 100/((ss_len==0) ? 1 : ss_len);
			int computed = 0;

			// generate circle
			int dpi = page_manager_.dpi();
			bitmap_ops_.Follow.resize_circle(dpi/10);

			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				GObject with_staffs = staff_system.gobject();
				GObject without_staffs = staff_system.without_staffs();
				GObjectList staff_system_objects = staff_system.staff_system_objects();
				GObjectList proc_staff_system_objects 
					= staff_system.processed_staff_system_objects();

				StaffList staffs = staff_system.staffs();
				StaffNode st_current = staffs.head();

				int st_len = staffs.length();

				int st_computing_step;
				if (st_len>0)
				{
					st_computing_step = ss_computing_step/st_len;
				}
				else
				{
					st_computing_step = ss_computing_step;
					computed+=st_computing_step;
				}

				while (st_current!=null)
				{
					BaseIvLink.computed_so_far(circle,computed);

					Staff staff = st_current.val();
					StaffLines staff_lines = staff.staff_lines();
					staff_lines.resize(staff.no_staff_lines());
					generate_removed_staff_lines(with_staffs,without_staffs,staff,mode);

					computed+=st_computing_step;
					st_current = staffs.next();
				}

				// flood fill remaining objects
				int x_dim = without_staffs.x_dim();
				int y_dim = without_staffs.y_dim();
				for (int y=0; y<y_dim; y++)
					for (int x=0; x<x_dim; x++)
					{
						if (without_staffs[y,x]==Bitmap.PixelEnum.set)
						{
							GObject isolated_gobject 
								= bitmap_ops_.Flood.flood_fill_basic(without_staffs,x,y);
							staff_system_objects.append(isolated_gobject);

							cleanup_staff_lines(isolated_gobject,staff_system,mode);
							GObject trimmed_isolated_gobject
								= bitmap_ops_.Flood.trim(isolated_gobject);
							proc_staff_system_objects.append(trimmed_isolated_gobject);
						}
					}

				ss_current = ssl.next();
			} 

			transfer_to_staff_areas(nssl,ssl);

			BaseIvLink.computed_so_far(circle,100);
		}





		protected void generate_crossed_staff_lines(GObject staff_system,
			GObject staff_system_copy, Staff staff,
			RemoveStaffsModeEnum mode)
		{
			int ss_x_org = staff_system_copy.x_org();
			int ss_y_org = staff_system_copy.y_org();
			int ss_x_dim = staff_system_copy.x_dim();
			int ss_y_dim = staff_system_copy.y_dim();

			StaffSeg left_staff_seg = staff.left_staff_seg();
			StaffSeg right_staff_seg = staff.right_staff_seg();

			int lx_start = (left_staff_seg.x_start()+left_staff_seg.x_end())/2;
			int rx_start = (right_staff_seg.x_start()+right_staff_seg.x_end())/2;

			StaffLines staff_lines = staff.staff_lines();
			SubBitmapArray1d staff_line_bitmaps=staff_lines.bitmaps();
			int no_lines = staff.no_staff_lines();
			staff_lines.resize(no_lines);

			for (int i=0; i<no_lines; i++)
			{
				// follow staff line
				StaffInfo staff_info = staff.info(); // make reference? // ****
				int avg_line_thick = staff_info.line_height();
				LineInfo line_info = new LineInfo(ss_x_dim,left_staff_seg,right_staff_seg,
					avg_line_thick,i);
				switch (mode)
				{
					case RemoveStaffsModeEnum.cross_straight:
						bitmap_ops_.Follow.follow_line_straight(staff_system_copy,line_info,
							staff_line_bitmaps[i]);
						break;

					case RemoveStaffsModeEnum.cross_wobble:
						bitmap_ops_.Follow.follow_line_wobble(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.vertical_requirement,
							false);
						break;

					case RemoveStaffsModeEnum.cross_track:
						bitmap_ops_.Follow.follow_line_track(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.vertical_requirement,
							false);
						break;

					case RemoveStaffsModeEnum.cross_wobble_circle:
						bitmap_ops_.Follow.follow_line_wobble(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.circle_requirement,
							false);
						break;

					case RemoveStaffsModeEnum.cross_track_circle:
						bitmap_ops_.Follow.follow_line_track(staff_system,staff_system_copy,
							line_info,staff_line_bitmaps[i],
							BopFollow.ModeEnum.circle_requirement,
							false);
						break;

				}
			}
		}


		protected void generate_all_staff_lines_crossed(StaffSystemList ssl,
			GObjectList nssl,
			RemoveStaffsModeEnum mode,
			int verbose_level,
			ComputingCircle circle)
		{
			Bitmap is_staff_line = bitmap_ops_.scratch();

			int ss_len = staff_systems_.length();
            int ss_computing_step = 100 / ((ss_len == 0) ? 1 : ss_len);
			int computed = 0;

			// generate circle
			int dpi = page_manager_.dpi();
			bitmap_ops_.Follow.resize_circle(dpi/10);

			BopFollow.ModeEnum remove_mode = BopFollow.ModeEnum.vertical_requirement; // default

			switch (mode)
			{
				case RemoveStaffsModeEnum.cross_straight:
				case RemoveStaffsModeEnum.cross_wobble:
				case RemoveStaffsModeEnum.cross_track:
					remove_mode = BopFollow.ModeEnum.vertical_requirement;
					break;

				case RemoveStaffsModeEnum.cross_wobble_circle:
				case RemoveStaffsModeEnum.cross_track_circle:
					remove_mode = BopFollow.ModeEnum.circle_requirement;
					break;
			}


			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				GObject with_staffs = staff_system.gobject();
				GObject without_staffs = staff_system.without_staffs();
				GObjectList staff_system_objects = staff_system.staff_system_objects();
				GObjectList proc_staff_system_objects 
					= staff_system.processed_staff_system_objects();

				int ss_x_org = with_staffs.x_org();
				int ss_y_org = with_staffs.y_org();
				int ss_y_dim = with_staffs.y_dim();
				int ss_y_bot = with_staffs.y_bot();

				StaffList staffs = staff_system.staffs();
				StaffNode st_current = staffs.head();

				int st_len = staffs.length();

				int st_computing_step;
				if (st_len>0)
				{
					st_computing_step = ss_computing_step/st_len;
				}
				else
				{
					st_computing_step = ss_computing_step;
					computed+=st_computing_step;
				}

				while (st_current!=null)
				{
					BaseIvLink.computed_so_far(circle,computed);

					Staff staff = st_current.val();
					StaffLines staff_lines = staff.staff_lines();
					staff_lines.resize(staff.no_staff_lines());
					generate_crossed_staff_lines(with_staffs,without_staffs,staff,mode);

					computed+=st_computing_step;
					st_current = staffs.next();
				}

				// flood fill remaining objects
				int x_dim = without_staffs.x_dim();
				int y_dim = without_staffs.y_dim();
				for (int y=0; y<y_dim; y++)
					for (int x=0; x<x_dim; x++)
					{
						if (without_staffs[y,x]==Bitmap.PixelEnum.set)
						{
							int glob_x = ss_x_org + x;
							int glob_y = ss_y_org + y;

							int glob_ym1 = (y>0) ? glob_y-1 : ss_y_org;
							int glob_yp1 = (y<ss_y_dim-1) ? glob_y+1 : ss_y_bot;

							if (!is_staff_line.pixel_is_set(glob_x,glob_y) &&
								!is_staff_line.pixel_is_set(glob_x,glob_ym1) &&
								!is_staff_line.pixel_is_set(glob_x,glob_yp1))
							{
								GObject isolated_gobject 
									= bitmap_ops_
									.cross_staff_lines_flood_fill(without_staffs,
									x,y,remove_mode);
								staff_system_objects.append(isolated_gobject);

								// include these to be consitent with remove method;
								cleanup_staff_lines(isolated_gobject,staff_system,mode);
								GObject trimmed_isolated_gobject
									= bitmap_ops_.Flood.trim(isolated_gobject);
								proc_staff_system_objects.append(trimmed_isolated_gobject);
							}
						}
					}

				ss_current = ssl.next();
			} 

			is_staff_line.clear_all();

			transfer_to_staff_areas(nssl,ssl);

			BaseIvLink.computed_so_far(circle,100);
		}



		  
		public void generate_all_staff_lines(StaffSystemList ssl,
			GObjectList nssl,
			RemoveStaffsModeEnum mode,
			int verbose_level,
			ComputingCircle circle)
		{
			switch (mode)
			{
				case RemoveStaffsModeEnum.remove_straight_prerau:
				case RemoveStaffsModeEnum.remove_straight_clarke:
				case RemoveStaffsModeEnum.remove_wobble:
				case RemoveStaffsModeEnum.remove_track:
				case RemoveStaffsModeEnum.remove_track_wide:
				case RemoveStaffsModeEnum.remove_wobble_simple:
				case RemoveStaffsModeEnum.remove_track_simple:
				case RemoveStaffsModeEnum.remove_wobble_prerau:
				case RemoveStaffsModeEnum.remove_track_prerau:
				case RemoveStaffsModeEnum.remove_wobble_circle:
				case RemoveStaffsModeEnum.remove_track_circle:
					generate_all_staff_lines_removed(ssl,nssl,mode,verbose_level,circle);
					break;


				case RemoveStaffsModeEnum.cross_straight:
				case RemoveStaffsModeEnum.cross_wobble:
				case RemoveStaffsModeEnum.cross_track:
				case RemoveStaffsModeEnum.cross_wobble_circle:
				case RemoveStaffsModeEnum.cross_track_circle:
					generate_all_staff_lines_crossed(ssl,nssl,mode,verbose_level,circle);
					break;
			}
		}





		protected void reset_given_staff_lines_removed(StaffSystemList ssl)
		{
			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				StaffList staffs = staff_system.staffs();
				StaffNode st_current = staffs.head();

				while (st_current!=null)
				{
					Staff staff = st_current.val();
					StaffLines staff_lines = staff.staff_lines();
	  
					int no_staff_lines = staff.no_staff_lines();
					SubBitmapArray1d line_bitmaps = staff_lines.bitmaps();
	  
					for (int i=0; i<no_staff_lines; i++)
						line_bitmaps[i].resize(0,0);
	  
	  
					st_current = staffs.next();
				}
				ss_current = ssl.next();
			} 
		}


		protected void reset_given_without_staffs(StaffSystemList ssl)
		{
			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				GObject gobject = staff_system.gobject();
				GObject without_staffs = staff_system.without_staffs();

				without_staffs.extract(gobject,0,0,gobject.x_dim(),gobject.y_dim());
				staff_system.staff_system_objects().reset();
				staff_system.processed_staff_system_objects().reset();

				ss_current = ssl.next();
			} 
		}


		protected void reset_staff_lines_removed()
		{
			bitmap_ops_.scratch().clear_all();

			reset_given_staff_lines_removed(staff_systems_);
			reset_given_without_staffs(staff_systems_);

			corrected_non_staff_systems_.reset();
			corrected_staff_systems_.reset();
		}


		public void generate_deform(StaffSystemList ssl,int max_x_dim,
			CorrectDeformModeEnum mode, double max_acc_thresh,
			int verbose_level,
			ComputingCircle circle)
		{
			deform_.generate_count(ssl,max_x_dim);
			deform_.generate_thick_line(max_acc_thresh);
			deform_.generate_thin_line_basic();
			deform_.fill_in_gaps();

			// double cd_max_acc_thresh=0.0; // unused // ****
			switch (mode)
			{
				case CorrectDeformModeEnum.deform_level_60:
				case CorrectDeformModeEnum.deform_level_70:
				case CorrectDeformModeEnum.deform_level_80:
					deform_.extend_ends_level();
					break;

				case CorrectDeformModeEnum.deform_angle_60:
				case CorrectDeformModeEnum.deform_angle_70:
				case CorrectDeformModeEnum.deform_angle_80:
					deform_.extend_ends_angle();
					break;
			}

			deform_.generate_transform();
			deform_.generate_smooth_transform();
		}


		public void correct_deform(Bitmap bitmap, 
			ComputingCircle circle)
		{
			// caution - undeformed page larger.

			Bitmap undeformed_page = bitmap_ops_.Deform.undeform(bitmap,deform_,circle);
			corrected_page_ = undeformed_page;
		}

		public void reset_deform()
		{
			deform_.reset();
		}

		protected void transfer_to_staff_system_objects(StaffSystem staff_system,
			GObject non_ss_ob)
		{
			int nsso_x_org = non_ss_ob.x_org();
			int nsso_y_org = non_ss_ob.y_org();

			GObjectList ssol = staff_system.staff_system_objects();
			GObjectList pssol = staff_system.processed_staff_system_objects();

			GObjectNode current_sso = ssol.head();
			GObjectNode current_psso = pssol.head();
  
			while (current_sso!=null)
			{
				GObject ss_ob = current_sso.val();
				int ss_ob_x_org = ss_ob.x_org();
				int ss_ob_y_org = ss_ob.y_org();

				if ((nsso_y_org<ss_ob_y_org) && (nsso_x_org<ss_ob_x_org)) 
				{
					current_sso = ssol.next();
					current_psso = pssol.next();
				}
				else
				{
					ssol.insert_before((GObject)non_ss_ob); // better if
					int nso_x_org = non_ss_ob.x_org();
					int nso_y_org = non_ss_ob.y_org();
					int nso_x_dim = non_ss_ob.x_dim();
					int nso_y_dim = non_ss_ob.y_dim();

					SubBitmap non_ss_ob_subbitmap = non_ss_ob;
					GObject pnon_ss_ob
						= new GObject(non_ss_ob_subbitmap,0,0,nso_x_dim,nso_y_dim);
					pnon_ss_ob.set_x_org(nso_x_org);
					pnon_ss_ob.set_y_org(nso_y_org);
	  
					pssol.insert_before(pnon_ss_ob);
					break;
				}

				if (current_sso==null) 
				{
					// identical to append, but more consitent
					ssol.tail();
					ssol.insert_after((GObject)non_ss_ob); // better if
					int nso_x_dim = non_ss_ob.x_dim();
					int nso_y_dim = non_ss_ob.y_dim();
					int nso_x_org = non_ss_ob.x_org();
					int nso_y_org = non_ss_ob.y_org();

					SubBitmap non_ss_ob_subbitmap = non_ss_ob;
					GObject pnon_ss_ob
						= new GObject(non_ss_ob_subbitmap,0,0,nso_x_dim,nso_y_dim);
					pnon_ss_ob.set_x_org(nso_x_org);
					pnon_ss_ob.set_y_org(nso_y_org);
	  
					// identical to append, but more consitent
					pssol.tail();
					pssol.insert_before(pnon_ss_ob);
				}
			}
		}


		public void transfer_to_staff_areas(GObjectList nssl, 
			StaffSystemList ssl)
		{
			GObjectNode current_nss = nssl.head();

			while (current_nss!=null)
			{
				bool match = false;

				GObject non_ss_ob = current_nss.val();
				int non_ss_ob_y_dim = non_ss_ob.y_dim();

				StaffSystemNode current_ss = ssl.head();
      
				while (current_ss!=null)
				{
					StaffSystem staff_system = current_ss.val();

					int staff_gap = staff_system.staff_info().gap_height();

					StaffList stl = staff_system.staffs();
					StaffNode current_st = stl.head();
	      
					while (current_st!=null)
					{
						Staff staff = current_st.val();
						SubBitmap staff_area = staff.staff_area();
						if (staff_area.rect_intersect(non_ss_ob))
						{
							current_nss = nssl.detach_forward(); // gobject intact
		      
							// add into correct position of staff object list
							transfer_to_staff_system_objects(staff_system,non_ss_ob);
		  
							match = true;
							break;
						}
		  
						current_st = stl.next();
					}
	  
					if (match) break;
					current_ss = ssl.next();
				}
      
				if (!match) current_nss = nssl.next();
			}
		}

		protected void transfer_small_to_staff_areas(GObjectList nssl, 
			StaffSystemList ssl)
		{
			GObjectNode current_nss = nssl.head();

			while (current_nss!=null)
			{
				bool match = false;

				GObject non_ss_ob = current_nss.val();
				int non_ss_ob_y_dim = non_ss_ob.y_dim();

				StaffSystemNode current_ss = ssl.head();
      
				while (current_ss!=null)
				{
					StaffSystem staff_system = current_ss.val();

					int staff_gap = staff_system.staff_info().gap_height();
					if (non_ss_ob_y_dim<staff_gap)
					{
						StaffList stl = staff_system.staffs();
						StaffNode current_st = stl.head();
	      
						while (current_st!=null)
						{
							Staff staff = current_st.val();
							SubBitmap staff_area = staff.staff_area();
							if (staff_area.rect_intersect(non_ss_ob))
							{
								current_nss = nssl.detach_forward(); // gobject intact
		      
								// add into correct position of staff object list
								transfer_to_staff_system_objects(staff_system,non_ss_ob);
		      
								match = true;
								break;
							}
		  
							current_st = stl.next();
						}
					}
					if (match) break;
					current_ss = ssl.next();
				}

				if (!match) current_nss = nssl.next();
			}
		}



		/*--------------------------------------------------------------------------*/


		public void split_on_title(double title_ymul, int title_thresh,
			int print_level)
		{
			int page_y_dim = page_manager_.current_fbitmap().y_dim();

			// init hproject data
			for (int y=0; y<page_y_dim; y++)
			{
				title_hproject_[y]=0;
				title_is_vspan_[y]=false;
			}

			GCharList gchar_list = gcharacters_;

			// project any gchars matching prefix "title" into histogram count
			GCharNode current_gc = gchar_list.head();
			while (current_gc!=null)
			{
				GChar gchar = current_gc.val();
				string gchar_name = gchar.name();
				int perc_pos = gchar_name.IndexOf('%');

				if (gchar_name.Substring(0,perc_pos)=="title")
				{
					int y_mid = (gchar.y_top() + gchar.y_bot())/2;
					int y_dim = gchar.y_dim();
					int y_extend = MathExtra.irint(((double)y_dim)*title_ymul);

					int yt = y_mid - (y_extend/2);
					int yb = y_mid + (y_extend/2);

					if (yt<0) { yt=0; }
					if (yb>=page_y_dim) { yb = page_y_dim-1; }

					for (int y=yt; y<=yb; y++)
					{
						title_hproject_[y]++;
					}
				}

				current_gc = gchar_list.next();
			}
  
    
			// threshold histogram projection
			for (int y=0; y<page_y_dim; y++)
			{
				if (title_hproject_[y]>=title_thresh)
				{
					title_is_vspan_[y] = true;
				}
			}

			// generate vspan list of title spans
			for (int y=0; y<page_y_dim; y++)
			{
				if (title_is_vspan_[y])
				{
					int y_start = y;

					while (title_is_vspan_[y])
					{
						y++;
						if (y==page_y_dim) break;
					}
					y--;

					int y_end = y;

					VSpan vspan = new VSpan(0,y_start,y_end);
					title_vspans_.append(vspan); // responsibility passed over
				}
			}

			if (print_level>=2)
			{
				VSpanNode current_vs = title_vspans_.head();
				while (current_vs!=null)
				{
					Console.Error.WriteLine(current_vs.val());
					current_vs=title_vspans_.next();
				}
			}

		}


		protected void staff_system_extent(StaffSystem staff_system,
			out int min_xl_staff, out int max_xr_staff,
			out int min_yt_staff, out int max_yb_staff)
		{
			Bitmap current_bitmap = page_manager_.current_fbitmap();
			int page_y_dim = current_bitmap.y_dim();
			int page_x_dim = current_bitmap.x_dim();

			min_xl_staff = page_x_dim-1;
			max_xr_staff = 0;
			min_yt_staff = page_y_dim-1;
			max_yb_staff = 0;

			StaffList stl = staff_system.staffs();
			StaffNode current_st = stl.head();
			while (current_st!=null)
			{
				Staff staff = current_st.val();
				GObject staff_area = staff.staff_area();
				int sa_xl = staff_area.x_left();
				int sa_xr = staff_area.x_right();
				int sa_yt = staff_area.y_top();
				int sa_yb = staff_area.y_bot();

				min_xl_staff = Math.Min(min_xl_staff,sa_xl);
				max_xr_staff = Math.Max(max_xr_staff,sa_xr);
				min_yt_staff = Math.Min(min_yt_staff,sa_yt);
				max_yb_staff = Math.Max(max_yb_staff,sa_yb);

				current_st = stl.next();
			}

			GObjectList ssol = staff_system.processed_staff_system_objects();
			GObjectNode current_sso = ssol.head();
			while (current_sso!=null)
			{
				GObject sso_gobject = current_sso.val();
				int sso_xl = sso_gobject.x_left();
				int sso_xr = sso_gobject.x_right();
				int sso_yt = sso_gobject.y_top();
				int sso_yb = sso_gobject.y_bot();
      
				min_xl_staff = Math.Min(min_xl_staff,sso_xl);
				max_xr_staff = Math.Max(max_xr_staff,sso_xr);
				min_yt_staff = Math.Min(min_yt_staff,sso_yt);
				max_yb_staff = Math.Max(max_yb_staff,sso_yb);
      
				current_sso = ssol.next();
			}
		}

		protected void subdivide_staff_system_on_title(StaffSystemList ssl,
			GObjectList nss,
			int y_subdivide)
		{
			GObjectList non_staff_systems = nss;

			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();

				GObject ss_gobject = staff_system.gobject();
				int ss_yt = ss_gobject.y_top();
				int ss_yb = ss_gobject.y_bot();
      
				if ((ss_yt<y_subdivide) && (ss_yb>=y_subdivide))
				{
					// staff system vertically spans title
					int ss_x_dim = ss_gobject.x_dim();

					int ssa_y_dim = y_subdivide - ss_yt;
					int ssb_y_dim = ss_yb - y_subdivide +1;

					GObject ssa_gobject 
						= new GObject(ss_gobject,0,0,ss_x_dim,ssa_y_dim);
					GObject ssb_gobject 
						= new GObject(ss_gobject,0,ssa_y_dim,ss_x_dim,ssb_y_dim);
					ssa_gobject.add_to_y_org(ss_yt);
					ssb_gobject.add_to_y_org(ss_yt);

					StaffSystem staff_system_above = new StaffSystem(ssa_gobject);
					StaffSystem staff_system_below = new StaffSystem(ssb_gobject);
	  
					// set staff infos
					StaffInfo ss_staff_info = staff_system.staff_info();
					staff_system_above.copy_staff_info(ss_staff_info);
					staff_system_below.copy_staff_info(ss_staff_info);

					// separate out staffs
					StaffList stl = staff_system.staffs();
					StaffNode current_st = stl.head();
					while (current_st!=null)
					{
						Staff staff = current_st.val();
						GObject staff_area = staff.staff_area();
						int sa_yt = staff_area.y_top();
						int sa_yb = staff_area.y_bot();

						if ((sa_yt<y_subdivide) && (sa_yb>=y_subdivide))
						{
							// staff area spans dividing line
							Console.Error.WriteLine("Warning: staff area vertically spans title");
							Console.Error.WriteLine("         staff area omitted");

							current_st = stl.delete_forward();
						}
						else 
						{
							if (sa_yt<y_subdivide)
							{
								// place staff in top list
								staff_system_above.staffs().append(staff);
							}
							else 
							{
								// place staff in bottom list
								staff_system_below.staffs().append(staff);
							}
							current_st = stl.detach_forward();
						}
					}

					// separate out staff system objects
					GObjectList ssol = staff_system.processed_staff_system_objects();
					GObjectNode current_sso = ssol.head();
					while (current_sso!=null)
					{
						GObject sso_gobject = current_sso.val();
						int sso_yt = sso_gobject.y_top();
						int sso_yb = sso_gobject.y_bot();

						if ((sso_yt<y_subdivide) && (sso_yb>=y_subdivide))
						{
							// object vertically spans subdivide		  
							int sso_x_dim = sso_gobject.x_dim();
		  
							int ssoa_y_dim = y_subdivide - sso_yt;
							int ssob_y_dim = sso_yb - y_subdivide +1;

							GObject ssoa_gobject 
								= new GObject((GObject)sso_gobject,
								0,0,sso_x_dim,ssoa_y_dim);
							GObject ssob_gobject 
								= new GObject((GObject)sso_gobject,
								0,ssoa_y_dim,sso_x_dim,ssob_y_dim);

							ssoa_gobject.add_to_y_org(sso_yt);
							ssob_gobject.add_to_y_org(sso_yt);

							GObjectList staff_system_objects_above
								= staff_system_above.processed_staff_system_objects();
							GObjectList staff_system_objects_below
								= staff_system_below.processed_staff_system_objects();

							insert_order_xy(staff_system_objects_above,ssoa_gobject);
							insert_order_xy(staff_system_objects_below,ssob_gobject);

							current_sso = ssol.delete_forward();
						}
						else
						{
							if (sso_yt<y_subdivide)
							{
								// place sso in top list
								GObjectList staff_system_objects_above
									= staff_system_above.processed_staff_system_objects();
								staff_system_objects_above.append(sso_gobject);
							}
							else 
							{
								// place sso in bottom list
								GObjectList staff_system_objects_below
									= staff_system_below.processed_staff_system_objects();
								staff_system_objects_below.append(sso_gobject);
							}
							current_sso = ssol.detach_forward();
						}
					}
	  
					// delete current staff system and replace with two new ones
					current_ss = ssl.delete_forward();

					if ((staff_system_above.staffs().length()>0) ||
						(staff_system_above.processed_staff_system_objects().length()>0))
					{
						if (ssl.length()>0)
						{
							if (current_ss!=null)
							{
								ssl.insert_before(staff_system_above);
							}
							else
							{
								ssl.append(staff_system_above);
							}
						}
						else
						{
							ssl.prepend(staff_system_above);
						}
					}

					if ((staff_system_below.staffs().length()>0) ||
						(staff_system_below.processed_staff_system_objects().length()>0))
					{
						if (ssl.length()>0)
						{
							if (current_ss!=null)
							{
								ssl.insert_before(staff_system_below);
							}
							else
							{
								ssl.append(staff_system_below);
							}
						}
						else
						{
							ssl.prepend(staff_system_below);
						}
					}
				}
				else
				{
					// check for any  staff system objects that may span divide
					GObjectList ssol = staff_system.processed_staff_system_objects();
					GObjectNode current_sso = ssol.head();
					while (current_sso!=null)
					{
						GObject sso_gobject = current_sso.val();
						int sso_yt = sso_gobject.y_top();
						int sso_yb = sso_gobject.y_bot();

						if ((sso_yt<y_subdivide) && (sso_yb>=y_subdivide))
						{
							// object vertically spans subdivide		  
							int sso_x_dim = sso_gobject.x_dim();
		  
							int ssoc_y_dim = y_subdivide - sso_yt;
							int nssoc_y_dim = sso_yb - y_subdivide +1;

							GObject ssoc_gobject 
								= new GObject((GObject)sso_gobject,
								0,0,sso_x_dim,ssoc_y_dim);
							GObject nssoc_gobject 
								= new GObject((GObject)sso_gobject,
								0,ssoc_y_dim,sso_x_dim,nssoc_y_dim);

							ssoc_gobject.add_to_y_org(sso_yt);
							nssoc_gobject.add_to_y_org(sso_yt);

							GObjectNode store_sso = current_sso;
							insert_order_xy(ssol,ssoc_gobject);
							ssol.set_position(store_sso);

							insert_order_xy(nss,nssoc_gobject);

							current_sso = ssol.delete_forward();
						}
						else
						{
							current_sso = ssol.next();
						}
					}

					current_ss = ssl.next();
				}
			}
		}

		public void subdivide_staff_systems_on_title(StaffSystemList ssl,
			GObjectList nss,
			int print_level)
		{
			// check for any staff systems that vertically span a title and
			// subdivide appropriately

			VSpanNode current_vs = title_vspans_.head();
			while (current_vs!=null)
			{
				VSpan vspan = current_vs.val();
				int curr_y_start = vspan.yt();

				subdivide_staff_system_on_title(ssl,nss,curr_y_start);

				current_vs = title_vspans_.next();
			}
		}


		protected bool staff_system_above_title(StaffSystemList ssl,
			int y_title_top)
		{
			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();
				GObject ss_gobject = staff_system.gobject();
				int ss_yt = ss_gobject.y_top();
				if (ss_yt<y_title_top)
				{
					return true;
				}

				current_ss = ssl.next();
			}

			return false;
		}




		protected void staff_areas_extent(StaffSystemList ssl,
			out int min_xl_staff, out int max_xr_staff,
			out int min_yt_staff, out int max_yb_staff)
		{
			Bitmap current_bitmap = page_manager_.current_fbitmap();
			int page_y_dim = current_bitmap.y_dim();
			int page_x_dim = current_bitmap.x_dim();

			min_xl_staff = page_x_dim-1;
			max_xr_staff = 0;
			min_yt_staff = page_y_dim-1;
			max_yb_staff = 0;

			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();
				StaffList stl = staff_system.staffs();
				StaffNode current_st = stl.head();
				while (current_st!=null)
				{
					Staff staff = current_st.val();
					GObject staff_area = staff.staff_area();
					int sa_xl = staff_area.x_left();
					int sa_xr = staff_area.x_right();
					int sa_yt = staff_area.y_top();
					int sa_yb = staff_area.y_bot();

					min_xl_staff = Math.Min(min_xl_staff,sa_xl);
					max_xr_staff = Math.Max(max_xr_staff,sa_xr);
					min_yt_staff = Math.Min(min_yt_staff,sa_yt);
					max_yb_staff = Math.Max(max_yb_staff,sa_yb);

					current_st = stl.next();
				}

				current_ss = ssl.next();
			}
		}


		public void filter_on_auto_bcrop(StaffSystemList ssl, GObjectList nssl, 
			double abcrop_ymul,
			int abcrop_thresh,
			int auto_bcrop)
		{
			int min_xl_staff, max_xr_staff;
			int min_yt_staff, max_yb_staff;

			staff_areas_extent(ssl,out min_xl_staff,out max_xr_staff,out min_yt_staff,out max_yb_staff);

			Bitmap current_bitmap = page_manager_.current_fbitmap();
			int page_x_dim = current_bitmap.x_dim();
			int page_y_dim = current_bitmap.y_dim();

			// hproject smallish non staff system objects that are below last staff area

			// init hproject data
			for (int y=0; y<page_y_dim; y++)
			{
				abcrop_hproject_[y]=0;
			}

			// project non staff system objects below lowest staff area into histogram 
			GObjectNode current_nss = nssl.head();
			while (current_nss!=null)
			{
				GObject gobject = current_nss.val();
				int x_dim = gobject.x_dim();
				int y_dim = gobject.y_dim();

				double dpi = (double)page_manager_.dpi();
				int noise_thresh = MathExtra.irint(dpi/250.0);
				int staff_x_thresh = MathExtra.irint(dpi/3.0);

				if (((x_dim>noise_thresh) && (y_dim>noise_thresh)) 
					&& (x_dim<staff_x_thresh))
				{
					int y_mid = (gobject.y_top() + gobject.y_bot())/2;
					int y_extend = MathExtra.irint(((double)y_dim)*abcrop_ymul);

					int yt = y_mid - (y_extend/2);
					int yb = y_mid + (y_extend/2);

					if (yt<0) { yt=0; }
					if (yb>=page_y_dim) { yb = page_y_dim-1; }

					for (int y=yt; y<=yb; y++)
					{
						abcrop_hproject_[y]++;
					}
				}

				current_nss = nssl.next();
			}
  
			// locate last value greater than abcrop_thresh

			int max_yb_nss = page_y_dim-1;
			for (int y=page_y_dim-1; y>max_yb_staff; y--)
			{
				if (abcrop_hproject_[y]>abcrop_thresh)
				{
					max_yb_nss=y;
					break;
				}
			}

			// calculate y_cutoff
			int y_cutoff = max_yb_nss + auto_bcrop;
			if (y_cutoff>=page_y_dim)
			{
				y_cutoff=page_y_dim-1;
			}

			// filter on y_cutoff value

			// go through staff systems deleting gobjects outside extent+bcrop

			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();
				GObjectList ssol = staff_system.processed_staff_system_objects();
				GObjectNode current_sso = ssol.head();
				while (current_sso!=null)
				{
					GObject gobject = current_sso.val();
					int yb = gobject.y_bot();
					if (yb>y_cutoff)
					{
						current_sso = ssol.delete_forward();
					}
					else
					{
						current_sso = ssol.next();
					}
				}

				current_ss = ssl.next();
			}

			// do the same for non staff system objects
			current_nss = nssl.head();
			while (current_nss!=null)
			{
				GObject gobject = current_nss.val();
				int yb = gobject.y_bot();
				if (yb>y_cutoff)
				{
					current_nss = nssl.delete_forward();
				}
				else
				{
					current_nss = nssl.next();
				}
			}

			// resize page
			int new_page_x_dim = page_x_dim;
			int new_page_y_dim = y_cutoff +1;
			current_bitmap.resize(new_page_x_dim,new_page_y_dim);
		}





		public void filter_on_bcrop(StaffSystemList ssl, GObjectList nssl, 
			int bcrop, int restrict_y_dim)
		{
			int min_xl_staff, max_xr_staff;
			int min_yt_staff, max_yb_staff;

			staff_areas_extent(ssl,out min_xl_staff,out max_xr_staff,out min_yt_staff,out max_yb_staff);

			Bitmap current_bitmap = page_manager_.current_fbitmap();
			int page_x_dim = current_bitmap.x_dim();
			int page_y_dim = current_bitmap.y_dim();

			int y_cutoff = (bcrop==0) ? page_y_dim-1 : max_yb_staff + bcrop;
			if ((restrict_y_dim>0) && (restrict_y_dim>max_yb_staff))
			{
				y_cutoff=restrict_y_dim-1;
			} 
			if (y_cutoff>=page_y_dim)
			{
				y_cutoff=page_y_dim-1;
			}

			// go through staff systems deleting gobjects outside extent+bcrop

			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();
				GObjectList ssol = staff_system.processed_staff_system_objects();
				GObjectNode current_sso = ssol.head();
				while (current_sso!=null)
				{
					GObject gobject = current_sso.val();
					int yb = gobject.y_bot();
					if (yb>y_cutoff)
					{
						current_sso = ssol.delete_forward();
					}
					else
					{
						current_sso = ssol.next();
					}
				}

				current_ss = ssl.next();
			}

			// do the same for non staff system objects
			GObjectNode current_nss = nssl.head();
			while (current_nss!=null)
			{
				GObject gobject = current_nss.val();
				int yb = gobject.y_bot();
				if (yb>y_cutoff)
				{
					current_nss = nssl.delete_forward();
				}
				else
				{
					current_nss = nssl.next();
				}
			}

			// resize page
			int new_page_x_dim = page_x_dim;
			int new_page_y_dim = y_cutoff +1;
			current_bitmap.resize(new_page_x_dim,new_page_y_dim);
		}

		public void filter_on_rcrop(StaffSystemList ssl, GObjectList nssl, 
			int rcrop, int restrict_x_dim)
		{
			int min_xl_staff, max_xr_staff;
			int min_yt_staff, max_yb_staff;

			staff_areas_extent(ssl,out min_xl_staff,out max_xr_staff,out min_yt_staff,out max_yb_staff);

			Bitmap current_bitmap = page_manager_.current_fbitmap();
			int page_x_dim = current_bitmap.x_dim();
			int page_y_dim = current_bitmap.y_dim();

			int x_cutoff = (rcrop==0) ? page_x_dim-1 : max_xr_staff + rcrop;
			if ((restrict_x_dim>0) && (restrict_x_dim>max_xr_staff))
			{
				x_cutoff=restrict_x_dim-1;
			}
			if (x_cutoff>=page_x_dim)
			{
				x_cutoff=page_x_dim-1;
			}

			// go through staff systems deleteing gobjects outside extent+rcrop

			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();
				GObjectList ssol = staff_system.processed_staff_system_objects();
				GObjectNode current_sso = ssol.head();
				while (current_sso!=null)
				{
					GObject gobject = current_sso.val();
					int xr = gobject.x_right();
					if (xr>x_cutoff)
					{
						current_sso = ssol.delete_forward();
					}
					else
					{
						current_sso = ssol.next();
					}
				}

				current_ss = ssl.next();
			}

			// do the same for non staff system objects
			GObjectNode current_nss = nssl.head();
			while (current_nss!=null)
			{
				GObject gobject = current_nss.val();
				int xr = gobject.x_right();
				if (xr>x_cutoff)
				{
					current_nss = nssl.delete_forward();
				}
				else
				{
					current_nss = nssl.next();
				}
			}

			// resize page
			int new_page_x_dim = x_cutoff +1;
			int new_page_y_dim = page_y_dim;
			current_bitmap.resize(new_page_x_dim,new_page_y_dim);
		}


		public void filter_on_lclear(StaffSystemList ssl, GObjectList nssl, 
			int lclear)
		{
			int min_xl_staff, max_xr_staff;
			int min_yt_staff, max_yb_staff;

			staff_areas_extent(ssl,out min_xl_staff,out max_xr_staff,out min_yt_staff,out max_yb_staff);

			Bitmap current_bitmap = page_manager_.current_fbitmap();
			int page_x_dim = current_bitmap.x_dim();
			int page_y_dim = current_bitmap.y_dim();

			int x_cutoff = min_xl_staff - lclear;
			if (x_cutoff<0)
			{
				x_cutoff=0;
			}

			// go through staff systems deleteing gobjects outside extent+rcrop

			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();
				GObjectList ssol = staff_system.processed_staff_system_objects();
				GObjectNode current_sso = ssol.head();
				while (current_sso!=null)
				{
					GObject gobject = current_sso.val();
					int xl = gobject.x_left();
					if (xl<x_cutoff)
					{
						current_sso = ssol.delete_forward();
					}
					else
					{
						current_sso = ssol.next();
					}
				}

				current_ss = ssl.next();
			}

			// do the same for non staff system objects
			GObjectNode current_nss = nssl.head();
			while (current_nss!=null)
			{
				GObject gobject = current_nss.val();
				int xl = gobject.x_left();
				if (xl<x_cutoff)
				{
					current_nss = nssl.delete_forward();
				}
				else
				{
					current_nss = nssl.next();
				}
			}
		}

		protected string dec(int v)
		{
			string str = v.ToString();

			return str;
		}

		public void output_split_on_title(string base_sep_filename,
			StaffSystemList ssl,
			GObjectList nssl,
			bool do_crop,
			ref int tune_no, ref int part_no)

		{
			int page_y_dim = page_manager_.current_fbitmap().y_dim();

			int start_y_sep = 0;
			int end_y_sep = page_y_dim-1;

			int prev_y_start = start_y_sep;

			VSpanNode head_vs = title_vspans_.head();
			VSpanNode current_vs = head_vs;
			while (current_vs!=null)
			{      
				VSpan vspan = current_vs.val();
				int curr_y_start = vspan.yt();
				if (curr_y_start>=page_y_dim)
				{
					curr_y_start = page_y_dim;
				}

				bool ss_above = true;
				if (current_vs==head_vs)
				{
					// check to see if any staffs exist above title
					ss_above = staff_system_above_title(ssl,curr_y_start);

					if (!ss_above)
					{
						// no staff systems above first title => tune starts new page
						// Elevate part number of area above title to be first part 
						//   of next tune
						tune_no++;
						part_no=0;
						if (do_crop) 
						{
							prev_y_start = curr_y_start;
							current_vs = title_vspans_.next();
							continue;
						}
					}
				}

				String tns = "_tune";
				String pns = "_part";
				tns += dec(tune_no);
				pns += dec(part_no);

				String seperate_filename = base_sep_filename + tns + pns + ".sep";

				output_separated_objects(seperate_filename,ssl,nssl,
					prev_y_start,curr_y_start-1);

				if (!ss_above)
				{
					part_no++;
				}
				else
				{
					tune_no++;
					part_no=0;
				}
      
				prev_y_start = curr_y_start;
				current_vs = title_vspans_.next();
			}

			if (prev_y_start<page_y_dim)
			{
				String tns = "_tune";
				String pns = "_part";
				tns += dec(tune_no);
				pns += dec(part_no);

				String seperate_filename = base_sep_filename + tns + pns + ".sep";

				output_separated_objects(seperate_filename,ssl,nssl,
					prev_y_start,end_y_sep);
				part_no++;
			}
		}


		/*--------------------------------------------------------------------------*/


		public GCharList corrected_gcharacters() 
		{ return corrected_gcharacters_; }

		public StaffSystemList corrected_staff_systems() 
		{ return corrected_staff_systems_; }

		public GObjectList corrected_non_staff_systems() 
		{ return corrected_non_staff_systems_; }

		public override int page_x_dim()
		{
			return ((object)corrected_page_==null) ? page_manager_.current_fbitmap().x_dim()
				: corrected_page_.x_dim(); }

		public override int page_y_dim()
		{
			return ((object)corrected_page_==null) ? page_manager_.current_fbitmap().y_dim()
				: corrected_page_.y_dim(); }

		public DeformData deform() { return deform_; }

		public Bitmap corrected_page() { return corrected_page_; } 

		public Bitmap corrected_bitmap() { return corrected_page_; } // identical to above in all but name! // *****

		public BitmapOps bitmap_ops() { return bitmap_ops_; }





	}
}
	    
