

using System;
using System.IO;

namespace Espresso.OMR
{
	class PageManager : BasePageManager
	{
		protected BaseMObjects base_mobjects_;


		public PageManager(int argc, string[] argv)
			: base(argc,argv)
		{
			base_mobjects_ = null;
		}

		public void set_base_mobjects(BaseMObjects bmobjects)
		{
			base_mobjects_ = bmobjects;
		}


		protected BinaryReader open_fstream(string filename)
		{
			FilenameManager.canonical_filename(filename); 

			if (filename!="")
			{
				// set filename
				string added_filename = filename + FilenameManager.added_ending();
				current_fbitmap_.set_filename(added_filename); 
      
				current_fbitmap_.set_valid(true); 
      
				if (!FilenameManager.check_full_filename())
				{
	  
					string error_mess = "Unable to open " + filename;
					current_fbitmap_.set_error(error_mess);
	  
					current_fbitmap_.resize(0,0);
					current_fbitmap_.set_valid(false);
					return null;
				}
			}

			string input_filename = FilenameManager.input_filename();
			BinaryReader brin = FilenameManager.open_fbitmap_input_pipe(input_filename);

			return brin;
		}



		protected int append_page_from_stream(BinaryReader brin, int y_offset)
		{
			int new_y_offset = y_offset;

			if (FilenameManager.file_type()==FilenameManager.FileTypeEnum.sep)
			{
				if (base_mobjects_.input_separated_objects(brin,y_offset))
				{
					if (base_mobjects_.empirical_filename_valid())
					{
						string emp_filename = base_mobjects_.empirical_filename();
						if (FilenameManager.check_empirical_filename(emp_filename))
						{
							// reassign specified empirical values
							BinaryReader brin_emp 
								= FilenameManager.open_empirical_input_pipe(emp_filename);

							ShareGlobals.source_file(brin_emp);
							FilenameManager.close_empirical_input_pipe();

						}
						else
						{
							Console.Error.WriteLine("Warning: unable to open empirical file " + emp_filename);
						}                 
					}

					int x_dim = current_fbitmap_.x_dim();
					int y_dim = current_fbitmap_.y_dim();
					dpi_ = ShareGlobals.scan_dpi_;
					((MObjects)base_mobjects_).reset_and_resize_sep(x_dim,y_dim);

					new_y_offset += y_dim;
				}
				else
				{
					current_fbitmap_.resize(0,0);
					current_fbitmap_.set_valid(false);
				}
			}
			else if (FilenameManager.file_type()==FilenameManager.FileTypeEnum.dsep)
			{

				if (base_mobjects_.input_dseparated_objects(brin))
				{
					// make fn with above // ****
					if (base_mobjects_.empirical_filename_valid())
					{
						string emp_filename = base_mobjects_.empirical_filename();
						if (FilenameManager.check_empirical_filename(emp_filename))
						{
							// reassign specified empirical values
							BinaryReader brin_emp 
								= FilenameManager.open_empirical_input_pipe(emp_filename);

							ShareGlobals.source_file(brin_emp);
							FilenameManager.close_empirical_input_pipe();

							/*
							if (!is.fail())
							  {
								ShareGlobals.source_file(is);
								FilenameManager.close_empirical_input_pipe();
							  }
							else
							  {
								cerr << "Warning: unable to read from empirical file ";
								cerr << emp_filename << endl;
							  }             
							*/

						}
						else
						{
							Console.Error.WriteLine("Warning: unable to open empirical file " + emp_filename);
						}                 
					}

					int x_dim = current_fbitmap_.x_dim();
					int y_dim = current_fbitmap_.y_dim();
					dpi_ = ShareGlobals.scan_dpi_;
					((MObjects)base_mobjects_).reset_and_resize_dsep(x_dim,y_dim);

					new_y_offset += y_dim;
				}
				else
				{
					current_fbitmap_.resize(0,0);
					current_fbitmap_.set_valid(false);
				}
			}
			else
			{
				// pbm
				current_fbitmap_.replace(brin);
				if (current_fbitmap_.valid())
				{
					int x_dim = current_fbitmap_.x_dim();
					int y_dim = current_fbitmap_.y_dim();
					dpi_ = ShareGlobals.scan_dpi_;
					base_mobjects_.reset_and_resize(x_dim,y_dim);

					new_y_offset += y_dim;
				}
			}

			return new_y_offset;
		}


		protected void append_page_from_stream(BinaryReader brin)
		{
			append_page_from_stream(brin,0);
		}

		public void simple_replace_page(string filename) 
		{
			Console.Error.WriteLine("Warning: PageManger.simple_replace_page() simplified version of replace_page() -- only loads uncompressed PBM");

			// open filename
	  
			FileStream fsin = new FileStream(filename, FileMode.Open, System.IO.FileAccess.Read);
			BinaryReader brin = new BinaryReader(fsin);

			current_fbitmap_.set_filename(filename);
			current_fbitmap_.set_valid(true);

			current_fbitmap_.replace(brin);
			if (current_fbitmap_.valid())
			{
				int x_dim = current_fbitmap_.x_dim();
				int y_dim = current_fbitmap_.y_dim();
				dpi_ = ShareGlobals.scan_dpi_;
				base_mobjects_.reset_and_resize(x_dim,y_dim);
			}
			else 
			{
				Console.Error.WriteLine("Failed to load " + filename);
			}

			// close filename
			brin.Close();
			fsin.Close();


		}

		public override void replace_page(string filename)
		{
			BinaryReader brin = open_fstream(filename);
			if (brin==null) { return; }
  
			append_page_from_stream(brin);

			FilenameManager.close_fbitmap_input_pipe();
		}

		public void concat_pages(StringList cp_filenames) 
		{
			// Console.Error.WriteLine("PageManager.concat_pages() not implemented.");
	  
	  
			int curr_y_offset = 0;

			if (cp_filenames.length()==0)
			{
				Stream base_stream = Console.OpenStandardInput();
				BinaryReader brin = new BinaryReader(base_stream);

				while (brin.PeekChar() != -1)
				{
					int new_y_offset = append_page_from_stream(brin,curr_y_offset);
					curr_y_offset = new_y_offset;
				}
			}
			else
			{
				StringNode current_cpf = cp_filenames.head();
				while (current_cpf!=null)
				{
					string cp_filename = current_cpf.val();

					BinaryReader brin = open_fstream(cp_filename);
					if (brin==null) { return; }

					int new_y_offset 
						= append_page_from_stream(brin,curr_y_offset);

					FilenameManager.close_fbitmap_input_pipe();

					curr_y_offset = new_y_offset;
					current_cpf = cp_filenames.next();
				}
			}
		}

	}
}


