using System;

namespace Espresso.OMR
{
	class SepGraphics
	{

		public enum StateEnum 
		{
			empty, bitmap, gcharacters, staff_systems, 
			non_staff_systems, 
			staff_areas, staff_areas_hproject, staff_areas_hpeaks,
			staff_objects, staff_segs, staff_segs_hproject, 
			staff_lines, staff_lines_hproject,
			corrected_bitmap, corrected_staff_objects,
			corrected_non_staff_systems,
			horizontal_count, vertical_count };


		protected Bitmap scratch_;

		protected StateEnum state_;
		protected bool is_definitive_;
		protected bool is_definitive_opposite_;
		protected bool grayed_out_;


		public SepGraphics(Bitmap scratch)
		{
			scratch_ = scratch;
			state_ = StateEnum.empty;
			is_definitive_ = false;
			is_definitive_opposite_ = false;
			grayed_out_ = false;
		}


		public void plot_bitmap_list(GObjectList sbl)
		{
			GObjectNode current;

			current = sbl.head();

			while (current!=null)
			{
				scratch_.inlay_bitmap((SubBitmap)current.val());
				current = sbl.next();
			}
		}


		/*---------------------------------------------------------------------------*/

		public void gray_out(int bl, int wl)
		{
			int tl = bl+wl;

			int true_px_dim = scratch_.x_dim();
			int true_py_dim = scratch_.y_dim();

			for (int by=0; by<true_py_dim; by++)
			{
				int sp = by % tl;

				for (int bx=0; bx<true_px_dim; bx++)
				{
					if (sp>=bl) scratch_.clear_pixel(bx,by);
					sp = (sp+1) % tl;
				}
			}

			grayed_out_ = true;
		}

		public void clear_all()
		{
			if (!grayed_out_)
			{
				scratch_.clear_all();
			}
			else
			{
				grayed_out_ = false;
			}
		}
      
		public void plot_line(int px1,int py1,
			int px2,int py2)
		{
			if (px1==px2)
			{
				// vertical line 
				int py_low,py_high;

				if (py1<py2)
				{
					py_low = py1;
					py_high = py2;
				}
				else
				{
					py_low = py2;
					py_high = py1;
				}

				for (int py=py_low; py<=py_high; py++)
					scratch_.set_pixel(px1,py);
			}
			else if (py1==py2)
			{
				// horizontal line 
				int px_low,px_high;

				if (px1<px2)
				{
					px_low = px1;
					px_high = px2;
				}
				else
				{
					px_low = px2;
					px_high = px1;
				}

				for (int px=px_low; px<=px_high; px++)
					scratch_.set_pixel(px,py1);
			}
		}

		public void plot_rectangle_spec_corners(int x_left,
			int y_top,
			int x_right, 
			int y_bot)
		{  
			plot_line(x_left,y_top,x_right,y_top);
			plot_line(x_right,y_top,x_right,y_bot);
			plot_line(x_left,y_bot,x_right,y_bot);
			plot_line(x_left,y_top,x_left,y_bot);
		}

		public void plot_rectangle(int x_left,int y_top,
			int x_dim,int y_dim)
		{
			int x_right = x_left + x_dim -1;
			int y_bot   = y_top + y_dim -1;
  
			plot_rectangle_spec_corners(x_left,y_top,x_right,y_bot);
		}

		public void plot_black_rectangle(int x_org, int y_org,
			int x_dim, int y_dim)
		{
			int x_end = x_org + x_dim;
			int y_end = y_org + y_dim;

			for (int y=y_org; y<y_end; y++)
				for (int x=x_org; x<x_end; x++)
					scratch_.set_pixel(x,y);
		}


		public void plot_graymap(byte[,] graymap,
			int gx_org, int gy_org,
			int gx_dim, int gy_dim)
		{
			int dy,gy;

			int dst_x_dim = scratch_.x_dim();
			int dst_y_dim = scratch_.y_dim();
			int ex_dim = gx_org+gx_dim;
			int ey_dim = gy_org+gy_dim;

			for (dy=gy_org,gy=0; gy<gy_dim; dy++,gy++)
				for (int dx=gx_org,gx=0; gx<gx_dim; dx++,gx++)
				{
					if (graymap[gy,gx]>0)
						scratch_.set_pixel(dx,dy);
				}
		}




		/*---------------------------------------------------------------------------*/


		public Projection calc_horizontal_count()
		{
			Projection y_project = new Projection(scratch_.y_dim());
			y_project.generate_projection(scratch_,Projection.AxisEnum.y_axis);
			return y_project;
		}

		public Projection calc_vertical_count()
		{
			Projection x_project = new Projection(scratch_.x_dim());
			x_project.generate_projection(scratch_,Projection.AxisEnum.x_axis);
			return x_project;
		}

		public void plot_horizontal_count(Projection y_project,
			int x_org, int y_org,
			bool filled_in)
		{
			if (filled_in)
			{
				for (int y=0; y<y_project.length(); y++)
					plot_line(x_org,y_org+y,x_org+y_project[y],y_org+y);
			}
			else
			{
				for (int y=1; y<y_project.length(); y++)
					plot_line(x_org+y_project[y-1],y_org+y,x_org+y_project[y],y_org+y);
			}
		}


		public void plot_screen_horizontal_count(Projection y_project,
			bool filled_in)
		{
			if (filled_in)
			{
				for (int y=0; y<y_project.length(); y++)
					plot_line(0,y,y_project[y],y);
			}
			else
			{
				for (int y=1; y<y_project.length(); y++)
					plot_line(y_project[y-1],y,y_project[y],y);
			}
		}

		public void plot_screen_vertical_count(Projection x_project,
			bool filled_in)
		{
			int lsb_yt = scratch_.y_dim() -1;

			if (filled_in)
			{
				for (int x=0; x<x_project.length(); x++)
					plot_line(x,lsb_yt,x,lsb_yt-x_project[x]);
			}
			else
			{
				for (int x=1; x<x_project.length(); x++)
					plot_line(x,lsb_yt-x_project[x-1],x,lsb_yt-x_project[x]);
			}
		}

		public void plot_horizontal_peaks(IntArray1d h_peak,
			int xl, int yt)
		{
			for (int y=0; y<h_peak.length(); y++)
			{
				if (h_peak[y]>0)
				{
					plot_line(xl,yt+y,xl+h_peak[y],yt+y);
				}
			}
		}

		public void plot_vertical_peaks(IntArray1d v_peak,
			int xl, int yb)
		{
			for (int x=0; x<v_peak.length(); x++)
			{
				if (v_peak[x]>0)
				{
					plot_line(xl+x,yb,xl+x,yb-v_peak[x]);
				}
			}
		}


		public void plot_gcharacters(GCharList gcl, GChar.SelectionEnum select)
		{
			GCharNode current_gc = gcl.head();
			while (current_gc!=null)
			{
				switch (select)
				{
					case GChar.SelectionEnum.gc_bitmap:
						scratch_.inlay_bitmap(current_gc.val());
						break;

					case GChar.SelectionEnum.gc_library_graymap:
						GChar gchar = current_gc.val();
						byte[,] library_graymap = gchar.library_graymap();
						int lg_x_org = gchar.x_org();
						int lg_y_org = gchar.y_org();
						int lg_x_dim = gchar.lg_x_dim();
						int lg_y_dim = gchar.lg_y_dim();
						plot_graymap(library_graymap,lg_x_org,lg_y_org,lg_x_dim,lg_y_dim);
						break;
				}

				current_gc = gcl.next();
			}
		}


		public void plot_gobjects(GObjectList gol)
		{
			GObjectNode current_go = gol.head();
			while (current_go!=null)
			{
				scratch_.inlay_bitmap(current_go.val());
				current_go = gol.next();
			}
		}

		public void plot_proc_staff_systems_objects(StaffSystemList ssl)
		{
			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				GObjectList staff_system_objects 
					= staff_system.processed_staff_system_objects();
				plot_gobjects(staff_system_objects);
      
				ss_current = ssl.next();
			}
		}

		public void plot_staff_systems(StaffSystemList ssl)
		{
			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				scratch_.inlay_bitmap(current_ss.val().gobject());
				current_ss = ssl.next();
			}
		}


		public void plot_staff_areas(StaffSystemList ssl)
		{
			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem ss = current_ss.val();
				StaffList stl = ss.potential_staffs();
				StaffNode current_st = stl.head();

				while (current_st!=null)
				{
					Staff staff = current_st.val();
					scratch_.inlay_bitmap(staff.staff_area());
	  
					current_st = stl.next();
				}

				current_ss = ssl.next();
			}
		}


		public void plot_staff_segs(StaffSystemList ssl, 
			int verbose_level)
		{
			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem ss = current_ss.val();
				GObject ss_gobject = ss.gobject();
				StaffList stl = ss.staffs();
				StaffNode current_st = stl.head();
				while (current_st!=null)
				{
					Staff staff = current_st.val();

					if (verbose_level<=4) scratch_.inlay_bitmap(staff.staff_area());

					if (verbose_level>=2)
					{
						GObject staff_area = staff.staff_area();
						int y_top = staff_area.y_org();
						int y_bot = y_top + staff_area.y_dim();
						int ss_x_org = ss_gobject.x_org();

						if (staff.has_left_staff_seg())
						{
							StaffSeg left_staff_seg = staff.left_staff_seg();

							if (verbose_level>4)
							{
								GObject lgobject = left_staff_seg.gobject();
								scratch_.inlay_bitmap(lgobject,ss_x_org,y_top);
							}

							int lx_start = ss_x_org + left_staff_seg.x_start();
							int lx_end = ss_x_org + left_staff_seg.x_end();
							plot_line(lx_start,y_top,lx_start,y_bot);
							plot_line(lx_end,y_top,lx_end,y_bot);
							plot_line(lx_start,y_top,lx_end,y_top);
							plot_line(lx_start,y_bot,lx_end,y_bot);
						}

						if (staff.has_right_staff_seg())
						{
							StaffSeg right_staff_seg = staff.right_staff_seg();
		  
							if (verbose_level>4)
							{
								GObject rgobject = right_staff_seg.gobject();
								scratch_.inlay_bitmap(rgobject,ss_x_org,y_top);
							}
		  
							int rx_start = ss_x_org + right_staff_seg.x_start();
							int rx_end = ss_x_org + right_staff_seg.x_end();
							plot_line(rx_start,y_top,rx_start,y_bot);
							plot_line(rx_end,y_top,rx_end,y_bot);
							plot_line(rx_start,y_top,rx_end,y_top);
							plot_line(rx_start,y_bot,rx_end,y_bot);
						}
					}

					current_st = stl.next();
				}
				current_ss = ssl.next();
			}
		}

		public void plot_staff_lines(StaffSystemList ssl, 
			int verbose_level)
		{
			StaffSystemNode current_ss = ssl.head();
			while (current_ss!=null)
			{
				StaffSystem ss = current_ss.val();
				GObject ss_gobject = ss.gobject();
				StaffList stl = ss.staffs();
				StaffNode current_st = stl.head();
				while (current_st!=null)
				{
					Staff staff = current_st.val();

					scratch_.inlay_bitmap(staff.staff_area());

					if (verbose_level>=2)
					{
						GObject staff_area = staff.staff_area();
						int y_top = staff_area.y_org();
						int y_bot = y_top + staff_area.y_dim();
	      
						StaffSeg left_staff_seg = staff.left_staff_seg();
						int ss_x_org = ss_gobject.x_org();
						int lx_start = ss_x_org + left_staff_seg.x_start();
						int lx_end = ss_x_org + left_staff_seg.x_end();
						plot_line(lx_start,y_top,lx_start,y_bot);
						plot_line(lx_end,y_top,lx_end,y_bot);
						plot_line(lx_start,y_top,lx_end,y_top);
						plot_line(lx_start,y_bot,lx_end,y_bot);
	      
						StaffSeg right_staff_seg = staff.right_staff_seg();
						int rx_start = ss_x_org + right_staff_seg.x_start();
						int rx_end = ss_x_org + right_staff_seg.x_end();
						plot_line(rx_start,y_top,rx_start,y_bot);
						plot_line(rx_end,y_top,rx_end,y_bot);
						plot_line(rx_start,y_top,rx_end,y_top);
						plot_line(rx_start,y_bot,rx_end,y_bot);
					}

					current_st = stl.next();
				}
				current_ss = ssl.next();
			}
		}

		/*--------------------*/


		public void plot_removed_staff_lines(StaffSystemList ssl)
		{
			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				StaffList staffs = staff_system.staffs();
				StaffNode st_current = staffs.head();

				while (st_current!=null)
				{
					Staff staff = st_current.val();
					StaffLines staff_lines = staff.staff_lines();
	  
					int no_staff_lines = staff.no_staff_lines();
					SubBitmapArray1d line_bitmaps = staff_lines.bitmaps();
	  
					for (int i=0; i<no_staff_lines; i++)
						scratch_.inlay_bitmap(line_bitmaps[i]);
	  
	  
					st_current = staffs.next();
				}
				ss_current = ssl.next();
			} 
		}



		public void plot_staff_without_lines(StaffSystemList ssl)
		{
			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				SubBitmap without_staffs = staff_system.without_staffs();
				scratch_.inlay_bitmap(without_staffs);
	  	  
				ss_current = ssl.next();
			} 
		}

		public void plot_bitmap(Bitmap bitmap, int x_org, int y_org)
		{ 
			scratch_.inlay_bitmap(bitmap,x_org,y_org); 
		}

		public void plot_bitmap(Bitmap bitmap)
		{ 
			scratch_.inlay_bitmap(bitmap,0,0); 
		}

		public void set_state(StateEnum state, bool is_def)
		{ 
			state_ = state; 
			is_definitive_ = is_def;
			is_definitive_opposite_ = false;
		}

		public void set_definitive_opposite(bool is_def_opp)
		{ is_definitive_opposite_ = is_def_opp; }

		public StateEnum state() { return state_; }
		public bool is_definitive() { return is_definitive_; }
		public bool is_definitive_opposite() { return is_definitive_opposite_; }

	}
}
