
using System;

namespace Espresso.OMR
{
  class SepLibrary
    {

	  protected static void parse_app_cmd_options(ref int argc, ref string[] argv)
	  {
		  int j=1;
		  int new_argc = argc;

		  for (int i=1; i<argc; i++)
		  {
			  if (argv[i].StartsWith("-"))
			  {
				  int arg_len = argv[i].Length;
				  if (arg_len>1)
				  {
					  if (argv[i].Equals("-remote"))
					  {
						  new_argc--;
						  FilenameManager.set_file_access("remote");
						  continue;
					  }
	      
					  if (argv[i].Equals("-local"))
					  {
						  new_argc--;
						  FilenameManager.set_file_access("local");
						  continue;
					  }

					  if (argv[i].Equals("-def_convert"))
					  {
						  new_argc--;
						  MObjects.set_definitive_convert(true);
						  continue;
					  }

					  if (argv[i].Equals("-def_comp"))
					  {
						  new_argc--;
						  MObjects.set_definitive_compare(true);
						  continue;
					  }

					  if (argv[i].Equals("-remove_def_text"))
					  {
						  new_argc--;
						  MObjects.set_remove_text_pending(true);
						  continue;
					  }

					  if (argv[i].Equals("-single_staff_lines"))
					  {
						  new_argc--;
						  MObjects.set_single_staff_lines(true);
						  continue;
					  }

					  if (argv[i].Equals("-insist_sl_num"))
					  {
						  i++;
						  new_argc--;
						  if (i<argc)
						  {
							  new_argc--;
							  int insist_sl_num = Int32.Parse(argv[i]);
							  MObjects.set_insist_sl_num(insist_sl_num);
							  continue;
						  }
					  }

				  }
			  }

			  argv[j++]=argv[i];
		  }

          // allocate fresh array for "argv" that is new_argc long
          string[] new_argv = new string[new_argc];
          for (int i = 0; i < new_argc; i++)
          {
              new_argv[i] = argv[i];
          }
          argv = new_argv;
		  argc = new_argc; 
	  }

      static public SepApplication general_process(int argc, string[] argv)
      {
          ShareGlobals.reset();

          parse_app_cmd_options(ref argc, ref argv);

          SepApplication sep_app = new SepApplication(argc, argv);

          sep_app.default_run();
          sep_app.default_write();

          return sep_app;
      }

      static public SepApplication custom_process(string input_fname, 
						     double cond_skew_val,
						     int verbose_level)
	{
	  // Loads in image file and processes with custom setting below
      // Natively handles PBM file format, others converted to PBM through a piped process using ImageMagick
	  // If filename has .sep extension, then the saved result of previous processing is loaded in instead.

	  int sub_argc = 25;
	  string[] sub_argv = { "seplib",
				  "-insist_sl_num", "5",
				  "-staff_system_mode", "byte",
				  "-staff_area_mode", "sa_locate_3_4",
				  "-staff_seg_pass_mode", "second_pass",
				  "-staff_seg_mode", "seg_1_3_4",
				  "-staff_line_mode", "sl_loose",
				  "-remove_staffs_mode", "remove_wobble",
				  "-correct_skew_mode", "rotate_pg",
				  "-conditional_skew", cond_skew_val.ToString(), 
				  "-print", verbose_level.ToString(),
				  "-input_filename", input_fname,
				  "-output_filename", "tmp/pracscore.sep" };

	  // customised run
	  ShareGlobals.reset();

	  parse_app_cmd_options(ref sub_argc,ref sub_argv);

	  Console.Out.Write("[" + input_fname + "] ");
	  SepApplication sep_app = new SepApplication(sub_argc,sub_argv);

	  sep_app.customised_run();
	  sep_app.customised_write();

	  return sep_app;
	}


     

    }
}      
