
using System;
using System.IO;

namespace Espresso.OMR
{
	class StaffInfo
	{
		protected int staff_height_;
		protected int gap_height_;
		protected int line_height_;
		protected double skew_; 


		public StaffInfo()
		{
			staff_height_ = 0;
			gap_height_ = 0;
			line_height_ = 0; 
			skew_ = 0.0;
		}


		public void set_staff_height(int height)
		{ 
			staff_height_ = height; 
		}

		public void set_gap_height(int height)
		{ 
			gap_height_ = height; 
		}

		public void set_line_height(int height)
		{ 
			line_height_ = height; 
		}

		public void set_skew(double skew)
		{ 
			skew_ = skew; 
		}

		public int staff_height() 
		{ 
			return staff_height_; 
		}

		public int gap_height() 
		{ 
			return gap_height_; 
		}

		public int line_height() 
		{ 
			return line_height_; 
		}

		public double skew() 
		{ 
			return (double)skew_; 
		}


		public StaffInfo ShallowCopy()
		{ 
			return (StaffInfo)MemberwiseClone();
		}

		public StaffInfo DeepCopy()
		{ 
			return ShallowCopy(); // equivalent for this class
		}


		/*---------------------------------------------------------------------------*/

		public void write(BinaryWriter bwout)

		  {
		    IOSupport.write(bwout,"staff_height = " + staff_height_.ToString());
		    IOSupport.write(bwout," , gap_height = " + gap_height_.ToString());
		    IOSupport.write(bwout," , line_height = " + line_height_.ToString());

		    // ensure skew is printed as a floating point number

		    string skew_str = String.Format("{0:F10}",skew_);
		    IOSupport.write_line(bwout," , skew = " + skew_str);
		  }


		public bool read(BinaryReader brin)
		  {
		    if (!IOSupport.read_token(brin,"staff_height")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out staff_height_)) return false;

		    if (!IOSupport.read_token(brin,",")) return false;

		    if (!IOSupport.read_token(brin,"gap_height")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out gap_height_)) return false;

		    if (!IOSupport.read_token(brin,",")) return false;

		    if (!IOSupport.read_token(brin,"line_height")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out line_height_)) return false;

		    if (!IOSupport.read_token(brin,",")) return false;

		    if (!IOSupport.read_token(brin,"skew")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;

		    double double_skew; // info.skew_ is a float because of g++ alignment bug
		    if (!IOSupport.read_double(brin,out double_skew)) return false;
		    else skew_=(float)double_skew;

		    return true;
		  }


	}
}



