using System;
using System.IO;

namespace Espresso.OMR
{

	class StaffSeg
	{

		protected GObject gobject_;
  
		protected int sa_x_org_;
		protected int sa_y_org_;
    
		protected int x_start_;
		protected int x_end_;
		protected int global_x_start_;
		protected int global_x_end_;
		protected StaffRegionArray1d lines_;
		protected StaffRegionArray1d gaps_;


		public StaffSeg(int sa_x_org, int sa_y_org)
		{
			gobject_ = new GObject();
			sa_x_org_ = sa_x_org;
			sa_y_org_ = sa_y_org;
			x_start_ = 0;
			x_end_ = 0;
			global_x_start_ = 0;
			global_x_end_ = 0;
			lines_ = new StaffRegionArray1d();
			gaps_ = new StaffRegionArray1d();
		}
    
		public StaffSeg(GObject gobject, int x_start, int x_end)
		{
			gobject_ = new GObject(gobject,x_start,0,x_end-x_start+1,gobject.y_dim(),GObject.Operation.nop);
			sa_x_org_ = gobject.x_org();
			sa_y_org_ = gobject.y_org();
			x_start_ = x_start;
			x_end_ = x_end;
			global_x_start_ = gobject.x_org()+x_start;
			global_x_end_ = gobject.x_org()+x_end;
			lines_ = new StaffRegionArray1d();
			gaps_ = new StaffRegionArray1d();
      
			generate_data();
		}


		public void generate_data()
		{
			gobject_.generate_y_project();
			Projection y_project = gobject_.y_project();
			int max_value = y_project.max_value();
			int thresh = 50*max_value/100;

			int sa_y_org = sa_y_org_;
			int y_dim = gobject_.y_dim();
	
			BoolArray1d is_staff_line = new BoolArray1d(y_dim);
			IntArray1d isl_len = new IntArray1d(y_dim); // wasteful of space but reduces code
			for (int i=0; i<y_dim; i++)
			{
				is_staff_line[i]=false;
			}

			// mark which entries are staff lines
			int no_staff_lines=0;
			int y=0;
			while (y<y_dim)
			{
				if (y_project[y]>=thresh)  // not really necessary
				{
					int y_start = y;
					while (y_project[y]>=thresh)
					{
						is_staff_line[y]=true;
						y++;
						if (y==y_dim) break;
					}
					int y_len = y - y_start;
					isl_len[no_staff_lines] = y_len;
					no_staff_lines++;
				}
				if (y==y_dim) break;
				is_staff_line[y++]=false;
			}

#if _StaffSegFilter
			double sl_median = Statistics.int_median(isl_len.data());
			int outlier_thresh = irint(sl_median*2.0); // twice as thick
	
			int no_constrained_sls = 0;
			for (i=0; i<no_staff_lines; i++)
			{
				if (isl_len[i]<=outlier_thresh) no_constrained_sls++;
			}
#else
			int outlier_thresh = int.MaxValue;
			int no_constrained_sls = no_staff_lines;
#endif

			lines_.resize(no_constrained_sls);
			if (no_constrained_sls>1) gaps_.resize(no_constrained_sls-1);
	
			int staff_line_no=0;
			int gap_no=0;
			y=0;
			while (y<y_dim)
			{
				if (is_staff_line[y])
				{
					int staff_line_start = sa_y_org+y;
					while (is_staff_line[y])
					{
						y++;
						if (y==y_dim) break;
					}
					int staff_line_end = sa_y_org+y-1;
					int sl_len = staff_line_end - staff_line_start +1;
		
					if (sl_len<=outlier_thresh)
					{
						lines_[staff_line_no].set_top_edge(staff_line_start);
						lines_[staff_line_no].set_bot_edge(staff_line_end);
						staff_line_no++;
		    
						if (staff_line_no==no_constrained_sls) break;
		    
						int gap_start = sa_y_org+y;
						while (!is_staff_line[y])
						{
							y++;
							if (y==y_dim) break;
						}
						int gap_end = sa_y_org+y-1;
		    
						gaps_[gap_no].set_top_edge(gap_start);
						gaps_[gap_no].set_bot_edge(gap_end);
						gap_no++;
					}
				}
				else y++;
			}
		}    

		public GObject gobject() { return gobject_; }
    
		public int x_start() { return x_start_; }
		public int x_end() { return x_end_;  }
    
		public int global_x_start() { return global_x_start_; }
		public int global_x_end() { return global_x_end_;  }
    
		public StaffRegionArray1d lines() { return lines_; }
		public StaffRegionArray1d gaps() { return gaps_; }


		static public void write_empty_staff_seg(BinaryWriter bwout)
		  {
		    IOSupport.write(bwout,"global_x_start = 0");
		    IOSupport.write_line(bwout," , global_x_end = 0");		    
		  }

		public void write(BinaryWriter bwout)
		  {	
		    IOSupport.write(bwout,"global_x_start = " + global_x_start_);
		    IOSupport.write_line(bwout," , global_x_end = " + global_x_end_);
		  }


		public bool read_staff_seg (BinaryReader brin)
		  {
		    if (!IOSupport.read_token(brin,"global_x_start")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out global_x_start_)) return false;
		    
		    x_start_ = global_x_start_ - sa_x_org_;
		    
		    if (!IOSupport.read_token(brin,",")) return false;
		    
		    if (!IOSupport.read_token(brin,"global_x_end")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out global_x_end_)) return false;
		    
		    x_end_ = global_x_end_ - sa_x_org_;
		    
		    return true;
		  }

	}
}
    

