using System;
using System.IO;
using System.Drawing;

namespace Espresso.OMR
{
	internal class Pixel
	{
		public enum PixelEnum { clear=0, set=1 };

		protected ByteArray2d bitmap_;  
    
		protected int x_off_;
		protected int byte_x_;
		protected byte bit_x_;
		protected int y_;

		public Pixel()
		{
			bitmap_ = new ByteArray2d();
			x_off_ = 0;
			byte_x_ = 0;
			bit_x_ = 0;
			y_ = 0;
		}

		public Pixel(Pixel pixel)
		{
			bitmap_ = pixel.bitmap_.ShallowCopy();
			x_off_ = pixel.x_off_;
			byte_x_ = pixel.byte_x_;
			bit_x_ = pixel.bit_x_;
			y_ = pixel.y_;
		}

		public Pixel(int x_off)
		{
			bitmap_ = new ByteArray2d();
			x_off_ = x_off;
			byte_x_ = 0;
			bit_x_ = 0;
			y_ = 0;
		}


		public static bool operator==(Pixel pixel_lhs, PixelEnum pixel_value_rhs)
		{
			PixelEnum pixel_value_lhs = ((pixel_lhs.bitmap_[pixel_lhs.y_,pixel_lhs.byte_x_] & pixel_lhs.bit_x_) > (int)PixelEnum.clear) ? PixelEnum.set : PixelEnum.clear;
			return pixel_value_lhs == pixel_value_rhs;

		}

		public static bool operator==(Pixel pixel_lhs, Pixel pixel_rhs)
		{
			PixelEnum pixel_value_rhs
				=((pixel_rhs.bitmap_[pixel_rhs.y_,pixel_rhs.byte_x_] & pixel_rhs.bit_x_)> (int)PixelEnum.clear) ? PixelEnum.set : PixelEnum.clear;
			PixelEnum pixel_value_lhs 
				= ((pixel_lhs.bitmap_[pixel_lhs.y_,pixel_lhs.byte_x_] & pixel_lhs.bit_x_) > (int)PixelEnum.clear) ? PixelEnum.set : PixelEnum.clear;

			return pixel_value_lhs == pixel_value_rhs;

		}

		public static bool operator!=(Pixel pixel_lhs, PixelEnum pixel_value_rhs)
		{
			return !(pixel_lhs == pixel_value_rhs);
		}

		public static bool operator!=(Pixel pixel_lhs, Pixel pixel_rhs)
		{
			return !(pixel_lhs == pixel_rhs);
		}

	

		public override bool Equals(object obj)
		{
			if (!(obj is Pixel))
			{
				return false;
			}
			return this == (Pixel) obj;
		}

		public override int GetHashCode() 
		{
			// not a particulary smart hashing function.  Is hashing used, this should be replaced
			return bitmap_.x_dim() + bitmap_.y_dim(); 
		}

		/*
		public Pixel operator =(byte pixel_value)
		{
			if (pixel_value==PixelEnum.clear) bitmap_[y_,byte_x_] &= ~bit_x_;
			else bitmap_[y_,byte_x_] |= bit_x_;

			return this;
		}

		public Pixel operator =(Pixel b_right)
		{
			int pixel_value 
				= (b_right.bitmap_[b_right.y_,b_right.byte_x_] & b_right.bit_x_);

			if (pixel_value==PixelEnum.clear) bitmap_[y_,byte_x_] &= ~bit_x_;
			else bitmap_[y_,byte_x_] |= bit_x_;

			return b_right; // instead of the more usual 'this'
		}
*/

		public static PixelEnum operator ~(Pixel pixel)
		{
			return ((byte)(pixel.bitmap_[pixel.y_,pixel.byte_x_] & pixel.bit_x_)>0) ? PixelEnum.clear : PixelEnum.set;
		}
    
		public void set_pixel()
		{
			// use stored values that represent pixel location (y_, x_byte_ and bit_x_)
			bitmap_[y_,byte_x_] = (byte)(bitmap_[y_,byte_x_] | bit_x_);
		}

		public void clear_pixel()
		{
			// use stored values that represent pixel location (y_, x_byte_ and bit_x_)
			bitmap_[y_,byte_x_] = (byte)(bitmap_[y_,byte_x_] & ~bit_x_);
		}



		public void set_x_off(int x_off) { x_off_ = x_off; }


	}


	/*-------------------*/

	internal class IncompletePixel : Pixel
	{
		public IncompletePixel()
			: base()
		{}

		public IncompletePixel(IncompletePixel incomplete_pixel)
			: base(incomplete_pixel)
		{
		}

		public IncompletePixel(int x_off)
			: base(x_off)
		{}

		public bool pixel_is_set(int x, int y)
		{
			int ox = x + x_off_;
			return (bitmap_[y,ox>>3] & (1<<(7-(ox%8))))>(int)PixelEnum.clear;
		}

		/*
		public Pixel operator[](int x)
		{
			IncompletePixel store_value = (IncompletePixel)(this);

			int ox = x + x_off_;
			store_value.byte_x_ = ox>>3;
			store_value.bit_x_ = 1<<(7-(ox%8));
	
			return (Pixel) this;
		}


			public Pixel operator[](int x)
			{
				int ox = x + x_off_;
				byte_x_ = ox>>3;
				bit_x_ = 1<<(7-(ox%8));
	
				return (Pixel) this;
			}
			*/

	}

	internal class Bitmap : IncompletePixel
	{
		protected int x_dim_;
		protected int y_dim_;
		protected byte x_left_mask_;
		protected byte x_right_mask_;  

		public Bitmap()
			: base()
		{}

		public Bitmap(Bitmap bitmap)
			: base(bitmap)
		{
			x_dim_ = bitmap.x_dim_;
			y_dim_ = bitmap.y_dim_;
			x_left_mask_ = bitmap.x_left_mask_;
			x_right_mask_ = bitmap.x_right_mask_;
		}

		public Bitmap(int x_dim, int y_dim, int x_off)
			: base(x_off)
		{
			x_dim_ = x_dim;
			y_dim_ = y_dim;

			if ((x_dim>0) && (y_dim>0))
			{
				bitmap_.resize((x_off+x_dim+7)>>3,y_dim);

				// compute x-edge extraction 
				x_left_mask_ = (byte)((1<<(8-x_off_))-1);
				int x_end = (x_off_+x_dim_-1)%8;
				int pad_zero = 255<<8;
				x_right_mask_ = (byte)((pad_zero>>(x_end+1))&255);
			}
		}

		public Bitmap(int x_dim, int y_dim)
			: this(x_dim,y_dim,0)
		{}

		public void extract(Bitmap src_bitmap, 
			int dx_org, int dy_org,
			int dx_dim, int dy_dim)
		{
			Bitmap dst_bitmap = this;
			ByteArray2d dst_array = dst_bitmap.bitmap_;
			ByteArray2d src_array = src_bitmap.bitmap_;

			int lx_off = (src_bitmap.x_off_+dx_org)%8;
			byte lx_left_mask = (byte)((1<<(8-lx_off))-1);
			int dx_end = (lx_off+dx_dim-1)%8;
			int pad_zero = 255<<8;
			byte lx_right_mask = (byte)((pad_zero>>(dx_end+1))&255);

			int dax_dim = dst_array.x_dim();
			int day_dim = dst_array.y_dim();
			int sax_org = (src_bitmap.x_off_+dx_org)>>3;
			int say_org = dy_org;

			if (x_off_+dx_dim<=8)
			{
				byte lx_lr_mask = (byte)(lx_left_mask & lx_right_mask);

				for (int dy=0,sy=say_org; dy<day_dim; dy++,sy++)
					dst_array[dy,0] = (byte)(src_array[sy,sax_org] & lx_lr_mask);
			}
			else
			{
				for (int dy=0,sy=say_org; dy<day_dim; dy++,sy++)
				{
					int dx,sx;

					dst_array[dy,0] = (byte)(src_array[sy,sax_org] & lx_left_mask);
					for (dx=1,sx=sax_org+1; dx<dax_dim-1; dx++,sx++)
						dst_array[dy,dx] = src_array[sy,sx];
					dst_array[dy,dx] = (byte)(src_array[sy,sx] & lx_right_mask);
				}
			}
		}


		public void extend(int x_dim, int y_dim)
		{ 
			bitmap_.extend((x_off_+x_dim+7)>>3,y_dim); 
			x_dim_ = x_dim;
			y_dim_ = y_dim;
		}

		public void resize(int x_dim, int y_dim)
		{
			x_dim_ = x_dim;
			y_dim_ = y_dim;
			int byte_x_dim = (x_off_+x_dim+7)>>3;
			bitmap_.resize(byte_x_dim,y_dim);

			x_left_mask_ = (byte)((1<<(8-x_off_))-1);
			int x_end = (x_off_+x_dim_-1)%8;
			int pad_zero = 255<<8;
			x_right_mask_ = (byte)((pad_zero>>(x_end+1))&255);
		}


		public void clear_all()
		{
			for (int y=0; y<y_dim_; y++)
				for (int x=0; x<bitmap_.x_dim(); x++)
					bitmap_[y,x]=0;
		}

		public void clear_area(int x_org, int y_org, 
			int x_dim, int y_dim)
		{
			// compute x-edge extraction 
			int x_off = x_org%8;
			byte x_left_mask = (byte)((1<<(8-x_off))-1);
			int x_end = (x_off+x_dim-1)%8;
			int pad_zero = 255<<8;
			byte x_right_mask = (byte)((pad_zero>>(x_end+1))&255);

			int ax_org = x_org>>3;
			int ay_org = y_org;
			int ax_dim = (x_off+x_dim+7)>>3;
			int ay_dim = y_dim;


			if (x_off+x_dim<=8)
			{
				byte x_lr_mask = (byte)(x_left_mask & x_right_mask);

				for (int y=ay_org; y<ay_org+ay_dim; y++)
					bitmap_[y,ax_org] = (byte)(bitmap_[y,ax_org] & ~x_lr_mask);
			}
			else
			{
				for (int y=ay_org; y<ay_org+ay_dim; y++)
				{
					int x;

					bitmap_[y,ax_org] = (byte)(bitmap_[y,ax_org] & ~x_left_mask);
					for (x=ax_org+1; x<ax_org+ax_dim-1; x++)
						bitmap_[y,x] = 0;
					bitmap_[y,x] = (byte)(bitmap_[y,x] & ~x_right_mask);
				}
			}


		}

		public void inlay_bitmap(Bitmap src_bitmap,int sx_org, int sy_org)
		{
			Bitmap dst_bitmap = this;
			ByteArray2d dst_array = dst_bitmap.bitmap_;
			ByteArray2d src_array = src_bitmap.bitmap_;

			int sax_dim = src_array.x_dim();
			int say_dim = src_array.y_dim();

			int dax_org = (dst_bitmap.x_off_+sx_org)>>3;
			int day_org = sy_org;

			byte sx_left_mask = src_bitmap.x_left_mask_;
			byte sx_right_mask = src_bitmap.x_right_mask_;

			if (src_bitmap.x_off_+src_bitmap.x_dim_<=8)
			{
				byte sx_lr_mask = (byte)(sx_left_mask & sx_right_mask);

				for (int sy=0,dy=day_org; sy<say_dim; sy++,dy++)
					dst_array[dy,dax_org] = (byte)(dst_array[dy,dax_org] | (src_array[sy,0] & sx_lr_mask));
			}
			else
			{
				for (int sy=0,dy=day_org; sy<say_dim; sy++,dy++)
				{
					int dx,sx;

					dst_array[dy,dax_org] = (byte)(dst_array[dy,dax_org] | (src_array[sy,0] & sx_left_mask));
					for (sx=1,dx=dax_org+1; sx<sax_dim-1; sx++,dx++)
						dst_array[dy,dx] = (byte)(dst_array[dy,dx] | src_array[sy,sx]);
					dst_array[dy,dx] = (byte)(dst_array[dy,dx]  | (src_array[sy,sx] & sx_right_mask));
				}
			}
		}

		public void inlay_bitmap(SubBitmap src_bitmap)
		{ 
			inlay_bitmap((Bitmap)src_bitmap,src_bitmap.x_org(),src_bitmap.y_org()); 
		}

		public void remove_bitmap(SubBitmap src_bitmap)
		{
			remove_bitmap((Bitmap)src_bitmap,src_bitmap.x_org(),src_bitmap.y_org()); 
		}

		public void remove_bitmap(Bitmap src_bitmap,
			int sx_org, int sy_org)
		{
			// it is assumed that the src_bitmap fits inside dst_bitmap
			Bitmap dst_bitmap = this;

			int src_x_dim = src_bitmap.x_dim();
			int src_y_dim = src_bitmap.y_dim();

			for (int sy=0,dy=sy_org; sy<src_y_dim; sy++,dy++)
				for (int sx=0,dx=sx_org; sx<src_x_dim; sx++,dx++)
				{
					if (src_bitmap[sy,sx]==PixelEnum.set) 
					{								
						dst_bitmap.clear_pixel(dx,dy);
					}
				}
		}

        
        public int x_off() { return x_off_; }
		public int x_dim() { return x_dim_; }
		public int y_dim() { return y_dim_; }
		public ByteArray2d array() { return bitmap_; }

		public Bitmap this [int y, int x]
		{
			get
			{
				// store y value (incase referred to as a Pixel in == or != operator)
				y_ = y; 

				// store x value encoded as byte_x_ and bit_x_ mask within byte
				int ox = x + x_off_;
				byte_x_ = ox>>3;
				bit_x_ = (byte)(1<<(7-(ox%8)));

				return this;

				// return bitmap_[y,x];
			}
			set
			{
				// work out [y,x] position in bitmap, but do not overwrite y_, byte_x_ etc.
				// This is so rhs Pixel can refer to another [y,x] location in the same
				// Bitmap.

				int lhs_y = y; 

				// work out x byte and mask
				int ox = x + x_off_;
				int lhs_byte_x = ox>>3;
				int lhs_bit_x = 1<<(7-(ox%8));

				// 'value' is a Pixel which stores the particular [y,x] position we are interested in 

				int pixel_byte
					= (value.bitmap_[value.y_,value.byte_x_] & value.bit_x_);

				if (pixel_byte==(int)PixelEnum.clear)
				{
					bitmap_[lhs_y,lhs_byte_x] = (byte)(bitmap_[lhs_y,lhs_byte_x] & ~lhs_bit_x);
				}
				else 
				{
					bitmap_[lhs_y,lhs_byte_x] = (byte)(bitmap_[lhs_y,lhs_byte_x] | lhs_bit_x);
				}
			}
		}

		public void set_pixel(int x, int y)
		{
			// work out x byte and mask
			int ox = x + x_off_;
			int byte_x = ox>>3;
			int bit_x = 1<<(7-(ox%8));
		
			bitmap_[y,byte_x] = (byte)(bitmap_[y,byte_x] | bit_x);
		
		}

		public void clear_pixel(int x, int y)
		{
			// work out x byte and mask
			int ox = x + x_off_;
			int byte_x = ox>>3;
			int bit_x = 1<<(7-(ox%8));
			
			bitmap_[y,byte_x] = (byte)(bitmap_[y,byte_x] & ~bit_x);			
		}

		/*
			public IncompletePixel operator [](int y)
			{ 
				y_=y;
				return (IncompletePixel) this;
			}
    

			*/
		/*
    
					public char* cdata() { return bitmap_.cdata(); }
					public char** pdata() { return bitmap_.pdata(); }
					*/
		/*
							public Bitmap operator =(Bitmap src_bitmap)
							{
								int sx_dim = (src_bitmap.x_dim()>>3)+1;
								int sy_dim = src_bitmap.y_dim();
								char** raw_data = src_bitmap.pdata(); // change this to chararray2d
      
								for (int y=0; y<sy_dim; y++)
									for (int x=0; x<sx_dim; x++)
										bitmap_[y,x]=raw_data[y,x];
      
								return src_bitmap; // instead of the more usual 'this'
							}
		  */
  
		public Bitmap DeepCopy()
		{
			Bitmap copy_bitmap = new Bitmap(x_dim_,y_dim_);
			bitmap_ = bitmap_.ShallowCopy(); // shallow copy of 2d array of 'byte' is equivalent to deepcopy

			return copy_bitmap;
		}

		public bool coord_in_bitmap(int x,int y)
		{
			if (x<0) return false;
			if (x>=x_dim_) return false;
			if (y<0) return false;
			if (y>=y_dim_) return false;
			return true;
		}
    
    
		public void replace(Bitmap bitmap)
		{ 
			bitmap_.replace(bitmap.bitmap_); 
			x_dim_ = bitmap.x_dim_;
			y_dim_ = bitmap.y_dim_;  
		}
    
		public void exchange(Bitmap bitmap) 
		{ 
			bitmap_.exchange(bitmap.bitmap_); 
	
			int temp_x_dim = x_dim_;
			x_dim_ = bitmap.x_dim_;
			bitmap.x_dim_ = temp_x_dim;
	
			int temp_y_dim = y_dim_;
			y_dim_ = bitmap.y_dim_;
			bitmap.y_dim_ = temp_y_dim;
		}

		public byte[,] data()
		{
			return bitmap_.data();
		}



		/*---------------------------------------------------------------------------*/


		public void fwrite(BinaryWriter brout)
		  {
		    // write out file in pbm format

		    int x_org=0;
		    int y_org=0;

		    int x_sub_dim = x_dim_;
		    int y_sub_dim = y_dim_;

		    Console.Out.WriteLine("Writing out file - Please wait");

		    string magic_no = "P4";

		    IOSupport.write_line(brout,magic_no);
		    IOSupport.write_line(brout,x_sub_dim.ToString() + " " + y_sub_dim.ToString());

		    int byte_x_org = (x_org +7)>>3;
		    int byte_x_dim = (x_sub_dim+7)>>3; 

		    // this can be improved by writing out a line at a time

		    for (int y=0; y<y_sub_dim; y++)
		      for (int x=0; x<byte_x_dim; x++)
			brout.Write(bitmap_[y_org+y,byte_x_org+x]);
		  }

		public void output_pbm(BinaryWriter brout)
		  {
		    // write out file in pbm format

		    IOSupport.write_line(brout,"P4");
		    IOSupport.write_line(brout,x_dim_.ToString() + " " + y_dim_.ToString());

		    int byte_x_dim = (x_dim_+7)>>3; 

		    for (int y=0; y<y_dim_; y++)
		      for (int x=0; x<byte_x_dim; x++)
			brout.Write(bitmap_[y,x]);


		    brout.Flush();
		  }


		/*---------------------------------------------------------------------------*/

		public virtual void write(BinaryWriter brout)
		  {

		    IOSupport.write_line(brout,"x_dim = " + x_dim_ + " , y_dim = " + y_dim_);

		    ByteArray2d byte_array = bitmap_;
		    int byte_x_dim = byte_array.x_dim();
		    int byte_y_dim = byte_array.y_dim();

		    int pbyte_x_dim = (x_off_+x_dim_+7)>>3;
		    if (pbyte_x_dim!=byte_x_dim)
		      {
			Console.Error.Write("byte_x_dim's different: ");
			Console.Error.Write(" x_off = " + x_off_.ToString() + " and ");
			Console.Error.Write(" x_dim = " + x_dim_.ToString());
			Console.Error.Write(" (calc byte xdim = " + pbyte_x_dim.ToString() + ") But ");
			Console.Error.WriteLine(" byte_x_dim = " + byte_x_dim.ToString());
		      }

			/*
		    if ((brout!=Console.Out) && (brout!=Console.Error))
		      {
			  */

			for (int y=0; y<byte_y_dim; y++)
			  for (int x=0; x<byte_x_dim; x++)
			    brout.Write(byte_array[y,x]);
			IOSupport.write_line(brout);

			/*
		      }
		    else IOSupport.write_line(brout,"[data]");
			*/

		    brout.Flush();
		  }


		public virtual bool read(BinaryReader brin)
		  {
		    // object (this) assumed to be empty prior to call

		    int x_dim;
		    if (!IOSupport.read_token(brin,"x_dim")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out x_dim)) return false;

		    if (!IOSupport.read_token(brin,",")) return false;
  
		    int y_dim;
		    if (!IOSupport.read_token(brin,"y_dim")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out y_dim)) return false;

		    resize(x_dim,y_dim);

		    // calculate left and right masks
		    x_left_mask_ =  (byte)((1<<(8-x_off_))-1);
		    int pad_zero = 255<<8;
		    int x_end = (x_off_+x_dim_-1)%8;
		    x_right_mask_ = (byte)((pad_zero>>(x_end+1))&255);

		    ByteArray2d byte_array = bitmap_;
		    int byte_x_dim = (x_off_+x_dim_+7)>>3;
		    int byte_y_dim = y_dim_;

		    // consume carriage return
		    byte dummy = brin.ReadByte();

			for (int y=0; y<byte_y_dim; y++)
			{
				for (int x=0; x<byte_x_dim; x++)
				{
					byte_array[y,x] = brin.ReadByte();
				}
			}

		    return true;
		  }

	}
}



    
