
using System;

namespace Espresso.OMR
{
	class BresenCircle
	{

		protected const double two_pi_ = 2.0 * Math.PI;
		protected readonly int radius_;
      
		protected PointArray1d second_oct_;
		protected int second_oct_len_;
		protected PointArray1d complete_circle_;
		protected int complete_circle_len_;
      

		public BresenCircle(int radius)
		{
			radius_ = radius;
			second_oct_ = new PointArray1d(radius/4);
			complete_circle_ = new PointArray1d();

			int x = 0;
			int y = radius;

			int d = 1-radius;
			int delta_e = 3;
			int delta_se = -2*radius+5;

			int p=0;
			second_oct_[p].x(x);
			second_oct_[p++].y(y);

			while (y>x)
			{
				if (d<0) // select east
				{
					d += delta_e;
					delta_e +=2;
					delta_se +=2;
					x++;
				}
				else     // select south east
				{
					d += delta_se;
					delta_e += 2;
					delta_se += 4;
					x++;
					y--;
				}
				if (p==second_oct_.length()) second_oct_.extend(2*p);
				second_oct_[p].x(x);
				second_oct_[p++].y(y);
			}

			// set up complete circle

			second_oct_len_ = p;
			complete_circle_.resize(second_oct_len_*8);
			complete_circle_len_ = complete_circle_.length();

			// direct copy
			for (p=0; p<second_oct_len_; p++)
				complete_circle_[p]=second_oct_[p];

			// mirror in y=x
			int first_oct_base = 2*second_oct_len_-1;
			for (p=0; p<second_oct_len_; p++)
			{
				x = complete_circle_[p].x();
				y = complete_circle_[p].y();
				complete_circle_[first_oct_base-p].x(y);
				complete_circle_[first_oct_base-p].y(x);
			}

			// mirror in y=0
			int fourth_quad_base = 4*second_oct_len_-1;
			for (p=0; p<2*second_oct_len_; p++)
			{
				x = complete_circle_[p].x();
				y = complete_circle_[p].y();
				complete_circle_[fourth_quad_base-p].x(x);
				complete_circle_[fourth_quad_base-p].y(-y);
			}

			// mirror in x=0
			int second_demi_base = 8*second_oct_len_-1;
			for (p=0; p<4*second_oct_len_; p++)
			{
				x = complete_circle_[p].x();
				y = complete_circle_[p].y();
				complete_circle_[second_demi_base-p].x(-x);
				complete_circle_[second_demi_base-p].y(y);
			}

		}

		/*  if (fmod(type,180.0)<90.0)  **
		{
		// hit on y
		int local_p = MathExtra.irint((double)radius_ * sin(two_pi_*theta/360.0));

		if (MathExtra.fmod(type,360.0)<180.0) final_p = quad_len-local_p;   // rhs
		else final_p = 3*quad_len+local_p;                        // lhs
		}
		else 
		{
		// hit on x
		int complete_circle_len = complete_circle_.len();
		int local_p = MathExtra.irint((double)radius_ * cos(two_pi_*theta/360.0));

		if (MathExtra.fmod(theta,360.0)<180.0) // rhs
		final_p = (local_p+complete_circle_len)%complete_circle_len;
		else final_p = 2*quad_len-local_p; // lhs
		} */


		private int find_index(double theta)
		{
			double type =MathExtra.fmod(theta+45.0,360.0);
			if (type<0) { type += 360.0; }

			int quad_len = 2*second_oct_len_;
			int final_p;

			if (MathExtra.fmod(type,180.0)<90.0) 
			{
				// hit on x (remember x now +90 rot => use sin)
				int local_p = MathExtra.irint((double)radius_ * Math.Sin(two_pi_*theta/360.0));

				if (MathExtra.fmod(type,360.0)<180.0) 
					final_p = (complete_circle_len_+local_p)%complete_circle_len_; // rhs
				else final_p = 2*quad_len-local_p;                               // lhs
			}
			else 
			{
				// hit on y (remember x now +90 rot => use cos)
				int local_p = MathExtra.irint((double)radius_ * Math.Cos(two_pi_*theta/360.0));

				if (MathExtra.fmod(theta,360.0)<180.0) final_p = quad_len-local_p;         // rhs
				else final_p = 3*quad_len+local_p;                               // lhs
			}

			return(final_p);
		}

  
		public void arc_range(double theta_start, double theta_end,
			out int p_low, out int p_high)
		{
			p_low = find_index(theta_start);
			p_high = find_index(theta_end);

			if (p_high<=p_low) p_high+=complete_circle_len_;
		}

		public int length() { return complete_circle_.length(); }

		public Point data(int p)
		{ 
			return complete_circle_[p%complete_circle_len_]; 
		}
	}
}
