using System;

namespace Espresso.OMR
{
	class BresenLine
	{

		protected PointArray1d line_;

		public BresenLine()
		{
			line_ = new PointArray1d();
		}

		public BresenLine(int x_start,int y_start, int x_end, int y_end)
			: this()
		{
			generate_line(x_start,y_start,x_end,y_end);
		}

		public void generate_line(int x_start,int y_start, 
			int x_end, int y_end)
		{
			int delta_x = x_end - x_start;
			int delta_y = y_end - y_start;
			int abs_delta_x = Math.Abs(delta_x);
			int abs_delta_y = Math.Abs(delta_y);
  
			int incr_dep, incr_var, diff;
			if (abs_delta_x>abs_delta_y)
			{
				incr_dep =  2*abs_delta_y;
				incr_var = 2*(abs_delta_y - abs_delta_x);
				diff = 2*abs_delta_y - abs_delta_x;
			}
			else
			{
				incr_dep =  2*abs_delta_x;
				incr_var = 2*(abs_delta_x - abs_delta_y);
				diff = 2*abs_delta_x - abs_delta_y;
			}

			int incr_x = (delta_x>0) ? 1 : -1;
			int incr_y = (delta_y>0) ? 1 : -1;

			Point pt = new Point(x_start,y_start);
			Point.IncrementDelegate dep_incr;
			Point.IncrementDelegate var_incr;
			int unit_dep, unit_var;
			int dep_len;

			if (abs_delta_x>abs_delta_y)
			{
				// oct 0,3,4,7
				dep_incr = new Point.IncrementDelegate(pt.x_incr);
				var_incr = new Point.IncrementDelegate(pt.y_incr);
				unit_dep = incr_x;
				unit_var = incr_y;
				dep_len = abs_delta_x+1;
			}
			else
			{
				// oct 1,2,5,6
				dep_incr = new Point.IncrementDelegate(pt.y_incr);
				var_incr = new Point.IncrementDelegate(pt.x_incr);
				unit_dep = incr_y;
				unit_var = incr_x;
				dep_len = abs_delta_y+1;
			}

			line_.resize(dep_len); 

			for (int d=0; d<dep_len; d++)
			{
				line_[d]=pt;
				if (diff <=0)
				{
					diff += incr_dep;
					dep_incr(unit_dep); // changes pt
				}
				else
				{
					diff += incr_var;
					dep_incr(unit_dep); // changes pt
					var_incr(unit_var); // changes pt
				}
			}
		}

		public int length() 
		{ 
			return line_.length(); 
		}

		public Point data(int p)
		{ 
			return line_[p]; // 'p' trusted to be in range
		} 

		public PointArray1d line()
		{ 
			return line_; 
		} 
	}
}


      
