

using System;
using System.IO;

namespace Espresso.OMR
{
	class FileBitmap : Bitmap
	{
		protected string filename_; 
		protected bool valid_;
		protected string error_;

		public FileBitmap()
			: base()
		{
			valid_ = false;
			filename_ = ""; 
			error_ = ""; 
		}

		/*    
		void read_file()
	  {
		fs = new FileStream(FILE_NAME, FileMode.Open, FileAccess.Read);
		BinaryReader r = new BinaryReader(fs);
		// Read data from Test.data.
		for (int i = 0; i < 11; i++) 
		  {
			Console.WriteLine(r.ReadInt32());
		  }
		w.Close();

	  }
		*/

		protected string get_text_line(BinaryReader brin , int limit, char eot)
		{
			char[] buffer = new char[limit];
			int i = 0;

			while (true)
			{	char b = brin.ReadChar();
				if (b==eot) { break; }

				buffer[i++] = b;

				if (i>=limit) { break; }
			}
			int length = i;

			return new string(buffer,0,length);
		}


		protected void get_next_text(BinaryReader brin ,out string text_str)
		{
			bool got_text_start = false;
			
			char b = brin.ReadChar();
			while (!got_text_start)
			{		  
				while ((b < '!') || (b>'~')) // i.e. not a printable character
				{
					b = brin.ReadChar();
				}

				// b = first printable character

				if (b=='#')
				{
					// skip to end of line then repeat above 
					string comment = get_text_line(brin,256,'\n');
					
#if EXTENDED_PBM_FORMAT
				  // extended feature to pbm format
				  if (comment[0]=='#')
				    {
				      // specifies global parameter setting
					ShareGlobals.initialise_variable(comment[1]);
					}
#endif
					b = brin.ReadChar(); 
				}
				else 
				{
					got_text_start=true;
				}
			}
  
			char[] text = new char[256];
			int i=0;

			while ((b >= '!') && (b <= '~'))
			{
				text[i]=b;
				b= brin.ReadChar();
				i++;

				if (i>=256) { break; }
			}
  
			int text_length = i;

			text_str = new string(text,0,text_length);
  
		} /* End of get next text */

		bool read_file_bitmap_header(BinaryReader brin,
			out string magic_no, 
			out int x_dim, out int y_dim)
		{
			get_next_text(brin,out magic_no);
  
			if (magic_no!="P4")
			{
				x_dim = 0;
				y_dim = 0;
				return false;
			} 

			string x_dim_str;
			string y_dim_str;
      
			get_next_text(brin,out x_dim_str);
			get_next_text(brin,out y_dim_str);

			x_dim = Int32.Parse(x_dim_str);
			y_dim = Int32.Parse(y_dim_str);

			return true;
		}

		protected bool read_file_bitmap_body(BinaryReader brin, int byte_x_dim, 
			int y_dim)
		{      
			int len = byte_x_dim*y_dim;
      
			byte[,] byte_array2d  = bitmap_.data();

			for (int y=0; y<y_dim; y++)
			{
				for (int x=0; x<byte_x_dim; x++)
				{
					byte_array2d[y,x] = brin.ReadByte();
				}
			}

			/* 
			  
			if (brin.Read(bitmap_.data(),0,len)!=len)
			{
				return false;
			}
			*/

			return true;
		}


		public void replace(BinaryReader brin)
		{
			// load file_bitmap

			string magic_no;
			if (!read_file_bitmap_header(brin, out magic_no, out x_dim_, out y_dim_))
			{
				error_ = "Unsupported magic number " + magic_no;
      
				x_dim_=y_dim_=0; 
				valid_=false;
				return;
			}

			int byte_x_dim=(x_dim_+7)>>3;
  
			bitmap_.resize(byte_x_dim, y_dim_);
  
			if (!read_file_bitmap_body(brin,byte_x_dim,y_dim_))
			{
				error_ = "Body of bitmap file " + filename_ + " corrupt";
      
				x_dim_=y_dim_=0;
				valid_=false;
				return;
			}
		}


		public void set_valid(bool valid) { valid_ = valid; }
		public void set_error(string error) { error_ = error; } 
		public void set_filename(String filename) { filename_ = filename; }
    
		public bool valid() { return valid_; }
		public string error() { return error_; }
		public string filename() { return filename_; }
	}
}

