using System;
using System.IO;
using System.Text.RegularExpressions;

namespace Espresso.OMR
{
	
	class FilenameManager
	{

		public enum FileTypeEnum 
		{
			unknown, scanned_bitmap, pbm, 
			dsep, sep, prim, ass, bass };
#if UNIX
        protected static Regex full_filename_prefix = new Regex("^/");
        protected static string dir_sep = "/";
#else
        protected static Regex full_filename_prefix = new Regex("^[A-Z]:\\\\");
        protected static string dir_sep = "\\"; 
#endif

		protected static InputFileAccess.Type file_access_ = InputFileAccess.Type.local;

		protected static string full_filename_;
		protected static string root_filename_;
		protected static string root_dir_;
		protected static string filename_;

		protected static string separate_filename_;
		protected static string primitive_filename_;
		protected static string assembly_filename_;
		protected static string bassembly_filename_;

		protected static string added_ending_ = "";
		protected static FileTypeEnum file_type_;
		protected static int op_no_ = 0;

		protected static InputFileAccess input_file_access_    = null;

		protected static InputFileAccess input_fbitmap_file_   = null;
		protected static InputFileAccess input_file_           = null;
		protected static InputFileAccess input_prim_file_      = null;
		protected static InputFileAccess input_ass_file_       = null;
		protected static InputFileAccess input_bass_file_      = null;
		protected static InputFileAccess input_glob_file_      = null;
		protected static InputFileAccess input_empirical_file_ = null;

		protected static FileStream output_file_          = null;
		protected static BinaryWriter output_bw_          = null;

      
		public static void set_file_access(string file_access)
		{

			if (file_access == "local")
			{
				file_access_ = InputFileAccess.Type.local;
				input_file_access_ = new LocalFileAccess();
				
			}
			else if (file_access == "remote")
			{
				file_access_ = InputFileAccess.Type.remote;
				input_file_access_ = new RemoteFileAccess();
			}
			else
			{
				Console.Error.Write("set_file_access: ");
				Console.Error.WriteLine("Attempting to set file access to unknown type.");
			}
		}

		/*
		public static void expand_reg_filename(string reg_filename,
					 stringArray1d filenames)
	  {
		string full_reg_filename;
		int break_pos;

		if (!reg_filename.contains(dir_sep,0))
		  {
			char fname_buf[256];
			string cwd = getcwd(fname_buf,254);
			if (cwd == null)
		  {
			cerr << "expand_reg_filename: call to 'getcwd' failed" << endl;
			exit(1);
		  }
			strcat(fname_buf,dir_sep);
			full_reg_filename =  cwd + reg_filename;
			break_pos = strlen(cwd);      
		  }
		else 
		  {
			full_reg_filename = reg_filename;
			break_pos=0;
		  }
  
		char cmd[256];
		ostrstream oss_expanded(cmd,256);
		oss_expanded << fexpand_str_ << "'" << full_reg_filename << "'" << ends;

#if defined _darwin || __linux__
		FILE expanded_file = popen(cmd,"r");
#else
		FILE expanded_file = popen(cmd,"rb");
#endif
		if (expanded_file==null) {
		  cerr << "Error opening pipe to expanded file:\n  " << cmd << endl;
		  exit(1);
		}

  #ifdef _Gpp2
		int efd = fileno(expanded_file);
		ifstream efs(efd);
#else
		filemm::streambuf *sb = new filemm::streambuf(expanded_file);
		istream efs = new istream(sb);
#endif

		int filename_num;
		efs >> filename_num;
		filenames.resize(filename_num);

		string expanded_filename;

		for (int i=0; i<filename_num; i++)
		  {
			efs >> expanded_filename;
			filenames[i] = (Substring&)expanded_filename.from(break_pos);
		  }

  #ifdef _Gpp2
		pclose(expanded_file);
#endif
	  }
  */


		protected static bool try_adding_compressed_ending(ref string poss_filename)
		{
			// try adding various compression filename extensions and see if you
			// get a hit.  Update poss_filename accordingly

			if (input_file_access_.file_exists(poss_filename + ".Z"))
			{
				poss_filename += ".Z";
				added_ending_ += ".Z";
				return true;
			}

			if (input_file_access_.file_exists(poss_filename + ".z"))
			{
				poss_filename += ".z";
				added_ending_ += ".z";
				return true;
			}
  
			if (input_file_access_.file_exists(poss_filename + ".gz"))
			{
				poss_filename += ".gz";
				added_ending_ += ".gz";
				return true;
			}

			return false;
	  

		}


		protected static int string_char_freq(string s, char c)
		{
			int count = 0;

			for (int i=0; i<s.Length; i++)
			{
				if (s[i]==c)
				{
					count++;
				}
			}

			return count;
		}

		protected static void complete_filename()
		{

			int root_break = full_filename_.LastIndexOf(dir_sep);
			string only_filename = full_filename_.Substring(root_break+1);
			int freq = string_char_freq(only_filename,'.');

			file_type_ = FileTypeEnum.scanned_bitmap;
			if (only_filename.LastIndexOf(".pbm")>0) file_type_ = FileTypeEnum.pbm;
			if (only_filename.LastIndexOf(".dsep")>0) file_type_ = FileTypeEnum.dsep;
			if (only_filename.LastIndexOf(".sep")>0) file_type_ = FileTypeEnum.sep;
			if (only_filename.LastIndexOf(".prim")>0) file_type_ = FileTypeEnum.prim;
			if (only_filename.LastIndexOf(".ass")>0) file_type_ = FileTypeEnum.ass;
			if (only_filename.LastIndexOf(".bass")>0) file_type_ = FileTypeEnum.bass;

			// **** 
			// Artificially terminated -- filename on command line
			//   must match exactly
			Console.Error.WriteLine("Warning: complete_filename() artifically terminated -- filename must match exactly");
			Console.Error.WriteLine("         Termination appears to have been introduced when upgrading C++ code from 2.95 to series 3.x");

			added_ending_ = "";
			return;

			/*
				  // ****
				  if (freq>=2) 
					{
					  added_ending_="";
					  return;
					}
      
				  if (freq==1)
					{
					  // check to see if filename provided exists as is
					  string cmd = fcheck_str_ + full_filename_;
					  if (system(cmd.chars())!=0) 
					{
					  added_ending_="";
					  return;
					}

					  (void) has_compressed_ending(full_filename_);
					  return;
					}

				  string bass_filename = full_filename_ + ".bass";
				  added_ending_ = ".bass";
				  string bass_cmd = fcheck_str_ + bass_filename;
				  if (system(bass_cmd.chars())!=0) 
					{
					  full_filename_ = bass_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = bass;
					  return;
					}
				  if (has_compressed_ending(bass_filename))
					{
					  full_filename_ = bass_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = bass;
					  return;
					}

				  string ass_filename = full_filename_ + ".ass";
				  added_ending_ = ".ass";
				  string ass_cmd = fcheck_str_ + ass_filename;
				  if (system(ass_cmd.chars())!=0) 
					{
					  full_filename_ = ass_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = ass;
					  return;
					}
				  if (has_compressed_ending(ass_filename))
					{
					  full_filename_ = ass_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = ass;
					  return;
					}

				  string prim_filename = full_filename_ + ".prim";
				  added_ending_ = ".prim";
				  string prim_cmd = fcheck_str_ + prim_filename;
				  if (system(prim_cmd.chars())!=0) 
					{
					  full_filename_ = prim_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = prim;
					  return;
					}
				  if (has_compressed_ending(prim_filename))
					{
					  full_filename_ = prim_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = prim;
					  return;
					}

				  string dsep_filename = full_filename_ + ".dsep";
				  added_ending_ = ".dsep";
				  string dsep_cmd = fcheck_str_ + dsep_filename;
				  if (system(dsep_cmd.chars())!=0) 
					{
					  full_filename_ = dsep_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = dsep;
					  return;
					}
				  if (has_compressed_ending(dsep_filename))
					{
					  full_filename_ = dsep_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = dsep;
					  return;
					}

				  string sep_filename = full_filename_ + ".sep";
				  added_ending_ = ".sep";
				  string sep_cmd = fcheck_str_ + sep_filename;
				  if (system(sep_cmd.chars())!=0) 
					{
					  full_filename_ = sep_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = sep;
					  return;
					}
				  if (has_compressed_ending(sep_filename))
					{
					  full_filename_ = sep_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = sep;
					  return;
					}

				  string pbm_filename = full_filename_ + ".pbm";
				  added_ending_ = ".pbm";
				  string pbm_cmd = fcheck_str_ + pbm_filename;
				  if (system(pbm_cmd.chars())!=0) 
					{
					  full_filename_ = pbm_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = pbm; // faster then going through 'convert'
					  return;
					}
				  if (has_compressed_ending(pbm_filename))
					{
					  full_filename_ = pbm_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = pbm;  // faster then going through 'convert'
					  return;
					}


				  string tiff_filename = full_filename_ + ".tiff";
				  added_ending_ = ".tiff";
				  string tiff_cmd = fcheck_str_ + tiff_filename;
				  if (system(tiff_cmd.chars())!=0) 
					{
					  full_filename_ = tiff_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = scanned_bitmap;
					  return;
					}
				  if (has_compressed_ending(tiff_filename))
					{
					  full_filename_ = tiff_filename;
					  filename_ = filename_ + added_ending_;
					  file_type_ = scanned_bitmap;
					  return;
					}

				  added_ending_ = "";
				  */
		}




		public static string full_pathname(string filename)
		{
			Console.Error.WriteLine("FilenameManager.full_pathname() not implemented.");
			Console.Error.WriteLine("  ... returning local filename unchanged!");

			return filename;

			/*
			string full_filename = new string("");

			if (filename != "")
			  {
				if ((!filename.contains(dir_sep,0) )
				&& (!filename.contains(drive_start,1)))
			  {
				char fname_buf[256];      
				string cwd = getcwd(fname_buf,256);
				if (cwd == null)
				  {
					cerr << "full_pathname: call to 'getcwd' failed" << endl;
					exit(1);
				  }
				strcat(fname_buf,dir_sep);
				full_filename =  cwd + filename;
			  }
				else full_filename = filename;
			  }

			return full_filename;
			*/
		}



		public static void canonical_filename(string filename)
		{
			op_no_ = 0;
			filename_ = filename;
			added_ending_ = "";

			if (filename!="")
			{
				if (!full_filename_prefix.IsMatch(filename))
				{
					string cwd = Directory.GetCurrentDirectory();
		  
					full_filename_ =  cwd + dir_sep + filename;
				}
				else 
				{
					full_filename_ = filename;
				}


				complete_filename();
      
				int len = full_filename_.Length;
				int skip_end;
				if (full_filename_.EndsWith(".gz")) skip_end = 4;
				else if (full_filename_.EndsWith(".z")) skip_end = 3;
				else if (full_filename_.EndsWith(".Z")) skip_end = 3;
				else skip_end = 1;
      
				int extension_break = full_filename_.LastIndexOf('.',len-skip_end);
      
				root_filename_ = full_filename_.Substring(0,extension_break);
				root_dir_ = root_filename_.Substring(root_filename_.LastIndexOf(dir_sep));
				separate_filename_  = root_filename_ + ".sep";
				primitive_filename_ = root_filename_ + ".prim";
				assembly_filename_  = root_filename_ + ".ass";
				bassembly_filename_ = root_filename_ + ".bass";
			}
			else
			{
				full_filename_ = "";
				file_type_ = FileTypeEnum.unknown;
				root_filename_ = "";
				root_dir_ = "";
				separate_filename_  = "";
				primitive_filename_ = "";
				assembly_filename_  = "";
				bassembly_filename_ = "";
			}
		}


		public static void override_separate_filename(string filename)
		{
			separate_filename_ = filename;
		}

	  
      


		public static bool filename_exists(string filename)
		{
			return input_file_access_.file_exists(filename);
		}

	  
		public static bool check_full_filename()
		{
			return filename_exists(full_filename_);
		}


		public static string root_filename()
		{
			return root_filename_;
		}


		public static string input_filename()
		{
			return full_filename_;
		}

		public static string separate_filename()
		{
			return separate_filename_;
		}

		public static string primitive_filename()
		{
			return primitive_filename_;
		}

		public static string assembly_filename()
		{
			return assembly_filename_;
		}


		public static string bassembly_filename()
		{
			return bassembly_filename_;
		}


		public static string generate_output_filename()
		{
			op_no_++;
			string output_filename = root_filename_ + "_" + op_no_.ToString() + ".pbm";

			return output_filename;
		}
      

		/*------------------*/

      

		public static BinaryReader open_fbitmap_input_pipe(string filename)
		{

            input_fbitmap_file_ = InputFileAccess.Create(file_access_);

			switch (file_type_)
			{
				case FileTypeEnum.scanned_bitmap:
				{
					// check ImageMagick's convert can be found
					/*
					int exit_status = system("fcheck_exec convert > /dev/null");
					if (exit_status!=0)
					  {
					cerr << "ImageMagick utility 'convert' is not installed"<< endl;
					exit(1);
					  }
					  */

					// convert  -type bilevel  -density 300 mapleLeaf.pdf mapleLeaf%d.pbm


					// oss_read << " | convert - pbm:- " << ends;
					Console.Error.WriteLine("Warning: open_fbitmap_input_pipe() only supports PBM format.");
					break;
				}
  
				case FileTypeEnum.bass:
				case FileTypeEnum.ass:
				case FileTypeEnum.prim:
				case FileTypeEnum.dsep:
				case FileTypeEnum.sep:
				case FileTypeEnum.pbm:  // faster then going through 'convert'
				default:
					// nothing needs doing! // ***** change switch statement to if?
					break;
			}

			return input_fbitmap_file_.open_file(filename);

	      
		}

		public static void close_fbitmap_input_pipe()
		{
			if (input_fbitmap_file_ != null)
			{
				input_fbitmap_file_.close_file();
			}
		}

      
		/*------------------*/

      
		BinaryReader open_input_pipe(string filename)
		{
			input_file_ = InputFileAccess.Create(file_access_);

			return input_file_.open_file(filename);
		}

		public static void close_input_pipe()
		{
			if (input_file_!=null)
				input_file_.close_file();
		}

      

		/*------------------*/

      
		public static BinaryReader open_prim_input_pipe(string filename)
		{
            input_prim_file_ = InputFileAccess.Create(file_access_);

			return input_prim_file_.open_file(filename);

		}

		public static void close_prim_input_pipe()
		{
			if (input_prim_file_!=null)
				input_prim_file_.close_file();
		}
      

		/*------------------*/

      
		public static BinaryReader open_ass_input_pipe(string filename)
		{
            input_ass_file_ = InputFileAccess.Create(file_access_);

			return input_ass_file_.open_file(filename);
		}

		public static void close_ass_input_pipe()
		{
			if (input_ass_file_!=null)
				input_ass_file_.close_file();
		}
      

		/*------------------*/

      
		public static BinaryReader open_bass_input_pipe(string filename)
		{
            input_bass_file_ = InputFileAccess.Create(file_access_);
		
			return input_bass_file_.open_file(filename);

	 
		}

		public static void close_bass_input_pipe()
		{
			if (input_bass_file_!=null)
				input_bass_file_.close_file();
		}
      

		/*------------------*/

      

		public static BinaryWriter open_output_pipe(string filename)
		{
		  
			Console.Error.WriteLine("Warning: open_output_pipe() using simplified, non-piped version!");
			Console.Error.WriteLine("         Should eventually support compression using pipe, but no longer takes pre-command");
			Console.Error.WriteLine("         No longer makes sense to write out data on remote site");

			output_file_ = new FileStream(filename,FileMode.Create);
			output_bw_ = new BinaryWriter(output_file_);
	  
			return output_bw_;

		}
	


		public static void close_output_pipe()
		{
			Console.Error.WriteLine("Warning: close_output_pipe() using simplified, none piped version!");

			output_bw_.Close();
			output_file_.Close();

		}


		public static BinaryReader open_glob_input_pipe(string filename)
		{
			string full_filename;
			if (!full_filename_prefix.IsMatch(filename))
			{ 
				string cwd = Directory.GetCurrentDirectory();
				full_filename =  cwd + dir_sep + filename;
			}
			else 
			{
				full_filename = filename;
			}

			input_glob_file_ = InputFileAccess.Create(file_access_);

			return input_glob_file_.open_file(full_filename);

	  
		}

		public static void close_glob_input_pipe()
		{
			if (input_glob_file_!=null)
			{
				input_glob_file_.close_file();
			}
		}



		public static bool check_empirical_filename(string filename)
		{
			string full_filename;
			if (!full_filename_prefix.IsMatch(filename))
			{
				full_filename =  root_dir_ + dir_sep + filename;
			}
			else full_filename = filename;

			return input_file_access_.file_exists(full_filename);

	 
		}



		public static BinaryReader open_empirical_input_pipe(string filename)
		{
			string full_filename;
			if (!full_filename_prefix.IsMatch(filename))
			{
				full_filename =  root_dir_ + dir_sep + filename;
			}
			else full_filename = filename;

            input_empirical_file_ = InputFileAccess.Create(file_access_);
	
			return input_empirical_file_.open_file(full_filename);

		}

		public static void close_empirical_input_pipe()
		{
			if (input_empirical_file_!=null)
			{
				input_empirical_file_.close_file();
			}
		}

		public static bool is_remote()
		{
			return file_access_ == InputFileAccess.Type.remote;
		}

		public static bool is_local()
		{
			return file_access_ == InputFileAccess.Type.local;
		}

		public static string added_ending() { return added_ending_; }

		public static bool fbitmap_input_is_open() {return input_fbitmap_file_!=null;}
		public static bool input_is_open() {return input_file_!=null;}
		public static bool prim_input_is_open() {return input_prim_file_!=null;}
		public static bool ass_input_is_open() {return input_ass_file_!=null;}
		public static bool bass_input_is_open() {return input_bass_file_!=null;}
		public static bool glob_input_is_open() {return input_glob_file_!=null;}
		public static bool empirical_input_is_open() {return input_empirical_file_!=null;}
	
		public static bool output_is_open() { return output_file_!=null; }
		public static FileTypeEnum file_type() { return file_type_; }

		public static FileStream output_file() {return output_file_;}

	}
}
      
