
using System;
using System.IO;

namespace Espresso.OMR
{
	class GChar : SubBitmap
	{
		public enum SelectionEnum { gc_bitmap, gc_library_graymap };

		protected string name_;
		protected double certainty_;
		protected byte[,] library_graymap_;
		protected int lg_x_dim_;
		protected int lg_y_dim_;

		public GChar()
		{}

		public GChar(string name, Bitmap src_bitmap, 
			int sx_org, int sy_org,
			int dx_dim, int dy_dim,
			double certainty,
			byte[,] library_graymap, 
			int lg_x_dim, int lg_y_dim,
			SubBitmap.Operation op)
		{}

		
		public GChar(string name, Bitmap src_bitmap, 
			int sx_org, int sy_org,
			int dx_dim, int dy_dim,
			double certainty,
			byte[,] library_graymap, 
			int lg_x_dim, int lg_y_dim)
			: this(name,src_bitmap,sx_org,sy_org,dx_dim,dy_dim,certainty,
		library_graymap,lg_x_dim,lg_y_dim,SubBitmap.Operation.nop)
		{}



		public GChar(string name, SubBitmap src_bitmap, 
			int sx_org, int sy_org,
			int dx_dim, int dy_dim,
			double certainty,
			byte[,] library_graymap, 
			int lg_x_dim, int lg_y_dim,
			SubBitmap.Operation op)
		{}

		public GChar(string name, SubBitmap src_bitmap, 
			int sx_org, int sy_org,
			int dx_dim, int dy_dim,
			double certainty,
			byte[,] library_graymap, 
			int lg_x_dim, int lg_y_dim)
			: this(name,src_bitmap,sx_org,sy_org,dx_dim,dy_dim,certainty,
			library_graymap,lg_x_dim,lg_y_dim,SubBitmap.Operation.nop)
		{}

		public GChar(string name, int sx_org, int sy_org,
			int dx_dim, int dy_dim,
			double certainty,
			byte[,] library_graymap, 
			int lg_x_dim, int lg_y_dim)
		{}
  
		public String name() { return name_; }
  
		public double certainty() { return certainty_; }
  
		public byte[,] library_graymap() 
		{ return library_graymap_; }
  
		public int lg_x_dim() { return lg_x_dim_; }
		public int lg_y_dim() { return lg_y_dim_; }



		/*---------------------------------------------------------------------------*/

		public bool read_gprim_as_gchar(BinaryReader brin, bool is_hetrogeneous)
		  {
		    //  if (is_hetrogeneous)  // ***** // (sort of redundent)

		    if (!IOSupport.read_token(brin,"prim_name")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_string(brin,out name_)) return false;

		    base.read(brin);

		    if (!IOSupport.read_token(brin,"certainty")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_double(brin,out certainty_)) return false;

		    return true;
		  }

		/*---------------------------------------------------------------------------*/


		public override void write(BinaryWriter brout)
		  {
		    base.write(brout);
  
		    IOSupport.write_line(brout,"name = " + name_);
		    IOSupport.write_line(brout,"certainty = " + certainty_.ToString());

		    int lg_x_dim = lg_x_dim_;
		    int lg_y_dim = lg_y_dim_;

		    // library_graymap is fairly redundent
		    lg_x_dim = 0;
		    lg_y_dim = 0;

		    IOSupport.write_line(brout,"lg_x_dim = " + lg_x_dim.ToString() + " , lg_y_dim = " + lg_y_dim.ToString());

		    for (int y=0; y<lg_y_dim; y++)
		      for (int x=0; x<lg_x_dim; x++)
				brout.Write(library_graymap_[y,x]);
		    IOSupport.write_line(brout);
		  }


		public override bool read(BinaryReader brin)
		  {
		    base.read(brin);

		    if (!IOSupport.read_token(brin,"name")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_string(brin,out name_)) return false;
  
		    if (!IOSupport.read_token(brin,"certainty")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_double(brin,out certainty_)) return false;

		    if (!IOSupport.read_token(brin,"lg_x_dim")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out lg_x_dim_)) return false;

		    if (!IOSupport.read_token(brin,",")) return false;

		    if (!IOSupport.read_token(brin,"lg_y_dim")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out lg_y_dim_)) return false;

		    // consume carriage return
		    byte dummy = brin.ReadByte();

		    // library_graymap is fairly redundent // ***

		    if ((lg_x_dim_>0) && (lg_y_dim_>0))
		      {
			library_graymap_ = null;
      
			for (int y=0; y<lg_y_dim_; y++)
			  for (int x=0; x<lg_x_dim_; x++)
			    dummy = brin.ReadByte();
		      }
		    else library_graymap_ = null;

		    return true;
		  }


		/*---------------------------------------------------------------------------*/

		// input and output operations for gcharacter conforming to gprim requirements

		public void output_gchar_as_gprim(BinaryWriter brout, bool is_hetrogeneous)
		  {
		    //  if (is_hetrogeneous)  // ***** // (sort of redundent)

		    IOSupport.write_line(brout,"prim_name = " + name_);

		    base.write(brout);

		    IOSupport.write_line(brout,"certainty = " + certainty_.ToString());
		  }


  
	}
}




  


  
