
using System;

namespace Espresso.OMR
{
	class GObject : SubBitmap
	{
		protected Projection x_project_;
		protected Projection y_project_;
		protected int area_;

		public GObject() 
			: base()
		{
			x_project_ = new Projection(0);
			y_project_ = new Projection(0);
			area_ = 0;	
		}

		public GObject(GObject gobject)
			: base(gobject)
		{
			x_project_ = gobject.x_project_.DeepCopy();
			y_project_ = gobject.y_project_.DeepCopy();
			area_ = gobject.area_;
		}

		public GObject(Bitmap src_bitmap, int sx_org, int sy_org,
			int dx_dim, int dy_dim)
			: base(src_bitmap,sx_org,sy_org,dx_dim,dy_dim)

		{
			x_project_ = new Projection(dx_dim);
			y_project_ = new Projection(dy_dim);
			area_ = 0;
		}

		public GObject(Bitmap src_bitmap, int sx_org, int sy_org,
			int dx_dim, int dy_dim, Operation op)
			: base(src_bitmap,sx_org,sy_org,dx_dim,dy_dim,op)

		{
			x_project_ = new Projection(dx_dim);
			y_project_ = new Projection(dy_dim); 
			area_ = 0;
		}

		public GObject(SubBitmap src_bitmap, int sx_org, int sy_org,
			int dx_dim, int dy_dim)
			: base(src_bitmap,sx_org,sy_org,dx_dim,dy_dim)

		{
			x_project_ = new Projection(dx_dim);
			y_project_ = new Projection(dy_dim);
			area_ = 0;
		}

		public GObject(SubBitmap src_bitmap, int sx_org, int sy_org,
			int dx_dim, int dy_dim, Operation op)
			: base(src_bitmap,sx_org,sy_org,dx_dim,dy_dim,op)
		{
			x_project_ = new Projection(dx_dim);
			y_project_ = new Projection(dy_dim);
			area_ = 0;
		}

		public GObject(GObject src_gobject, Operation op)
			: base(src_gobject,0,0,src_gobject.x_dim(),src_gobject.y_dim(),op)
		{
			x_project_ = new Projection(src_gobject.x_dim());
			y_project_ = new Projection(src_gobject.y_dim());
			area_ = 0;

			x_org_ = src_gobject.x_org();
			y_org_ = src_gobject.y_org();
		}

		public GObject(int sx_org, int sy_org,int dx_dim, int dy_dim)
			: base(sx_org,sy_org,dx_dim,dy_dim)

		{
			x_project_ = new Projection(dx_dim);
			y_project_ = new Projection(dy_dim);
			area_ = 0;
		}

		public new void resize(int x_dim, int y_dim)
		{
			base.resize(x_dim,y_dim); // falls straight throught to Bitmap::
			x_project_.resize(x_dim);
			y_project_.resize(y_dim);
		}

		public void resize_projects(int x_dim, int y_dim)
		{
			x_project_.resize(x_dim);
			y_project_.resize(y_dim);
		}

		public void generate_x_project()
		{
			Bitmap this_bitmap = this;

			x_project_.generate_projection(this_bitmap,Projection.AxisEnum.x_axis);
		}

		public void generate_y_project()
		{
			Bitmap this_bitmap = this;

			y_project_.generate_projection(this_bitmap,Projection.AxisEnum.y_axis);
		}

		public void generate_area()
		{
			Bitmap this_bitmap = this;
	
			for (int y=0; y<y_dim_; y++)
			{
				for (int x=0; x<x_dim_; x++)
				{
					if (this_bitmap[y,x]==PixelEnum.set) 
					{
						area_++;
					}
				}
			}
		}


		public Projection x_project() { return x_project_; }
		public Projection y_project() { return y_project_; }
    
		public int area() { return area_; }

	}
}

