
using System;

namespace Espresso.OMR
{


    /*------------------*/


    class BitmapNode
    {
        public object val_;
        public BitmapNode next_;
        public BitmapNode prev_;

        public BitmapNode(Bitmap val)
        {
            val_ = val; // boxed if val-bitmap
        }

        public BitmapNode next() { return next_; }
        public BitmapNode prev() { return prev_; }

        public Bitmap val() { return (Bitmap)val_; } // unboxed if val bitmap

    }


    class BitmapList
    {

        private BitmapNode head_;
        private BitmapNode current_;
        private BitmapNode tail_;

        private int length_;


        public BitmapList()
        {
            head_ = null;
            current_ = null;
            tail_ = null;
            length_ = 0;
        }

        public void reset()
        {
            head_ = tail_ = current_ = null;
            length_ = 0;
        }

        public void append(Bitmap val)
        {
            BitmapNode new_node = new BitmapNode(val);

            if (tail_ != null)
            {
                tail_.next_ = new_node;
                new_node.prev_ = tail_;
            }
            else
            {
                tail_ = head_ = current_ = new_node;
                new_node.prev_ = null;
            }

            new_node.next_ = null; // should already be null from contructor
            tail_ = new_node;

            length_++;
        }

        public void prepend(Bitmap val)
        {
            BitmapNode new_node = new BitmapNode(val);

            if (head_ != null)
            {
                head_.prev_ = new_node;
                new_node.next_ = head_;
            }
            else
            {
                head_ = tail_ = current_ = new_node;
                new_node.next_ = null;
            }

            new_node.prev_ = null; // should already be null from contructor
            head_ = new_node;

            length_++;
        }

        public void insert_before(Bitmap val)
        {
            BitmapNode new_node = new BitmapNode(val);

            new_node.prev_ = current_.prev_;
            new_node.next_ = current_;

            current_.prev_ = new_node;

            if (current_ == head_) head_ = new_node;
            else new_node.prev_.next_ = new_node;

            length_++;
        }


        public void insert_after(Bitmap val)
        {
            BitmapNode new_node = new BitmapNode(val);

            new_node.next_ = current_.next_;
            new_node.prev_ = current_;

            current_.next_ = new_node;

            if (current_ == tail_) tail_ = new_node;
            else new_node.next_.prev_ = new_node;

            length_++;
        }

        public BitmapNode delete_forward()
        {
            if (current_ == null) return current_; // this should never happen

            if (current_ == head_) head_ = current_.next_;
            else current_.prev_.next_ = current_.next_;

            if (current_ == tail_) tail_ = current_.prev_;
            else current_.next_.prev_ = current_.prev_;

            BitmapNode new_current = current_.next_;
            current_.next_ = current_.prev_ = null;
            // nothing now links to current_
            current_ = new_current;

            length_--;

            return current_;
        }

        public BitmapNode delete_backward()
        {
            if (current_ == null) return current_; // this should never happen

            if (current_ == head_) head_ = current_.next_;
            else current_.prev_.next_ = current_.next_;

            if (current_ == tail_) tail_ = current_.prev_;
            else current_.next_.prev_ = current_.prev_;

            BitmapNode new_current = current_.prev_;
            current_.next_ = current_.prev_ = null;
            // nothing now links to current_
            current_ = new_current;

            length_--;

            return current_;
        }


        public BitmapNode detach_forward()
        {
            if (current_ == null) return current_; // this should never happen

            if (current_ == head_) head_ = current_.next_;
            else current_.prev_.next_ = current_.next_;

            if (current_ == tail_) tail_ = current_.prev_;
            else current_.next_.prev_ = current_.prev_;

            BitmapNode new_current = current_.next_;
            current_.next_ = current_.prev_ = null;

            current_.val_ = null; // assume data already known to outside object
            current_ = new_current;

            length_--;

            return current_;
        }

        public BitmapNode detach_backward()
        {
            if (current_ == null) return current_; // this should never happen

            if (current_ == head_) head_ = current_.next_;
            else current_.prev_.next_ = current_.next_;

            if (current_ == tail_) tail_ = current_.prev_;
            else current_.next_.prev_ = current_.prev_;

            BitmapNode new_current = current_.prev_;
            current_.next_ = current_.prev_ = null;

            current_.val_ = null; // assume data already known to outside object
            current_ = new_current;

            length_--;

            return current_;
        }


        public BitmapNode head()
        {
            current_ = head_;
            return head_;
        }

        public BitmapNode tail()
        {
            current_ = tail_;
            return tail_;
        }

        public BitmapNode next()
        {
            if (current_ == null) return null;
            current_ = current_.next();
            return current_;
        }

        public BitmapNode prev()
        {
            if (current_ == null) return null;
            current_ = current_.prev();
            return current_;
        }

        public BitmapNode curr()
        {
            return current_;
        }

        public void set_position(BitmapNode pos)
        {
            current_ = pos;
        }

        public int length() { return length_; }


    }
}
    
