
using System;

namespace Espresso.OMR
{
  class ByteArray2d
    {
      protected byte[,] array_;
      protected int x_dim_;
      protected int y_dim_;

      private int print_width_;

      public ByteArray2d()
	{ 
	  array_ = null; 
	  x_dim_ = y_dim_ = 0; 
	  print_width_ = 0;
	}

      public ByteArray2d(int x_dim, int y_dim)
	{ 
	  x_dim_ = x_dim;
	  y_dim_ = y_dim;
	  print_width_ = 0;

	  if ((x_dim>0) && (y_dim>0))
	    {
	      array_ = new byte[y_dim,x_dim];
	    }
	  else 
	    {
	      array_ = null;
	    }
	}


      public void resize(int x_dim, int y_dim)
	{ 
	  x_dim_ = x_dim;
	  y_dim_ = y_dim;

	  if ((x_dim>0) && (y_dim>0))
	    {
	      array_ = new byte[y_dim,x_dim];
      	    }
	  else 
	    {
	      array_=null;
	    }
	}


      public void extend(int x_dim, int y_dim)
	{ 
  
	  byte[,] new_array = new byte[y_dim,x_dim];
  
	  int min_x_dim = Math.Min(x_dim,x_dim_);
	  int min_y_dim = Math.Min(y_dim,y_dim_);

	  for (int y=0; y<min_y_dim; y++)
	    for (int x=0; x<min_x_dim; x++)
	      {
	        new_array[y,x]=array_[y,x];
	      }

	  array_ = new_array;
	  x_dim_ = x_dim;
	  y_dim_ = y_dim;
	}

      /*

      ostream& operator << (ostream& os, ByteArray2d& array)
	{ 
	  int i;
	
	  os << " ";
	  for (i=0; i<array.x_dim_*array.print_width_; i++)
	    cout << "-";
	  os << endl;
  
	  for (int y=0; y<array.y_dim_; y++)
	    {
	      os << "|";
	      for (int x=0; x<array.x_dim_; x++)
		{
		  os.width(array.print_width_);
		  os << array.array_[y][x];
		}
	      os << "|" << endl;
	    }
  
	  os << " ";
	  for (i=0; i<array.x_dim_*array.print_width_; i++)
	    cout << "-";
	  os << endl;
  
	  return os;
	}
      */

 
      public int x_dim() { return x_dim_; }
      public int y_dim() { return y_dim_; }
 
      public void print_width(int w) { print_width_=w; }
      public int print_width() { return print_width_; }

	/*
      public byte* cdata() { return array_[0]; }
      public byte** pdata() { return array_; }
	*/

      public void replace(ByteArray2d array)
	{
	  array_ = array.array_;
	  x_dim_ = array.x_dim_;
	  y_dim_ = array.y_dim_;
	}

      public void exchange(ByteArray2d array)
	{
	  int temp_x_dim = x_dim_;
	  x_dim_ = array.x_dim_;
	  array.x_dim_ = temp_x_dim;

	  int temp_y_dim = y_dim_;
	  y_dim_ = array.y_dim_;
	  array.y_dim_ = temp_y_dim;

	  byte[,] temp_array = array_;
	  array_ = array.array_;
	  array.array_ = temp_array;
	}

      public ByteArray2d ShallowCopy() 
      { 

	  ByteArray2d array_copy = new ByteArray2d(x_dim_,y_dim_);

	  for (int y=0; y<y_dim_; y++)
	    {
	      for (int x=0; x<x_dim_; x++)
		{
		  array_copy[y,x] = array_[y,x];
		}
	    }

	  return array_copy;
      }


      public byte this [int y, int x]
 	{
  	  get
  	  {
	    return array_[y,x];
  	  }
  	  set
  	  {
   	    array_[y,x] = value;
  	  }
 	}

      public byte[,] data() { return array_; }

    }
}
