
using System;

namespace Espresso.OMR
{


                      /*------------------*/


  class GObjectNode
    {
      public object val_;
      public GObjectNode next_;
      public GObjectNode prev_;
      
      public GObjectNode(GObject val)
	
	{
	  val_ = val; // boxed if val-gobject
	}
      
      public GObjectNode next() { return next_; }
      public GObjectNode prev() { return prev_; }
	
      public GObject val() { return (GObject)val_; } // unboxed if val gobject
	
    }
  

  class GObjectList
  {

    private GObjectNode head_;
    private GObjectNode current_;
    private GObjectNode tail_;

    private int length_;


    public GObjectList()
      {
	head_ = null;
	current_ = null;
	tail_ = null;
	length_ = 0;
      }

    public void reset()
      {
	head_ = tail_ = current_ = null;
	length_=0;
      }

    public void append(GObject val)
      {
	GObjectNode new_node = new GObjectNode(val);
	
	if (tail_!=null) 
	  {
	    tail_.next_ = new_node;
	    new_node.prev_ = tail_;
	  }
	else 
	  {
	    tail_ = head_ = current_ = new_node;
	    new_node.prev_ = null;
	  }
	
	new_node.next_ = null; // should already be null from contructor
	tail_ = new_node;
	
	length_++;
      }

    public void prepend(GObject val)
      {
	GObjectNode new_node = new GObjectNode(val);

	if (head_!=null) 
	  {
	    head_.prev_ = new_node;
	    new_node.next_ = head_;
	  }
	else 
	  {
	    head_ = tail_ = current_ = new_node;
	    new_node.next_ = null;
	  }

	new_node.prev_ = null; // should already be null from contructor
	head_ = new_node;

	length_++;
      }

    public void insert_before(GObject val)
      {
	GObjectNode new_node = new GObjectNode(val);

	new_node.prev_ = current_.prev_;
	new_node.next_ = current_;

	current_.prev_ = new_node;

	if (current_==head_) head_ = new_node;
	else new_node.prev_.next_ = new_node;

	length_++;
      }


    public void insert_after(GObject val)
      {
	GObjectNode new_node = new GObjectNode(val);

	new_node.next_ = current_.next_;
	new_node.prev_ = current_;

	current_.next_ = new_node;

	if (current_==tail_) tail_ = new_node;
	else new_node.next_.prev_=new_node;

	length_++;
      }

    public GObjectNode delete_forward()
      {
	if (current_==null) return current_; // this should never happen

	if (current_==head_) head_ = current_.next_;
	else current_.prev_.next_ = current_.next_;

	if (current_==tail_) tail_ = current_.prev_;
	else current_.next_.prev_ = current_.prev_;

	GObjectNode new_current = current_.next_;
	current_.next_ = current_.prev_=null;
	// nothing now links to current_
	current_ = new_current;

	length_--;

	return current_;
      }

    public GObjectNode delete_backward()
      {
	if (current_==null) return current_; // this should never happen

	if (current_==head_) head_ = current_.next_;
	else current_.prev_.next_ = current_.next_;

	if (current_==tail_) tail_ = current_.prev_;
	else current_.next_.prev_ = current_.prev_;

	GObjectNode new_current = current_.prev_;
	current_.next_ = current_.prev_ = null;
	// nothing now links to current_
	current_ = new_current;

	length_--;

	return current_;
      }


    public GObjectNode detach_forward()
      {
	if (current_==null) return current_; // this should never happen

	if (current_==head_) head_ = current_.next_;
	else current_.prev_.next_ = current_.next_;

	if (current_==tail_) tail_ = current_.prev_;
	else current_.next_.prev_ = current_.prev_;

	GObjectNode new_current = current_.next_;
	current_.next_ = current_.prev_=null;

	current_.val_ = null; // assume data already known to outside object
	current_ = new_current;

	length_--;

	return current_;
      }

    public GObjectNode detach_backward()
      {
	if (current_==null) return current_; // this should never happen

	if (current_==head_) head_ = current_.next_;
	else current_.prev_.next_ = current_.next_;

	if (current_==tail_) tail_ = current_.prev_;
	else current_.next_.prev_ = current_.prev_;

	GObjectNode new_current = current_.prev_;
	current_.next_ = current_.prev_ = null;

	current_.val_ = null; // assume data already known to outside object
	current_ = new_current;

	length_--;

	return current_;
      }


    public GObjectNode head() 
      { 
	current_ = head_;
	return head_;
      }
    
    public GObjectNode tail() 
      { 
	current_ = tail_;
	return tail_;
      }

    public GObjectNode next() 
      { 
	if (current_==null) return null;
	current_=current_.next();
	return current_;
      }

    public GObjectNode prev()
      { 
	if (current_==null) return null;
	current_=current_.prev();
	return current_;
      }

    public GObjectNode curr()
      {
	return current_;
      }

    public void set_position(GObjectNode pos)
      {
	current_=pos;
      }

    public int length() { return length_; }


  }
}
    
