
using System;

namespace Espresso.OMR
{


                      /*------------------*/


  class StringNode
    {
      public object val_;
      public StringNode next_;
      public StringNode prev_;
      
      public StringNode(String val)
	
	{
	  val_ = val; // boxed if val-string
	}
      
      public StringNode next() { return next_; }
      public StringNode prev() { return prev_; }
	
      public String val() { return (String)val_; } // unboxed if val string
	
    }
  

  class StringList
  {

    private StringNode head_;
    private StringNode current_;
    private StringNode tail_;

    private int length_;


    public StringList()
      {
	head_ = null;
	current_ = null;
	tail_ = null;
	length_ = 0;
      }

    public void reset()
      {
	head_ = tail_ = current_ = null;
	length_=0;
      }

    public void append(String val)
      {
	StringNode new_node = new StringNode(val);
	
	if (tail_!=null) 
	  {
	    tail_.next_ = new_node;
	    new_node.prev_ = tail_;
	  }
	else 
	  {
	    tail_ = head_ = current_ = new_node;
	    new_node.prev_ = null;
	  }
	
	new_node.next_ = null; // should already be null from contructor
	tail_ = new_node;
	
	length_++;
      }

    public void prepend(String val)
      {
	StringNode new_node = new StringNode(val);

	if (head_!=null) 
	  {
	    head_.prev_ = new_node;
	    new_node.next_ = head_;
	  }
	else 
	  {
	    head_ = tail_ = current_ = new_node;
	    new_node.next_ = null;
	  }

	new_node.prev_ = null; // should already be null from contructor
	head_ = new_node;

	length_++;
      }

    public void insert_before(String val)
      {
	StringNode new_node = new StringNode(val);

	new_node.prev_ = current_.prev_;
	new_node.next_ = current_;

	current_.prev_ = new_node;

	if (current_==head_) head_ = new_node;
	else new_node.prev_.next_ = new_node;

	length_++;
      }


    public void insert_after(String val)
      {
	StringNode new_node = new StringNode(val);

	new_node.next_ = current_.next_;
	new_node.prev_ = current_;

	current_.next_ = new_node;

	if (current_==tail_) tail_ = new_node;
	else new_node.next_.prev_=new_node;

	length_++;
      }

    public StringNode delete_forward()
      {
	if (current_==null) return current_; // this should never happen

	if (current_==head_) head_ = current_.next_;
	else current_.prev_.next_ = current_.next_;

	if (current_==tail_) tail_ = current_.prev_;
	else current_.next_.prev_ = current_.prev_;

	StringNode new_current = current_.next_;
	current_.next_ = current_.prev_=null;
	// nothing now links to current_
	current_ = new_current;

	length_--;

	return current_;
      }

    public StringNode delete_backward()
      {
	if (current_==null) return current_; // this should never happen

	if (current_==head_) head_ = current_.next_;
	else current_.prev_.next_ = current_.next_;

	if (current_==tail_) tail_ = current_.prev_;
	else current_.next_.prev_ = current_.prev_;

	StringNode new_current = current_.prev_;
	current_.next_ = current_.prev_ = null;
	// nothing now links to current_
	current_ = new_current;

	length_--;

	return current_;
      }


    public StringNode detach_forward()
      {
	if (current_==null) return current_; // this should never happen

	if (current_==head_) head_ = current_.next_;
	else current_.prev_.next_ = current_.next_;

	if (current_==tail_) tail_ = current_.prev_;
	else current_.next_.prev_ = current_.prev_;

	StringNode new_current = current_.next_;
	current_.next_ = current_.prev_=null;

	current_.val_ = null; // assume data already known to outside object
	current_ = new_current;

	length_--;

	return current_;
      }

    public StringNode detach_backward()
      {
	if (current_==null) return current_; // this should never happen

	if (current_==head_) head_ = current_.next_;
	else current_.prev_.next_ = current_.next_;

	if (current_==tail_) tail_ = current_.prev_;
	else current_.next_.prev_ = current_.prev_;

	StringNode new_current = current_.prev_;
	current_.next_ = current_.prev_ = null;

	current_.val_ = null; // assume data already known to outside object
	current_ = new_current;

	length_--;

	return current_;
      }


    public StringNode head() 
      { 
	current_ = head_;
	return head_;
      }
    
    public StringNode tail() 
      { 
	current_ = tail_;
	return tail_;
      }

    public StringNode next() 
      { 
	if (current_==null) return null;
	current_=current_.next();
	return current_;
      }

    public StringNode prev()
      { 
	if (current_==null) return null;
	current_=current_.prev();
	return current_;
      }

    public StringNode curr()
      {
	return current_;
      }

    public void set_position(StringNode pos)
      {
	current_=pos;
      }

    public int length() { return length_; }


  }
}
    
