
using System;
using System.IO;
using System.Text;

namespace Espresso.OMR
{
	class IOSupport
	{


		public static void swap_short(ref ushort s)
		{
			ushort copy = s;
			s = (ushort)((copy & 0x00ff)<<8);
			s |= (ushort)((copy & 0xff00)>>8);
		}

		public static void swap_integer(ref uint s)
		{
			uint copy = s;
			s =  (copy & 0x000000ff)<<24;
			s |= (copy & 0x0000ff00)<<8;
			s |= (copy & 0x00ff0000)>>8;
			s |= (copy & 0xff000000)>>24;
		}


		public static void error(String got, String wanted)
		{
			Console.Error.WriteLine("Error reading in:  Got " + got + ", wanted " + wanted + ".");
		}

		public static void error(int value, String report)
		{
			Console.Error.WriteLine("Error reading in number:  Got " + value.ToString() + ", " + report);
		}

		public static void error(String report)
		{
			Console.Error.WriteLine("Error readin in: " + report);
		}


		public static bool write_bytes(BinaryWriter brout, byte[] byte_buffer, int num_bytes)
		{
			try 
			{
				brout.Write(byte_buffer,0,num_bytes);
				return true;
			}
			catch (IOException)
			{
				Console.Error.WriteLine("Error: unsuccessful attempt to write "+ num_bytes.ToString() + " bytes.");
				return false;
			}
		}


		public static bool write(BinaryWriter bwout, String word)
		{
			try
			{
				bwout.Write(word.ToCharArray());
				return true;
			}
			catch (IOException)
			{
				error("error occured trying to write out word '" + word + "'");
				return false;
			}	
		}

		public static bool write_line(BinaryWriter bwout, String line)
		{
			try
			{
				bwout.Write(line.ToCharArray());
				bwout.Write('\n');
				return true;
			}
			catch (IOException)
			{
				error("error occured trying to write out line '" + line + "'");
				return false;
			}	
		}

		public static bool write_line(BinaryWriter bwout)
		{
			return write_line(bwout,"");
		}

		public static bool read_bytes(BinaryReader brin, byte[] byte_buffer, int num_bytes)
		{
			int num_bytes_read = brin.Read(byte_buffer,0,num_bytes);
			if (num_bytes_read != num_bytes)
			{	
				Console.Error.WriteLine("Error: unsuccessful attempt to read "+ num_bytes.ToString() + " bytes.");
				return false;
			}
	 
			return true;
		}

		public static bool read_bytes(BinaryReader brin, out short bytes)
		{
			bytes = brin.ReadInt16();
			return true;
		}

		public static bool read_bytes(BinaryReader brin, out int bytes)
		{
			bytes = brin.ReadInt32();
			return true;
		}

		protected static bool read_word(BinaryReader brin, out String word)
		{
			// skip over any white space at start
			while (true)
			{

				try 
				{
					char check_char = (char)brin.PeekChar();

					if ((check_char!=' ') && (check_char!='\n') && (check_char!='\t')) 
					{
						break;
					}

					brin.ReadChar(); // consume white space character

				}
				catch (IOException)
				{
					error("error occured scanning for start of word");
					word = "";
					return false;
				}
			}

			// now read in the word until next white space encountered
			StringBuilder word_buffer = new StringBuilder();

			while (true)
			{

				try 
				{
					char check_char = (char)brin.PeekChar();

					if ((check_char==' ') || (check_char=='\n') || (check_char=='\t')) 
					{
						break;
					}

					char read_char = brin.ReadChar();

					word_buffer.Append(read_char);

				}
				catch (IOException)
				{
					error("error occured before end of word reached");
					word = "";
					return false;
				}
			}

			word = word_buffer.ToString();
			return true;
		}

		public static bool read_token(BinaryReader brin, string given_token)
		{
			string next_token;

			if (read_word(brin,out next_token))
			{
				if (next_token!=given_token)
				{
					error(next_token,given_token);
					return false;
				}
				return true;
			}

			error("did not get a string.");
			return false;
		}




		public static bool read_string(BinaryReader brin, out string word)
		{
			string attempt_word;
			if (read_word(brin,out attempt_word))
			{
				word = attempt_word;
				return true;
			}

			error("did not get a string.");
			word = "";
			return false;
		}

		public static bool read_line(BinaryReader brin, out String word)
		{
			StringBuilder word_buffer = new StringBuilder();
    
			while (true)
			{

				try 
				{
					char check_char = brin.ReadChar();

					if (check_char=='\n') 
					{
						break;
					}

					word_buffer.Append(check_char);

				}
				catch (IOException)
				{
					error("error occured before end of word reached");
					word = "";
					return false;
				}
			}

			word = word_buffer.ToString();
			return true;
		}


		public static bool read_null_terminated_string(BinaryReader brin, out String word)
		{
    
			StringBuilder word_buffer = new StringBuilder();

			while (true)
			{
				try
				{
					char check_char = brin.ReadChar();

					word_buffer.Append(check_char);
					if (check_char=='\0') break;
				}
				catch (IOException)
				{
	      
					error("error occured before end of null terminated string");
					word = "";
					return false;
				}
				
			}

			word = word_buffer.ToString();
			return true;
		}

		public static bool read_char(BinaryReader brin, out char c)
		{
			try
			{
				c = brin.ReadChar();
				return true;
			}
			catch (IOException)
			{
				error("did not get a char.");
				c = '\0';
				return false;
			}
		}


		public static bool read_cardinal(BinaryReader brin, out int val)
		{
			string attempt_val_str;
	 
			if (read_word(brin,out attempt_val_str))
			{
				int attempt_val = Int32.Parse(attempt_val_str);

				if (attempt_val<0)
				{
					error(attempt_val,"wanted a positive number.");
					val = 0;
					return false;
				}
				val = attempt_val;
				return true;
			}

			error("did not get a number.");
			val = 0;
			return false;
		}


		public static bool read_double(BinaryReader brin, out double val)
		{

			string attempt_val_str;
			if (read_word(brin,out attempt_val_str))
			{      
				val = Double.Parse(attempt_val_str);
				return true;
			}


			error("did not get a double.");
			val = 0.0;
			return false;
		}

	}
}


      
