
using System;

namespace Espresso.OMR
{

	class Projection : IntArray1d
	{
		public enum AxisEnum { x_axis, y_axis };
		public enum DirectionEnum { forwards, backwards };
    
		protected int max_value_;
		protected int average_;
    
		public Projection(int len)
			: base(len)
		{
			max_value_ = 0;
			average_ = 0;
		}

		public void generate_projection(Bitmap bitmap, AxisEnum axis)
		{
			IntArray1d array = (IntArray1d)(this);
	
			switch (axis)
			{
				case AxisEnum.x_axis:
					for (int x=0; x<bitmap.x_dim(); x++)
					{
						array[x]=0;
						for (int y=0; y<bitmap.y_dim(); y++)
							if (bitmap[y,x]==Pixel.PixelEnum.set) array[x]++;
		
						if (array[x]>max_value_) max_value_=array[x];
						average_ += array[x];
					}
					if (bitmap.x_dim()>0)
					{
						average_ /= bitmap.x_dim();
					}
					break;
      
				case AxisEnum.y_axis:
					for (int y=0; y<bitmap.y_dim(); y++)
					{
						array[y]=0;
						for (int x=0; x<bitmap.x_dim(); x++)
							if (bitmap[y,x]==Pixel.PixelEnum.set) array[y]++;

						if (array[y]>max_value_) max_value_=array[y];
						average_ += array[y];
					}
					if (bitmap.y_dim()>0)
					{
						average_ /= bitmap.y_dim();
					}
					break; 
			}
		}


		public int find_max_tp(int pos, DirectionEnum direction)
		{
			IntArray1d array = (IntArray1d)(this);

			int max_tp = array[pos];

			switch (direction)
			{
				case DirectionEnum.forwards:
					if (pos==0) pos++;
					if (pos<=len_-1)
						while (array[pos-1]>=array[pos])
						{
							if (pos==len_-1) break;
							pos++;
						}

					if (pos<len_-1)
						while (array[pos]<=array[pos+1])
						{
							pos++;
							if (pos==len_-1) break;
						}

					max_tp = pos;
					break;

				case DirectionEnum.backwards:
					if (pos==len_-1) pos--;
					if (pos>=0)
						while (array[pos+1]>=array[pos])
						{
							if (pos==0) break;
							pos--;
						}

					if (pos>0)
						while (array[pos]<=array[pos-1])
						{
							pos--;
							if (pos==0) break;
						}

					max_tp = pos;
					break;

			}

			return max_tp;
		}




		public int find_min_tp(int pos, DirectionEnum direction)
		{
			IntArray1d array = (IntArray1d)(this); 

			int min_tp = array[pos];

			switch (direction)
			{
				case DirectionEnum.forwards:
					if (pos==0) pos++;
					if (pos<=len_-1)
						while (array[pos-1]<=array[pos])
						{
							if (pos==len_-1) break;
							pos++;
						}

					if (pos<len_-1)
						while (array[pos]>=array[pos+1])
						{
							pos++;
							if (pos==len_-1) break;
						}

					min_tp = pos;

					break;

				case DirectionEnum.backwards:
					if (pos==len_-1) pos--;
					if (pos>=0)
						while (array[pos+1]<=array[pos])
						{
							if (pos==0) break;
							pos--;
						}

					if (pos>0)
						while (array[pos]>=array[pos-1])
						{
							pos--;
							if (pos==0) break;
						}

					min_tp = pos;

					break;

			}

			return min_tp;
		}

		public Projection DeepCopy()
		{
			Projection proj_copy = new Projection(len_);
			for (int i=0; i<len_; i++) 
			{
				proj_copy[i] = array_[i];
			}

			proj_copy.max_value_ = max_value_;
			proj_copy.average_ = average_;

			return proj_copy;
		}



		public int max_value() { return max_value_; }
		public int average() { return average_; }
	}
}
  
