
using System;
using System.IO;

namespace Espresso.OMR
{
	class SubBitmap : Bitmap
	{

		public enum Operation { nop, remove };

		protected int x_org_;
		protected int y_org_;

		public SubBitmap()
			: base()
		{
			x_org_ = 0; 
			y_org_ = 0;
		}

		public SubBitmap(SubBitmap sub_bitmap)
			: base(sub_bitmap)
		{
			x_org_ = sub_bitmap.x_org_; 
			y_org_ = sub_bitmap.y_org_;
		}

		public SubBitmap(int sx_org, int sy_org, int dx_dim, int dy_dim)
			: base(dx_dim,dy_dim,sx_org%8)
		{
			x_org_ = sx_org;
			y_org_ = sy_org;
		}

		public SubBitmap(Bitmap src_bitmap, int dx_org, int dy_org, int dx_dim, int dy_dim)
			: base(dx_dim,dy_dim,dx_org%8)
		{
			x_org_ = dx_org;
			y_org_ = dy_org;

			extract(src_bitmap,dx_org,dy_org,dx_dim,dy_dim);
		}

		public SubBitmap(Bitmap src_bitmap, int dx_org, int dy_org,
						 int dx_dim, int dy_dim, Operation op)
			: base(dx_dim,dy_dim,dx_org%8)
		{
			x_org_ = dx_org;
			y_org_ = dy_org;

			extract(src_bitmap,dx_org,dy_org,dx_dim,dy_dim);
			if (op==Operation.remove) src_bitmap.clear_area(dx_org,dy_org,dx_dim,dy_dim);
		}

		public SubBitmap(SubBitmap src_bitmap, 
			int dx_org, int dy_org,int dx_dim, int dy_dim)
			: base(dx_dim,dy_dim,(src_bitmap.x_off_+dx_org)%8)

		{
			x_org_ = dx_org;
			y_org_ = dy_org;

			extract(src_bitmap,dx_org,dy_org,dx_dim,dy_dim);
		}

		public SubBitmap(SubBitmap src_bitmap, 
			int dx_org, int dy_org,int dx_dim, int dy_dim, Operation op)
			: base(dx_dim,dy_dim,(src_bitmap.x_off_+dx_org)%8)

		{
			x_org_ = dx_org;
			y_org_ = dy_org;

			extract(src_bitmap,dx_org,dy_org,dx_dim,dy_dim);
			if (op==Operation.remove) src_bitmap.clear_area(dx_org,dy_org,dx_dim,dy_dim);
		}


		/*
		SubBitmap(char* new_bytes, int x_off, 
			  int x_org, int y_org,
			  int x_dim, int y_dim)
		  : Bitmap(x_dim,y_dim,x_off),
		  x_org_(x_org), y_org_(y_org)
		  {
		char** pdata = bitmap_.pdata();
		if (pdata!=nil)
		  {
			char*  old_bytes = bitmap_.cdata();
	    
			int byte_x_dim = (x_off + x_dim + 7) >> 3;
			char* line = new_bytes;
			for (int i=0; i<y_dim; i++,line+=byte_x_dim)
			  pdata[i]=line;
	    
			delete old_bytes;
		  }
		  }
		*/

		public bool rect_intersect(SubBitmap other_bitmap)
		{
			int xl = x_org_;
			int xr = xl + x_dim_;
			int yt = y_org_;
			int yb = yt + y_dim_;
	
			int oxl = other_bitmap.x_org();
			int oxr = oxl + other_bitmap.x_dim();
			int oyt = other_bitmap.y_org();
			int oyb = oyt + other_bitmap.y_dim();
	
			return (((oxl<=xr) && (oxr>=xl)) && ((oyt<=yb) && (oyb>=yt)));
		}
    
    
		public bool rect_inside(SubBitmap other_bitmap)
		{
			int xl = x_org_;
			int xr = xl + x_dim_ -1;
			int yt = y_org_;
			int yb = yt + y_dim_ -1;
	
			int oxl = other_bitmap.x_org();
			int oxr = oxl + other_bitmap.x_dim() -1;
			int oyt = other_bitmap.y_org();
			int oyb = oyt + other_bitmap.y_dim() -1;
	
			return (((oxr<=xr) && (oxl>=xl)) && ((oyb<=yb) && (oyt>=yt)));
		}


		public int rect_overlap(SubBitmap other_bitmap)
		{
			int xl = x_org_;
			int xr = xl + x_dim_;
			int yt = y_org_;
			int yb = yt + y_dim_;

			int oxl = other_bitmap.x_org();
			int oxr = oxl + other_bitmap.x_dim();
			int oyt = other_bitmap.y_org();
			int oyb = oyt + other_bitmap.y_dim();

			int cropped_xl = Math.Max(xl,oxl);
			int cropped_xr = Math.Min(xr,oxr);
			int cropped_yt = Math.Max(yt,oyt);
			int cropped_yb = Math.Min(yb,oyb);

			int cropped_x_dim = cropped_xr - cropped_xl +1; 
			int cropped_y_dim = cropped_yb - cropped_yt +1;

			int overlap = cropped_x_dim * cropped_y_dim;

			// Make result -ve if two rects do not overlap.
			// orthogonal misses negative, but diagonals are not
			// => make diagonal ones negative also

			if ((cropped_x_dim<0) && (cropped_y_dim<0))
			{
				overlap = -overlap;
			}

			return overlap;
		}

		/*-------------------*/

		public void copy_sub_bitmap(SubBitmap src_bitmap)
		{
			SubBitmap dst_bitmap = this;

			int src_x_off = src_bitmap.x_off();
			int src_x_org = src_bitmap.x_org();
			int src_y_org = src_bitmap.y_org();
			int src_x_dim = src_bitmap.x_dim();
			int src_y_dim = src_bitmap.y_dim();

			dst_bitmap.set_x_off(src_x_off);
			dst_bitmap.set_x_org(src_x_org);
			dst_bitmap.set_y_org(src_y_org);

			resize(src_x_dim,src_y_dim);
			extract(src_bitmap,0,0,src_x_dim,src_y_dim);
		}

		public void relay_prim(SubBitmap rl_prim)
		{
			// the prim to relay may lie outside the src_bitmap
			// 'if' checks prevent any illegal accesses

			SubBitmap src_bitmap = this;

			int src_x_org = src_bitmap.x_org();
			int src_y_org = src_bitmap.y_org();

			int src_x_dim = src_bitmap.x_dim();
			int src_y_dim = src_bitmap.y_dim();

			int rl_x_dim = rl_prim.x_dim();
			int rl_y_dim = rl_prim.y_dim();
			int rl_x_org = rl_prim.x_org();
			int rl_y_org = rl_prim.y_org();

			int x_diff = rl_x_org - src_x_org;
			int y_diff = rl_y_org - src_y_org;

			for (int by=0, sy=y_diff; by<rl_y_dim; by++,sy++)
				for (int bx=0, sx=x_diff; bx<rl_x_dim; bx++,sx++)
				{
					if (((sx>=0) && (sx<src_x_dim)) &&
						((sy>=0) && (sy<src_y_dim)))
					{
						if (rl_prim[by,bx]==PixelEnum.set) 
						{
							src_bitmap.set_pixel(sx,sy);
						}
					}
				}
		}


		public int x_org() { return x_org_; }
		public int y_org() { return y_org_; }
		public void add_to_x_org(int x_add) { x_org_+=x_add; }
		public void add_to_y_org(int y_add) { y_org_+=y_add; }
		public void set_x_org(int x_org) { x_org_=x_org; }
		public void set_y_org(int y_org) { y_org_=y_org; }

		public int x_left() { return x_org_; }
		public int x_right() { return x_org_+x_dim_-1; }
		public int y_top() { return y_org_; }
		public int y_bot() { return y_org_+y_dim_-1; }

		public bool point_inside(int x_pt, int y_pt)
		{
			int xl = x_org_;
			int xr = xl + x_dim_ -1;
			int yt = y_org_;
			int yb = yt + y_dim_ -1;

			return (((x_pt<=xr) && (x_pt>=xl)) && ((y_pt<=yb) && (y_pt>=yt)));
		}


		/*---------------------------------------------------------------------------*/


		public override void write(BinaryWriter bwout)
		  {
		    IOSupport.write_line(bwout,"x_off = " + x_off_);
		    IOSupport.write_line(bwout,"x_org = " + x_org_ + " , y_org = " + y_org_);
		    base.write(bwout);

		  }

		public override bool read(BinaryReader brin)
		  {
		    if (!IOSupport.read_token(brin,"x_off")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out x_off_)) return false;

		    if (!IOSupport.read_token(brin,"x_org")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out x_org_)) return false;

		    if (!IOSupport.read_token(brin,",")) return false;

		    if (!IOSupport.read_token(brin,"y_org")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out y_org_)) return false;

		    base.read(brin);

		    return true;
		  }


	}
}
    


