using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace Espresso.Utils.View
{
    public partial class PreferencesForm : Form
    {
        public PreferencesForm()
        {
            InitializeComponent();
            
            FillFields();

            this.bt_Cancel.Click += new EventHandler(bt_Cancel_Click);
            this.bt_OK.Click += new EventHandler(bt_OK_Click);
        }

        void bt_OK_Click(object sender, EventArgs e)
        {
            Preferences p = Preferences.Instance;

            p.SetProperty("PageFoldAngle", (int)nm_Angle.Value);
            p.SetProperty("TransitionDelay", (decimal)nm_Delay.Value);
            p.SetProperty("RollHeight", (int)nm_RollHeight.Value);
            p.SetProperty("TransitionType", (string)cb_TransitionType.SelectedItem);
            p.SetProperty("StopHalfWayThrough", (bool)cb_StopHalfWay.Checked);
            p.SetProperty("InstantTransition", (bool)cb_InstantTransition.Checked);
            p.SetProperty("ChangeSpeedHalfWayThrough", (bool)cb_SlowDownHalfWay.Checked);
            p.SetProperty("FirstHalfDelay", nm_FirstHalf.Value);
            p.SetProperty("SecondHalfDelay", nm_SecondHalf.Value);
            p.SetProperty("UseDefaultPathToStorageFolder", cb_UseDefaultStoragePath.Checked);
            p.SetProperty("PathToStorageFolder", tx_StoragePath.Text);
            p.SetProperty("DPI", tb_Dpi.Value);
            p.SetProperty("DocumentPageLimit", Convert.ToInt32(tx_PageLimit.Text));
            p.SetProperty("LimitFoldMaxHeight", cb_LimitFoldHeight.Checked);

            if (rb_LeftForwarRightBack.Checked == true)
            {
                p.SetProperty("ControlStrategy", "LeftNextRightPrevious");
            }
            else if (rb_MenuChoice.Checked == true)
            {
                p.SetProperty("ControlStrategy", "MenuChoice");
            }
            else if (rb_SingleClickForwardDoubleBack.Checked == true)
            {
                p.SetProperty("ControlStrategy", "SingleNextDoublePrevious");
            }

            p.Save();
            this.Close();
        }

        void bt_Cancel_Click(object sender, EventArgs e)
        {
            Cancel();
            this.Close();
        }

        private void Cancel()
        {
            Preferences.Instance.Load();
        }

        private void FillFields()
        {
            try
            {
                Preferences p = Preferences.Instance;

                nm_Angle.Value = (int)p.GetProperty("PageFoldAngle");
                nm_Delay.Value = (decimal)p.GetProperty("TransitionDelay");
                nm_RollHeight.Value = (int)p.GetProperty("RollHeight");
                cb_TransitionType.SelectedItem = (string)p.GetProperty("TransitionType");
                cb_StopHalfWay.Checked = (bool)p.GetProperty("StopHalfWayThrough");
                cb_InstantTransition.Checked = (bool)p.GetProperty("InstantTransition");
                cb_SlowDownHalfWay.Checked = (bool)p.GetProperty("ChangeSpeedHalfWayThrough");
                nm_FirstHalf.Value = (decimal)p.GetProperty("FirstHalfDelay");
                nm_SecondHalf.Value = (decimal)p.GetProperty("SecondHalfDelay");
                cb_UseDefaultStoragePath.Checked = (bool)p.GetProperty("UseDefaultPathToStorageFolder");
                tx_StoragePath.Text = (string)p.GetProperty("PathToStorageFolder");
                tb_Dpi.Value = (int)p.GetProperty("DPI");
                tx_PageLimit.Text = Convert.ToString(p.GetProperty("DocumentPageLimit"));
                cb_LimitFoldHeight.Checked = (bool)p.GetProperty("LimitFoldMaxHeight");
            }
            catch (InvalidCastException)
            {

            }
            finally
            {

                if ((bool)(Preferences.Instance.GetProperty("InstantTransition")) == true)
                {
                    nm_Delay.Enabled = false;
                }
                else
                {
                    nm_Delay.Enabled = true;
                }

                string strategy = (string)Preferences.Instance.GetProperty("ControlStrategy");

                if (strategy.Equals("LeftNextRightPrevious"))
                {
                    rb_LeftForwarRightBack.Checked = true;
                }
                else if (strategy.Equals("MenuChoice"))
                {
                    rb_MenuChoice.Checked = true;
                }
                else if (strategy.Equals("SingleNextDoublePrevious"))
                {
                    rb_SingleClickForwardDoubleBack.Checked = true;
                }
            }
        }

        private void cb_InstantTransition_CheckedChanged(object sender, EventArgs e)
        {
            if (cb_InstantTransition.Checked)
                nm_Delay.Enabled = false;
            else
                nm_Delay.Enabled = true;
        }

        private void cb_SlowDownHalfWay_CheckedChanged(object sender, EventArgs e)
        {
            if (cb_SlowDownHalfWay.Checked)
            {
                nm_FirstHalf.Enabled = true;
                nm_SecondHalf.Enabled = true;
                nm_Delay.Enabled = false;
                cb_InstantTransition.Enabled = false;
            }
            else
            {
                nm_FirstHalf.Enabled = false;
                nm_SecondHalf.Enabled = false;
                nm_Delay.Enabled = true;
                cb_InstantTransition.Enabled = true;
            }
        }

        private void cb_TransitionType_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (cb_TransitionType.SelectedItem.Equals("Simple line"))
            {
                nm_Angle.Enabled = false;
                nm_RollHeight.Enabled = false;
                cb_LimitFoldHeight.Enabled = false;
            }
            else if (cb_TransitionType.SelectedItem.Equals("Page turn"))
            {
                nm_Angle.Enabled = true;
                nm_RollHeight.Enabled = false;
                cb_LimitFoldHeight.Enabled = true;
            }
            else if (cb_TransitionType.SelectedItem.Equals("Page roll"))
            {
                nm_Angle.Enabled = false;
                nm_RollHeight.Enabled = true;
                cb_LimitFoldHeight.Enabled = false;
            }
        }

        private void cb_UseDefaultStoragePath_CheckedChanged(object sender, EventArgs e)
        {
            bt_SelectFolder.Enabled = !cb_UseDefaultStoragePath.Checked;
            tx_StoragePath.Enabled = !cb_UseDefaultStoragePath.Checked;
        }

        private void bt_SelectFolder_Click(object sender, EventArgs e)
        {
            storageFolderDialog.Description = "Please select a folder for cache storage";
            DialogResult res = storageFolderDialog.ShowDialog();

            if (res == DialogResult.OK)
            {
                tx_StoragePath.Text = storageFolderDialog.SelectedPath;
            }
        }
    }
}