﻿using System;
using System.Collections.Generic;
using System.Text;
using Espresso.Utils;

namespace Espresso.Viewer.Model
{
    /// <summary>
    /// A fragment of a document.
    /// </summary>
    public class DocumentFragment
    {
        /// <summary>
        /// Document Fragment conversion resolution
        /// </summary>
        public int DPI = (int)Preferences.Instance.GetProperty("DPI");

        private List<Page> m_Pages;
        
        /// <summary>
        /// All the pages in the fragment
        /// </summary>
        public List<Page> Pages
        {
            get { return m_Pages; }
            private set { m_Pages = value; }
        }

        private int m_CurrentPageNumber;

        public int CurrentPageNumber
        {
            get { return m_CurrentPageNumber; }

            set
            {
                m_CurrentPageNumber = value;

                if (CurrentPageChanged != null)
                {
                    CurrentPageChanged(new CurrentPageChangedEventArgs(value));
                }
            }
        }

        public delegate void CurrentPageChangedEventHandler(CurrentPageChangedEventArgs args);

        public event CurrentPageChangedEventHandler CurrentPageChanged;

        public DocumentFragment()
        {
            this.Pages = new List<Page>();
        }

        public DocumentFragment(int numPages)
        {
            this.Pages = new List<Page>(numPages);
        }

        /// <summary>
        /// Can we go to a given page number?
        /// </summary>
        /// <param name="pageNumber"></param>
        /// <returns></returns>
        public bool CanGoToPage(int pageNumber)
        {
            bool canGo = false;

            if (pageNumber < 1)
            {
                return false;
            }

            foreach(Page aPage in this.Pages)
            {
                canGo = canGo || (aPage.PageNumber == pageNumber - 1); // TODO: DANGEROUS
            }

            return canGo;
        }

        public void addPage(Page aPage)
        {
            this.Pages.Add(aPage);
        }

        public class CurrentPageChangedEventArgs : EventArgs
        {
            public int currentPageNumber;

            public CurrentPageChangedEventArgs(int pageNumber)
            {
                currentPageNumber = pageNumber;
            }
        }
    }
}
