using System;
using System.Collections.Generic;
using System.Text;

namespace Espresso.Viewer.Model
{
    public class Menu
    {
        protected List<MenuItem> m_Items = new List<MenuItem>();

        public IEnumerable<MenuItem> Items
        {
            get { return m_Items; }
        }

        public void AddItem(MenuItem item)
        {
            m_Items.Add(item);
        }

        public void RemoveItem(MenuItem item)
        {
            m_Items.Remove(item);
        }

        private MenuItem m_CurrentItem;

        /// <summary>
        /// The item that is currently selected.
        /// </summary>
        public MenuItem CurrentItem
        {
            get { return m_CurrentItem; }
            set { m_CurrentItem = value; }
        }

        private bool m_IsVisible = false;

        public bool IsVisible
        {
            get { return m_IsVisible; }
            set { m_IsVisible = value; }
        }
	
        public MenuItem NextItem
        {
            get
            {
                int curIndex = m_Items.IndexOf(CurrentItem);

                if (curIndex == m_Items.Count - 1)
                { 
                    //Last item is the current one => set the first item to the current one.
                    return m_Items[0];
                }
                else
                {
                    return m_Items[curIndex + 1];
                }
            }
        }

        public void ChooseNextItem()
        {
            this.CurrentItem = this.NextItem;
        }

        /// <summary>
        /// Resets the current item to the first item in the menu
        /// </summary>
        public void Reset()
        {
            this.CurrentItem = m_Items[0];
        }
    }
}
