using System;
using System.Collections.Generic;
using System.Text;

namespace Espresso.Viewer.Model
{
    /// <summary>
    /// A class that describes an item in a menu
    /// </summary>
    public class MenuItem
    {
        private string m_Text;

        /// <summary>
        /// The text that appears in the menu
        /// </summary>
        public string Text
        {
            get { return m_Text; }
            set { m_Text = value; }
        }

        protected VoidDelegate m_Action;

        /// <summary>
        /// Called when the item is selected in the menu.
        /// </summary>
        public void Select()
        {
            if (m_Action != null)
                m_Action();
        }

        /// <summary>
        /// Registers the action callback and sets the text
        /// </summary>
        /// <param name="action">A void method with no arguments that will be invoked when 
        ///                      the item is selected</param>
        /// <param name="text">The text on the item</param>
        public MenuItem(VoidDelegate action, string text)
        {
            m_Action = action;
            Text = text;
        }
    }

    public delegate void VoidDelegate();
}
