using System;
using System.Collections.Generic;
using System.Text;

using System.Xml;
using System.Xml.Serialization;

namespace Espresso.Viewer.Model
{
    /// <summary>
    /// The class that reprensents a system in a document.
    /// Stores the rectangle coordinates and possibly some meta-info like bookmarks
    /// </summary>
    [XmlRoot("systemRectangle")]
    public class SystemRectangle
    {
        private Rect m_Rectangle;

        /// <summary>
        /// The boundaries of the system withing a page
        /// </summary>
        [XmlElement("rectangle")]
        public Rect Rectangle
        {
            get { return m_Rectangle; }
            set { m_Rectangle = value; }
        }

        private Page m_Page;

        /// <summary>
        /// The page that this system belongs to
        /// </summary>
        [XmlIgnore]
        public Page Page
        {
            get { return m_Page; }
            internal set { m_Page = value; }
        }

        /// <summary>
        /// Sets the page and the rectangle
        /// </summary>
        /// <param name="page"></param>
        /// <param name="rectangle"></param>
        public SystemRectangle(Page page, Rect rectangle)
        {
            this.Page = page;
            this.Rectangle = rectangle;
        }

        /// <summary>
        /// Creates a system that occupies the whole page
        /// </summary>
        /// <param name="page">Page that the system belongs to</param>
        public SystemRectangle(Page page)
            :
            this(page, new Rect(0, 0, page.Width, page.Height)) { }

        public SystemRectangle() : this(null) { }

        public override string ToString()
        {
            return this.Rectangle.ToString();
        }
    }
}
