using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace Espresso.Viewer.View
{
    public class CompositePainter
    {
        protected List<IPainter> children;

        private static CompositePainter m_Instance;

        /// <summary>
        /// Singleton instance
        /// </summary>
        public static CompositePainter Instance
        {
            get
            {
                if (m_Instance == null)
                    m_Instance = new CompositePainter();

                return m_Instance;
            }
        }

        private CompositePainter()
        {
            children = new List<IPainter>();
        }

        /// <summary>
        /// Removes a leaf painter from the tree
        /// </summary>
        /// <param name="painter"></param>
        /// <returns></returns>
        public IPainter RemovePainter(IPainter painter)
        {
            if (painter == null)
                return null;

            IPainter retVal = children.Find(
                new Predicate<IPainter>(delegate(IPainter p)
                    {
                        if (p.Name.Equals(painter.Name))
                            return true;
                        else
                            return false;
                    }));

            if (retVal != null)
                children.Remove(retVal);
            
            return retVal;
        }

        /// <summary>
        /// Adds a painter to the tree.
        /// </summary>
        /// <param name="painter"></param>
        public void AddPainter(IPainter painter)
        {
            children.Add(painter);
        }

        /// <summary>
        /// Adds a painter to the beginning of the list of painters
        /// </summary>
        /// <param name="painter"></param>
        public void AddPainterToStart(IPainter painter)
        {
            List<IPainter> newList = new List<IPainter>();
            newList.Add(painter);
            newList.AddRange(children);

            children = newList;
        }

        /// <summary>
        /// Resets its invalid region, calls paint on its children and unions its
        /// invalid region with the invalid regions of the children
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        public void Paint(object sender, System.Windows.Forms.PaintEventArgs args)
        {
            foreach (IPainter p in children)
            {
                p.Paint(sender, args);
            }
        }

        //TODO: Change that to return dynamic values, otherwise there will be weird behaviour
        //      when addind and removing composite painters to/from composite painters
        public string Name
        {
            get { return "CompositePainter"; }
        }
    }
}
