using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using Espresso.Viewer.Model;

namespace Espresso.Viewer.View
{
    public class PageNumberPainter : PainterDecorator
    {
        private int m_PageNumber;

        private int m_TotalPages;

        private Document m_Document;

        public PageNumberPainter(AbstractPainter decoratee, Document document)
            : base(decoratee)
        {
            m_Document = document;
            document.CurrentPageChanged += UpdatePageNumber;

            m_PageNumber = document.CurrentPageNumber;
            m_TotalPages = document.Pages.Count;
        }

        public override void Paint(object sender, PaintEventArgs args)
        {
            base.Paint(sender, args); 
            
            Graphics g = args.Graphics;
            string str = "Page " + m_PageNumber + "/" + m_TotalPages;
            g.DrawString(str, new Font(FontFamily.GenericSansSerif, 15), Brushes.DarkSlateBlue, new PointF(10, 10));
        }

        public override string Name
        {
            get { return "PageNumberPainter"; }
        }

        private void UpdatePageNumber(CurrentPageChangedEventArgs args)
        {
            m_PageNumber = args.currentPageNumber;
        }
    }
}
