using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;
using Espresso.Utils;
using Espresso.Viewer.Model;

namespace Espresso.Viewer.View
{
    public class PageRollTransitionPainter : AbstractPainter
    {
        private int ROLL_HEIGHT{
            get
            {
                return (int)Preferences.Instance.GetProperty("RollHeight");
            }
        }

        public PageRollTransitionPainter()
            : base(null, null) { }

        public PageRollTransitionPainter(Bitmap oldImage, Bitmap newImage) : base(oldImage, newImage) { }

        // TODO: Added By Jojo
        public PageRollTransitionPainter(Bitmap oldImage, Bitmap newImage, Document document) : base(oldImage, newImage, document) { }

        public override void Paint(object sender, System.Windows.Forms.PaintEventArgs args)
        {
            Point A, B, C, D;
            Control control = (Control)sender;
            Rectangle drawRect = GetDrawRectangle(control);
            Matrix m = new Matrix();

            //TODO: JOJO
            //List<Rectangle> canvas = GetDrawRectangleList(control, 3);
            

            if (this.m_bitBuffer == null || this.m_bitBuffer.Size != control.Size
               || this.ScaledNewImage == null || this.ScaledOldImage == null)
            {
                InitialiseBuffers(control);
            }

            g.FillRectangle(Brushes.DarkGray, new Rectangle(new Point(0, 0), m_bitBuffer.Size));
            
            if (this.CompleteRatio == 0)
            {
                //for(int i = 0; i < canvas.Count; i++) 
                //{
                   // Rectangle drawingRect = canvas[i];

                    //Draw the white background
                    g.FillRectangle(Brushes.White, drawRect);

                    //Draw the complete old page
                    Rectangle oldSrcRect = new Rectangle(0, 0, ScaledOldImage.Width, ScaledOldImage.Height);
                    g.DrawImage(ScaledOldImage, drawRect, oldSrcRect, GraphicsUnit.Pixel);

                    this.InvalidRegion = drawRect;

                    //reload
                    //Rectangle scaled = GetDrawRectangle(control, 3);
                    //this.OldImage = this.Document.Pages[i].Bitmap;
                    //this.ScaledOldImage = (Bitmap)OldImage.GetThumbnailImage(scaled.Width, scaled.Height, null, System.IntPtr.Zero);
                //}
            }
            else if (this.CompleteRatio == 1)
            {
                //foreach (Rectangle drawingRect in canvas)
                //{
                    //Draw the white background
                    g.FillRectangle(Brushes.White, drawRect);

                    //Draw the complete new page
                    Rectangle newSrcRect = new Rectangle(0, 0, ScaledNewImage.Width, ScaledNewImage.Height);
                    g.DrawImage(ScaledNewImage, drawRect, newSrcRect, GraphicsUnit.Pixel);

                    this.InvalidRegion = drawRect;
                //}
            }
            else
            {
                double H = drawRect.Height;
                double W = drawRect.Width;
                double X = drawRect.X;
                double Y = drawRect.Y;
                double T = H * CompleteRatio;
                
                //A value by which the turning page should be offset down in order
                //to make it disappear completely when the transition is complete
                double offRatio = ROLL_HEIGHT * (T / H);

                T += offRatio;

                //Need to do this because sometimes scaled images end up being null by the time
                // we get to this point in execution. Need to investigate into this more.
                if (this.m_bitBuffer == null || this.m_bitBuffer.Size != control.Size
                || this.ScaledNewImage == null || this.ScaledOldImage == null)
                {
                    InitialiseBuffers(control);
                }

                Rectangle oldSrcRect = new Rectangle(0, 0, ScaledOldImage.Width, ScaledOldImage.Height);
                Rectangle newSrcRect = new Rectangle(0, 0, ScaledNewImage.Width, (int)T + 1);
                Rectangle newDestRect = new Rectangle((int)X, (int)Y, (int)W, (int)T + 1); 

                //Draw the white background
                g.FillRectangle(Brushes.White, drawRect);

                //Draw the complete old page
                g.DrawImage(ScaledOldImage, drawRect, oldSrcRect, GraphicsUnit.Pixel);

                //Draw part of the new image
                g.DrawImage(ScaledNewImage, newDestRect, newSrcRect, GraphicsUnit.Pixel);
                
                A = new Point((int)X, (int)T);
                B = new Point((int)(X + W), (int)T);
                C = new Point((int)B.X, (int)(B.Y - ROLL_HEIGHT));
                D = new Point((int)A.X, (int)(A.Y - ROLL_HEIGHT));

                Point[] pts = new Point[4] { A, B, C, D };
                g.FillPolygon(Brushes.White, pts, FillMode.Winding);
                //g.DrawPolygon(new Pen(Brushes.Gray, 1), pts);

                if (T != 0)
                {
                    Point BC = new Point(C.X, (int)(C.Y + ROLL_HEIGHT / 2));
                    Point AD = new Point(D.X, (int)(D.Y + ROLL_HEIGHT / 2));
                    GraphicsPath path = new GraphicsPath(new Point[] { A, B, BC, C, D, AD },
                                                         new byte[] { 1, 1, 1, 1, 1, 1 });

                    PathGradientBrush pgBrush = new PathGradientBrush(path);
                    pgBrush.CenterColor = Color.White;

                    pgBrush.SurroundColors = new Color[] { Color.FromArgb(40, Color.Black), 
                                                           Color.FromArgb(40, Color.Black), 
                                                           Color.FromArgb(127, Color.White), 
                                                           Color.FromArgb(180, Color.Black),
                                                           Color.FromArgb(180, Color.Black), 
                                                           Color.FromArgb(127, Color.White) 
                                                           };
                    Region r = new Region(path);

                    g.FillRegion(pgBrush, r);
                }

                this.InvalidRegion = new Rectangle(new Point(0, 0), m_bitBuffer.Size);
            }

            args.Graphics.DrawImage(this.m_bitBuffer, new Point(0, 0));
        }

        public override string Name
        {
            get { return "PageRollPainter"; }
        }
    }
}
