using System;
using System.Collections.Generic;
using System.Text;
using Espresso.Viewer.Model;

namespace Espresso.Viewer.View
{
    public class PainterBuilder : IBuilder<IPainter>
    {
        protected Document m_Document;

        protected Menu m_Menu;

        public PainterBuilder(Document document, Menu menu)
        {
            m_Document = document;
            m_Menu = menu;
        }

        public IPainter Build(string description)
        {
            AbstractPainter painter;

            if (description.Equals("Page turn"))
            {
                painter = new PageTurnTransitionPainter();
            }
            else if (description.Equals("Page roll"))
            {
                painter = new PageRollTransitionPainter();
                // TODO: JOJO
                //painter.Document = m_Document;
            }
            else
            {
                painter = new WipeTransitionPainter();
            }

            //Wrap in decorators
            painter = new MenuPainter(painter, m_Menu);
            
            painter = new PageNumberPainter(painter, m_Document);

            painter.NewImage = m_Document.Pages[m_Document.CurrentPageNumber - 1].Bitmap;
            painter.OldImage = painter.NewImage;

            return painter;
        }
    }
}
