using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Configuration;
using System.Threading;
using Espresso.Viewer.Controller;
using Espresso.Viewer.Model;
using Espresso.Utils;
using Espresso;
using Espresso.Utils.View;
using Espresso.Viewer.View;

namespace Latte
{
    /// <summary>
    /// The main form that is displayed to the user. Containts
    /// </summary>
    public partial class MainForm : Form
    {
        private const string CAPTION = "Latte -- A Light and Fluffy Digital Music Stand";

        private bool m_keyStrokeProcessed = true;

        private bool m_InFullScreenMode = false;

        private bool m_ControllerPaused = false;
        //A temporary measure
        private string m_CurrentOpenDocument = "";

        // TODO: JOJO
        IViewerController controller;
        //DocumentViewerController controller;

        /// <summary>
        /// Sets the caption on the window. Registers the menu items events.
        /// </summary>
        public MainForm()
        {
            Thread.CurrentThread.Name = "Main";

            InitializeComponent();

            this.fileOpen.Click += new System.EventHandler(this.fileOpen_Click);
            this.KeyPress += new KeyPressEventHandler(MainForm_KeyPress);

            menuItem_FullScreen.Checked = false;

            this.Text = CAPTION;

            progressPanel.Visible = false;

            Preferences.Instance.Load();
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            Preferences.Instance.Save();

            base.OnClosing(e);
        }

        [STAThread]
        static void Main()
        {
            Application.Run(new MainForm());
        }

        private void CloseCurrentDocument()
        {

        }

        private void OpenDocument(string filePath, bool overWriteCache)
        {
            this.Cursor = Cursors.WaitCursor;

            //Create an empty document, register the event and open the file
            Document d = new Document();
            //d.ProcessedAPage += d_ProcessedAPage;

            bool openedDocument = false;
            ThreadStart ts = new ThreadStart(delegate()
            {
                openedDocument = d.OpenAPdf(filePath, overWriteCache);
            });

            Thread openDocumentThread = new Thread(ts);
            openDocumentThread.Start();

            this.Invalidate();

            Application.DoEvents();
            openDocumentThread.Join();
            
            /*
            MethodCall<bool, string> asyncOpen = d.OpenAPdf;
            asyncOpen.BeginInvoke(filePath, delegate(IAsyncResult result)
              {
                  asyncOpen.EndInvoke(result);
              }, null);
            */

            if (openedDocument)
            {
                DocumentViewerController docViewController = new DocumentViewerController(d, 3);
                controller = new DocumentFragmentViewerController(d, docViewController, this.espressoViewer1);
                
                //controller = new ViewerController(d, this.espressoViewer1);

                //Create auto-updatable fields
                FieldUpdater<IPainter> painterUpdater =
                    new FieldUpdater<IPainter>("TransitionType",
                                               controller,
                                               new PainterBuilder(d, controller.Menu));

                FieldUpdater<MouseStrategy> mouseUpdater =
                    new FieldUpdater<MouseStrategy>("ControlStrategy",
                                                    controller,
                                                    new MouseStrategyBuilder(controller));

                //FieldUpdater<ViewStrategy> viewUpdater =
                //    new FieldUpdater<ViewStrategy>("ViewStrategy",
                //                                    controller,
                //                                    new ViewStrategyBuilder());

                controller.Reset();
            }

            this.Cursor = Cursors.Arrow;

           // d.ProcessedAPage -= d_ProcessedAPage;
        }

        private void d_ProcessedAPage(object sender, ProcessedPageEventArgs e)
        {
            this.progressPanel.Value = (int)(e.ProcessedPageNumber / e.TotalPages * 100);
            this.Invalidate();
        }

        /// <summary>
        /// Closes the current document, prompts the user for a pdf and
        /// displays it if the user selects one.
        /// </summary>
        private void OpenAFile()
        {
            //Prompt user for a file
            OpenFileDialog dialog = new OpenFileDialog();
            dialog.Filter = "PDF Files|*.pdf|All files|*.*";
            dialog.FilterIndex = 1;

            //If there is a file selected, open it
            if (dialog.ShowDialog() == DialogResult.OK)
            {
                CloseCurrentDocument();

                string pdfFilename = dialog.FileName;

                dialog.Dispose();
                progressPanel.Reset(0, 10);
                progressPanel.Visible = true;

                Thread.Sleep(300);

                if (pdfFilename.Equals(m_CurrentOpenDocument))
                {
                    OpenDocument(pdfFilename, true);
                }
                else
                {
                    OpenDocument(pdfFilename, false);
                    m_CurrentOpenDocument = pdfFilename;
                }

                progressPanel.Visible = false;
            }
        }

        private void fileOpen_Click(object sender, System.EventArgs e)
        {
            OpenAFile();
        }

        /// <summary>
        /// Closes the current document.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void fileClose_Click(object sender, System.EventArgs e)
        {
            CloseCurrentDocument();
        }

        /// <summary>
        /// Processes the keys that <see cref="ProcessDialogKey"/> can not process. Namely 
        /// number keys (which set the number of columns).
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_KeyPress(object sender, System.Windows.Forms.KeyPressEventArgs e)
        {
            if (!m_keyStrokeProcessed)
            {
                char key = e.KeyChar;

                if ((key >= '1') && (key <= '9'))
                {
                    int no_cols = key - '0';
                }
            }
        }

        /// <summary>
        /// Processes the key events and calls apropriate methods on viewerPanel.
        /// </summary>
        /// <param name="keyData"></param>
        /// <returns></returns>
        protected override bool ProcessDialogKey(Keys keyData)
        {
            m_keyStrokeProcessed = true;

            switch (keyData)
            {
                case Keys.PageDown:
                    this.controller.Timer.AlterSpeed(0.5f);
                    break;
                case Keys.PageUp:
                    this.controller.Timer.AlterSpeed(2f);
                    break;
                case Keys.End:
                    this.controller.GoToLastPage(false);
                    break;
                case Keys.Home:
                    this.controller.GoToFirstPage(false);
                    break;
                case Keys.Down:
                    this.controller.GoToNextPage(false);
                    break;
                case Keys.Up:
                    this.controller.GoToPreviousPage(false);
                    break;
                case Keys.Left:
                    this.controller.GoToPreviousPage(true);
                    break;
                case Keys.Right:
                    this.controller.GoToNextPage(true);
                    break;
                case Keys.Space:
                    this.controller.GoToNextPage(true);
                    break;
                case Keys.Control | Keys.A:
                    // viewerPanel.SelectAllStaffSystems();
                    break;
                case Keys.Control | Keys.O:
                    OpenAFile();
                    break;
                case Keys.Q | Keys.Control:
                    this.Close();
                    break;
                case Keys.F:
                    if (m_InFullScreenMode)
                    {
                        ChangeViewMode(false);
                    }
                    else
                    {
                        ChangeViewMode(true);
                    }
                    break;
                case Keys.P:
                    if (m_ControllerPaused)
                    {
                        this.controller.Timer.Resume();
                        m_ControllerPaused = false;
                    }
                    else
                    {
                        this.controller.Timer.Pause();
                        m_ControllerPaused = true;
                    }
                    break;
                case Keys.Z:
                    controller.ToggleMenu();
                    break;
                case Keys.X:
                    controller.ChooseNextMenuItem();
                    break;
                case Keys.C:
                    controller.SelectMenuItem();
                    break;
                default:
                    m_keyStrokeProcessed = false; // let KeyPress event handler handle this keystroke.
                    break;
            }

            return base.ProcessDialogKey(keyData);
        }

        private void preferences_Click(object sender, EventArgs e)
        {
            PreferencesForm form = new PreferencesForm();
            form.Show();
        }

        private void menuItem4_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void menuItem_FullScreen_Click(object sender, EventArgs e)
        {
            if (menuItem_FullScreen.Checked)
            {
                ChangeViewMode(false);
            }
            else
            {
                ChangeViewMode(true);
            }
        }

        /// <summary>
        /// If <see cref="fullScreen"/> is true, sets the app into full screen mode;
        /// sets to normal window otherwise
        /// </summary>
        /// <param name="fullScreen"></param>
        private void ChangeViewMode(bool fullScreen)
        {
            if (fullScreen)
            {
                this.FormBorderStyle = FormBorderStyle.None;
                this.WindowState = FormWindowState.Maximized;
                m_InFullScreenMode = true;
                menuItem_FullScreen.Checked = true;
            }
            else
            {
                this.FormBorderStyle = FormBorderStyle.Sizable;
                this.WindowState = FormWindowState.Normal;
                m_InFullScreenMode = false;
                menuItem_FullScreen.Checked = false;
            }
        }

        private void mi_NextPage_Click(object sender, EventArgs e)
        {
            controller.GoToNextPage(false);
        }

        private void mi_PreviousPage_Click(object sender, EventArgs e)
        {
            controller.GoToPreviousPage(false);
        }

        private void mi_FirstPage_Click(object sender, EventArgs e)
        {
            controller.GoToFirstPage(false);
        }

        private void mi_LastPage_Click(object sender, EventArgs e)
        {
            controller.GoToLastPage(false);
        }
    }
}